package npc.model;

import ai.L2DCSiegeMonsterAI;
import events.clanhall.CHSiegeManager;
import javolution.util.FastMap;
import L2jFT.Config;
import static L2jFT.game.ai.CtrlIntention.AI_INTENTION_ACTIVE;
import L2jFT.game.ai.L2CharacterAI;
import L2jFT.game.model.L2Character;
import L2jFT.game.model.L2Party;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2Summon;
import L2jFT.game.model.instances.L2NpcInstance;
import L2jFT.game.model.instances.L2SummonInstance;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.templates.L2NpcTemplate;

import java.util.ArrayList;

public class L2DCSiegeMonsterInstance extends L2NpcInstance
{
	private boolean _overhit;

	/** Stores the extra (over-hit) damage done to the L2NpcInstance when the attacker uses an over-hit enabled skill */
	private double _overhitDamage;

	/** Stores the attacker who used the over-hit enabled skill on the L2NpcInstance */
	private L2Character _overhitAttacker;

	private int _homeX;
	private int _homeY;
	private int _homeZ;
	private int _homeH;

	protected final class RewardInfo
	{
		protected L2Character attacker;
		protected int dmg = 0;

		public RewardInfo(final L2Character attacker, final int dmg)
		{
			this.attacker = attacker;
			this.dmg = dmg;
		}

		public void addDamage(int dmg)
		{
			if(dmg < 0)
				dmg = 0;
			this.dmg += dmg;
		}

		@Override
		public int hashCode()
		{
			return attacker.getObjectId();
		}
	}

	/**
	 * Constructor of L2MonsterInstance (use L2Character and L2NpcInstance constructor).<BR><BR>
	 * <p/>
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Call the L2Character constructor to set the _template of the L2MonsterInstance (copy skills from template to object and link _calculators to NPC_STD_CALCULATOR) </li>
	 * <li>Set the name of the L2MonsterInstance</li>
	 * <li>Create a RandomAnimation Task that will be launched after the calculated delay if the server allow it </li><BR><BR>
	 *
	 * @param objectId	  Identifier of the object to initialized
	 * @param L2NpcTemplate Template to apply to the NPC
	 */
	public L2DCSiegeMonsterInstance(int objectId, L2NpcTemplate template)
	{
		super(objectId, template);
	}

	@Override
	public L2CharacterAI getAI()
	{
		if(_ai == null)
			_ai = new L2DCSiegeMonsterAI(this);
		return _ai;
	}

	@Override
	public boolean isAutoAttackable(L2Character attacker)
	{
		if(!CHSiegeManager.getInstance().getSiege(34).isInProgress())
			return false;

		if(attacker.isPlayer)
		{
			if(!CHSiegeManager.getInstance().getSiege(34).getAttackers().contains(((L2Player) attacker).getClan()))
				return false;
		}
		else if(attacker.isPet || attacker.isSummon)
			if(!CHSiegeManager.getInstance().getSiege(34).getAttackers().contains(attacker.getOwner().getClan()))
				return false;

		return true;
	}

	@Override
	public void doDie(L2Character killer)
	{
		calculateRewards(killer);
		super.doDie(killer);
	}

	private void calculateRewards(final L2Character lastAttacker)
	{
		final L2Player player = lastAttacker.getPlayer();
		if(player == null || _aggroList == null || _aggroList.isEmpty())
			return;

		final int npcID = getTemplate().npcId;

		FastMap<L2Character, RewardInfo> rewards = new FastMap<L2Character, RewardInfo>().setShared(true);

		for(final AggroInfo info : _aggroList.values())
		{
			L2Character attacker = info.attacker;
			L2Character owner = null;

			if(attacker instanceof L2Summon)
				owner = attacker.getOwner();

			int damage = info.damage;

			if(npcID <= 0 || damage <= 1)
				continue;

			synchronized (this)
			{
				RewardInfo reward = rewards.get(attacker);
				if(reward == null)
					rewards.put(attacker, new RewardInfo(attacker, damage));
				else
					reward.addDamage(damage);

				if(owner != null)
				{
					int levelsDiff = owner.getLevel() - attacker.getLevel();
					if(levelsDiff < 0)
						levelsDiff = 0;
					damage = damage - levelsDiff * 50;
					if(damage < 0)
						damage = 0;
					reward = rewards.get(owner);
					if(reward == null)
						rewards.put(owner, new RewardInfo(owner, damage));
					else
						reward.addDamage(damage);
				}
			}
		}

		for(FastMap.Entry<L2Character, RewardInfo> e = rewards.head(), end = rewards.tail(); e != null && (e = e.getNext()) != end && e != null;)
		{
			final RewardInfo reward = e.getValue();
			if(reward == null)
				continue;

			final L2Character attacker = reward.attacker;
			L2Character master = attacker;

			if(attacker.isSummon)
			{
				master = attacker.getOwner();
				rewards.remove(attacker);
			}

			if(master == null)
				continue;

			if(master.isDead())
				continue;

			L2Party attackerParty = null;
			if(master.isPlayer)
				attackerParty = ((L2Player) master).getParty();

			final int maxHp = getMaxHp();

			if(attackerParty == null)
			{
				int damage = reward.dmg;
				float penalty = 0;
				if(master.isPlayer)
				{
					final L2Summon summon = master.getPet();
					if(summon != null && summon.isSummon)
						penalty = ((L2SummonInstance) summon).getExpPenalty();
				}

				if(damage > maxHp)
					damage = maxHp;

				if(damage > 0)
				{
					final int diff = master.getLevel() - getLevel();

					long xp = 0;
					long sp = 0;

					if(master.knowsObject(this))
					{
						final long[] tmp = calculateExpAndSp(diff, damage);
						xp = tmp[0];
						xp *= 1 - penalty;
						sp = tmp[1];
					}

					if(isOverhit() && player == getOverhitAttacker())
					{
						int overHitExp = calculateOverhitExp(xp);
						player.sendPacket(new SystemMessage(SystemMessage.OVER_HIT));
						player.sendPacket(new SystemMessage(SystemMessage.ACQUIRED_S1_BONUS_EXPERIENCE_THROUGH_OVER_HIT).addNumber(overHitExp));
						xp += overHitExp;
					}

					master.addExpAndSp(xp, sp);
				}
				rewards.remove(master);
			}
			else
			{
				int partyDmg = 0;
				float partyMul = 1.f;

				final ArrayList<L2Player> rewardedMembers = new ArrayList<L2Player>();

				for(final L2Player partyMember : attackerParty.getPartyMembers())
				{
					if(partyMember == null || partyMember.isDead())
						continue;

					final RewardInfo ai = rewards.get(partyMember);
					if(ai != null)
						partyDmg += ai.dmg;

					final L2Summon pet = partyMember.getPet();
					if(pet != null)
					{
						final RewardInfo aiPet = rewards.get(pet);
						if(aiPet != null)
							partyDmg += aiPet.dmg;
						rewards.remove(pet);
					}

					if(!partyMember.isDead())
						rewardedMembers.add(partyMember);

					rewards.remove(partyMember);
				}

				if(partyDmg < maxHp)
					partyMul = (float) partyDmg / maxHp;
				else
					partyDmg = maxHp;

				if(partyDmg > 0)
				{
					long xp = 0;
					long sp = 0;

					if(master.knowsObject(this))
					{
						final int diff = attackerParty.getLevel() - getLevel();
						final long[] tmp = calculateExpAndSp(diff, partyDmg);
						xp = tmp[0];
						sp = tmp[1];
					}

					xp *= partyMul;
					sp *= partyMul;

					// Check for an over-hit enabled strike
					// (When in party, the over-hit exp bonus is given to the whole party and splitted proportionally through the party members)
					if(isOverhit() && player == getOverhitAttacker())
					{
						int overHitExp = calculateOverhitExp(xp);
						attackerParty.broadcastToPartyMembersInRange(player, new SystemMessage(SystemMessage.ACQUIRED_S1_BONUS_EXPERIENCE_THROUGH_OVER_HIT).addNumber(overHitExp), 2700);
						xp += overHitExp;
					}

					attackerParty.distributeXpAndSp(xp, sp, rewardedMembers, lastAttacker, getLevel());
				}
			}
		}
		rewards = null;
	}

	private long[] calculateExpAndSp(long diff, final long damage)
	{
		long xp = (long) (getExpReward() / getTemplate().baseHpMax * damage);
		long sp = (long) (getSpReward() / getTemplate().baseHpMax * damage);

		if(Config.OVER_LEVEL_PENALTY)
			diff = Math.abs(diff);

		if(diff > 5)
		{
			final double mod = Math.pow(.83, diff - 5);
			xp *= mod;
			sp *= mod;
		}

		if(xp < 0)
			xp = 0;
		if(sp < 0)
			sp = 0;

		return new long[] { xp, sp };
	}

	@Override
	public void setOverhitValues(final L2Character attacker, final double damage)
	{
		// Calculate the over-hit damage
		// Ex: mob had 10 HP left, over-hit skill did 50 damage total, over-hit damage is 40
		final double overhitDmg = (getCurrentHp() - damage) * -1;
		if(overhitDmg < 0)
		{
			// we didn't killed the mob with the over-hit strike. (it wasn't really an over-hit strike)
			// let's just clear all the over-hit related values
			setOverhitEnabled(false);
			_overhitDamage = 0;
			_overhitAttacker = null;
			return;
		}
		setOverhitEnabled(true);
		_overhitDamage = overhitDmg;
		_overhitAttacker = attacker;
	}

	@Override
	public void setOverhitEnabled(final boolean status)
	{
		_overhit = status;
	}

	public L2Character getOverhitAttacker()
	{
		return _overhitAttacker;
	}

	public double getOverhitDamage()
	{
		return _overhitDamage;
	}

	@Override
	public boolean isOverhit()
	{
		return _overhit;
	}

	public int calculateOverhitExp(final long normalExp)
	{
		double overhitPercentage = getOverhitDamage() * 100 / getMaxHp();
		if(overhitPercentage > 25)
			overhitPercentage = 25;
		final double overhitExp = overhitPercentage / 100 * normalExp;
		return (int) Math.round(overhitExp);
	}

	public int getHomeX()
	{
		return _homeX;
	}

	public int getHomeY()
	{
		return _homeY;
	}

	public int getHomeZ()
	{
		return _homeZ;
	}

	public int getHomeH()
	{
		return _homeH;
	}

	@Override
	public void onSpawn()
	{
		_homeX = getX();
		_homeY = getY();
		_homeZ = getZ();
		_homeH = getHeading();
		super.onSpawn();
	}

	public void returnHome()
	{
		if(!isInRangeSq(_homeX, _homeY, 40000))
		{
			clearAggroList();
			moveToLocation(_homeX, _homeY, _homeZ, 0, false);
		}
	}

	public void teleportHome()
	{
		clearAggroList();
		setAttackTimeout(Integer.MAX_VALUE);
		if(hasAI())
		{
			getAI().setAttackTarget(null);
			getAI().clientStopMoving(null);
			getAI().setIntention(AI_INTENTION_ACTIVE);
		}
		teleToLocation(_homeX, _homeY, _homeZ);
	}

	@Override
	public boolean hasRandomAnimation()
	{
		return false;
	}

	@Override
	public boolean isInvul()
	{
		return false;
	}

	@Override
	public boolean isFearImmune()
	{
		return true;
	}
}
