package L2jFT.game;

import L2jFT.Config;
import L2jFT.game.instancemanager.DayNightSpawnManager;
import L2jFT.game.model.L2Character;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2World;
import L2jFT.game.model.L2WorldRegion;
import L2jFT.game.model.entity.SevenSigns;
import L2jFT.game.network.serverpackets.L2GameServerPacket;
import L2jFT.game.network.serverpackets.SunRise;
import L2jFT.game.network.serverpackets.SunSet;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.logging.Logger;

@SuppressWarnings( { "nls", "unqualified-field-access", "boxing" })
public class GameTimeController
{
	protected static final Logger _log = Logger.getLogger(GameTimeController.class.getName());

	public static final int TICKS_PER_SECOND = 10;
	public static final int MILLIS_IN_TICK = 1000 / TICKS_PER_SECOND;

	private static GameTimeController _instance = new GameTimeController();

	protected static int _gameTicks;
	protected static long _gameStartTime;
	protected static boolean _isNight;

	protected TimerThread _timer;

	public int _moveTimestamp;

	/**
	 * one ingame day is 240 real minutes
	 */
	public static GameTimeController getInstance()
	{
		return _instance;
	}

	private GameTimeController()
	{
		_gameStartTime = System.currentTimeMillis() - 3600000; // offset so that the server starts a day begin
		_gameTicks = 3600000 / MILLIS_IN_TICK; // offset so that the server starts a day begin

		_timer = new TimerThread();
		_timer.start();

		ThreadPoolManager.getInstance().scheduleGeneralAtFixedRate(new TimerWatcher(), 0, 1000);
		ThreadPoolManager.getInstance().scheduleGeneralAtFixedRate(new BroadcastSunState(), 120000, 600000);
	}

	public boolean isNowNight()
	{
		return _isNight;
	}

	public int getGameTime()
	{
		return _gameTicks / (TICKS_PER_SECOND * 10);
	}

	public static int getGameTicks()
	{
		return _gameTicks;
	}

	protected void moveObjects()
	{
		// Check if the position has alreday be calculated
		if(_gameTicks - _moveTimestamp < 5)
			return;

		// Set the timer of last position update to now
		_moveTimestamp = _gameTicks;

		Iterator<L2WorldRegion> it = L2World.getActiveRegions().iterator();
		//noinspection WhileLoopReplaceableByForEach
		while(it.hasNext())
		{
			L2WorldRegion region = it.next();
			if(region != null && region.isActive())
				for(Enumeration<L2Character> i = region.getCharacters().elements(); i.hasMoreElements();)
				{
					L2Character cha = i.nextElement();
					if(cha == null || !cha.isMoving && !cha.isFollow || !cha.hasAI() || cha.isMovementDisabled())
						continue;
					cha.updatePosition(_gameTicks);
				}
		}

		for(L2Character cha : L2World.getCustomMoveObjects())
		{
			if(cha == null || !cha.isMoving && !cha.isFollow || !cha.hasAI() || cha.isMovementDisabled())
				continue;
			cha.updatePosition(_gameTicks);
		}
	}

	public void stopTimer()
	{
		_timer._stop = true;
	}

	class TimerThread extends Thread
	{
		protected boolean _stop;
		protected Exception _error;

		public TimerThread()
		{
			super("TimerThread");
			setDaemon(true);
			setPriority(MAX_PRIORITY);
			_stop = false;
			_error = null;
		}

		@Override
		public void run()
		{
			try
			{
				for(;;)
				{
					long runtime = System.currentTimeMillis() - _gameStartTime;
					_gameTicks = (int) (runtime / MILLIS_IN_TICK);
					moveObjects();
					runtime = System.currentTimeMillis() - _gameStartTime;
					int sleepTime = MILLIS_IN_TICK - (int) runtime % MILLIS_IN_TICK;

					if(sleepTime > 0)
						sleep(sleepTime);

					if(_stop)
						throw new Exception("stopTimer() was called");
				}
			}
			catch(Exception e)
			{
				_error = e;
			}
		}
	}

	class TimerWatcher implements Runnable
	{
		public void run()
		{
			if(!_timer.isAlive())
			{
				String time = (new SimpleDateFormat("HH:mm:ss")).format(new Date());
				_log.warning(time + " TimerThread stop with following error. restart it.");
				if(_timer._error != null)
					_timer._error.printStackTrace();

				_timer = new TimerThread();
				_timer.start();
			}
		}
	}

	public class BroadcastSunState implements Runnable
	{
		public void run()
		{
			int h = getGameTime() / 60 % 24; // Time in hour
			boolean tempIsNight = h < Config.DAY_STATUS_SUN_RISE_AT || h >= Config.DAY_STATUS_SUN_SET_AT;

			if(tempIsNight != isNowNight()) // If diff day/night state
				_isNight = tempIsNight; // Set current day/night varible to value of temp varible
			else
				return; // Do nothing if same state

			DayNightSpawnManager.getInstance().notifyChangeMode();

			if(!Config.DAY_STATUS_FORCE_CLIENT_UPDATE) // Do not force client to update their day/night status
				return;

			if(SevenSigns.getInstance().isSealValidationPeriod()) // Do nothing if in Seal validation period 
				return;

			// Force client day/night cycle to sync everyone's day/night state
			L2GameServerPacket packet;
			if(isNowNight()) // If is now night
			{
				packet = new SunSet(); // Set client day/night state to night
				if(Config.DEBUG)
					_log.fine("SunSet");
			}
			else
			{
				packet = new SunRise(); // Set client day/night state to day
				if(Config.DEBUG)
					_log.fine("SunRise");
			}

			for(L2Player player : L2World.getAllPlayers())
			{
				player.sendPacket(packet);
				player.checkDayNightMessages();
			}
		}
	}
}