package L2jFT.game;

import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigInteger;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.security.GeneralSecurityException;
import java.security.KeyFactory;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.RSAKeyGenParameterSpec;
import java.security.spec.RSAPublicKeySpec;
import java.util.ArrayList;
import java.util.Map;
import java.util.Random;
import java.util.logging.Logger;

import javolution.util.FastList;
import javolution.util.FastMap;
import L2jFT.Config;
import L2jFT.extensions.multilang.CustomMessage;
import L2jFT.extensions.scripts.Functions;
import L2jFT.game.network.gameserverpackets.AuthRequest;
import L2jFT.game.network.gameserverpackets.BanIP;
import L2jFT.game.network.gameserverpackets.BlowFishKey;
import L2jFT.game.network.gameserverpackets.ChangeAccessLevel;
import L2jFT.game.network.gameserverpackets.ChangePassword;
import L2jFT.game.network.gameserverpackets.CheckSessionKey;
import L2jFT.game.network.gameserverpackets.GameServerBasePacket;
import L2jFT.game.network.gameserverpackets.PlayerAuthRequest;
import L2jFT.game.network.gameserverpackets.PlayerInGame;
import L2jFT.game.network.gameserverpackets.PlayerLogout;
import L2jFT.game.network.gameserverpackets.RequestBanIPList;
import L2jFT.game.network.gameserverpackets.Restart;
import L2jFT.game.network.gameserverpackets.ServerStatus;
import L2jFT.game.network.gameserverpackets.TestAnswer;
import L2jFT.game.network.gameserverpackets.TestPacket;
import L2jFT.game.network.gameserverpackets.UnbanIP;
import L2jFT.game.network.gameserverpackets.AuthRequest.AdvIP;
import L2jFT.game.network.loginserverpackets.AuthResponse;
import L2jFT.game.network.loginserverpackets.BanIPList;
import L2jFT.game.network.loginserverpackets.ChangePasswordResponse;
import L2jFT.game.network.loginserverpackets.CheckSessionKeyResponse;
import L2jFT.game.network.loginserverpackets.InitLS;
import L2jFT.game.network.loginserverpackets.KickPlayer;
import L2jFT.game.network.loginserverpackets.LoginServerFail;
import L2jFT.game.network.loginserverpackets.PlayerAuthResponse;
import L2jFT.game.network.loginserverpackets.BanIPList.BannedIp;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2World;
import L2jFT.game.network.L2GameClient;
import L2jFT.game.network.L2GameClient.GameClientState;
import L2jFT.game.network.serverpackets.AuthLoginFail;
import L2jFT.game.network.serverpackets.CharSelectInfo;
import L2jFT.game.network.serverpackets.GameGuardQuery;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.login.crypt.NewCrypt;
import L2jFT.util.Util;

public class LoginServerThread extends Thread
{
	protected static Logger _log = Logger.getLogger(LoginServerThread.class.getName());

	/** The LoginServerThread singleton */
	private static LoginServerThread _instance;

	/** {@see L2jFT.login.LoginServer#PROTOCOL_REV } */
	private static final int REVISION = 0x0102;
	private RSAPublicKey _publicKey;
	private String _hostname;
	private int _port;
	private int _gamePort;
	Socket _loginSocket;
	private InputStream _in;
	private OutputStream _out;
	private boolean _shutdown = false;

	/**
	 * The BlowFish engine used to encrypt packets<br>
	 * It is first initialized with a unified key:<br>
	 * "_;v.]05-31!|+-%xT!^[$\00"<br>
	 * <br>
	 * and then after handshake, with a new key sent by<br>
	 * loginserver during the handshake. This new key is stored<br>
	 * in {@link #_blowfishKey}
	 */
	private NewCrypt _blowfish;
	private byte[] _blowfishKey;
	private byte[] _hexID;
	private boolean _acceptAlternate;
	private int _requestID;
	private int _serverID;
	private boolean _reserveHost;
	private int _maxPlayer;
	private final ArrayList<WaitingClient> _waitingClients;
	Map<String, L2GameClient> _accountsInGameServer;
	private int _status;
	private String _serverName;
	private String _gameExternalHost;
	private String _gameInternalHost;
	private ArrayList<AdvIP> _ips = new ArrayList<AdvIP>();
	private ArrayList<BannedIp> _baniplist = new ArrayList<BannedIp>();
	private Boolean _requested;
	int _tryes = 0;
	boolean _answered = false;
	TestPacketThread _tpt;

	private class WaitingClient
	{
		public int timestamp;
		public String account;
		public L2GameClient clientThread;
		public SessionKey session;

		public WaitingClient(String acc, L2GameClient client, SessionKey key)
		{
			account = acc;
			timestamp = GameTimeController.getGameTicks();
			clientThread = client;
			session = key;
		}
	}

	public void addWaitingClientAndSendRequest(String acc, L2GameClient client, SessionKey key)
	{
		WaitingClient wc = new WaitingClient(acc, client, key);
		_waitingClients.add(wc);
		PlayerAuthRequest par = new PlayerAuthRequest(acc, key);
		try
		{
			sendPacket(par);
		}
		catch(IOException e)
		{
			_log.warning("Error while sending player auth request");
			e.printStackTrace();
		}
	}

	public void sendLogout(String account)
	{
		PlayerLogout pl = new PlayerLogout(account);
		try
		{
			sendPacket(pl);
		}
		catch(IOException e)
		{
			_log.warning("Error while sending logout packet to login");
			e.printStackTrace();
		}
	}

	public void addGameServerLogin(String account, L2GameClient client)
	{
		L2GameClient oldClient = _accountsInGameServer.remove(account);
		if(oldClient != null)
			oldClient.closeNow(true);
		_accountsInGameServer.put(account, client);
	}

	public void sendCheckSessionKey(String account, SessionKey key)
	{
		CheckSessionKey csk = new CheckSessionKey(account, key);
		try
		{
			sendPacket(csk);
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}

	public void sendAccessLevel(String account, int level, String comments, int banTime)
	{
		ChangeAccessLevel cal = new ChangeAccessLevel(account, level, comments, banTime);
		try
		{
			sendPacket(cal);
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}

	public boolean sendBanIP(String ip, String admin)
	{
		BanIP bi = new BanIP(ip, admin);
		try
		{
			sendPacket(bi);
		}
		catch(IOException e)
		{
			e.printStackTrace();
			return false;
		}
		return true;
	}

	public boolean sendUnbanIP(String ip)
	{
		UnbanIP ubi = new UnbanIP(ip);
		try
		{
			sendPacket(ubi);
		}
		catch(IOException e)
		{
			e.printStackTrace();
			return false;
		}
		return true;
	}

	public void RequestBanIPList()
	{
		RequestBanIPList rbil = new RequestBanIPList();
		_requested = true;
		try
		{
			sendPacket(rbil);
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}

	public ArrayList<BannedIp> getBanIPList()
	{
		return _baniplist;
	}

	public Boolean isRequested()
	{
		return _requested;
	}

	public LoginServerThread()
	{
		super("LoginServerThread");
		_port = Config.GAME_SERVER_LOGIN_PORT;
		_gamePort = Config.PORT_GAME;
		_hostname = Config.GAME_SERVER_LOGIN_HOST;
		_hexID = Config.HEX_ID;
		if(_hexID == null)
		{
			_hexID = generateHex(16);
			Config.saveHexid(hexToString(_hexID));
		}
		_acceptAlternate = Config.ACCEPT_ALTERNATE_ID;
		_requestID = Config.REQUEST_ID;
		_reserveHost = Config.RESERVE_HOST_ON_LOGIN;
		_gameExternalHost = Config.EXTERNAL_HOSTNAME;
		_gameInternalHost = Config.INTERNAL_HOSTNAME;
		_ips = Config.GAMEIPS;
		_waitingClients = new ArrayList<WaitingClient>();
		_accountsInGameServer = new FastMap<String, L2GameClient>();
	}

	public static LoginServerThread getInstance()
	{
		if(_instance == null)
			_instance = new LoginServerThread();
		return _instance;
	}

	public void changePassword(String accName, String oldPass, String newPass)
	{
		ChangePassword cp = new ChangePassword(accName, oldPass, newPass);
		try
		{
			sendPacket(cp);
		}
		catch(IOException e)
		{
			if(Config.DEBUG)
				e.printStackTrace();
		}
	}

	private void changePasswordResult(ChangePasswordResponse cp)
	{
		L2GameClient c = _accountsInGameServer.get(cp.getAccount());

		if(c == null)
			return;

		L2Player pl = c.getActiveChar();

		if(pl == null)
			return;

		if(cp.isChanged())
			Functions.show(new CustomMessage("scripts.commands.user.password.ResultTrue", pl), pl);
		else
			Functions.show(new CustomMessage("scripts.commands.user.password.ResultFalse", pl), pl);
	}

	@Override
	public void run()
	{
		while(!_shutdown)
		{
			int lengthHi = 0;
			int lengthLo = 0;
			int length = 0;
			boolean checksumOk = false;
			try
			{
				// Connection
				_log.info("Connecting to login on " + _hostname + ":" + _port);
				_loginSocket = new Socket(_hostname, _port);
				_in = _loginSocket.getInputStream();
				_out = new BufferedOutputStream(_loginSocket.getOutputStream());

				//init Blowfish
				_blowfishKey = generateHex(40);
				_blowfish = new NewCrypt("_;v.]05-31!|+-%xT!^[$\00");
				while(!_shutdown)
				{
					if(_loginSocket.isInputShutdown() || _loginSocket.isClosed())
						break;
					if(_in.available() == 0)
					{
						try
						{
							sleep(200);
							if(Config.PING && _tryes == 20)
								sendTestPacket();
							_tryes++;
						}
						catch(Exception e)
						{}
						continue;
					}
					_tryes = 0;

					lengthLo = _in.read();
					lengthHi = _in.read();
					length = lengthHi * 256 + lengthLo;

					if(lengthHi < 0)
					{
						_log.finer("LoginServerThread: Login terminated the connection.");
						break;
					}

					byte[] incoming = new byte[length];
					incoming[0] = (byte) lengthLo;
					incoming[1] = (byte) lengthHi;

					int receivedBytes = 0;
					int newBytes = 0;
					while(newBytes != -1 && receivedBytes < length - 2)
					{
						newBytes = _in.read(incoming, 2, length - 2);
						receivedBytes = receivedBytes + newBytes;
					}

					if(receivedBytes != length - 2)
					{
						_log.warning("Incomplete Packet is sent to the server, closing connection.(GS)");
						break;
					}

					byte[] decrypt = new byte[length - 2];
					System.arraycopy(incoming, 2, decrypt, 0, decrypt.length);
					// decrypt if we have a key
					decrypt = _blowfish.decrypt(decrypt);
					checksumOk = NewCrypt.verifyChecksum(decrypt);

					if(!checksumOk)
					{
						_log.warning("Incorrect packet checksum, ignoring packet (GS)");
						break;
					}

					if(Config.DEBUG)
						_log.warning("[C]\n" + printData(decrypt, decrypt.length));

					int packetType = decrypt[0] & 0xff;
					switch(packetType)
					{
						case 0x00:
							InitLS init = new InitLS(decrypt);
							if(Config.DEBUG)
								_log.info("Init received");
							if(init.getRevision() != REVISION)
								//TODO: revision mismatch
								_log.warning("Revision mismatch between LS and GS");
							try
							{
								KeyFactory kfac = KeyFactory.getInstance("RSA");
								BigInteger modulus = new BigInteger(init.getRSAKey());
								RSAPublicKeySpec kspec1 = new RSAPublicKeySpec(modulus, RSAKeyGenParameterSpec.F4);
								_publicKey = (RSAPublicKey) kfac.generatePublic(kspec1);
								if(Config.DEBUG)
									_log.info("RSA key set up");
							}
							catch(GeneralSecurityException e)
							{
								_log.warning("Troubles whil init the public key send by login");
							}
							//send the blowfish key through the rsa encryption
							BlowFishKey bfk = new BlowFishKey(_blowfishKey, _publicKey);
							sendPacket(bfk);
							if(Config.DEBUG)
								_log.info("Sent new blowfish key");
							//now, only accept paket with the new encryption
							_blowfish = new NewCrypt(_blowfishKey);
							if(Config.DEBUG)
								_log.info("Changed blowfish key");
							AuthRequest ar = new AuthRequest(_requestID, _acceptAlternate, _hexID, _gameExternalHost, _gameInternalHost, _gamePort, _reserveHost, _maxPlayer, _ips);
							sendPacket(ar);
							if(Config.DEBUG)
								_log.info("Sent AuthRequest to login");
							break;
						case 0x01:
							LoginServerFail lsf = new LoginServerFail(decrypt);
							lsf.getReason(); //TODO: manage errors from the loginServer
							_log.info("Damn! Registeration Failed: " + lsf.getReason());
							// login will close the connection here
							break;
						case 0x02:
							AuthResponse aresp = new AuthResponse(decrypt);
							_serverID = aresp.getServerId();
							_serverName = aresp.getServerName();
							// Config.saveHexid(hexToString(_hexID));
							_log.info("Registered on login as Server " + _serverID + " : " + _serverName);
							//TODO: save hexID
							ServerStatus st = new ServerStatus();
							if(Config.SERVER_LIST_BRACKET)
								st.addAttribute(ServerStatus.SERVER_LIST_SQUARE_BRACKET, ServerStatus.ON);
							else
								st.addAttribute(ServerStatus.SERVER_LIST_SQUARE_BRACKET, ServerStatus.OFF);
							if(Config.SERVER_LIST_CLOCK)
								st.addAttribute(ServerStatus.SERVER_LIST_CLOCK, ServerStatus.ON);
							else
								st.addAttribute(ServerStatus.SERVER_LIST_CLOCK, ServerStatus.OFF);
							if(Config.SERVER_LIST_TESTSERVER)
								st.addAttribute(ServerStatus.TEST_SERVER, ServerStatus.ON);
							else
								st.addAttribute(ServerStatus.TEST_SERVER, ServerStatus.OFF);
							if(Config.SERVER_GMONLY)
								st.addAttribute(ServerStatus.SERVER_LIST_STATUS, ServerStatus.STATUS_GM_ONLY);
							else
								st.addAttribute(ServerStatus.SERVER_LIST_STATUS, ServerStatus.STATUS_AUTO);
							sendPacket(st);
							if(L2World.getAllPlayers().size() > 0)
							{
								FastList<String> playerList = new FastList<String>();
								for(L2Player player : L2World.getAllPlayers())
									playerList.add(player.getAccountName());
								PlayerInGame pig = new PlayerInGame(playerList);
								sendPacket(pig);
							}
							break;
						case 0x03:
							PlayerAuthResponse par = new PlayerAuthResponse(decrypt);
							String account = par.getAccount();
							WaitingClient wcToRemove = null;
							synchronized (_waitingClients)
							{
								for(WaitingClient wc : _waitingClients)
									if(wc.account.equals(account))
										wcToRemove = wc;
							}
							if(wcToRemove != null)
							{
								if(par.isAuthed())
								{
									if(Config.DEBUG)
										_log.info("Login accepted player " + wcToRemove.account + " waited(" + (GameTimeController.getGameTicks() - wcToRemove.timestamp) + "ms)");
									PlayerInGame pig = new PlayerInGame(par.getAccount());
									sendPacket(pig);
									wcToRemove.clientThread.setState(GameClientState.AUTHED);
									wcToRemove.clientThread.setBonus(par.getBonus());
									CharSelectInfo cl = new CharSelectInfo(wcToRemove.account, wcToRemove.clientThread.getSessionId().playOkID1);
									wcToRemove.clientThread.getConnection().sendPacket(cl);
									wcToRemove.clientThread.setSessionId(wcToRemove.session);
									wcToRemove.clientThread.setCharSelection(cl.getCharInfo());
								}
								else
								{
									_log.warning("session key is not correct. closing connection");
									wcToRemove.clientThread.getConnection().sendPacket(new AuthLoginFail(1));
									wcToRemove.clientThread.closeNow(false);
								}
								_waitingClients.remove(wcToRemove);
							}
							break;
						case 0x04:
							KickPlayer kp = new KickPlayer(decrypt);
							kickAccount(kp.getAccount());
							break;
						case 0x05:
							BanIPList bil = new BanIPList(decrypt);
							_requested = false;
							_baniplist = bil.getBanIPList();
							break;
						case 0x06:
							sendTestAnswer();
							break;
						case 0x07:
							_answered = true;
							break;
						case 0x08:
							ChangePasswordResponse cp;
							cp = new ChangePasswordResponse(decrypt);
							changePasswordResult(cp);
							break;
						case 0x0a:
							CheckSessionKeyResponse cskr = new CheckSessionKeyResponse(decrypt);
							L2GameClient client = _accountsInGameServer.get(cskr.getAccount());
							client.setChecking(false);
							client.setPassed(cskr.isPassed());
							break;

					}
				}
			}
			catch(UnknownHostException e)
			{
				e.printStackTrace();
			}
			catch(IOException e)
			{
				_log.info("Disconnected from Login Trying to reconnect");
			}
			finally
			{
				try
				{
					_loginSocket.close();
				}
				catch(Exception e)
				{}
			}

			try
			{
				Thread.sleep(5000);
			}
			catch(InterruptedException e)
			{
				//e.printStackTrace();
			}
		}
	}

	private String hexToString(byte[] hex)
	{
		return new BigInteger(hex).toString(16);
	}

	public void kickAccount(String account)
	{
		L2GameClient client = _accountsInGameServer.remove(account);

		if(client != null)
		{
			if(client.getActiveChar() != null)
			{
				client.getActiveChar().sendPacket(new SystemMessage(SystemMessage.ANOTHER_PERSON_HAS_LOGGED_IN_WITH_THE_SAME_ACCOUNT));
				ThreadPoolManager.getInstance().scheduleGeneral(new KickTask(client), 1000);
			}
			else
			{
				client.closeNow(false);
				LoginServerThread.getInstance().sendLogout(account);
			}
		}
	}

	private class KickTask implements Runnable
	{
		private final L2GameClient _client;

		public KickTask(L2GameClient client)
		{
			_client = client;
		}

		public void run()
		{
			_client.getActiveChar().logout(false, false);
			//_accountsInGameServer.remove(_account);
			//_accountsInGameServer.get(account).closeNow(false);
			LoginServerThread.getInstance().sendLogout(_client.getLoginName());
		}
	}

	public void removeAccount(String account)
	{
		_accountsInGameServer.remove(account);
	}

	public static byte[] generateHex(int size)
	{
		byte[] array = new byte[size];
		Random rnd = new Random();
		for(int i = 0; i < size; i++)
			array[i] = (byte) rnd.nextInt(256);
		if(Config.DEBUG)
			_log.fine("Generated random String:  \"" + array + "\"");
		return array;
	}

	/**
	 * @param sl
	 * @throws IOException
	 */
	private void sendPacket(GameServerBasePacket sl) throws IOException
	{
		if(_loginSocket.isClosed())
			return;
		byte[] data = sl.getContent();
		NewCrypt.appendChecksum(data);
		if(Config.DEBUG)
			_log.finest("[S]\n" + Util.printData(data));
		data = _blowfish.crypt(data);

		int len = data.length + 2;
		synchronized (_out) //avoids tow threads writing in the mean time
		{
			try
			{
				_out.write(len & 0xff);
				_out.write(len >> 8 & 0xff);
				_out.write(data);
				_out.flush();
			}
			catch(SocketException e)
			{}
		}
	}

	private String fillHex(int data, int digits)
	{
		String number = Integer.toHexString(data);

		for(int i = number.length(); i < digits; i++)
			number = "0" + number;
		return number;
	}

	private String printData(byte[] data, int len)
	{
		StringBuffer result = new StringBuffer();

		int counter = 0;

		for(int i = 0; i < len; i++)
		{
			if(counter % 16 == 0)
				result.append(fillHex(i, 4) + ": ");

			result.append(fillHex(data[i] & 0xff, 2) + " ");
			counter++;
			if(counter == 16)
			{
				result.append("   ");

				int charpoint = i - 15;
				for(int a = 0; a < 16; a++)
				{
					int t1 = data[charpoint++];
					if(t1 > 0x1f && t1 < 0x80)
						result.append((char) t1);
					else
						result.append('.');
				}

				result.append("\n");
				counter = 0;
			}
		}

		int rest = data.length % 16;
		if(rest > 0)
		{
			for(int i = 0; i < 17 - rest; i++)
				result.append("   ");

			int charpoint = data.length - rest;
			for(int a = 0; a < rest; a++)
			{
				int t1 = data[charpoint++];
				if(t1 > 0x1f && t1 < 0x80)
					result.append((char) t1);
				else
					result.append('.');
			}
			result.append("\n");
		}
		return result.toString();
	}

	/**
	 * @param maxPlayer The maxPlayer to set.
	 */
	public void setMaxPlayer(int maxPlayer)
	{
		_maxPlayer = maxPlayer;
	}

	/**
	 * @return Returns the maxPlayer.
	 */
	public int getMaxPlayer()
	{
		return _maxPlayer;
	}

	/**
	 * @param server_gm_only
	 */
	public void sendServerStatus(int id, int value)
	{
		ServerStatus ss = new ServerStatus();
		ss.addAttribute(id, value);
		try
		{
			sendPacket(ss);
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}

	/**
	 * @return
	 */
	public String getStatusString()
	{
		return ServerStatus.statusString[_status];
	}

	/**
	 * @return
	 */
	public boolean isClockShown()
	{
		// TODO Auto-generated method stub
		return false;
	}

	/**
	 * @return
	 */
	public boolean isBracketShown()
	{
		// TODO Auto-generated method stub
		return false;
	}

	/**
	 * @return Returns the serverName.
	 */
	public String getServerName()
	{
		return _serverName;
	}

	public static class SessionKey
	{
		public int playOkID1;
		public int playOkID2;
		public int loginOkID1;
		public int loginOkID2;

		public SessionKey(int loginOK1, int loginOK2, int playOK1, int playOK2)
		{
			playOkID1 = playOK1;
			playOkID2 = playOK2;
			loginOkID1 = loginOK1;
			loginOkID2 = loginOK2;
		}

		@Override
		public String toString()
		{
			return "PlayOk: " + playOkID1 + " " + playOkID2 + " LoginOk:" + loginOkID1 + " " + loginOkID2;
		}
	}

	public class CheckGG extends Thread
	{
		L2GameClient _client;

		public CheckGG(L2GameClient client)
		{
			_client = client;
		}

		@Override
		public void run()
		{
			if(!_client.isGameGuardOk())
			{
				if(Config.DEBUG)
					_log.warning(_client.getLoginName() + " not using GameGuard. Closing connection");
				_client.getConnection().sendPacket(new AuthLoginFail(1));
				_client.closeNow(false);
			}
		}
	}

	public void sendGameGuardQuery(L2GameClient client)
	{
		ThreadPoolManager.getInstance().scheduleGeneral(new CheckGG(client), 5 * 1000);
		GameGuardQuery ggq = new GameGuardQuery();
		client.getConnection().sendPacket(ggq);
	}

	private void sendTestPacket()
	{
		try
		{
			if(_tpt == null)
			{
				TestPacket tp = new TestPacket();
				sendPacket(tp);
				_tpt = new TestPacketThread();
				ThreadPoolManager.getInstance().scheduleGeneral(_tpt, Config.PING_TIME * 1000);
			}
		}
		catch(Exception e)
		{}
	}

	private void sendTestAnswer()
	{
		try
		{
			TestAnswer tp = new TestAnswer();
			sendPacket(tp);
		}
		catch(Exception e)
		{}
	}

	public class TestPacketThread implements Runnable
	{
		public void run()
		{
			try
			{
				if(!_answered)
					_loginSocket.close();
				_tryes = 0;
				_answered = false;
				_tpt = null;
			}
			catch(Exception e)
			{}
		}
	}

	public void shutdown()
	{
		try
		{
			sendPacket(new Restart());
		}
		catch(IOException e)
		{}
		_shutdown = true;
	}
}
