package L2jFT.game.ai;

import static L2jFT.game.ai.CtrlIntention.AI_INTENTION_ACTIVE;
import static L2jFT.game.ai.CtrlIntention.AI_INTENTION_ATTACK;
import static L2jFT.game.ai.CtrlIntention.AI_INTENTION_IDLE;

import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.Future;
import java.util.logging.Logger;

import javolution.util.FastList;
import L2jFT.Config;
import L2jFT.game.GameTimeController;
import L2jFT.game.ThreadPoolManager;
import L2jFT.game.geodata.GeoData;
import L2jFT.game.model.L2Character;
import L2jFT.game.model.L2Object;
import L2jFT.game.model.L2Playable;
import L2jFT.game.model.L2Skill;
import L2jFT.game.model.L2Spawn;
import L2jFT.game.model.L2Summon;
import L2jFT.game.model.L2World;
import L2jFT.game.model.L2WorldRegion;
import L2jFT.game.model.L2Skill.SkillType;
import L2jFT.game.model.instances.L2ChestInstance;
import L2jFT.game.model.instances.L2FestivalMonsterInstance;
import L2jFT.game.model.instances.L2MinionInstance;
import L2jFT.game.model.instances.L2NpcInstance;
import L2jFT.game.model.instances.L2RaidBossInstance;
import L2jFT.game.model.instances.L2TamedBeastInstance;
import L2jFT.game.network.serverpackets.MagicSkillUse;
import L2jFT.game.tables.TerritoryTable;
import L2jFT.util.Location;
import L2jFT.util.Rnd;

/**
 * This class manages AI of L2NpcInstance.<BR><BR>
 */
public class DefaultAI extends L2CharacterAI implements Runnable
{
	public static enum TaskType
	{
		MOVE,
		ATTACK,
		CAST,
		BUFF
	}

	public static class Task
	{
		public TaskType type;
		public L2Skill skill;
		public L2Character target;
		public int x;
		public int y;
		public int z;
	}
	
	public class Teleport implements Runnable
	{
		Location _destination;

		public Teleport(Location destination)
		{
			_destination = destination;
		}

		public void run()
		{
			_thisActor.teleToLocation(_destination);
		}
	}
	
	protected static Logger _log = Logger.getLogger(DefaultAI.class.getName());
	
	protected int MAX_ATTACK_TIMEOUT = 40000 / GameTimeController.MILLIS_IN_TICK; // int millis->ticks
	protected int TELEPORT_TIMEOUT = 10000 / GameTimeController.MILLIS_IN_TICK; // int millis->ticks
	protected int MAX_PATHFIND_FAILS = 5;
	protected int AI_TASK_DELAY = 1000;
	
	/** Current L2NpcInstance */
	protected L2NpcInstance _thisActor;
	
	/** The L2NpcInstance AI task executed every 1s (call onEvtThink method)*/
	protected Future _aiTask;
	
	/** The flag used to indicate that a thinking action is in progress */
	private boolean _thinking = false;
	
	protected int _aggroRange;
	protected boolean _aggressive;
	
	/** The L2NpcInstance aggro counter */
	protected int _globalAggro;

	protected long _randomAnimationEnd;
	protected boolean _isMobile = true;
	protected int _pathfind_fails;
	
	protected double _att_hp;
	protected double _att_mp;
	protected double _att_max_hp;
	protected double _att_max_mp;
	protected double _def_hp;
	protected double _def_mp;
	protected double _def_max_hp;
	protected double _def_max_mp;

	/** Список заданий */
	protected ArrayList<Task> _task_list = new ArrayList<Task>();

	/** Номер текущего задания */
	protected int _def_step;

	/** Показывает, есть ли задания */
	protected boolean _def_think = false;

	protected L2Skill _r_skill;
	protected ArrayList<L2Skill> _d_skill = new ArrayList<L2Skill>();
	protected ArrayList<L2Skill> _t_skill = new ArrayList<L2Skill>();

	public L2Skill[] _heal;
	public L2Skill[] _buff;
	public L2Skill[] _teleport;
	public L2Skill[] _mute;
	public L2Skill[] _paralyze;
	public L2Skill[] _root;
	public L2Skill[] _sleep;
	public L2Skill[] _stun;
	public L2Skill[] _cancel;
	public L2Skill[] _debuff;
	public L2Skill[] _dot;
	public L2Skill[] _poison;
	public L2Skill[] _bleed;
	public L2Skill[] _mdot;
	public L2Skill[] _manadam;
	public L2Skill[] _mdam;
	public L2Skill[] _pdam;
	public L2Skill[] _drain;
	public L2Skill[] _drainsoul;
	public L2Skill[] _healpercent;
	public L2Skill[] _hot;

	public L2Skill[] _dot_skills;
	public L2Skill[] _buff_skills;
	public L2Skill[] _dam_skills;
	public L2Skill[] _debuff_skills;

	protected FastList<L2NpcInstance> ActiveFriendTargets(boolean active, boolean attack)
	{
		FastList<L2NpcInstance> ActiveFriends = new FastList<L2NpcInstance>();
		L2WorldRegion region = L2World.getRegion(_actor);
		if(region == null)
			return ActiveFriends;
		for(L2NpcInstance npc : region.getNpcs().values())
			if(npc != null && npc.hasAI())
				if(!npc.isDead() && (attack && npc.getAI().getIntention() == AI_INTENTION_ATTACK || active && npc.getAI().getIntention() == AI_INTENTION_ACTIVE))
					if(GeoData.getInstance().canSeeTarget(_actor, npc))
						ActiveFriends.add(npc);
		return ActiveFriends;
	}

	/**
	 * Сложение массивов скиллов
	 */
	public static L2Skill[] msum(L2Skill[] array1, L2Skill[] array2)
	{
		L2Skill[] result = new L2Skill[array1.length + array2.length];

		for(int i = 0; i < array1.length; i++)
			result[i] = array1[i];

		for(int i = 0; i < array2.length; i++)
			result[i + array1.length] = array2[i];

		return result;
	}

	public DefaultAI(L2Character actor)
	{
		super(actor);

		_globalAggro = -10; // 10 seconds timeout of ATTACK after respawn
		_thisActor = (L2NpcInstance) _actor;
		_thisActor.setAttackTimeout(Integer.MAX_VALUE);
		_aggroRange = _thisActor.getAggroRange();
		_aggressive = _thisActor.isAggressive();

		_heal = _thisActor.getTemplate().getSkillsByType(SkillType.HEAL);
		_buff = _thisActor.getTemplate().getSkillsByType(SkillType.BUFF);
		_teleport = _thisActor.getTemplate().getSkillsByType(SkillType.TELEPORT_NPC);
		_mute = _thisActor.getTemplate().getSkillsByType(SkillType.MUTE);
		_paralyze = _thisActor.getTemplate().getSkillsByType(SkillType.PARALYZE);
		_root = _thisActor.getTemplate().getSkillsByType(SkillType.ROOT);
		_sleep = _thisActor.getTemplate().getSkillsByType(SkillType.SLEEP);
		_stun = _thisActor.getTemplate().getSkillsByType(SkillType.STUN);
		_cancel = _thisActor.getTemplate().getSkillsByType(SkillType.CANCEL);
		_debuff = _thisActor.getTemplate().getSkillsByType(SkillType.DEBUFF);
		_dot = _thisActor.getTemplate().getSkillsByType(SkillType.DOT);
		_poison = _thisActor.getTemplate().getSkillsByType(SkillType.POISON);
		_bleed = _thisActor.getTemplate().getSkillsByType(SkillType.BLEED);
		_mdot = _thisActor.getTemplate().getSkillsByType(SkillType.MDOT);
		_manadam = _thisActor.getTemplate().getSkillsByType(SkillType.MANADAM);
		_mdam = _thisActor.getTemplate().getSkillsByType(SkillType.MDAM);
		_pdam = _thisActor.getTemplate().getSkillsByType(SkillType.PDAM);
		_drain = _thisActor.getTemplate().getSkillsByType(SkillType.DRAIN);
		_drainsoul = _thisActor.getTemplate().getSkillsByType(SkillType.DRAIN_SOUL);
		_healpercent = _thisActor.getTemplate().getSkillsByType(SkillType.HEAL_PERCENT);
		_hot = _thisActor.getTemplate().getSkillsByType(SkillType.HOT);

		_dot_skills = msum(_dot, _poison);
		_dot_skills = msum(_dot_skills, _bleed);
		_dot_skills = msum(_dot_skills, _mdot);

		_buff_skills = msum(_heal, _buff);
		_buff_skills = msum(_buff_skills, _healpercent);

		_dam_skills = msum(_manadam, _mdam);
		_dam_skills = msum(_dam_skills, _pdam);
		_dam_skills = msum(_dam_skills, _drain);
		_dam_skills = msum(_dam_skills, _drainsoul);

		_debuff_skills = msum(_teleport, _mute);
		_debuff_skills = msum(_debuff_skills, _paralyze);
		_debuff_skills = msum(_debuff_skills, _root);
		_debuff_skills = msum(_debuff_skills, _sleep);
		_debuff_skills = msum(_debuff_skills, _cancel);
		_debuff_skills = msum(_debuff_skills, _debuff);
	}

	public void run()
	{
		onEvtThink();
	}

	@Override
	public void startAITask()
	{
		if(_aiTask == null)
		{
			_aiTask = ThreadPoolManager.getInstance().scheduleAiAtFixedRate(this, Rnd.get(AI_TASK_DELAY), AI_TASK_DELAY);
			setIntention(CtrlIntention.AI_INTENTION_ACTIVE);
		}
	}

	@Override
	public void stopAITask()
	{
		if(_aiTask != null)
		{
			setIntention(CtrlIntention.AI_INTENTION_IDLE);
			_aiTask.cancel(false);
			_aiTask = null;
		}
	}

	/**
	 * Определяет, может ли этот тип АИ видеть персонажей в режиме Silent Move.
	 * @param target L2Playable цель
	 * @return true если цель не видна в режиме Silent Move
	 */
	protected boolean isSilentMoveNotVisible(L2Playable target)
	{
		return target.isSilentMoving();
	}

	protected boolean autoAttackCondition(L2Playable target)
	{
		if(!_aggressive)
			return false;
		if(target.getNonAggroTime() > System.currentTimeMillis())
			return false;
		if(target.isAlikeDead())
			return false;
		if(!_thisActor.isInRange(target, _aggroRange))
			return false;
		if(Math.abs(target.getZ() - _actor.getZ()) > 400)
			return false;
		if(target.isInvul() || !target.isVisible() || isSilentMoveNotVisible(target))
			return false;
		if(_thisActor.getFactionId().equalsIgnoreCase("varka_silenos_clan") && target.getOwner().getVarka() > 0)
			return false;
		if(_thisActor.getFactionId().equalsIgnoreCase("ketra_orc_clan") && target.getOwner().getKetra() > 0)
			return false;
		if(_thisActor.isInZonePeace())
			return false;
		if(!GeoData.getInstance().canSeeTarget(_actor, target))
			return false;
		return _aggressive;
	}

	@Override
	protected void onEvtForgetObject(L2Object object)
	{
		if(object instanceof L2Character)
		{
			_thisActor.removeFromAggrolist((L2Character) object);
			if(_thisActor.noTarget())
				setIntention(AI_INTENTION_ACTIVE);
		}
	}

	protected L2Character findTarget()
	{
		if(_thisActor.isConfused() && _attack_target != null)
		{
			if(_thisActor.getHating(_attack_target) == 0)
				_thisActor.addDamageHate(_attack_target, 0, 1);
			return _attack_target;
		}

		// Add all autoAttackable L2Character in L2NpcInstance Aggro Range to its _aggroList with 0 damage and 1 hate
		// A L2NpcInstance isn't aggressive during 10s after its spawn because _globalAggro is set to -10
		if(_globalAggro >= 0)
		{
			if(_aggroRange > 0)
			{
				L2WorldRegion region = L2World.getRegion(_actor);
				if(region != null)
				{
					Collection<L2Playable> playables = region.getPlayables().values();
					if(playables.size() > 0)
					{
						// Чтобы не дергать их много раз в цикле. Перепроверяются потому что есть скиллы убирающие агрессивность.
						_aggroRange = _thisActor.getAggroRange();
						_aggressive = _thisActor.isAggressive();
						for(L2Playable target : playables)
						{
							if(target == null || target.getOwner() == null)
								continue;

							if(_thisActor instanceof L2FestivalMonsterInstance && !target.getOwner().isFestivalParticipant())
								continue;

							if(autoAttackCondition(target) && _thisActor.getHating(target) == 0)
								_thisActor.addDamageHate(target, 0, 1);
						}
					}
				}
			}

			L2Character hated = _thisActor.getMostHated();
			if(hated != null)
				return hated;
		}

		return null;
	}

	protected boolean randomAnimation()
	{
		if(!_actor.isMoving && !_actor.isFollow)
			// Анимации мобов
			if(_actor.hasRandomAnimation() && Rnd.chance(Config.RND_ANIMATION_RATE))
			{
				_randomAnimationEnd = System.currentTimeMillis();
				_thisActor.onRandomAnimation();
				return true;
			}
		return false;
	}

	protected boolean randomWalk()
	{
		if(!_actor.isMoving && !_actor.isFollow)
			if(maybeMoveToHome())
				return true;
		return false;
	}

	protected boolean thinkActive()
	{
		if(_thisActor.isDead())
			return true;

		// Update every 1s the _globalAggro counter to come close to 0
		if(_globalAggro < 0)
			_globalAggro++;
		else if(_globalAggro > 0)
			_globalAggro--;

		if(_randomAnimationEnd - System.currentTimeMillis() > 0)
			return true;

		L2Character target = findTarget();
		if(target != null)
		{
			_thisActor.setRunning();
			setIntention(CtrlIntention.AI_INTENTION_ATTACK, target);
			return true;
		}

		if(_def_think)
		{
			doTask();
			return true;
		}

		// If this is a festival monster or chest, then it remains in the same location
		if(!_thisActor.isMonster || _thisActor instanceof L2FestivalMonsterInstance || _thisActor instanceof L2ChestInstance || _thisActor instanceof L2TamedBeastInstance)
			return false;

		if(randomAnimation())
			return true;

		if(randomWalk())
			return true;

		return false;
	}

	private void thinkAttack()
	{
		L2Character _temp_attack_target = _attack_target;

		// Проверки на возможные подвисания
		if(_def_think)
			if(_task_list.size() <= _def_step)
				_def_think = false;

		if(!(_def_think && _task_list.get(_def_step).type == TaskType.MOVE))
		{
			// Если не атаковали в течении _attack_timeout - перейти на шаг, выставить _attack_timeout/4
			if(_thisActor.getAttackTimeout() < GameTimeController.getGameTicks() && _thisActor.isRunning())
			{
				_thisActor.setWalking();
				_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT / 4 + GameTimeController.getGameTicks());
				return;
			}

			// Если уже перешел на шаг и снова таймаут - прекратить атаку
			if(_temp_attack_target == null || _temp_attack_target.isAlikeDead() || _thisActor.getAttackTimeout() < GameTimeController.getGameTicks() || !_thisActor.isInRange(_temp_attack_target, 4000))
			{
				if(_temp_attack_target != null)
					_thisActor.removeFromAggrolist(_temp_attack_target);
				_thisActor.setAttackTimeout(Integer.MAX_VALUE);
				_temp_attack_target = null;
				clientStopMoving(null);
				setIntention(AI_INTENTION_ACTIVE);
				return;
			}
		}

		// back raidbosses to spawned point if needed
		if(_thisActor instanceof L2RaidBossInstance && _thisActor.getSpawnedX() != 0 && !_thisActor.isInRange(_thisActor.getSpawnedX(), _thisActor.getSpawnedY(), Config.MAX_PURSUE_RANGE))
		{
			teleportHome();
			return;
		}

		if(doTask())
			createNewTask();
	}

	protected boolean doTask()
	{
		L2Character _temp_attack_target = _attack_target;

		if(_task_list.size() <= _def_step)
			_def_think = false;

		if(!_def_think)
			return true;

		// Установка цели, указанной в задании
		if(_task_list.get(_def_step).type == TaskType.ATTACK || _task_list.get(_def_step).type == TaskType.CAST)
		{
			_temp_attack_target = _task_list.get(_def_step).target;
			if(_temp_attack_target == _thisActor)
				return true; // Пересоздать задание
			_attack_target = _temp_attack_target;
		}

		// Для атаки "на ходу"
		int speeddiff = Math.max(0, _actor.isBehindTarget(_temp_attack_target) ? (int) (_thisActor.getMoveSpeed() - _temp_attack_target.getMoveSpeed()) : 0) + 20;

		switch(_task_list.get(_def_step).type)
		{
			// Задание "прибежать в заданные координаты"
			case MOVE:
			{
				if(!_thisActor.isInRange(_task_list.get(_def_step).x, _task_list.get(_def_step).y, 50))
				{
					if(_actor.isMovementDisabled())
					{
						_def_think = false;
						return true;
					}

					if(!_thisActor.moveToLocation(_task_list.get(_def_step).x, _task_list.get(_def_step).y, _task_list.get(_def_step).z, 0, true))
					{
						_thisActor.broadcastPacketToOthers(new MagicSkillUse(_thisActor, _thisActor, 2036, 1, 500, 600000));
						ThreadPoolManager.getInstance().scheduleAi(new Teleport(new Location(_task_list.get(_def_step).x, _task_list.get(_def_step).y, _task_list.get(_def_step).z)), 500);
						return false;
					}
				}
				else
				{
					_def_step += 1;
					// Если заданий больше нет - определить новое
					if(_def_step >= _task_list.size())
					{
						_def_think = false;
						return true;
					}
				}
			}
				break;
			// Задание "добежать - ударить"
			case ATTACK:
			{
				// Преследовать до тех пор пока не приблизиться на расстояние атаки, потом атаковать
				if(_thisActor.getRealDistance(_temp_attack_target) > _thisActor.getPhysicalAttackRange() + speeddiff || !GeoData.getInstance().canSeeTarget(_actor, _temp_attack_target))
				{
					if(_thisActor.isMovementDisabled())
					{
						_def_think = false;
						return true;
					}

					if(!_isMobile)
					{
						_def_think = false;
						return true;
					}
					
					if(!GeoData.getInstance().canSeeTarget(_actor, _temp_attack_target))
					{
						if(!_thisActor.moveToLocation(_temp_attack_target.getX(), _temp_attack_target.getY(), _temp_attack_target.getZ(), 0, true))
							_pathfind_fails++;
					}
					else
					{
						if(!_thisActor.moveToLocation(_temp_attack_target.getX(), _temp_attack_target.getY(), _temp_attack_target.getZ(), _thisActor.getPhysicalAttackRange(), true))
							_pathfind_fails++;
					}
					
					if(_pathfind_fails >= MAX_PATHFIND_FAILS && GameTimeController.getGameTicks() - (_thisActor.getAttackTimeout() - MAX_ATTACK_TIMEOUT) < TELEPORT_TIMEOUT && _thisActor.isInRange(_temp_attack_target, 2000))
					{
						_pathfind_fails = 0;
						_thisActor.broadcastPacketToOthers(new MagicSkillUse(_thisActor, _thisActor, 2036, 1, 500, 600000));
						ThreadPoolManager.getInstance().scheduleAi(new Teleport(GeoData.getInstance().moveCheckForAI(_temp_attack_target, _thisActor)), 500);
						return false;
					}

					// Периодически прекращать преследовать и заного оценивать обстановку
					if(Rnd.chance(10))
					{
						_def_think = false;
						return true;
					}
				}
				else
				{
					_pathfind_fails = 0;

					// Обновление таймаута
					_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT + GameTimeController.getGameTicks());

					clientStopMoving(null);

					// Атаковать
					_thisActor.doAttack(_temp_attack_target);

					// Следующее задание
					_def_step += 1;

					// Если заданий больше нет - определить новое
					if(_def_step >= _task_list.size())
					{
						_def_think = false;
						return true;
					}
				}
			}
				break;
			// Задание "добежать - атаковать скиллом"
			case CAST:
			{
				// Преследовать до тех пор пока не приблизиться на расстояние атаки, потом атаковать
				if(_thisActor.getRealDistance(_temp_attack_target) > _task_list.get(_def_step).skill.getCastRange() + speeddiff || !GeoData.getInstance().canSeeTarget(_actor, _temp_attack_target))
				{
					if(_thisActor.isMovementDisabled())
					{
						_def_think = false;
						return true;
					}

					if(!_isMobile)
					{
						_def_think = false;
						return true;
					}
					
					if(!GeoData.getInstance().canSeeTarget(_actor, _temp_attack_target))
					{
						if(!_thisActor.moveToLocation(_temp_attack_target.getX(), _temp_attack_target.getY(), _temp_attack_target.getZ(), 0, true))
							_pathfind_fails++;
					}
					else
					{
						if(!_thisActor.moveToLocation(_temp_attack_target.getX(), _temp_attack_target.getY(), _temp_attack_target.getZ(), _task_list.get(_def_step).skill.getCastRange(), true))
							_pathfind_fails++;
					}

					if(_pathfind_fails >= MAX_PATHFIND_FAILS && GameTimeController.getGameTicks() - (_thisActor.getAttackTimeout() - MAX_ATTACK_TIMEOUT) < TELEPORT_TIMEOUT && _thisActor.isInRange(_temp_attack_target, 2000))
					{
						_pathfind_fails = 0;
						_thisActor.broadcastPacketToOthers(new MagicSkillUse(_thisActor, _thisActor, 2036, 1, 500, 600000));
						ThreadPoolManager.getInstance().scheduleAi(new Teleport(GeoData.getInstance().moveCheckForAI(_temp_attack_target, _thisActor)), 500);
						return false;
					}

					// Периодически прекращать преследовать и заного оценивать обстановку
					if(Rnd.chance(10))
					{
						_def_think = false;
						return true;
					}
				}
				else
				{
					_pathfind_fails = 0;

					// Обновление таймаута
					_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT + GameTimeController.getGameTicks());

					clientStopMoving(null);

					// Использовать скилл на цель
					_thisActor.doCast(_task_list.get(_def_step).skill, _temp_attack_target, true);

					// Следующее задание
					_def_step += 1;

					// Если заданий больше нет - определить новое
					if(_def_step >= _task_list.size())
					{
						_def_think = false;
						return true;
					}
				}
			}
				break;
			// Задание "добежать - применить скилл"
			case BUFF:
			{
				L2Character target = _task_list.get(_def_step).target;
				// Преследовать до тех пор пока не приблизиться на расстояние атаки, потом атаковать
				if(_thisActor.getRealDistance(target) > _task_list.get(_def_step).skill.getCastRange() + speeddiff || !GeoData.getInstance().canSeeTarget(_actor, target))
				{
					if(_thisActor.isMovementDisabled())
					{
						_def_think = false;
						return true;
					}

					if(!_isMobile)
					{
						_def_think = false;
						return true;
					}
					
					if(!GeoData.getInstance().canSeeTarget(_actor, _temp_attack_target))
					{
						if(!_thisActor.moveToLocation(_temp_attack_target.getX(), _temp_attack_target.getY(), _temp_attack_target.getZ(), 0, true))
							_pathfind_fails++;
					}
					else
					{
						if(!_thisActor.moveToLocation(_temp_attack_target.getX(), _temp_attack_target.getY(), _temp_attack_target.getZ(), _task_list.get(_def_step).skill.getCastRange(), true))
							_pathfind_fails++;
					}

					if(_pathfind_fails >= MAX_PATHFIND_FAILS && GameTimeController.getGameTicks() - (_thisActor.getAttackTimeout() - MAX_ATTACK_TIMEOUT) < TELEPORT_TIMEOUT && _thisActor.isInRange(target, 2000))
					{
						_pathfind_fails = 0;
						_thisActor.broadcastPacketToOthers(new MagicSkillUse(_thisActor, _thisActor, 2036, 1, 500, 600000));
						ThreadPoolManager.getInstance().scheduleAi(new Teleport(GeoData.getInstance().moveCheckForAI(target, _thisActor)), 500);
						return false;
					}
				}
				else
				{
					_pathfind_fails = 0;

					clientStopMoving(null);

					// Использовать скилл на цель
					_thisActor.doCast(_task_list.get(_def_step).skill, target, true);

					// Следующее задание
					_def_step += 1;

					// Если заданий больше нет - определить новое
					if(_def_step >= _task_list.size())
					{
						_def_think = false;
						return true;
					}
				}
			}
				break;
		}

		return false;
	}

	protected boolean createNewTask()
	{
		return false;
	}

	@Override
	protected void onIntentionAttack(L2Character target)
	{
		// Удаляем все задания
		clearTasks();

		_attack_target = target;
		changeIntention(AI_INTENTION_ATTACK, target, null);
		onEvtThink();
	}

	@Override
	protected void onEvtThink()
	{
		if(_thinking || _thisActor.isActionsDisabled() || _thisActor.isAfraid() || _thisActor.isDead())
			return;

		if(_randomAnimationEnd - System.currentTimeMillis() > 0)
			return;

		_thinking = true;
		try
		{
			if(_intention == AI_INTENTION_ACTIVE || _intention == AI_INTENTION_IDLE)
				thinkActive();
			else if(_intention == AI_INTENTION_ATTACK)
				thinkAttack();
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		finally
		{
			_thinking = false;
		}
	}

	@Override
	protected void onEvtDead()
	{
		stopAITask();

		_thisActor.clearAggroList();

		// 10 seconds timeout of ATTACK after respawn
		_globalAggro = -10;

		_thisActor.setAttackTimeout(Integer.MAX_VALUE);

		// Удаляем все задания
		clearTasks();

		super.onEvtDead();
	}

	private long lastFactionNotifyTime = 0;

	@Override
	protected void onEvtAttacked(L2Character attacker, Boolean canBeAttackedByMob)
	{
		if((canBeAttackedByMob == null || !canBeAttackedByMob) && !(attacker instanceof L2Playable))
			return;

		// call friend's
		if(System.currentTimeMillis() - lastFactionNotifyTime > 10000)
		{
			ThreadPoolManager.getInstance().scheduleAi(new notifyFaction(attacker), 100);
			lastFactionNotifyTime = System.currentTimeMillis();
		}

		_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT + GameTimeController.getGameTicks());

		// Set the _globalAggro to 0 to permit attack even just after spawn
		if(_globalAggro < 0)
			_globalAggro = 0;

		// 1 хейт обавляется хозяину суммона, чтобы после смерти суммона моб накинулся на хозяина.
		if(attacker instanceof L2Summon)
			_thisActor.addDamageHate(attacker.getOwner(), 0, 1);

		// Add the attacker to the _aggroList of the actor
		_thisActor.addDamageHate(attacker, 0, 1, canBeAttackedByMob);
		_thisActor.setRunning();

		if(_intention != AI_INTENTION_ATTACK)
			setIntention(CtrlIntention.AI_INTENTION_ATTACK, attacker);
	}

	@Override
	protected void onEvtAggression(L2Character target, int aggro)
	{
		if(!(target instanceof L2Playable))
			return;

		if(_actor instanceof L2TamedBeastInstance && target.getPlayer().getTrainedBeast() != null && target.getPlayer().getTrainedBeast().equals(_actor))
			return;

		// Для предотвращения "вытаскивания" лидера хейтом.
		if(_thisActor.isMonster && _thisActor.hasMinions() && aggro > 0)
			for(L2MinionInstance minion : _thisActor.getSpawnedMinions())
			{
				minion.addDamageHate(target, 0, Rnd.get(aggro));
				minion.setRunning();
				minion.getAI().setIntention(CtrlIntention.AI_INTENTION_ATTACK, target);
			}

		_thisActor.setAttackTimeout(MAX_ATTACK_TIMEOUT + GameTimeController.getGameTicks());

		// Set the _globalAggro to 0 to permit attack even just after spawn
		if(_globalAggro < 0)
			_globalAggro = 0;

		// 1 хейт обавляется хозяину суммона, чтобы после смерти суммона моб накинулся на хозяина.
		if(target.isSummon)
			_thisActor.addDamageHate(target.getOwner(), 0, 1);

		_thisActor.addDamageHate(target, 0, aggro);
		_thisActor.setRunning();

		if(_intention != CtrlIntention.AI_INTENTION_ATTACK)
			setIntention(CtrlIntention.AI_INTENTION_ATTACK, target);
	}

	public class notifyFaction implements Runnable
	{
		private final L2Character attacker;

		notifyFaction(L2Character attacker)
		{
			this.attacker = attacker;
		}

		public void run()
		{
			if(attacker == null)
				return;
			try
			{
				String faction_id = _thisActor.getFactionId();
				if(faction_id == null)
					return;
				for(L2NpcInstance npc : ActiveFriendTargets(true, false))
				{
					if(!Rnd.chance(Config.ONE_FACTION_CALL_CHANCE))
						continue;
					if(npc == null || !faction_id.equalsIgnoreCase(npc.getFactionId()))
						continue;
					if(Math.abs(attacker.getZ() - npc.getZ()) > 400)
						continue;
					if(!_thisActor.isInRange(npc, npc.getFactionRange()))
						continue;
					if(!GeoData.getInstance().canSeeTarget(_thisActor, npc))
						continue;
					npc.getAI().notifyEvent(CtrlEvent.EVT_AGGRESSION, attacker, Rnd.get(1, 100));
				}
			}
			catch(Throwable t)
			{
				t.printStackTrace();
			}
		}
	}

	protected boolean maybeMoveToHome()
	{
		boolean randomWalk = _actor.hasRandomWalk();
		if(_thisActor.getSpawnedX() == 0)
			return false;

		// Random walk or not?
		if(randomWalk && (!Config.RND_WALK || !Rnd.chance(Config.RND_WALK_RATE)))
			return false;

		int spawnX = _thisActor.getSpawnedX();
		int spawnY = _thisActor.getSpawnedY();
		int spawnZ = _thisActor.getSpawnedZ();

		if(!randomWalk && _thisActor.isInRangeZ(spawnX, spawnY, spawnZ, Config.MAX_DRIFT_RANGE))
			return false;

		int x = spawnX + Rnd.get(2 * Config.MAX_DRIFT_RANGE) - Config.MAX_DRIFT_RANGE;
		int y = spawnY + Rnd.get(2 * Config.MAX_DRIFT_RANGE) - Config.MAX_DRIFT_RANGE;
		int z = GeoData.getInstance().getHeight(x, y, spawnZ);

		L2Spawn spawn = _thisActor.getSpawn();
		if(spawn != null && spawn.getLocation() != 0 && !TerritoryTable.getInstance().getLocation(spawn.getLocation()).isInside(x, y))
			return false;

		_thisActor.setWalking();
		_thisActor.moveToLocation(x, y, z, 0, true);

		return true;
	}

	public void returnHome()
	{
		_thisActor.clearAggroList();
		setIntention(AI_INTENTION_ACTIVE);

		// Удаляем все задания
		clearTasks();

		// Координаты точки спауна
		int x = _thisActor.getSpawnedX();
		int y = _thisActor.getSpawnedY();
		int z = _thisActor.getSpawnedZ();

		// Прибежать в заданную точку и переключиться в состояние AI_INTENTION_ACTIVE

		Task task = new Task();
		task.type = TaskType.MOVE;
		task.x = x;
		task.y = y;
		task.z = z;
		_task_list.add(task);
		_def_think = true;
	}

	@Override
	public void teleportHome()
	{
		if(_thisActor.getSpawn() != null && _thisActor.getSpawnedX() != 0)
		{
			_thisActor.broadcastPacketToOthers(new MagicSkillUse(_thisActor, _thisActor, 2036, 1, 500, 0));
			_thisActor.teleToLocation(_thisActor.getSpawnedX(), _thisActor.getSpawnedY(), GeoData.getInstance().getHeight(_thisActor.getSpawnedX(), _thisActor.getSpawnedY(), _thisActor.getSpawnedZ()));
			setIntention(AI_INTENTION_ACTIVE);
		}
	}

	@Override
	public boolean isActive()
	{
		return _aiTask != null;
	}
	
	public void clearTasks()
	{
		_def_step = 0;
		_def_think = false;
		_task_list.clear();
	}
}