package L2jFT.game.instancemanager;

import java.awt.Polygon;
import java.awt.Shape;
import java.sql.ResultSet;
import java.util.logging.Logger;

import javolution.util.FastMap;
import L2jFT.Config;
import L2jFT.database.DatabaseUtils;
import L2jFT.database.FiltredPreparedStatement;
import L2jFT.database.L2DatabaseFactory;
import L2jFT.database.ThreadConnection;
import L2jFT.util.Rnd;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.entity.DimensionalRift;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.model.instances.L2NpcInstance;
import L2jFT.game.network.serverpackets.NpcHtmlMessage;

public class DimensionalRiftManager
{

	private static Logger _log = Logger.getLogger(DimensionalRiftManager.class.getName());
	private static DimensionalRiftManager _instance;
	private FastMap<Byte, FastMap<Byte, DimensionalRiftRoom>> _rooms = new FastMap<Byte, FastMap<Byte, DimensionalRiftRoom>>();
	private final short DIMENSIONAL_FRAGMENT_ITEM_ID = 7079;

	public static DimensionalRiftManager getInstance()
	{
		if(_instance == null)
			_instance = new DimensionalRiftManager();

		return _instance;
	}

	public DimensionalRiftManager()
	{
		load();
	}

	public DimensionalRiftRoom getRoom(byte _type, byte room)
	{
		return _rooms.get(_type).get(room);
	}

	private void load()
	{
		ThreadConnection con = null;
		FiltredPreparedStatement s = null;
		ResultSet rs = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			s = con.prepareStatement("SELECT * FROM dimensional_rift WHERE 1 < 2");
			rs = s.executeQuery();

			while(rs.next())
			{
				// 0 waiting room, 1 recruit, 2 soldier, 3 officer, 4 captain , 5 commander, 6 hero
				byte type = rs.getByte("type");
				byte room_id = rs.getByte("room_id");

				//coords related
				int xMin = rs.getInt("xMin");
				int xMax = rs.getInt("xMax");
				int yMin = rs.getInt("yMin");
				int yMax = rs.getInt("yMax");
				int z1 = rs.getInt("zMin");
				int z2 = rs.getInt("zMax");
				int xT = rs.getInt("xT");
				int yT = rs.getInt("yT");
				int zT = rs.getInt("zT");
				boolean isBossRoom = rs.getByte("boss") > 0;

				if(!_rooms.containsKey(type))
					_rooms.put(type, new FastMap<Byte, DimensionalRiftRoom>());

				_rooms.get(type).put(room_id, new DimensionalRiftRoom(xMin, xMax, yMin, yMax, z1, z2, xT, yT, zT, isBossRoom));
			}

		}
		catch(Exception e)
		{
			_log.warning("Can't load Dimension Rift zones.\n" + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, s, rs);
		}

		int typeSize = _rooms.keySet().size();
		int roomSize = 0;

		for(Byte b : _rooms.keySet())
			roomSize += _rooms.get(b).keySet().size();

		_log.info("DimensionalRiftManager: Loaded " + typeSize + " room types with " + roomSize + " rooms.");
	}

	public void reload()
	{
		for(Byte b : _rooms.keySet())
			_rooms.get(b).clear();

		_rooms.clear();
		load();
	}

	public boolean checkIfInRiftZone(int x, int y, int z, boolean ignorePeaceZone)
	{
		if(ignorePeaceZone)
			return _rooms.get((byte) 0).get((byte) 1).checkIfInZone(x, y, z);
		return _rooms.get((byte) 0).get((byte) 1).checkIfInZone(x, y, z) && !_rooms.get((byte) 0).get((byte) 0).checkIfInZone(x, y, z);
	}

	public boolean checkIfInPeaceZone(int x, int y, int z)
	{
		return _rooms.get((byte) 0).get((byte) 0).checkIfInZone(x, y, z);
	}

	public void teleportToWaitingRoom(L2Player player)
	{
		int[] coords = getRoom((byte) 0, (byte) 0).getTeleportCoords();
		player.teleToLocation(coords[0], coords[1], coords[2]);
	}

	public void start(L2Player player, byte type, L2NpcInstance npc)
	{
		boolean canPass = true;
		if(!player.isInParty())
		{
			showHtmlFile(player, "data/html/rift/NoParty.htm", npc);
			return;
		}

		if(player.getParty().getPartyLeaderOID() != player.getObjectId())
		{
			showHtmlFile(player, "data/html/rift/NotPartyLeader.htm", npc);
			return;
		}

		if(player.getParty().isInDimensionalRift())
		{
			showHtmlFile(player, "data/html/rift/Cheater.htm", npc);

			if(!player.isGM())
				_log.warning("Player " + player.getName() + "(" + player.getObjectId() + ") was cheating in dimension rift area!");

			return;
		}

		if(player.getParty().getMemberCount() < Config.RIFT_MIN_PARTY_SIZE)
		{
			showHtmlFile(player, "data/html/rift/SmallParty.htm", npc);
			return;
		}

		for(L2Player p : player.getParty().getPartyMembers())
			if(!checkIfInPeaceZone(p.getX(), p.getY(), p.getZ()))
				canPass = false;

		if(!canPass)
		{
			showHtmlFile(player, "data/html/rift/NotInWaitingRoom.htm", npc);
			return;
		}

		L2ItemInstance i;
		for(L2Player p : player.getParty().getPartyMembers())
		{
			i = p.getInventory().getItemByItemId(DIMENSIONAL_FRAGMENT_ITEM_ID);

			if(i == null)
			{
				canPass = false;
				break;
			}

			if(i.getCount() > 0)
				if(i.getCount() <= getNeededItems(type))
					canPass = false;
		}

		if(!canPass)
		{
			showHtmlFile(player, "data/html/rift/NoFragments.htm", npc);
			return;
		}

		for(L2Player p : player.getParty().getPartyMembers())
		{
			i = p.getInventory().getItemByItemId(DIMENSIONAL_FRAGMENT_ITEM_ID);
			p.getInventory().destroyItem(i.getObjectId(), getNeededItems(type), true, true);
		}

		new DimensionalRift(player.getParty(), type, (byte) Rnd.get(1, 9));
	}

	public void killRift(DimensionalRift d)
	{
		d.getTimerTask().cancel();
		d.setTimerTask(null);

		d.getTimer().cancel();
		d.setTimer(null);
	}

	public class DimensionalRiftRoom
	{
		private final int _zMin;
		private final int _zMax;
		private final int[] _teleportCoords;
		private final Shape _s;
		private final boolean _isBossRoom;

		public DimensionalRiftRoom(int xMin, int xMax, int yMin, int yMax, int zMin, int zMax, int xT, int yT, int zT, boolean isBossRoom)
		{
			_zMin = zMin;
			_zMax = zMax;
			_teleportCoords = new int[] { xT, yT, zT };
			_isBossRoom = isBossRoom;
			_s = new Polygon(new int[] { xMin, xMax, xMax, xMin }, new int[] {
					yMin, yMin, yMax, yMax }, 4);
		}

		public int[] getTeleportCoords()
		{
			return _teleportCoords;
		}

		public boolean checkIfInZone(int x, int y, int z)
		{
			return _s.contains(x, y) && z >= _zMin && z <= _zMax;
		}

		public boolean isBossRoom()
		{
			return _isBossRoom;
		}
	}

	private int getNeededItems(byte type)
	{
		switch(type)
		{
			case 1:
				return Config.RIFT_ENTER_COST_RECRUIT;
			case 2:
				return Config.RIFT_ENTER_COST_SOLDIER;
			case 3:
				return Config.RIFT_ENTER_COST_OFFICER;
			case 4:
				return Config.RIFT_ENTER_COST_CAPTAIN;
			case 5:
				return Config.RIFT_ENTER_COST_COMMANDER;
			case 6:
				return Config.RIFT_ENTER_COST_HERO;
			default:
				return 999999;
		}
	}

	public void showHtmlFile(L2Player player, String file, L2NpcInstance npc)
	{
		NpcHtmlMessage html = new NpcHtmlMessage(npc.getObjectId());
		html.setFile(file);
		html.replace("%t_name%", npc.getName());
		player.sendPacket(html);
	}
}