package L2jFT.game.model;

import L2jFT.Config;
import L2jFT.util.CharPosition;
import L2jFT.util.Location;
import L2jFT.util.Rnd;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.model.instances.L2NpcInstance;
import L2jFT.game.network.serverpackets.Earthquake;
import L2jFT.game.network.serverpackets.ExRedSky;
import L2jFT.game.network.serverpackets.InventoryUpdate;
import L2jFT.game.network.serverpackets.Ride;
import L2jFT.game.network.serverpackets.SocialAction;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.network.serverpackets.ValidateLocation;
import L2jFT.game.tables.ItemTable;
import L2jFT.game.tables.SkillTable;

public class CursedWeapon
{
	private final String _name;
	private final int _itemId;
	private final short _skillId;
	private final int _skillMaxLevel;
	private int _dropRate;
	private int _durationMin;
	private int _durationMax;
	private int _durationLost;
	private int _disapearChance;
	private int _stageKills;

	public enum CursedWeaponState
	{
		NONE,
		ACTIVATED,
		DROPPED,
	}

	private CursedWeaponState _state = CursedWeaponState.NONE;

	private int _nbKills = 0;
	private long _endTime = 0;

	private int _playerId = 0;
	private L2Player _player = null;
	private L2ItemInstance _item = null;
	private int _playerKarma = 0;
	private int _playerPkKills = 0;

	private int _x = 0;
	private int _y = 0;
	private int _z = 0;

	public CursedWeapon(int itemId, short skillId, String name)
	{
		_name = name;
		_itemId = itemId;
		_skillId = skillId;
		_skillMaxLevel = SkillTable.getInstance().getMaxLevel(_skillId, 0);
	}

	public void initWeapon()
	{
		_state = CursedWeaponState.NONE;
		_endTime = 0;
		_player = null;
		_playerId = 0;
		_playerKarma = 0;
		_playerPkKills = 0;
		_item = null;
		_nbKills = 0;
	}

	// Drop of CursedWeapon
	public boolean dropIt(L2NpcInstance attackable, L2Player killer)
	{
		boolean success = false;

		if(attackable != null)
		{
			if(Rnd.get(100000000) <= _dropRate)
			{
				_item = ItemTable.getInstance().createItem(_itemId, killer.getObjectId(), attackable.getNpcId(), "CursedWeapon.dropIt");
				if(_item != null)
				{
					_player = null;
					_playerId = 0;
					_playerKarma = 0;
					_playerPkKills = 0;
					_state = CursedWeaponState.DROPPED;

					if(_endTime == 0)
						_endTime = System.currentTimeMillis() + getRndDuration() * 60000;

					attackable.dropItem(killer, _item);
					_item.setDropTime(0);

					// check if it still dropped (keep in mind AutoLoot)
					if(isDropped())
					{
						_x = _item.getX();
						_y = _item.getY();
						_z = _item.getZ();
					}

					// RedSky and Earthquake
					ExRedSky packet = new ExRedSky(10);
					Earthquake eq = new Earthquake(killer.getX(), killer.getY(), killer.getZ(), 30, 12);
					for(L2Player aPlayer : L2World.getAllPlayers())
					{
						aPlayer.sendPacket(packet);
						aPlayer.sendPacket(eq);
					}
					success = true;
				}
			}
		}
		else if(!Rnd.chance(_disapearChance))
		{
			L2ItemInstance oldItem = _player.getInventory().findItemByItemId(_itemId);
			if(oldItem == null)
				return false;

			int oldCount = oldItem.getCount();
			_player.sendPacket(new ValidateLocation(_player));

			L2ItemInstance dropedItem = _player.getInventory().dropItem(oldItem, oldCount);
			if(dropedItem == null)
				return false;

			_player.setKarma(_playerKarma);
			_player.setPkKills(_playerPkKills);
			_player.setCursedWeaponEquippedId(0);
			_player.removeSkill(SkillTable.getInstance().getInfo(_skillId, _player.getSkillLevel(_skillId)), false);
			_player.abortAttack();

			_playerId = 0;
			_playerKarma = 0;
			_playerPkKills = 0;
			_state = CursedWeaponState.DROPPED;

			final CharPosition dropPos = Rnd.coordsRandomize(_player, 70);
			_x = dropPos.x;
			_y = dropPos.y;
			_z = _player.getZ() + 25;

			dropedItem.dropMe(_player, _x, _y, _z + 25);
			dropedItem.setDropTime(0);
			_item = dropedItem;

			InventoryUpdate iu = new InventoryUpdate();
			iu.addRemovedItem(oldItem);

			_player.sendPacket(iu);
			_player.sendPacket(new SystemMessage(SystemMessage.YOU_HAVE_DROPPED_S1).addItemName(dropedItem.getItemId()));

			_player.refreshExpertisePenalty();
			_player.broadcastUserInfo();

			Earthquake eq = new Earthquake(_player.getX(), _player.getY(), _player.getZ(), 30, 12);
			_player.broadcastPacket(eq);

			success = true;
		}
		return success;
	}

	private void giveSkill()
	{
		int level = 1 + _nbKills / _stageKills;
		if(level > _skillMaxLevel)
			level = _skillMaxLevel;

		L2Skill skill = SkillTable.getInstance().getInfo(_skillId, level);
		_player.addSkill(skill, false);

		if(Config.DEBUG)
			System.out.println(_player + " has been awarded with skill " + skill + " with activeClass " + _player.getActiveClass());
	}

	// вызывается при загрузке оружия
	public boolean reActivate()
	{
		if(getTimeLeft() <= 0)
		{
			if(_playerId != 0)
				// to be sure, that cursed weapon will deleted in right way
				_state = CursedWeaponState.ACTIVATED;
			return false;
		}
		else if(_playerId == 0)
		{
			if(_x == 0 && _y == 0 && _z == 0)
				return false;

			_item = ItemTable.getInstance().createItem(_itemId, _itemId, _itemId, "CursedWeapon.reActivate");
			if(_item == null)
				return false;

			_item.dropMe(null, _x, _y, _z);
			_item.setDropTime(0);

			_state = CursedWeaponState.DROPPED;
		}
		else
			_state = CursedWeaponState.ACTIVATED;
		return true;
	}

	public void activate(L2Player player, L2ItemInstance item)
	{
		// оружие уже в руках игрока или новый игрок
		if(_state != CursedWeaponState.ACTIVATED || _playerId != player.getObjectId())
		{
			_playerKarma = player.getKarma();
			_playerPkKills = player.getPkKills();
		}

		_state = CursedWeaponState.ACTIVATED;
		_player = player;
		_playerId = player.getObjectId();

		player.setCursedWeaponEquippedId(_itemId);
		player.setKarma(9999999);
		player.setPkKills(_nbKills);

		if(player.isInParty())
			player.getParty().oustPartyMember(player);

		if(player.isMounted())
		{
			player.broadcastPacket(new Ride(player, Ride.ACTION_DISMOUNT, 0));
			player.setMountType((short) 0);
		}

		if(_endTime == 0)
			_endTime = System.currentTimeMillis() + getRndDuration() * 60000;

		giveSkill();

		_item = item;
		player.getInventory().equipItemAndRecord(_item);
		player.sendPacket(new SystemMessage(SystemMessage.YOU_ARE_EQUIPPED_WITH_S1).addItemName(_item.getItemId()));

		player.setCurrentHpMp(player.getMaxHp(), player.getMaxMp());
		player.setCurrentCp(player.getMaxCp());
		player.broadcastPacket(new SocialAction(player.getObjectId(), 17));

		player.broadcastUserInfo();
	}

	public void increaseKills()
	{
		_nbKills++;

		_player.setPkKills(_nbKills);
		_player.broadcastUserInfo();

		if(_nbKills % _stageKills == 0 && _nbKills <= _stageKills * (_skillMaxLevel - 1))
			giveSkill();

		// Reduce time-to-live
		_endTime -= _durationLost * 60000;
	}

	public void setDisapearChance(int disapearChance)
	{
		_disapearChance = disapearChance;
	}

	public void setDropRate(int dropRate)
	{
		_dropRate = dropRate;
	}

	public void setDurationMin(int duration)
	{
		_durationMin = duration;
	}

	public void setDurationMax(int duration)
	{
		_durationMax = duration;
	}

	public void setDurationLost(int durationLost)
	{
		_durationLost = durationLost;
	}

	public void setStageKills(int stageKills)
	{
		_stageKills = stageKills;
	}

	public void setNbKills(int nbKills)
	{
		_nbKills = nbKills;
	}

	public void setPlayerId(int playerId)
	{
		_playerId = playerId;
	}

	public void setPlayerKarma(int playerKarma)
	{
		_playerKarma = playerKarma;
	}

	public void setPlayerPkKills(int playerPkKills)
	{
		_playerPkKills = playerPkKills;
	}

	public void setState(CursedWeaponState state)
	{
		_state = state;
	}

	public void setEndTime(long endTime)
	{
		_endTime = endTime;
	}

	public void setPlayer(L2Player player)
	{
		_player = player;
	}

	public void setItem(L2ItemInstance item)
	{
		_item = item;
	}

	public void setX(int x)
	{
		_x = x;
	}

	public void setY(int y)
	{
		_y = y;
	}

	public void setZ(int z)
	{
		_z = z;
	}

	public CursedWeaponState getState()
	{
		return _state;
	}

	public boolean isActivated()
	{
		return _state == CursedWeaponState.ACTIVATED;
	}

	public boolean isDropped()
	{
		return _state == CursedWeaponState.DROPPED;
	}

	public long getEndTime()
	{
		return _endTime;
	}

	public String getName()
	{
		return _name;
	}

	public int getItemId()
	{
		return _itemId;
	}

	public L2ItemInstance getItem()
	{
		return _item;
	}

	public short getSkillId()
	{
		return _skillId;
	}

	public int getPlayerId()
	{
		return _playerId;
	}

	public L2Player getPlayer()
	{
		return _player;
	}

	public int getPlayerKarma()
	{
		return _playerKarma;
	}

	public int getPlayerPkKills()
	{
		return _playerPkKills;
	}

	public int getNbKills()
	{
		return _nbKills;
	}

	public int getStageKills()
	{
		return _stageKills;
	}

	public int getX()
	{
		return _x;
	}

	public int getY()
	{
		return _y;
	}

	public int getZ()
	{
		return _z;
	}

	public int getRndDuration()
	{
		if(_durationMin > _durationMax)
			_durationMax = 2 * _durationMin;
		return Rnd.get(_durationMin, _durationMax);
	}

	public boolean isActive()
	{
		return _state == CursedWeaponState.ACTIVATED || _state == CursedWeaponState.DROPPED;
	}

	public int getLevel()
	{
		if(_nbKills > _stageKills * _skillMaxLevel)
			return _skillMaxLevel;
		return _nbKills / _stageKills;
	}

	public long getTimeLeft()
	{
		return _endTime - System.currentTimeMillis();
	}

	public Location getWorldPosition()
	{
		if(isActivated())
		{
			if(_player != null && _player.isOnline())
				return new Location(_player.getX(), _player.getY(), _player.getZ());
		}
		else if(isDropped())
			if(_item != null)
				return new Location(_item.getX(), _item.getY(), _item.getZ());

		return null;
	}
}
