package L2jFT.game.model;

import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

import L2jFT.Config;
import L2jFT.database.DatabaseUtils;
import L2jFT.database.FiltredPreparedStatement;
import L2jFT.database.L2DatabaseFactory;
import L2jFT.database.ThreadConnection;
import L2jFT.game.idfactory.IdFactory;
import L2jFT.game.instancemanager.CastleManager;
import L2jFT.game.instancemanager.CursedWeaponsManager;
import L2jFT.game.model.entity.Hero;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.model.instances.L2ItemInstance.ItemLocation;
import L2jFT.game.model.instances.L2PetInstance;
import L2jFT.game.model.inventory.listeners.ArmorSetListener;
import L2jFT.game.model.inventory.listeners.BowListener;
import L2jFT.game.model.inventory.listeners.ChangeRecorder;
import L2jFT.game.model.inventory.listeners.CommonItemListener;
import L2jFT.game.model.inventory.listeners.FormalWearListener;
import L2jFT.game.model.inventory.listeners.ItemAugmentationListener;
import L2jFT.game.model.inventory.listeners.ItemPassiveSkillsListener;
import L2jFT.game.model.inventory.listeners.PaperdollListener;
import L2jFT.game.model.inventory.listeners.StatsListener;
import L2jFT.game.network.serverpackets.InventoryUpdate;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.tables.ItemTable;
import L2jFT.game.tables.PetDataTable;
import L2jFT.game.templates.L2EtcItem;
import L2jFT.game.templates.L2EtcItem.EtcItemType;
import L2jFT.game.templates.L2Item;
import L2jFT.util.Log;
import L2jFT.util.Util;

/**
 * This class manages inventory
 *
 * @version $Revision: 1.13.2.9.2.12 $ $Date: 2005/03/29 23:15:15 $
 */
public abstract class Inventory
{
	protected static final Logger _log = Logger.getLogger(Inventory.class.getName());

	public static final byte PAPERDOLL_UNDER = 0;
	public static final byte PAPERDOLL_LEAR = 1;
	public static final byte PAPERDOLL_REAR = 2;
	public static final byte PAPERDOLL_NECK = 3;
	public static final byte PAPERDOLL_LFINGER = 4;
	public static final byte PAPERDOLL_RFINGER = 5;
	public static final byte PAPERDOLL_HEAD = 6;
	public static final byte PAPERDOLL_RHAND = 7;
	public static final byte PAPERDOLL_LHAND = 8;
	public static final byte PAPERDOLL_GLOVES = 9;
	public static final byte PAPERDOLL_CHEST = 10;
	public static final byte PAPERDOLL_LEGS = 11;
	public static final byte PAPERDOLL_FEET = 12;
	public static final byte PAPERDOLL_BACK = 13;
	public static final byte PAPERDOLL_LRHAND = 14;
	public static final byte PAPERDOLL_FACE = 15;
	public static final byte PAPERDOLL_HAIR = 16;
	public static final byte PAPERDOLL_DHAIR = 17;
	public static final byte PAPERDOLL_CLOAK = 18;

	//Speed percentage mods
	public static final double MAX_ARMOR_WEIGHT = 12000;

	private final L2ItemInstance[] _paperdoll;
	private final CopyOnWriteArrayList<PaperdollListener> _paperdollListeners;

	// protected to be accessed from child classes only
	protected final List<L2ItemInstance> _items;
	private int _totalWeight;
	private boolean refreshingListeners;

	// used to quickly check for using of items of special type
	private int _wearedMask;

	// Castle Lord circlets, WARNING: position == castle.id !
	public static final Short[] _castleLordCirclets = { 0, // no castle - no circlet.. :)
			6838, // Circlet of Gludio
			6835, // Circlet of Dion
			6839, // Circlet of Giran
			6837, // Circlet of Oren
			6840, // Circlet of Aden
			6834, // Circlet of Innadril
			6836, // Circlet of Goddard
			8182, // Circlet of Rune
			8183, // Circlet of Schuttgart
	};

	/**
	 * Constructor of the inventory
	 */
	protected Inventory()
	{
		_paperdoll = new L2ItemInstance[0x13];
		_items = new Vector<L2ItemInstance>();
		_paperdollListeners = new CopyOnWriteArrayList<PaperdollListener>();
		addPaperdollListener(new BowListener(this));
		addPaperdollListener(new FormalWearListener(this));
		addPaperdollListener(new ArmorSetListener(this));
		addPaperdollListener(new StatsListener(this));
		addPaperdollListener(new ItemPassiveSkillsListener(this));
		addPaperdollListener(new ItemAugmentationListener(this));
		addPaperdollListener(new CommonItemListener(this));
	}

	public abstract L2Character getOwner();

	protected abstract ItemLocation getBaseLocation();

	protected abstract ItemLocation getEquipLocation();

	/**
	 * Returns the ownerID of the inventory
	 * @return int
	 */
	public int getOwnerId()
	{
		return getOwner() == null ? 0 : getOwner().getObjectId();
	}

	/**
	 * Returns the instance of new ChangeRecorder
	 * @return ChangeRecorder
	 */
	public ChangeRecorder newRecorder()
	{
		return new ChangeRecorder(this);
	}

	/**
	 * Returns the quantity of items in the inventory
	 * @return int
	 */
	public int getSize()
	{
		return _items.size();
	}

	/**
	 * Returns the list of items in inventory
	 * @return L2ItemInstance : items in inventory
	 */
	public L2ItemInstance[] getItems()
	{
		return _items.toArray(new L2ItemInstance[_items.size()]);
	}

	public List<L2ItemInstance> getItemsList()
	{
		return _items;
	}

	public L2ItemInstance addItem(int id, long count, int source, String create_type)
	{
		L2ItemInstance i = ItemTable.getInstance().createItem(id, getOwnerId(), source, create_type);
		i.setCount(count);
		return addItem(i, true, true);
	}

	/**
	 * Adds an item in inventory and update database
	 * @param newItem : L2ItemInstance
	 * @return L2ItemInstance equal to parameter newItem.
	 */
	public L2ItemInstance addItem(L2ItemInstance newItem)
	{
		return addItem(newItem, true, true);
	}

	protected L2ItemInstance addItem(L2ItemInstance newItem, boolean dbUpdate)
	{
		return addItem(newItem, dbUpdate, true);
	}

	/**
	 * Adds item in inventory and returns parameter newItem
	 * @param newItem : L2ItemInstance designating the item to add
	 * @param dbUpdate : boolean
	 * @return L2ItemInstance corresponding to the new item or the updated item in inventory
	 */
	protected L2ItemInstance addItem(L2ItemInstance newItem, boolean dbUpdate, boolean log)
	{
		if(newItem == null)
			return null;

		if(newItem.getItemId() >= 8600 && newItem.getItemId() <= 8614)
		{
			Util.handleIllegalPlayerAction(getOwner().getPlayer(), "tried to pickup herb into inventory", "Inventory[179]", 1);
			return null;
		}

		if(newItem.getCount() < 0)
		{
			_log.warning("AddItem: count < 0 owner:" + getOwner().getName());
			Thread.dumpStack();
			return null;
		}
		L2ItemInstance result = newItem;
		boolean stackableFound = false;
		if(log)
			Log.add("Inventory|" + getOwner().getName() + "|Get item|" + result.getItemId() + "|" + result.getCount() + "|" + result.getObjectId(), "items");

		// If stackable, search item in inventory in order to add to current quantity
		if(newItem.isStackable())
		{
			int itemId = newItem.getItemId();
			L2ItemInstance old = findItemByItemId(itemId);
			if(old != null)
			{
				// add new item quantity to existing stack
				old.setCount(old.getFullCount() + newItem.getFullCount());
				// reset new item to null
				if(log)
					Log.add("Inventory|" + getOwner().getName() + "|join item from-to|" + result.getItemId() + "|" + newItem.getObjectId() + "|" + old.getObjectId(), "items");
				newItem.setCount(0);
				newItem.setOwnerId(0);
				newItem.setLocation(ItemLocation.VOID);

				L2World.removeObject(newItem);
				IdFactory.getInstance().releaseId(newItem.getObjectId());
				// say stack has been found
				stackableFound = true;

				// say item in inventory has been modified
				old.setLastChange(L2ItemInstance.MODIFIED);

				// If database has to be updated (dbUpdate = true), update old item in inventory and destroy new item
				if(dbUpdate)
				{
					old.updateDatabase();
					newItem.updateDatabase();
				}
				result = old;
			}
		}

		// If item hasn't be found in inventory
		if(!stackableFound)
		{
			// Add item in inventory
			synchronized (_items)
			{
				if(getItemByObjectId(newItem.getObjectId()) == null)
					_items.add(newItem);
				else if(log)
					Log.add("Inventory|" + getOwner().getName() + "|add double link to item in inventory list!|" + newItem.getItemId() + "|" + newItem.getObjectId(), "items");
			}

			if(newItem.getOwnerId() != getOwnerId() || dbUpdate)
			{
				newItem.setOwnerId(getOwnerId());
				newItem.setLocation(getBaseLocation());
				newItem.setLastChange(L2ItemInstance.ADDED);
			}
			// If database wanted to be updated, update item
			if(dbUpdate)
				newItem.updateDatabase();
		}

		// FIXME при чтении инвентаря dbUpdate = false, это происходит только при логине игрока, но ведь могут и сломать... :)
		if(dbUpdate && CursedWeaponsManager.getInstance().isCursed(result.getItemId()) && getOwner().isPlayer)
			CursedWeaponsManager.getInstance().checkPlayer((L2Player) getOwner(), result);

		// Refresh weigth
		refreshWeight();
		return result;
	}

	/**
	 * Returns the item in the paperdoll slot
	 * @return L2ItemInstance
	 */
	public L2ItemInstance getPaperdollItem(int slot)
	{
		return _paperdoll[slot];
	}

	/**
	 * Returns the ID of the item in the paperdol slot
	 * @param slot : int designating the slot
	 * @return int designating the ID of the item
	 */
	public int getPaperdollItemId(int slot)
	{
		L2ItemInstance item = _paperdoll[slot];
		if(item != null)
			return item.getItemId();
		else if(slot == PAPERDOLL_HAIR)
		{
			item = _paperdoll[PAPERDOLL_DHAIR];
			if(item != null)
				return item.getItemId();
		}
		return 0;
	}

	/**
	 * Returns the objectID associated to the item in the paperdoll slot
	 * @param slot : int pointing out the slot
	 * @return int designating the objectID
	 */
	public int getPaperdollObjectId(int slot)
	{
		L2ItemInstance item = _paperdoll[slot];
		if(item != null)
			return item.getObjectId();
		else if(slot == PAPERDOLL_HAIR)
		{
			item = _paperdoll[PAPERDOLL_DHAIR];
			if(item != null)
				return item.getObjectId();
		}
		return 0;
	}

	/**
	 * Adds new inventory's paperdoll listener
	 * @param listener pointing out the listener
	 */
	public synchronized void addPaperdollListener(PaperdollListener listener)
	{
		_paperdollListeners.add(listener);
	}

	/**
	 * Removes a paperdoll listener
	 * @param listener pointing out the listener to be deleted
	 */
	public synchronized void removePaperdollListener(PaperdollListener listener)
	{
		_paperdollListeners.remove(listener);
	}

	/**
	 * Equips an item in the given slot of the paperdoll.
	 * <U><I>Remark :</I></U> The item <B>HAS TO BE</B> already in the inventory
	 * @param slot : int pointing out the slot of the paperdoll
	 * @param item : L2ItemInstance pointing out the item to add in slot
	 * @return L2ItemInstance designating the item placed in the slot before
	 */
	public L2ItemInstance setPaperdollItem(short slot, L2ItemInstance item)
	{
		L2ItemInstance old = _paperdoll[slot];
		if(old != item)
		{
			if(old != null)
			{
				_paperdoll[slot] = null;
				// Put old item from paperdoll slot to base location
				old.setLocation(getBaseLocation());
				old.setLastChange(L2ItemInstance.MODIFIED);
				// Get the mask for paperdoll
				int mask = 0;
				for(int i = 0; i < PAPERDOLL_LRHAND; i++)
				{
					L2ItemInstance pi = _paperdoll[i];
					if(pi != null)
						mask |= pi.getItem().getItemMask();
				}
				_wearedMask = mask;
				// Notify all paperdoll listener in order to unequip old item in slot
				for(PaperdollListener listener : _paperdollListeners)
					listener.notifyUnequipped(slot, old);
				old.updateDatabase();
			}
			// Add new item in slot of paperdoll
			if(item != null)
			{
				_paperdoll[slot] = item;
				item.setLocation(getEquipLocation(), slot);
				item.setLastChange(L2ItemInstance.MODIFIED);
				_wearedMask |= item.getItem().getItemMask();
				for(PaperdollListener listener : _paperdollListeners)
					listener.notifyEquipped(slot, item);
				item.updateDatabase();
			}
		}
		return old;
	}

	/**
	 * Return the mask of weared item
	 * @return int
	 */
	public int getWearedMask()
	{
		return _wearedMask;
	}

	public synchronized void unEquipItem(L2ItemInstance item)
	{
		short pdollSlot = item.getEquipSlot();
		if(pdollSlot < 0)
			return;

		if(pdollSlot == PAPERDOLL_LRHAND)
		{
			setPaperdollItem(PAPERDOLL_LHAND, null);
			setPaperdollItem(PAPERDOLL_RHAND, null); // this should be the same as in LRHAND
		}
		setPaperdollItem(pdollSlot, null);
	}

	/**
	 * Unequips item in body slot and returns alterations.
	 * @param slot : int designating the slot of the paperdoll
	 * @param item - предмет который снимаем, нужен только при побитовом сложении
	 * @return L2ItemInstance[] : list of changes
	 */
	public L2ItemInstance[] unEquipItemInBodySlotAndRecord(int slot, L2ItemInstance item)
	{
		ChangeRecorder recorder = newRecorder();
		try
		{
			unEquipItemInBodySlot(slot, item);
		}
		finally
		{
			removePaperdollListener(recorder);
		}
		return recorder.getChangedItems();
	}

	/**
	 * Sets item in slot of the paperdoll to null value
	 * @param pdollSlot : int designating the slot
	 * @return L2ItemInstance designating the item in slot before change
	 */
	public L2ItemInstance unEquipItemInSlot(short pdollSlot)
	{
		return setPaperdollItem(pdollSlot, null);
	}

	/**
	 * Unepquips item in slot and returns alterations
	 * @param slot : int designating the slot
	 * @return L2ItemInstance[] : list of items altered
	 */
	public L2ItemInstance[] unEquipItemInSlotAndRecord(short slot)
	{
		ChangeRecorder recorder = newRecorder();
		try
		{
			unEquipItemInSlot(slot);
		}
		finally
		{
			removePaperdollListener(recorder);
		}
		return recorder.getChangedItems();
	}

	/**
	 * Unequips item in slot (i.e. equips with default value)
	 * @param slot : int designating the slot
	 * @param item - предмет, нужен только при побитовом сложении.
	 */
	private void unEquipItemInBodySlot(int slot, L2ItemInstance item)
	{
		if(Config.DEBUG)
			_log.fine("--- unequip body slot:" + slot);
		byte pdollSlot = -1;

		switch(slot)
		{
			case L2Item.SLOT_L_EAR:
				pdollSlot = PAPERDOLL_LEAR;
				break;
			case L2Item.SLOT_R_EAR:
				pdollSlot = PAPERDOLL_REAR;
				break;
			case L2Item.SLOT_NECK:
				pdollSlot = PAPERDOLL_NECK;
				break;
			case L2Item.SLOT_R_FINGER:
				pdollSlot = PAPERDOLL_RFINGER;
				break;
			case L2Item.SLOT_L_FINGER:
				pdollSlot = PAPERDOLL_LFINGER;
				break;
			case L2Item.SLOT_HAIR:
				pdollSlot = PAPERDOLL_HAIR;
				break;
			case L2Item.SLOT_FACE:
				pdollSlot = PAPERDOLL_FACE;
				break;
			case L2Item.SLOT_DHAIR:
				setPaperdollItem(PAPERDOLL_HAIR, null);
				setPaperdollItem(PAPERDOLL_FACE, null);// this should be the same as in DHAIR
				pdollSlot = PAPERDOLL_DHAIR;
				break;
			case L2Item.SLOT_HEAD:
				pdollSlot = PAPERDOLL_HEAD;
				break;
			case L2Item.SLOT_R_HAND:
				pdollSlot = PAPERDOLL_RHAND;
				break;
			case L2Item.SLOT_L_HAND:
				pdollSlot = PAPERDOLL_LHAND;
				break;
			case L2Item.SLOT_GLOVES:
				pdollSlot = PAPERDOLL_GLOVES;
				break;
			case L2Item.SLOT_CHEST: // fall through
			case L2Item.SLOT_FULL_ARMOR:
			case L2Item.SLOT_FORMAL_WEAR:
				pdollSlot = PAPERDOLL_CHEST;
				break;
			case L2Item.SLOT_LEGS:
				pdollSlot = PAPERDOLL_LEGS;
				break;
			case L2Item.SLOT_BACK:
				pdollSlot = PAPERDOLL_BACK;
				break;
			case L2Item.SLOT_FEET:
				pdollSlot = PAPERDOLL_FEET;
				break;
			case L2Item.SLOT_UNDERWEAR:
				pdollSlot = PAPERDOLL_UNDER;
				break;
			case L2Item.SLOT_CLOAK:
				pdollSlot = PAPERDOLL_CLOAK;
				break;
			case L2Item.SLOT_LR_HAND:
				setPaperdollItem(PAPERDOLL_LHAND, null);
				setPaperdollItem(PAPERDOLL_RHAND, null);// this should be the same as in LRHAND
				pdollSlot = PAPERDOLL_LRHAND;
				break;
			case L2Item.SLOT_L_EAR | L2Item.SLOT_R_EAR:
				if(item == null)
				{
					_log.warning("You cant use bitmatrix slot without item!!!");
					Thread.dumpStack();
					return;
				}
				if(getPaperdollObjectId(PAPERDOLL_LEAR) == item.getObjectId())
					pdollSlot = PAPERDOLL_LEAR;
				else if(getPaperdollObjectId(PAPERDOLL_REAR) == item.getObjectId())
					pdollSlot = PAPERDOLL_REAR;
				else
				{
					_log.warning("Incorrect slot request...");
					Thread.dumpStack();
				}
				break;
			case L2Item.SLOT_L_FINGER | L2Item.SLOT_R_FINGER:
				if(item == null)
				{
					_log.warning("You cant use bitmatrix slot without item!!!");
					Thread.dumpStack();
					return;
				}
				if(getPaperdollObjectId(PAPERDOLL_LFINGER) == item.getObjectId())
					pdollSlot = PAPERDOLL_LFINGER;
				else if(getPaperdollObjectId(PAPERDOLL_RFINGER) == item.getObjectId())
					pdollSlot = PAPERDOLL_RFINGER;
				else
				{
					_log.warning("Incorrect slot request...");
					Thread.dumpStack();
				}
				break;
			default:
				_log.warning("Requested invalid body slot!!! " + slot + " for item " + item.getItemId());
				Thread.dumpStack();
		}
		if(pdollSlot >= 0)
			setPaperdollItem(pdollSlot, null);
	}

	/**
	 * Equips item and returns list of alterations
	 * @param item : L2ItemInstance corresponding to the item
	 * @return L2ItemInstance[] : list of alterations
	 */
	public L2ItemInstance[] equipItemAndRecord(L2ItemInstance item)
	{
		ChangeRecorder recorder = newRecorder();
		try
		{
			equipItem(item);
		}
		finally
		{
			removePaperdollListener(recorder);
		}
		return recorder.getChangedItems();
	}

	/**
	 * Equips item in slot of paperdoll.
	 * @param item : L2ItemInstance designating the item and slot used.
	 */
	public synchronized void equipItem(L2ItemInstance item)
	{
		int itemId = item.getItemId();

		if(getOwner().isPlayer && getOwner().getName() != null)
		{
			L2Player owner = (L2Player) getOwner();

			// Hero items
			if(item.isHeroItem() && !owner.isHero())
			{
				owner.sendMessage("Only a Hero can wear this item.", "Только герой может носить это.");
				return;
			}

			if(itemId >= 7850 && itemId <= 7859) // Clan Oath Armor
				if(owner.getClan() == null || owner.getPledgeType() != L2Clan.SUBUNIT_ACADEMY)
				{
					owner.sendPacket(new SystemMessage(SystemMessage.THIS_ITEM_CAN_ONLY_BE_WORN_BY_A_MEMBER_OF_THE_CLAN_ACADEMY));
					return;
				}

			if(itemId >= 7860 && itemId <= 7879) // Clan Apella armor
				if(owner.getPledgeClass() < L2Player.RANK_BARON)
				{
					owner.sendMessage("You need a pledge class of Baron or higher to wear this item.", "Ваш ранк должен быть Барон или выше, что бы носить этот вид брони.");
					return;
				}

			if(Arrays.asList(_castleLordCirclets).contains(itemId))
				if(owner.getClan() == null || itemId != _castleLordCirclets[owner.getClan().getHasCastle()])
				{
					owner.sendMessage("This circlet can only be worn by " + CastleManager.getInstance().getCastleByIndex(Arrays.asList(_castleLordCirclets).indexOf(itemId)).getName() + " Castle owner clan member.", "Эту корону может носить только член клана, владеющего замком " + CastleManager.getInstance().getCastleByIndex(Arrays.asList(_castleLordCirclets).indexOf(itemId)).getName());
					return;
				}
		}

		int targetSlot = item.getItem().getBodyPart();

		// Нельзя одевать оружие, если уже одето проклятое оружие
		if(CursedWeaponsManager.getInstance().isCursed(getPaperdollItemId(PAPERDOLL_LRHAND)) && (targetSlot == L2Item.SLOT_LR_HAND || targetSlot == L2Item.SLOT_L_HAND || targetSlot == L2Item.SLOT_R_HAND))
			return;

		switch(targetSlot)
		{
			case L2Item.SLOT_LR_HAND:
			{
				if(setPaperdollItem(PAPERDOLL_LHAND, null) != null)
				{
					// exchange 2h for 2h
					setPaperdollItem(PAPERDOLL_RHAND, null);
					setPaperdollItem(PAPERDOLL_LHAND, null);
				}
				else
					setPaperdollItem(PAPERDOLL_RHAND, null);
				setPaperdollItem(PAPERDOLL_RHAND, item);
				setPaperdollItem(PAPERDOLL_LRHAND, item);
				break;
			}
			case L2Item.SLOT_L_HAND:
			{
				if(!(item.getItem() instanceof L2EtcItem) || item.getItem().getItemType() != EtcItemType.ARROW && item.getItem().getItemType() != EtcItemType.BAIT)
				{
					L2ItemInstance old1 = setPaperdollItem(PAPERDOLL_LRHAND, null);
					if(old1 != null)
						setPaperdollItem(PAPERDOLL_RHAND, null);
				}
				setPaperdollItem(PAPERDOLL_LHAND, null);
				setPaperdollItem(PAPERDOLL_LHAND, item);
				break;
			}
			case L2Item.SLOT_R_HAND:
			{
				if(_paperdoll[PAPERDOLL_LRHAND] != null)
				{
					setPaperdollItem(PAPERDOLL_LRHAND, null);
					setPaperdollItem(PAPERDOLL_LHAND, null);
					setPaperdollItem(PAPERDOLL_RHAND, null);
				}
				else
					setPaperdollItem(PAPERDOLL_RHAND, null);
				setPaperdollItem(PAPERDOLL_RHAND, item);
				break;
			}
			case L2Item.SLOT_L_EAR:
			case L2Item.SLOT_R_EAR:
			case L2Item.SLOT_L_EAR | L2Item.SLOT_R_EAR:
			{
				if(_paperdoll[PAPERDOLL_LEAR] == null)
				{
					item.setBodyPart(L2Item.SLOT_L_EAR);
					setPaperdollItem(PAPERDOLL_LEAR, item);
				}
				else if(_paperdoll[PAPERDOLL_REAR] == null)
				{
					item.setBodyPart(L2Item.SLOT_R_EAR);
					setPaperdollItem(PAPERDOLL_REAR, item);
				}
				else
				{
					item.setBodyPart(L2Item.SLOT_L_EAR);
					setPaperdollItem(PAPERDOLL_LEAR, null);
					setPaperdollItem(PAPERDOLL_LEAR, item);
				}
				break;
			}
			case L2Item.SLOT_L_FINGER:
			case L2Item.SLOT_R_FINGER:
			case L2Item.SLOT_L_FINGER | L2Item.SLOT_R_FINGER:
			{
				if(_paperdoll[PAPERDOLL_LFINGER] == null)
				{
					item.setBodyPart(L2Item.SLOT_L_FINGER);
					setPaperdollItem(PAPERDOLL_LFINGER, item);
				}
				else if(_paperdoll[PAPERDOLL_RFINGER] == null)
				{
					item.setBodyPart(L2Item.SLOT_R_FINGER);
					setPaperdollItem(PAPERDOLL_RFINGER, item);
				}
				else
				{
					item.setBodyPart(L2Item.SLOT_L_FINGER);
					setPaperdollItem(PAPERDOLL_LFINGER, null);
					setPaperdollItem(PAPERDOLL_LFINGER, item);
				}
				break;
			}
			case L2Item.SLOT_NECK:
				setPaperdollItem(PAPERDOLL_NECK, item);
				break;
			case L2Item.SLOT_FULL_ARMOR:
				setPaperdollItem(PAPERDOLL_CHEST, null);
				setPaperdollItem(PAPERDOLL_LEGS, null);
				setPaperdollItem(PAPERDOLL_CHEST, item);
				break;
			case L2Item.SLOT_CHEST:
				setPaperdollItem(PAPERDOLL_CHEST, item);
				break;
			case L2Item.SLOT_LEGS:
			{
				// handle full armor
				L2ItemInstance chest = getPaperdollItem(PAPERDOLL_CHEST);
				if(chest != null && chest.getBodyPart() == L2Item.SLOT_FULL_ARMOR)
					setPaperdollItem(PAPERDOLL_CHEST, null);

				if(getPaperdollItemId(PAPERDOLL_CHEST) == 6408)
					setPaperdollItem(PAPERDOLL_CHEST, null);
				setPaperdollItem(PAPERDOLL_LEGS, null);
				setPaperdollItem(PAPERDOLL_LEGS, item);
				break;
			}
			case L2Item.SLOT_FEET:
				if(getPaperdollItemId(PAPERDOLL_CHEST) == 6408)
					setPaperdollItem(PAPERDOLL_CHEST, null);
				setPaperdollItem(PAPERDOLL_FEET, item);
				break;
			case L2Item.SLOT_GLOVES:
				if(getPaperdollItemId(PAPERDOLL_CHEST) == 6408)
					setPaperdollItem(PAPERDOLL_CHEST, null);
				setPaperdollItem(PAPERDOLL_GLOVES, item);
				break;
			case L2Item.SLOT_HEAD:
				if(getPaperdollItemId(PAPERDOLL_CHEST) == 6408)
					setPaperdollItem(PAPERDOLL_CHEST, null);
				setPaperdollItem(PAPERDOLL_HEAD, item);
				break;
			case L2Item.SLOT_HAIR:
				if(setPaperdollItem(PAPERDOLL_DHAIR, null) != null)
				{
					setPaperdollItem(PAPERDOLL_DHAIR, null);
					setPaperdollItem(PAPERDOLL_HAIR, null);
					setPaperdollItem(PAPERDOLL_FACE, null);
				}
				else
					setPaperdollItem(PAPERDOLL_HAIR, null);
				setPaperdollItem(PAPERDOLL_HAIR, item);
				break;
			case L2Item.SLOT_FACE:
				if(setPaperdollItem(PAPERDOLL_DHAIR, null) != null)
				{
					setPaperdollItem(PAPERDOLL_DHAIR, null);
					setPaperdollItem(PAPERDOLL_HAIR, null);
					setPaperdollItem(PAPERDOLL_FACE, null);
				}
				else
					setPaperdollItem(PAPERDOLL_FACE, null);
				setPaperdollItem(PAPERDOLL_FACE, item);
				break;
			case L2Item.SLOT_DHAIR:
				if(setPaperdollItem(PAPERDOLL_HAIR, null) != null)
				{
					setPaperdollItem(PAPERDOLL_HAIR, null);
					setPaperdollItem(PAPERDOLL_FACE, null);
				}
				else
					setPaperdollItem(PAPERDOLL_FACE, null);
				setPaperdollItem(PAPERDOLL_DHAIR, item);
				break;
			case L2Item.SLOT_UNDERWEAR:
				setPaperdollItem(PAPERDOLL_UNDER, item);
				break;
			case L2Item.SLOT_BACK:
				setPaperdollItem(PAPERDOLL_BACK, item);
				break;
			case L2Item.SLOT_CLOAK:
				setPaperdollItem(PAPERDOLL_CLOAK, item);
				break;
			case L2Item.SLOT_FORMAL_WEAR:
				//При одевании свадебного платья руки не трогаем
				setPaperdollItem(PAPERDOLL_LEGS, null);
				setPaperdollItem(PAPERDOLL_CHEST, null);
				setPaperdollItem(PAPERDOLL_HEAD, null);
				setPaperdollItem(PAPERDOLL_FEET, null);
				setPaperdollItem(PAPERDOLL_GLOVES, null);
				setPaperdollItem(PAPERDOLL_CHEST, item);
				break;
			default:
				_log.warning("unknown body slot:" + targetSlot);
		}

		//if(item != null)
		//Log.add("Inventory|" + getOwner().getName() + "|Equip|" + item.getItemId() + "|slot " + targetSlot + "|" + item.getObjectId(), "items"); // сомнительной ценности лог

		if(getOwner().isPlayer)
			((L2Player) getOwner()).AutoShot();
	}

	/**
	 * Returns the item from inventory by using its <B>itemId</B>
	 * @param itemId : int designating the ID of the item
	 * @return L2ItemInstance designating the item or null if not found in inventory
	 */
	public L2ItemInstance getItemByItemId(int itemId)
	{
		synchronized (_items)
		{
			for(L2ItemInstance temp : _items)
				if(temp.getItemId() == itemId)
					return temp;
		}
		return null;
	}

	/**
	 * Return the L2ItemInstance corresponding to the item identifier.<BR><BR>
	 * @param itemId : int designating the ID of the item
	 */
	public L2ItemInstance findItemByItemId(int itemId)
	{
		return getItemByItemId(itemId);
	}

	public L2ItemInstance[] getAllItemsById(int itemId)
	{
		ArrayList<L2ItemInstance> ar = new ArrayList<L2ItemInstance>();
		synchronized (_items)
		{
			for(L2ItemInstance i : _items)
				if(i.getItemId() == itemId)
					ar.add(i);
		}
		return ar.toArray(new L2ItemInstance[ar.size()]);
	}

	public int getPaperdollAugmentationId(int slot)
	{
		L2ItemInstance item = _paperdoll[slot];
		if(item != null && item.getAugmentation() != null)
			return item.getAugmentation().getAugmentationId();
		return 0;
	}

	public L2ItemInstance getItemByItemIdExclude(int itemId, Set<L2ItemInstance> excludeObjIds)
	{
		synchronized (_items)
		{
			for(L2ItemInstance temp : _items)
				if(temp.getItemId() == itemId && !excludeObjIds.contains(temp))
					return temp;
		}
		return null;
	}

	/**
	 * Returns item from inventory by using its <B>objectId</B>
	 * @param objectId : int designating the ID of the object
	 * @return L2ItemInstance designating the item or null if not found in inventory
	 */
	public L2ItemInstance getItemByObjectId(Integer objectId)
	{
		synchronized (_items)
		{
			for(L2ItemInstance temp : _items)
				if(temp.getObjectId() == objectId)
					return temp;
		}
		return null;
	}

	/**
	 * Destroy item from inventory by using its <B>objectID</B> and updates database
	 * @param objectId : int pointing out the objectID of the item
	 * @param count : long designating the quantity of item to destroy
	 * @param update : boolean if true, send InventoryUpdate
	 * @return L2ItemInstance designating the item up-to-date
	 */
	public L2ItemInstance destroyItem(int objectId, long count, boolean update, boolean toLog)
	{
		L2ItemInstance item = getItemByObjectId(objectId);
		return destroyItem(item, count, update, toLog);
	}

	/**
	 * Destroy item from inventory and updates database
	 * @param item : L2ItemInstance item to destroy
	 * @param count : long designating the quantity of item to destroy
	 * @param update : boolean if true, send InventoryUpdate
	 * @return L2ItemInstance designating the item up-to-date
	 */
	public L2ItemInstance destroyItem(L2ItemInstance item, long count, boolean update, boolean toLog)
	{
		if(count < 0)
		{
			_log.warning("DestroyItem: count < 0 owner:" + getOwner().getName());
			Thread.dumpStack();
			return null;
		}
		if(item != null && toLog)
		{
			Log.LogItem(getOwner(), null, Log.DeleteItem, item, count);
			Log.add("Inventory|" + getOwner().getName() + "|Destroys item|" + item.getItemId() + "|" + count + "|" + item.getObjectId(), "items");
		}
		if(item == null)
			return null;

		if(item.getCount() <= count)
		{
			if(item.getCount() < count && toLog)
				Log.add("!Inventory|" + getOwner().getName() + "|Destroys item|" + item.getItemId() + "|" + count + " but item count " + item.getCount() + "|" + item.getObjectId(), "items");
			removeItemFromInventory(item, true);
			// check destroy pet controls items
			if(PetDataTable.isPetControlItem(item))
				PetDataTable.deletePet(item, getOwner());
		}
		else
		{
			item.setCount(item.getCount() - count);
			item.setLastChange(L2ItemInstance.MODIFIED);
			item.updateDatabase();
		}

		refreshWeight();

		if(update)
			if(item.getCount() == 0)
				getOwner().sendPacket((new InventoryUpdate()).addRemovedItem(item));
			else
				getOwner().sendPacket((new InventoryUpdate()).addModifiedItem(item));

		return item;
	}

	// we need this one cuz warehouses send itemId only
	/**
	 * Destroy item from inventory by using its <B>itemID</B> and updates database
	 * @param itemId : int pointing out the itemID of the item
	 * @param count : long designating the quantity of item to destroy
	 * @param update : boolean if true, send InventoryUpdate
	 * @return L2ItemInstance designating the item up-to-date
	 */
	public L2ItemInstance destroyItemByItemId(int itemId, long count, boolean update, boolean toLog)
	{
		L2ItemInstance item = findItemByItemId(itemId);
		Log.LogItem(getOwner(), Log.Sys_DeleteItem, item, count);
		return destroyItem(item, count, update, toLog);
	}

	/**
	 * Destroy item from inventory and from database.
	 * @param item : L2ItemInstance designating the item to remove from inventory
	 * @param clearCount : boolean : if true, set the item quantity to 0
	 */
	private void removeItemFromInventory(L2ItemInstance item, boolean clearCount)
	{
		if(getOwner().isPlayer)
		{
			L2Player player = (L2Player) getOwner();
			player.removeItemFromShortCut(item.getObjectId());
			if(item.isEquipped())
				unEquipItem(item);
		}

		synchronized (_items)
		{
			_items.remove(item);
		}

		if(clearCount)
		{
			item.setCount(0);
			L2World.removeObject(item);
			IdFactory.getInstance().releaseId(item.getObjectId());
		}
		item.setOwnerId(0);
		item.setLocation(ItemLocation.VOID);
		item.setLastChange(L2ItemInstance.REMOVED);
		item.updateDatabase(true);
	}

	/**
	 * Drop item from inventory by using its <B>objectID</B>
	 * @param objectId : int pointing out the objectID
	 * @param count : int designating the quantity of item to drop
	 * @return L2ItemInstance
	 */
	public L2ItemInstance dropItem(int objectId, long count)
	{
		return dropItem(getItemByObjectId(objectId), count);
	}

	/**
	 * Флаг нужен для определения удалять ли инкрустацию
	 */
	public L2ItemInstance dropItem(L2ItemInstance item, long count, boolean whflag)
	{
		item.setWhFlag(whflag);
		return dropItem(item, count);
	}

	/**
	 * Drop item from inventory by using <B>object L2ItemInstance</B><BR><BR>
	 * <U><I>Concept :</I></U><BR>
	 * item equipped are unequipped
	 * <LI>If quantity of items in inventory after drop is negative or null, change location of item</LI>
	 * <LI>Otherwise, change quantity in inventory and create new object with quantity dropped</LI>
	 * @param oldItem : L2ItemInstance designating the item to drop
	 * @param count : int designating the quantity of item to drop
	 * @return L2ItemInstance designating the item dropped
	 */
	public L2ItemInstance dropItem(L2ItemInstance oldItem, long count)
	{
		if(getOwner() != null && getOwner().isPlayer && ((L2Player) getOwner()).getPlayerAccess() != null && ((L2Player) getOwner()).getPlayerAccess().BlockInventory)
			return null;
		if(count < 0)
		{
			_log.warning("DropItem: count < 0 owner:" + getOwner().getName());
			return null;
		}
		if(oldItem == null)
		{
			_log.warning("DropItem: item id does not exist in inventory");
			return null;
		}

		Log.LogItem(getOwner(), null, Log.Drop, oldItem, count);

		if(oldItem.getCount() <= count || oldItem.getCount() <= 1)
		{
			Log.add("Inventory|" + getOwner().getName() + "|Drop item|" + oldItem.getItemId() + "|" + count + "|" + oldItem.getObjectId(), "items");
			if(Config.DEBUG)
				_log.fine(" count = count  --> remove");
			removeItemFromInventory(oldItem, false);
			refreshWeight();

			//check drop pet controls items
			if(PetDataTable.isPetControlItem(oldItem))
				PetDataTable.unSummonPet(oldItem, getOwner());
			return oldItem;
		}
		if(Config.DEBUG)
			_log.fine(" count != count  --> reduce");
		oldItem.setCount(oldItem.getCount() - count);
		oldItem.setLastChange(L2ItemInstance.MODIFIED);
		L2ItemInstance newItem = ItemTable.getInstance().createItem(oldItem.getItemId(), getOwner().getObjectId(), oldItem.getOwnerId(), "Inventory.dropItem");
		newItem.setCount(count);
		oldItem.updateDatabase();
		refreshWeight();
		Log.add("Inventory|" + getOwner().getName() + "|Split item from-to|" + oldItem.getItemId() + "|" + oldItem.getObjectId() + "|" + newItem.getObjectId(), "items");
		Log.add("Inventory|" + getOwner().getName() + "|Drop item|" + newItem.getItemId() + "|" + count + "|" + newItem.getObjectId(), "items");
		return newItem;
	}

	/**
	 * Refresh the weight of equipment loaded
	 */
	private void refreshWeight()
	{
		int weight = 0;

		synchronized (_items)
		{
			for(L2ItemInstance element : _items)
				weight += element.getItem().getWeight() * element.getCount();
		}

		_totalWeight = weight;
		// notify char for overload checking
		if(getOwner().isPlayer)
			((L2Player) getOwner()).refreshOverloaded();
		else if(getOwner().isPet)
			((L2PetInstance) getOwner()).broadcastPetInfo();
	}

	/**
	 * Returns the totalWeight.
	 * @return int
	 */
	public int getTotalWeight()
	{
		return _totalWeight;
	}

	/**
	 * Return the L2ItemInstance of the arrows needed for this bow.<BR><BR>
	 * @param bow : L2Item designating the bow
	 * @return L2ItemInstance pointing out arrows for bow
	 */
	public L2ItemInstance findArrowForBow(L2Item bow)
	{
		int arrowsId = 0;

		switch(bow.getCrystalType())
		{
			default: // broken weapon.csv ??
			case L2Item.CRYSTAL_NONE:
				arrowsId = 17; // Wooden arrow
				break;
			case L2Item.CRYSTAL_D:
				arrowsId = 1341; // Bone arrow
				break;
			case L2Item.CRYSTAL_C:
				arrowsId = 1342; // Fine steel arrow
				break;
			case L2Item.CRYSTAL_B:
				arrowsId = 1343; // Silver arrow
				break;
			case L2Item.CRYSTAL_A:
				arrowsId = 1344; // Mithril arrow
				break;
			case L2Item.CRYSTAL_S:
				arrowsId = 1345; // Shining arrow
				break;
		}

		// Get the L2ItemInstance corresponding to the item identifier and return it
		return findItemByItemId(arrowsId);
	}

	/**
	 * Delete item object from world
	 */
	public void deleteMe()
	{
		ArrayList<L2ItemInstance> items = new ArrayList<L2ItemInstance>(_items);
		synchronized (_items)
		{
			_items.clear();
		}
		try
		{
			updateDatabase(items, true);
		}
		catch(Throwable t)
		{
			_log.log(Level.SEVERE, "deletedMe()", t);
		}
		for(L2ItemInstance inst : items)
			try
			{
				L2World.removeObject(inst);
			}
			catch(Throwable t)
			{
				_log.log(Level.SEVERE, "deletedMe()", t);
			}
	}

	/**
	 * Update database with items in inventory
	 */
	public void updateDatabase()
	{
		updateDatabase(_items, false);
	}

	public void updateDatabase(boolean commit)
	{
		updateDatabase(_items, commit);
	}

	/**
	 * Update database with item
	 * @param items : ArrayList &lt;L2ItemInstance&gt; pointing out the list of items
	 */
	private void updateDatabase(List<L2ItemInstance> items, boolean commit)
	{
		if(getOwner() != null)
			synchronized (_items)
			{
				for(L2ItemInstance inst : items)
					inst.updateDatabase(commit);
			}
	}

	/**
	 * Функция для проверки "одетости" клановой брони.
	 */
	public void validateClanArmorLocation()
	{
		for(L2ItemInstance item : getItemsList())
			if(item.getItemId() >= 7860 && item.getItemId() <= 7879)
				if(getOwner().isPlayer)
				{
					L2Player owner = (L2Player) getOwner();
					if(owner.getPledgeClass() < L2Player.RANK_BARON)
						unEquipItem(item);

					//TODO: отсылка inventory update???
				}
	}

	/**
	 * Get back items in inventory from database
	 */
	public void restore()
	{
		final int OWNER = getOwner().getObjectId();

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		FiltredPreparedStatement delete = null;
		ResultSet rset = null;

		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT object_id FROM items WHERE owner_id=? AND (loc=? OR loc=?) ORDER BY object_id DESC");
			statement.setInt(1, OWNER);
			statement.setString(2, getBaseLocation().name());
			statement.setString(3, getEquipLocation().name());
			rset = statement.executeQuery();

			L2ItemInstance item;
			L2ItemInstance newItem;
			while(rset.next())
			{
				int objectId = rset.getInt(1);
				item = L2ItemInstance.restoreFromDb(objectId);
				if(item == null)
					continue;
				if(getOwner().isPlayer)
				{
					L2Player owner = (L2Player) getOwner();
					int itemId = item.getItemId();
					if(Arrays.asList(Hero.getHeroItems()).contains(itemId)) // Hero items
						if(!owner.isHero())
							item.setLocation(ItemLocation.INVENTORY);

					if(Arrays.asList(_castleLordCirclets).contains(itemId)) // Castle Lord circlets
						if(owner.getClan() == null || itemId != _castleLordCirclets[owner.getClan().getHasCastle()])
						{
							removeItemFromInventory(item, true);
							continue;
						}
				}
				newItem = addItem(item, false, false);
				if(newItem == null)
					continue;
				newItem.setOwner(getOwner());
				if(item.isEquipped())
					equipItem(item);
				if(newItem == item)
					L2World.addObject(item); // add this item to the world
				else
				{ //we had another stack before, so update them to remove the duplicate
					item.updateDatabase();
					newItem.updateDatabase();
				}
			}
			DatabaseUtils.closeDatabaseSR(statement, rset);

			// Delayed add
			statement = con.prepareStatement("SELECT * FROM items_delayed WHERE owner_id=? AND payment_status=0");
			delete = con.prepareStatement("UPDATE items_delayed SET payment_status=1 WHERE payment_id=?");
			statement.setInt(1, OWNER);
			rset = statement.executeQuery();

			while(rset.next())
			{
				final short ITEM_ID = rset.getShort("item_id");
				final int ITEM_COUNT = rset.getInt("count");
				final short ITEM_ENCHANT = rset.getShort("enchant_level");
				final int PAYMENT_ID = rset.getInt("payment_id");

				item = ItemTable.getInstance().createItem(ITEM_ID, OWNER, 0, "delayed_add");
				if(item == null)
					continue;
				if(item.isStackable())
					item.setCount(ITEM_COUNT);
				else
					item.setEnchantLevel(ITEM_ENCHANT);
				item.setLocation(ItemLocation.INVENTORY);

				newItem = addItem(item, true, false);
				if(newItem == null)
					continue;
				Log.add("<add owner_id=" + OWNER + " item_id=" + ITEM_ID + " count=" + ITEM_COUNT + " enchant_level=" + ITEM_ENCHANT + " payment_id=" + PAYMENT_ID + "/>", "delayed_add");

				delete.setInt(1, PAYMENT_ID);
				delete.execute();

				if(newItem == item)
					L2World.addObject(item);
				else
				{
					item.updateDatabase();
					newItem.updateDatabase();
				}
			}

			refreshWeight();
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not restore inventory for player " + getOwner().getName() + ":", e);
		}
		finally
		{
			DatabaseUtils.closeStatement(delete);
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
	}

	/**
	 * Refresh all listeners
	 */
	public void refreshListeners()
	{
		setRefreshingListeners(true);
		for(int i = 0; i < _paperdoll.length; i++)
		{
			L2ItemInstance item = getPaperdollItem(i);
			if(item == null)
				continue;
			for(PaperdollListener listener : _paperdollListeners)
			{
				listener.notifyUnequipped(i, item);
				listener.notifyEquipped(i, item);
			}
		}
		setRefreshingListeners(false);
	}

	public L2ItemInstance getPaperdollItemBySlot(int slot)
	{
		byte pdollSlot = -1;

		switch(slot)
		{
			case L2Item.SLOT_L_EAR:
				pdollSlot = PAPERDOLL_LEAR;
				break;
			case L2Item.SLOT_R_EAR:
				pdollSlot = PAPERDOLL_REAR;
				break;
			case L2Item.SLOT_NECK:
				pdollSlot = PAPERDOLL_NECK;
				break;
			case L2Item.SLOT_R_FINGER:
				pdollSlot = PAPERDOLL_RFINGER;
				break;
			case L2Item.SLOT_L_FINGER:
				pdollSlot = PAPERDOLL_LFINGER;
				break;
			case L2Item.SLOT_HAIR:
				pdollSlot = PAPERDOLL_HAIR;
				break;
			case L2Item.SLOT_FACE:
				pdollSlot = PAPERDOLL_FACE;
				break;
			case L2Item.SLOT_DHAIR:
				pdollSlot = PAPERDOLL_DHAIR;
				break;
			case L2Item.SLOT_HEAD:
				pdollSlot = PAPERDOLL_HEAD;
				break;
			case L2Item.SLOT_R_HAND:
				pdollSlot = PAPERDOLL_RHAND;
				break;
			case L2Item.SLOT_L_HAND:
				pdollSlot = PAPERDOLL_LHAND;
				break;
			case L2Item.SLOT_GLOVES:
				pdollSlot = PAPERDOLL_GLOVES;
				break;
			case L2Item.SLOT_CHEST:
			case L2Item.SLOT_FULL_ARMOR:
				pdollSlot = PAPERDOLL_CHEST;
				break;
			case L2Item.SLOT_LEGS:
				pdollSlot = PAPERDOLL_LEGS;
				break;
			case L2Item.SLOT_BACK:
				pdollSlot = PAPERDOLL_BACK;
				break;
			case L2Item.SLOT_FEET:
				pdollSlot = PAPERDOLL_FEET;
				break;
			case L2Item.SLOT_UNDERWEAR:
				pdollSlot = PAPERDOLL_UNDER;
				break;
			case L2Item.SLOT_CLOAK:
				pdollSlot = PAPERDOLL_CLOAK;
				break;
			case L2Item.SLOT_LR_HAND:
				pdollSlot = PAPERDOLL_LRHAND;
				break;
		}

		if(pdollSlot >= 0)
			return _paperdoll[pdollSlot];
		return null;
	}

	public boolean isRefreshingListeners()
	{
		return refreshingListeners;
	}

	public void setRefreshingListeners(boolean refreshingListeners)
	{
		this.refreshingListeners = refreshingListeners;
	}
}