package L2jFT.game.model;

import L2jFT.game.ThreadPoolManager;
import L2jFT.game.network.serverpackets.ExOlympiadSpelledInfo;
import L2jFT.game.network.serverpackets.MagicEffectIcons;
import L2jFT.game.network.serverpackets.PartySpelled;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.skills.Env;
import L2jFT.game.skills.Stats;
import L2jFT.game.skills.effects.EffectTemplate;
import L2jFT.game.skills.effects.ResistType;
import L2jFT.game.skills.funcs.Func;
import L2jFT.game.skills.funcs.FuncTemplate;
import L2jFT.game.skills.funcs.Lambda;

import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

public abstract class L2Effect implements Comparable
{
	protected static final Logger _log = Logger.getLogger(L2Effect.class.getName());

	public static enum EffectState
	{
		CREATED,
		ACTING,
		FINISHING
	}

	public static enum EffectType
	{
		BUFF,
		DMG_OVER_TIME,
		DMG_OVER_TIME_LETHAL,
		HEAL_OVER_TIME,
		COMBAT_POINT_HEAL_OVER_TIME,
		MANA_DMG_OVER_TIME,
		MANA_HEAL_OVER_TIME,
		RELAXING,
		STUN,
		ROOT,
		SLEEP,
		HATE,
		FAKE_DEATH,
		UNAGGRO,
		CONFUSE_MOB_ONLY,
		MUTE,
		MUTE_PHYSICAL,
		MUTE_ALL,
		FEAR,
		SILENT_MOVE,
		SEED,
		PARALYZE,
		BLOCK_STAT_FUNCS,
		BLOCK_MAGIC_PHYSIC,
		MANAHEAL,
		HEAL,
		MANAHEAL_PERCENT,
		HEAL_PERCENT,
		LETHALSTRIKE1,
		LETHALSTRIKE2,
		TURNER,
		PETRIFICATION,
		BETRAY,
		WATCHER_GAZE,
		BLESS_NOBLESSE,
		MEDITATION,
		CIRCLE,
		BIG_HEAD,
		INTERRUPT,
		SALVATION,
		BUFFIMMUNITY,
		BATTLE_FORCE,
		SPELL_FORCE
	}

	private static final Func[] _emptyFunctionSet = new Func[0];

	protected final L2Character _effector;
	protected final L2Character _effected;

	protected final L2Skill _skill;

	//or the items that was used.
	//private final L2Item _item;

	// the value of an update
	private final Lambda _lambda;

	// the current state
	protected EffectState _state;

	// period, milliseconds
	private long _period;
	private long _periodStartTime;

	// function templates
	private final FuncTemplate[] _funcTemplates;
	private final Stats[] _stats;

	private final ResistType _resistType;

	// counter
	protected int _count;

	// abnormal effect mask
	private int _abnormalEffect;

	public final class EffectTask implements Runnable
	{
		protected final long _delay;
		protected final long _rate;

		EffectTask(long delay, long rate)
		{
			_delay = delay;
			_rate = rate;
		}

		public long getRate()
		{
			return _rate;
		}

		public void run()
		{
			try
			{
				scheduleEffect();
			}
			catch(Throwable e)
			{
				_log.log(Level.SEVERE, "", e);
			}
		}
	}

	private ScheduledFuture _currentFuture;
	private EffectTask _currentTask;

	/** The Identifier of the stack group */
	private final String _stackType;

	/** The position of the effect in the stack group */
	private final int _stackOrder;

	private boolean _inUse = false;

	private boolean _skillMastery = false;

	protected L2Effect(Env env, EffectTemplate template)
	{
		_state = EffectState.CREATED;
		_skill = env.skill;
		_effector = env.character;
		_effected = env.target;
		_lambda = template._lambda;
		_funcTemplates = template._funcTemplates;
		_stats = template._stats;
		_count = template._counter;
		_period = template.getPeriod();

		//Check for skill mastery duration time increase
		if(_effector.getSkillMastery(getSkill().getId()) == 2)
		{
			if(_count > 1)
				_count *= 2;
			else
				_period *= 2;
			_skillMastery = true;
			_effector.removeSkillMastery(getSkill().getId());
		}

		_abnormalEffect = template._abnormalEffect;
		_stackType = template._stackType;
		_stackOrder = template._stackOrder;
		_resistType = template._resistType;
		_periodStartTime = System.currentTimeMillis();
	}

	public long getPeriod()
	{
		return _period;
	}

	public void setPeriod(long time)
	{
		_period = time;
	}

	public int getCount()
	{
		return _count;
	}

	public void setCount(int newcount)
	{
		_count = newcount;
	}

	public long getTime()
	{
		return System.currentTimeMillis() - _periodStartTime;
	}

	public long getPeriodStartTime()
	{
		return _periodStartTime;
	}

	public boolean getInUse()
	{
		return _inUse;
	}

	public void setInUse(boolean inUse)
	{
		_inUse = inUse;
		if(_inUse)
			scheduleEffect();
		else
			_state = EffectState.FINISHING;
	}

	public String getStackType()
	{
		return _stackType;
	}

	public int getStackOrder()
	{
		return _stackOrder;
	}

	public final L2Skill getSkill()
	{
		return _skill;
	}

	public final L2Character getEffector()
	{
		return _effector;
	}

	public final L2Character getEffected()
	{
		return _effected;
	}

	public final double calc()
	{
		Env env = new Env();
		env.character = _effector;
		env.target = _effected;
		env.skill = _skill;
		return _lambda.calc(env);
	}

	private synchronized void startEffectTask(long duration)
	{
		stopEffectTask();
		_currentTask = new EffectTask(duration, -1);
		_currentFuture = ThreadPoolManager.getInstance().scheduleEffect(_currentTask, duration);
		updateEffects();
	}

	private synchronized void startEffectTaskAtFixedRate(long delay, long rate)
	{
		stopEffectTask();
		_currentTask = new EffectTask(delay, rate);
		_currentFuture = ThreadPoolManager.getInstance().scheduleEffectAtFixedRate(_currentTask, delay, rate);
		updateEffects();
	}

	/**
	 * Stop the L2Effect task and send Server->Client update packet.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Cancel the effect in the the abnormal effect map of the L2Character </li>
	 * <li>Stop the task of the L2Effect, remove it and update client magic icon </li><BR><BR>
	 *
	 */
	public final void exit()
	{
		if(_state != EffectState.CREATED)
		{
			_state = EffectState.FINISHING;
			scheduleEffect();
		}
		else
			_state = EffectState.FINISHING;
	}

	/**
	 * Stop the task of the L2Effect, remove it and update client magic icon.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Cancel the task </li>
	 * <li>Stop and remove L2Effect from L2Character and update client magic icon </li><BR><BR>
	 *
	 */
	private synchronized void stopEffectTask()
	{
		if(_currentFuture != null)
		{
			// Cancel the task
			_currentFuture.cancel(false);
			_currentFuture = null;
			_currentTask = null;

			_effected.removeEffect(this);
			updateEffects();
		}
	}

	/** returns effect type */
	public abstract EffectType getEffectType();

	/** Notify started */
	public void onStart()
	{
		if(_abnormalEffect != 0)
			getEffected().startAbnormalEffect(_abnormalEffect);
	}

	/**
	 * Cancel the effect in the the abnormal effect map of the effected L2Character.<BR><BR>
	 */
	public void onExit()
	{
		if(_abnormalEffect != 0)
			getEffected().stopAbnormalEffect(_abnormalEffect);
	}

	/** Return true for continuation of this effect */
	public abstract boolean onActionTime();

	public final void rescheduleEffectOnePeriod()
	{
		if(_state != EffectState.ACTING)
			scheduleEffect();
		else
		{
			_currentFuture.cancel(false);
			_currentTask = new EffectTask(getPeriod(), -1);
			_currentFuture = ThreadPoolManager.getInstance().scheduleGeneral(_currentTask, getPeriod());
			_effected.updateEffectIcons();
		}
	}

	public final void scheduleEffect()
	{
		if(_state == EffectState.CREATED)
		{
			_state = EffectState.ACTING;
			onStart();

			if(_skill.getId() != 60) // Fake Death не отображается
				getEffected().sendPacket(new SystemMessage(SystemMessage.YOU_CAN_FEEL_S1S_EFFECT).addSkillName(_skill.getId(), _skill.getLevel()));

			_effected.updateStats(); // Обрабатываем отображение статов

			if(_count > 1)
				startEffectTaskAtFixedRate(_skill.isToggle() ? _period : 5, _period);
			else
				startEffectTask(_period);
			return;
		}

		if(_state == EffectState.ACTING)
		{
			if(_count > 0)
			{
				_count--;
				if(onActionTime())
					return;
			}
			_state = EffectState.FINISHING;
		}

		if(_state == EffectState.FINISHING)
		{
			// Cancel the effect in the the abnormal effect map of the L2Character
			onExit();

			//If the time left is equal to zero, send the message
			if(_count == 0)
				getEffected().sendPacket(new SystemMessage(SystemMessage.THE_EFFECT_OF_S1_HAS_WORN_OFF).addSkillName(_skill.getId(), _skill.getLevel()));
			// Stop the task of the L2Effect, remove it and update client magic icon
			stopEffectTask();
		}
	}

	public void updateEffects()
	{
		_effected.updateStats();
	}

	public Func[] getStatFuncs()
	{
		if(_funcTemplates == null)
			return _emptyFunctionSet;
		Func[] funcs = new Func[_funcTemplates.length];
		for(int i = 0; i < funcs.length; i++)
		{
			Env env = new Env();
			env.character = _effector;
			env.target = _effected;
			env.skill = _skill;
			Func f = _funcTemplates[i].getFunc(env, this); // effect is owner 
			funcs[i] = f;
		}
		return funcs;
	}

	public final void addIcon(MagicEffectIcons mi)
	{
		if(_currentTask == null || _skill == null)
			return;
		if(_state != EffectState.ACTING)
			return;
		if(_currentTask.getRate() > 0 || _skill.isToggle())
			mi.addEffect(_skill.getDisplayId(), _skill.getDisplayLevel(), MagicEffectIcons.INFINITIVE_EFFECT);
		else if(_currentFuture != null)
			try
			{
				mi.addEffect(_skill.getDisplayId(), _skill.getDisplayLevel(), (int) _currentFuture.getDelay(TimeUnit.SECONDS));
			}
			catch(NullPointerException e)
			{}
	}

	public final void addPartySpelledIcon(PartySpelled ps)
	{
		if(_currentTask == null || _currentFuture == null)
			return;
		if(_state != EffectState.ACTING)
			return;
		int duration = _skill.isToggle() ? MagicEffectIcons.INFINITIVE_EFFECT : (int) _currentFuture.getDelay(TimeUnit.SECONDS);
		ps.addPartySpelledEffect(_skill.getDisplayId(), _skill.getDisplayLevel(), duration);
	}

	protected int getLevel()
	{
		return _skill.getLevel();
	}

	public boolean containsStat(Stats stat)
	{
		if(_stats != null)
			for(Stats s : _stats)
				if(s == stat)
					return true;
		return false;
	}

	public final ResistType getResistType()
	{
		return _resistType;
	}

	public final void addOlympiadSpelledIcon(L2Player player, ExOlympiadSpelledInfo os)
	{
		EffectTask task = _currentTask;
		ScheduledFuture future = _currentFuture;
		if(task == null || future == null)
			return;
		if(_state != EffectState.ACTING)
			return;
		L2Skill sk = getSkill();
		os.addSpellRecivedPlayer(player);
		os.addEffect(sk.getId(), getLevel(), (int) future.getDelay(TimeUnit.SECONDS));
	}

	public void finishEffect()
	{
		_state = EffectState.FINISHING;
	}

	public boolean isSkillMasteryEffect()
	{
		return _skillMastery;
	}

	public int compareTo(Object obj)
	{
		if(obj.equals(this))
			return 0;
		return 1;
	}
}
