package L2jFT.game.model;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import javolution.util.FastList;
import L2jFT.Config;
import L2jFT.extensions.listeners.MethodInvokeListener;
import L2jFT.extensions.listeners.PropertyChangeListener;
import L2jFT.extensions.listeners.engine.DefaultListenerEngine;
import L2jFT.extensions.listeners.engine.ListenerEngine;
import L2jFT.extensions.listeners.events.L2Object.TerritoryChangeEvent;
import L2jFT.extensions.listeners.events.MethodEvent;
import L2jFT.extensions.listeners.events.PropertyEvent;
import L2jFT.extensions.scripts.Events;
import L2jFT.extensions.scripts.Script;
import L2jFT.extensions.scripts.ScriptObject;
import L2jFT.extensions.scripts.Scripts;
import L2jFT.game.ai.L2CharacterAI;
import L2jFT.game.geodata.GeoData;
import L2jFT.game.idfactory.IdFactory;
import L2jFT.game.instancemanager.MercTicketManager;
import L2jFT.game.model.L2Zone.ZoneType;
import L2jFT.game.model.instances.L2BoatInstance;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.model.instances.L2NpcInstance;
import L2jFT.game.network.serverpackets.StopMove;
import L2jFT.game.tables.MapRegionTable;
import L2jFT.game.tables.NpcTable;
import L2jFT.game.tables.TerritoryTable;

public abstract class L2Object
{
	private static final Logger _log = Logger.getLogger(L2Object.class.getName());

	public boolean isPlayer;
	public boolean isPet;
	public boolean isSummon;
	public boolean isMonster;
	public boolean isNpc;
	public boolean isCharacter;
	public boolean isRaid;

	/** Object identifier */
	protected int _objectId;
	protected Integer _shortObjectId = null;

	/** Object localisation : Used for items/chars that are seen in the world */
	protected int _x;
	protected int _y;
	protected int _z;

	protected int _l;

	private int _polyid;
	private String _polytype;
	private int _polyRadius;
	private int _polyHeight;

	protected FastList<L2Territory> _territories = null;

	/** Object visibility */
	protected boolean _hidden;

	/**
	 * Constructor of L2Object.<BR><BR>
	 * @param objectId ид этого объекта
	 */
	public L2Object(Integer objectId)
	{
		_objectId = objectId;
	}

	public Object callScripts(String _class, String method, Object[] args)
	{
		return callScripts(_class, method, args, null);
	}

	public Object callScripts(String _class, String method, Object[] args, HashMap<String, Object> variables)
	{
		if(L2jFT.extensions.scripts.Scripts.loading)
			return null;

		ScriptObject o;

		Script scriptClass = Scripts.getInstance().getClasses().get(_class);

		if(scriptClass == null)
		{
			_log.info("Script class " + _class + " not found");
			return null;
		}

		try
		{
			o = scriptClass.newInstance();
		}
		catch(Exception e)
		{
			e.printStackTrace();
			return null;
		}

		if(variables != null)
			for(Map.Entry<String, Object> obj : variables.entrySet())
				try
				{
					o.setProperty(obj.getKey(), obj.getValue());
				}
				catch(Exception e)
				{}

		try
		{
			o.setProperty("self", this);
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}

		return o.invokeMethod(method, args);
	}

	/**
	 * Return the identifier of the L2Object.<BR><BR>
	 *
	 * @ - deprecated?
	 */
	@Override
	public final int hashCode()
	{
		return _objectId;
	}

	/**
	 * @return the identifier of the L2Object.<BR><BR>
	 */
	public final int getObjectId()
	{
		return _objectId;
	}

	public final Integer getShortObjectId()
	{
		return _shortObjectId;
	}

	public final void setShortObjectId(Integer soid)
	{
		_shortObjectId = soid;
	}

	/**
	 * @return the x position of the L2Object.<BR><BR>
	 */
	public final int getX()
	{
		return _x;
	}

	/**
	 * @return the y position of the L2Object.<BR><BR>
	 */
	public final int getY()
	{
		return _y;
	}

	/**
	 * @return the z position of the L2Object.<BR><BR>
	 */
	public int getZ()
	{
		return _z;
	}

	public int getGeoZ(int x, int y, int z)
	{
		if(isPlayer)
		{
			L2Player activechar = (L2Player) this;
			if(activechar.isFlying() || activechar.isInWater() || activechar.isInBoat())
				return z;
			return GeoData.getInstance().getHeight(x, y, z);
		}
		else if(isNpc)
		{
			L2Spawn spawn = ((L2NpcInstance) this).getSpawn();
			if(spawn != null && spawn.getLocx() == 0 && spawn.getLocy() == 0)
				return GeoData.getInstance().getHeight(x, y, z);
			return z;
		}
		else if(this instanceof L2BoatInstance)
			return z;
		return GeoData.getInstance().getHeight(x, y, z);
	}

	public void setPolyInfo(String polytype, int polyid)
	{
		_polytype = polytype;
		_polyid = polyid;
		_polyRadius = _polyid > 0 ? NpcTable.getTemplate(_polyid).collisionRadius : 0;
		_polyHeight = _polyid > 0 ? NpcTable.getTemplate(_polyid).collisionHeight : 0;
	}

	public void setPolyInfo(String polytype, String polyid)
	{
		_polytype = polytype;
		_polyid = Integer.parseInt(polyid);
		_polyRadius = _polyid > 0 ? NpcTable.getTemplate(_polyid).collisionRadius : 0;
		_polyHeight = _polyid > 0 ? NpcTable.getTemplate(_polyid).collisionHeight : 0;
	}

	/**
	 * Set the x,y,z position of the L2Object and if necessary modify its _worldRegion.<BR><BR>
	 *
	 * <B><U> Example of use </U> :</B><BR><BR>
	 * <li> Update position during and after movement, or after teleport </li><BR>
	 *
	 * @param x new x coord
	 * @param y new y coord
	 * @param z new z coord
	 */
	public void setXYZ(int x, int y, int z)
	{
		if(!L2World.validCoords(x, y))
			if(isPlayer)
			{
				_log.warning("Player " + this + " (" + _objectId + ") at bad coords: (" + _x + ", " + _y + ").");
				L2Player player = (L2Player) this;
				player.abortAttack();
				player.abortCast();
				player.sendActionFailed();
				player.broadcastPacket(new StopMove((L2Player) this));
				player.teleToLocation(MapRegionTable.getInstance().getTeleToLocation(((L2Player) this), MapRegionTable.TeleportWhereType.ClosestTown));
				return;
			}
			else if(this instanceof L2NpcInstance)
			{
				if(Config.DEBUG)
					_log.warning("Object " + this + " (" + _objectId + ") at bad coords: (" + _x + ", " + _y + ").");
				L2Spawn spawn = ((L2NpcInstance) this).getSpawn();
				if(spawn == null)
					return;
				if(spawn.getLocx() != 0)
				{
					x = spawn.getLocx();
					y = spawn.getLocy();
					z = spawn.getLocz();
				}
				else
				{
					int p[] = TerritoryTable.getInstance().getRandomPoint(spawn.getLocation());
					x = p[0];
					y = p[1];
					z = p[2];
				}
			}
			else if(isCharacter)
			{
				decayMe();
				return;
			}

		_x = x;
		_y = y;
		_z = z;

		L2World.addVisibleObject(this, null);
		updateTerritories();
	}

	public void updateTerritories()
	{
		//FastList<L2Territory> current_territories = L2World.getActiveTerritory(_x, _y);
		FastList<L2Territory> current_territories = L2World.getTerritory(_x, _y);
		FastList<L2Territory> new_territories = new FastList<L2Territory>();
		FastList<L2Territory> old_territories = new FastList<L2Territory>();

		if(_territories == null)
			new_territories = current_territories;
		else
		{
			for(FastList.Node<L2Territory> n = current_territories.head(), end = current_territories.tail(); (n = n.getNext()) != end;)
			{
				L2Territory terr = n.getValue();
				if(!_territories.contains(terr))
					new_territories.add(terr);
			}

			if(!_territories.isEmpty())
				for(FastList.Node<L2Territory> n = _territories.head(), end = _territories.tail(); (n = n.getNext()) != end;)
				{
					L2Territory terr = n.getValue();
					if(!current_territories.contains(terr))
						old_territories.add(terr);
				}
		}

		if(current_territories.size() > 0)
			_territories = current_territories;
		else
			_territories = null;

		for(FastList.Node<L2Territory> n = old_territories.head(), end = old_territories.tail(); (n = n.getNext()) != end;)
			n.getValue().doLeave(this);

		for(FastList.Node<L2Territory> n = new_territories.head(), end = new_territories.tail(); (n = n.getNext()) != end;)
			n.getValue().doEnter(this);

		firePropertyChanged(new TerritoryChangeEvent(old_territories, new_territories, this));
	}

	/**
	 * Set the x,y,z position of the L2Object and make it invisible.<BR><BR>
	 *
	 * <B><U> Concept</U> :</B><BR><BR>
	 * A L2Object is invisble if <B>_hidden</B> = true<BR><BR>
	 *
	 * <B><U> Example of use </U> :</B><BR><BR>
	 * <li> Create a Door</li>
	 * <li> Restore L2Player</li><BR>
	 *
	 * @param x new x coord
	 * @param y new y coord
	 * @param z new z coord
	 */
	public final void setXYZInvisible(int x, int y, int z)
	{
		if(x > L2World.MAP_MAX_X)
			x = L2World.MAP_MAX_X - 5000;
		if(x < L2World.MAP_MIN_X)
			x = L2World.MAP_MIN_X + 5000;
		if(y > L2World.MAP_MAX_Y)
			y = L2World.MAP_MAX_Y - 5000;
		if(y < L2World.MAP_MIN_Y)
			y = L2World.MAP_MIN_Y + 5000;

		if(z < -16000 || z > 16000)
			z = 16000;

		_x = x;
		_y = y;
		_z = z;

		_hidden = true;

		updateTerritories();
	}

	public boolean isPolymorphed()
	{
		return _polytype != null;
	}

	public String getPolytype()
	{
		return _polytype;
	}

	public int getPolyid()
	{
		return _polyid;
	}

	public int getPolyRadius()
	{
		return _polyRadius;
	}

	public int getPolyHeight()
	{
		return _polyHeight;
	}

	/**
	 * Return the visibilty state of the L2Object. <BR><BR>
	 *
	 * <B><U> Concept</U> :</B><BR><BR>
	 * A L2Object is invisble if <B>_hidden</B>=true or <B>_worldregion</B>==null <BR><BR>
	 *
	 * @return true if visible
	 */
	public final boolean isVisible()
	{
		return !_hidden;
	}

	/**
	 * Init a dropped L2ItemInstance and add it in the world as a visible object.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Set the x,y,z position of the L2ItemInstance dropped and update its _worldregion </li>
	 * <li>Add the L2ItemInstance dropped to _visibleObjects of its L2WorldRegion</li>
	 * <li>Add the L2ItemInstance dropped in the world as a <B>visible</B> object</li><BR><BR>
	 *
	 * <FONT COLOR=#FF0000><B> <U>Caution</U> : This method DOESN'T ADD the object to _allObjects of L2World </B></FONT><BR><BR>
	 *
	 * <B><U> Assert </U> :</B><BR><BR>
	 * <li> this instanceof L2ItemInstance</li>
	 * <li> _worldRegion == null <I>(L2Object is invisible at the beginning)</I></li><BR><BR>
	 *
	 * <B><U> Example of use </U> :</B><BR><BR>
	 * <li> Drop item</li>
	 * <li> Call Pet</li><BR>
	 *
	 * @param dropper Char that dropped item
	 * @param x drop x coord
	 * @param y drop y coord
	 * @param z drop z coord
	 */
	public final void dropMe(L2Character dropper, int x, int y, int z)
	{
		// Set the x,y,z position of the L2ItemInstance dropped and update its _worldregion
		_hidden = false;

		_x = x;
		_y = y;
		_z = getGeoZ(x, y, z);

		updateTerritories();

		L2World.addVisibleObject(this, dropper);
	}

	public final void spawnMe(int x, int y, int z)
	{
		if(x > L2World.MAP_MAX_X)
			x = L2World.MAP_MAX_X - 5000;
		if(x < L2World.MAP_MIN_X)
			x = L2World.MAP_MIN_X + 5000;
		if(y > L2World.MAP_MAX_Y)
			y = L2World.MAP_MAX_Y - 5000;
		if(y < L2World.MAP_MIN_Y)
			y = L2World.MAP_MIN_Y + 5000;

		_x = x;
		_y = y;
		_z = getGeoZ(x, y, z);

		updateTerritories();

		spawnMe();
	}

	/**
	 * Добавляет обьект в мир, добавляет в текущий регион. Делает обьект видимым.
	 */
	public void spawnMe()
	{
		// Set the x,y,z position of the L2Object spawn and update its _worldregion
		_hidden = false;

		// Add the L2Oject spawn in the _allobjects of L2World
		L2World.addObject(this);
		L2World.addVisibleObject(this, null);
	}

	public void toggleVisible()
	{
		if(isVisible())
			decayMe();
		else
			spawnMe();
	}

	/**
	 * Do Nothing.<BR><BR>
	 *
	 * <B><U> Overriden in </U> :</B><BR><BR>
	 * <li> L2GuardInstance :  Set the home location of its L2GuardInstance </li>
	 * <li> L2NpcInstance    :  Reset the Spoiled flag </li><BR><BR>
	 *
	 */
	public void onSpawn()
	{}

	/**
	 * Remove a L2ItemInstance from the world and send server->client GetItem packets.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Send a Server->Client Packet GetItem to player that pick up and its _knowPlayers member </li>
	 * <li>Remove the L2Object from the world</li><BR><BR>
	 *
	 * <FONT COLOR=#FF0000><B> <U>Caution</U> : This method DOESN'T REMOVE the object from _allObjects of L2World </B></FONT><BR><BR>
	 *
	 * <B><U> Assert </U> :</B><BR><BR>
	 * <li> this instanceof L2ItemInstance</li>
	 * <li> _worldRegion != null <I>(L2Object is visible at the beginning)</I></li><BR><BR>
	 *
	 * <B><U> Example of use </U> :</B><BR><BR>
	 * <li> Do Pickup Item : PCInstance and Pet</li><BR><BR>
	 *
	 * @param player Player that pick up the item
	 */
	public final void pickupMe(L2Character player)
	{
		// Create a server->client GetItem packet to pick up the L2ItemInstance
		//player.broadcastPacket(new GetItem((L2ItemInstance) this, player.getObjectId()));

		// if this item is a mercenary ticket, remove the spawns!
		if(this instanceof L2ItemInstance)
		{
			int itemId = ((L2ItemInstance) this).getItemId();
			/*if(itemId >= 3960 && itemId <= 3972 // Gludio
			 || itemId >= 3973 && itemId <= 3985 // Dion
			 || itemId >= 3986 && itemId <= 3998 // Giran
			 || itemId >= 3999 && itemId <= 4011 // Oren
			 || itemId >= 4012 && itemId <= 4026 // Aden
			 || itemId >= 5205 && itemId <= 5215 // Innadril
			 || itemId >= 6779 && itemId <= 6833 // Goddard
			 || itemId >= 7973 && itemId <= 8029 // Rune
			 || itemId >= 7918 && itemId <= 7972 // Schuttgart
			 )*/
			if(itemId >= 3960 && itemId <= 4026 || itemId >= 5205 && itemId <= 5214 || itemId >= 6038 && itemId <= 6306 || itemId >= 6779 && itemId <= 6833 || itemId >= 7918 && itemId <= 8029)
				MercTicketManager.getInstance().removeTicket((L2ItemInstance) this);
		}

		// this can synchronize on others instancies, so it's out of
		// synchronized, to avoid deadlocks
		// Remove the L2ItemInstance from the world
		_hidden = true;
		L2World.removeVisibleObject(this);
	}

	/**
	 * Удаляет обьект из текущего региона, делая его невидимым.
	 */
	public void decayMe()
	{
		_hidden = true;
		L2World.removeVisibleObject(this);
	}

	public void onAction(L2Player player)
	{
		if(Events.onAction(player, this))
			return;

		player.sendActionFailed();
	}

	public void onActionShift(L2Player player)
	{
		if(Events.onActionShift(player, this))
			return;

		player.sendActionFailed();
	}

	public void onForcedAttack(L2Player player)
	{
		player.sendActionFailed();
	}

	public boolean isAttackable()
	{
		return false;
	}

	public abstract boolean isAutoAttackable(L2Character attacker);

	public boolean isMarker()
	{
		return false;
	}

	public void refreshID()
	{
		int newObjectId = IdFactory.getInstance().getNextId();
		int oldObjectId = getObjectId();
		L2World.removeObject(this);
		_objectId = newObjectId;
		IdFactory.getInstance().releaseId(oldObjectId);
	}

	public String getL2ClassShortName()
	{
		return getClass().getName().replaceAll("^.*\\.(.*?)$", "$1");
	}

	public final double getDistance(int x, int y)
	{
		double dx = x - getX();
		double dy = y - getY();
		return Math.sqrt(dx * dx + dy * dy);
	}

	public final double getDistance(int x, int y, int z)
	{
		double dx = x - getX();
		double dy = y - getY();
		double dz = z - getZ();
		return Math.sqrt(dx * dx + dy * dy + dz * dz);
	}

	public final boolean isInRange(L2Object obj, int range)
	{
		if(obj == null)
			return false;
		long dx = Math.abs(obj.getX() - _x);
		if(dx > range)
			return false;
		long dy = Math.abs(obj.getY() - _y);
		if(dy > range)
			return false;
		long dz = Math.abs(obj.getZ() - _z);
		return dz <= 1500 && dx * dx + dy * dy <= range * range;
	}

	public final boolean isInRangeZ(L2Object obj, int range)
	{
		if(obj == null)
			return false;
		long dx = Math.abs(obj.getX() - _x);
		if(dx > range)
			return false;
		long dy = Math.abs(obj.getY() - _y);
		if(dy > range)
			return false;
		long dz = Math.abs(obj.getZ() - _z);
		return dz <= range && dx * dx + dy * dy + dz * dz <= range * range;
	}

	public final boolean isInRange(int x, int y, int range)
	{
		return (x - _x) * (x - _x) + (y - _y) * (y - _y) <= range * range;
	}

	public final boolean isInRangeSq(int x, int y, int range)
	{
		return (x - _x) * (x - _x) + (y - _y) * (y - _y) <= range;
	}

	public final boolean isInRangeZ(int x, int y, int z, int range)
	{
		return (x - _x) * (x - _x) + (y - _y) * (y - _y) + (z - _z) * (z - _z) <= range * range;
	}

	public final boolean isInRangeZSq(int x, int y, int z, int range)
	{
		return (x - _x) * (x - _x) + (y - _y) * (y - _y) + (z - _z) * (z - _z) <= range;
	}

	public final double getDistance(L2Object obj)
	{
		if(obj == null)
			return 0;
		double dx = obj.getX() - _x;
		double dy = obj.getY() - _y;
		return Math.sqrt(dx * dx + dy * dy);
	}

	public final double getRealDistance(L2Object obj)
	{
		if(obj == null)
			return 0;
		double dx = obj.getX() - _x;
		double dy = obj.getY() - _y;
		double distance = Math.sqrt(dx * dx + dy * dy);
		if(isCharacter)
			distance -= ((L2Character) this).getTemplate().collisionRadius;
		if(obj.isCharacter)
			distance -= ((L2Character) obj).getTemplate().collisionRadius;
		return distance;
	}

	public final double getDistance3D(L2Object obj)
	{
		if(obj == null)
			return 0;
		double dx = obj.getX() - getX();
		double dy = obj.getY() - getY();
		double dz = obj.getZ() - getZ();
		return Math.sqrt(dx * dx + dy * dy + dz * dz);
	}

	public final double getRealDistance3D(L2Object obj)
	{
		return getRealDistance3D(obj, false);
	}

	public final double getRealDistance3D(L2Object obj, boolean ignoreZ)
	{
		if(obj == null)
			return 0;

		double dx = obj.getX() - _x;
		double dy = obj.getY() - _y;
		double dz = obj.getZ() - _z;
		double distance = Math.sqrt(dx * dx + dy * dy + (ignoreZ ? 0 : dz * dz));
		if(isCharacter)
			distance -= ((L2Character) this).getTemplate().collisionRadius;
		if(obj.isCharacter)
			distance -= ((L2Character) obj).getTemplate().collisionRadius;
		return distance > 0 ? distance : 0;
	}

	public final long getSqDistance(int x, int y)
	{
		int dx = x - _x;
		int dy = y - _y;
		return dx * dx + dy * dy;
	}

	public final long getSqDistance(L2Object obj)
	{
		if(obj == null)
			return 0;

		long dx = obj.getX() - _x;
		long dy = obj.getY() - _y;

		return dx * dx + dy * dy;
	}

	public L2Player getPlayer()
	{
		if(isPlayer)
			return (L2Player) this;
		if(this instanceof L2Summon)
			return ((L2Summon) this).getOwner();
		return null;
	}

	/**
	 * Проверяет наличие объекта в мире
	 * @return true если оъект есть в мире
	 */
	public boolean isInWorld()
	{
		return L2World.findObject(_objectId) != null;
	}

	public int getHeading()
	{
		return 0;
	}

	public float getMoveSpeed()
	{
		return 0;
	}

	public boolean isInZonePeace()
	{
		return isInZone(ZoneType.peace_zone) && !isInZoneBattle();
	}

	public boolean isInZoneBattle()
	{
		return isInZone(ZoneType.battle_zone) || isInZone(ZoneType.OlympiadStadia);
	}

	public boolean isInZone(ZoneType type)
	{
		//FastList<L2Territory> territories = L2World.getActiveTerritory(_x, _y);
		if(_territories != null)
			for(L2Territory terr : _territories)
				if(terr.getZone() != null && terr.getZone().isActive() && terr.getZone().getType() == type)
					return true;
		return false;
	}

	public boolean isInZone(L2Zone zone)
	{
		//FastList<L2Territory> territories = L2World.getActiveTerritory(_x, _y);
		if(_territories != null)
			for(L2Territory terr : _territories)
				if(terr.getZone() != null && terr.getZone() == zone)
					return true;
		return false;
	}

	public boolean isInZoneIncludeZ(ZoneType type)
	{
		//FastList<L2Territory> territories = L2World.getActiveTerritory(_x, _y);
		if(_territories != null)
			for(L2Territory terr : _territories)
				if(terr.getZone() != null && terr.getZone().isActive() && terr.getZone().getType() == type && _z > terr.getZmin() && _z < terr.getZmax())
					return true;
		return false;
	}

	public L2Zone getZone(ZoneType type)
	{
		//FastList<L2Territory> territories = L2World.getActiveTerritory(_x, _y);
		if(_territories != null)
			for(L2Territory terr : _territories)
				if(terr.getZone() != null && terr.getZone().isActive() && terr.getZone().getType() == type)
					return terr.getZone();
		return null;
	}

	public boolean isActionBlocked(String action)
	{
		//FastList<L2Territory> territories = L2World.getActiveTerritory(_x, _y);
		if(_territories != null)
			for(L2Territory terr : _territories)
				if(terr.getZone() != null && terr.getZone().isActive() && terr.getZone().isActionBlocked(action))
					return true;
		return false;
	}

	public FastList<L2Territory> getTerritories()
	{
		return _territories;
	}

	L2WorldRegion _currentRegion;

	public L2WorldRegion getCurrentRegion()
	{
		return _currentRegion;
	}

	public void setCurrentRegion(L2WorldRegion region)
	{
		_currentRegion = region;
	}

	public L2CharacterAI getAI()
	{
		return null;
	}

	public boolean hasAI()
	{
		return false;
	}

	public boolean inObserverMode()
	{
		return false;
	}

	public void startAttackStanceTask()
	{}

	// --------------------------- Listeners system test -----------------------------
	private DefaultListenerEngine<L2Object> listenerEngine;

	public void addPropertyChangeListener(PropertyChangeListener listener)
	{
		getListenerEngine().addPropertyChangeListener(listener);
	}

	public void removePropertyChangeListener(PropertyChangeListener listener)
	{
		getListenerEngine().removePropertyChangeListener(listener);
	}

	public void addPropertyChangeListener(String value, PropertyChangeListener listener)
	{
		getListenerEngine().addPropertyChangeListener(value, listener);
	}

	public void removePropertyChangeListener(String value, PropertyChangeListener listener)
	{
		getListenerEngine().removePropertyChangeListener(value, listener);
	}

	public void firePropertyChanged(String value, Object oldValue, Object newValue)
	{
		getListenerEngine().firePropertyChanged(value, this, oldValue, newValue);
	}

	public void firePropertyChanged(PropertyEvent event)
	{
		getListenerEngine().firePropertyChanged(event);
	}

	public void addProperty(String property, Object value)
	{
		getListenerEngine().addProperty(property, value);
	}

	public Object getProperty(String property)
	{
		return getListenerEngine().getProperty(property);
	}

	public void addMethodInvokeListener(MethodInvokeListener listener)
	{
		getListenerEngine().addMethodInvokedListener(listener);
	}

	public void addMethodInvokeListener(String methodName, MethodInvokeListener listener)
	{
		getListenerEngine().addMethodInvokedListener(methodName, listener);
	}

	public void removeMethodInvokeListener(MethodInvokeListener listener)
	{
		getListenerEngine().removeMethodInvokedListener(listener);
	}

	public void removeMethodInvokeListener(String methodName, MethodInvokeListener listener)
	{
		getListenerEngine().removeMethodInvokedListener(methodName, listener);
	}

	public void fireMethodInvoked(MethodEvent event)
	{
		getListenerEngine().fireMethodInvoked(event);
	}

	public void fireMethodInvoked(String methodName, Object[] args)
	{
		getListenerEngine().fireMethodInvoked(methodName, this, args);
	}

	private ListenerEngine<L2Object> getListenerEngine()
	{
		if(listenerEngine == null)
			listenerEngine = new DefaultListenerEngine<L2Object>(this);
		return listenerEngine;
	}
	// ------------------------- Listeners system test end ---------------------------
}