package L2jFT.game.model;

import java.util.ArrayList;

import L2jFT.Config;
import L2jFT.game.model.base.Experience;
import L2jFT.game.model.entity.DimensionalRift;
import L2jFT.game.model.entity.SevenSignsFestival;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.model.instances.L2NpcInstance;
import L2jFT.game.model.instances.L2SummonInstance;
import L2jFT.game.network.serverpackets.ExMPCCClose;
import L2jFT.game.network.serverpackets.ExMPCCOpen;
import L2jFT.game.network.serverpackets.InventoryUpdate;
import L2jFT.game.network.serverpackets.L2GameServerPacket;
import L2jFT.game.network.serverpackets.PartySmallWindowAdd;
import L2jFT.game.network.serverpackets.PartySmallWindowAll;
import L2jFT.game.network.serverpackets.PartySmallWindowDelete;
import L2jFT.game.network.serverpackets.PartySmallWindowDeleteAll;
import L2jFT.game.network.serverpackets.PartySmallWindowUpdate;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.tables.ItemTable;
import L2jFT.util.Log;
import L2jFT.util.Rnd;

public class L2Party
{
	//private static Logger _log = Logger.getLogger(L2Party.class.getName());

	private final ArrayList<L2Player> _members = new ArrayList<L2Player>();
	private int _partyLvl = 0;
	private int _partySqLevelSum = 0;
	private int _itemDistribution = 0;
	private int _itemOrder = 0;
	private int _partyDiffLevels = 0;
	private DimensionalRift _dr;
	private L2CommandChannel _commandChannel;

	public static final int ITEM_LOOTER = 0;
	public static final int ITEM_RANDOM = 1;
	public static final int ITEM_RANDOM_SPOIL = 2;
	public static final int ITEM_ORDER = 3;
	public static final int ITEM_ORDER_SPOIL = 4;

	public float _rateExp;
	public float _rateSp;
	public float _rateDrop;
	public float _rateAdena;
	public float _rateSpoil;

	/**
	 * constructor ensures party has always one member - leader
	 * @param leader создатель парти
	 * @param itemDistribution режим распределения лута
	 */
	public L2Party(L2Player leader, int itemDistribution)
	{
		_itemDistribution = itemDistribution;
		_members.add(leader);
		_partyLvl = leader.getLevel();
		_partySqLevelSum = _partyLvl * _partyLvl;
	}

	/**
	 * @return number of party members
	 */
	public int getMemberCount()
	{
		return _members.size();
	}

	/**
	 * @return all party members
	 */
	public ArrayList<L2Player> getPartyMembers()
	{
		return _members;
	}

	public L2Player getRandomMember()
	{
		return _members.get(Rnd.get(_members.size()));
	}

	/**
	 * @return random member from party
	 */
	private L2Player getRandomMemberInRange(L2Player player, L2ItemInstance item, int range)
	{
		ArrayList<L2Player> ret = new ArrayList<L2Player>();

		for(L2Player member : _members)
			if(member.isInRange(player, range) && !member.isDead() && member.getInventory().validateCapacity(item) && member.getInventory().validateWeight(item))
				ret.add(member);

		if(ret.size() > 0)
			return ret.get(Rnd.get(ret.size()));

		return null;
	}

	/**
	 * @return next item looter
	 */
	private L2Player getNextLooterInRange(L2Player player, @SuppressWarnings("unused")
	L2ItemInstance item, int range)
	{
		while(true)
		{
			int looter = _itemOrder;
			_itemOrder++;
			if(_itemOrder > _members.size() - 1)
				_itemOrder = 0;

			L2Player ret = looter < _members.size() ? _members.get(looter) : player;

			if(ret.isInRange(player, range) && !ret.isDead())
				return ret;
		}
	}

	/**
	 * true if player is party leader
	 * @param player
	 * @return
	 */
	public boolean isLeader(L2Player player)
	{
		return _members.get(0).equals(player);
	}

	/**
	 * Returns the Object ID for the party leader to be used as a unique identifier of this party
	 * @return int
	 */
	public int getPartyLeaderOID()
	{
		return _members.get(0).getObjectId();
	}

	/**
	 * Возвращает лидера партии
	 * @return L2Player Лидер партии
	 */
	public L2Player getPartyLeader()
	{
		return _members.get(0);
	}

	/**
	 * Broadcasts packet to every party member 
	 * @param msg packet to broadcast
	 */
	public void broadcastToPartyMembers(L2GameServerPacket msg)
	{
		synchronized (_members)
		{
			for(L2Player member : _members)
				member.sendPacket(msg);
		}
	}

	/**
	 * Рассылает текстовое сообщение всем членам группы
	 * @param msg сообщение
	 */
	public void broadcastMessageToPartyMembers(String msg)
	{
		broadcastToPartyMembers(SystemMessage.sendString(msg));
	}

	/**
	 * Send a Server->Client packet to all other L2Player of the Party.<BR><BR>
	 */
	public void broadcastToPartyMembers(L2Player player, L2GameServerPacket msg)
	{
		synchronized (_members)
		{
			for(L2Player member : _members)
				if(!member.equals(player))
					member.sendPacket(msg);
		}
	}

	public void broadcastToPartyMembersInRange(L2Player player, L2GameServerPacket msg, int range)
	{
		synchronized (_members)
		{
			for(L2Player member : _members)
				if(player.isInRange(member, range))
					member.sendPacket(msg);
		}
	}

	public boolean containsMember(L2Character cha)
	{
		L2Player player = cha.getPlayer();
		if(player == null)
			return false;

		for(L2Player member : _members)
			if(member.getObjectId() == cha.getObjectId())
				return true;

		return false;
	}

	/**
	 * adds new member to party
	 * @param player L2Player to add
	 */
	public void addPartyMember(L2Player player)
	{
		//sends new member party window for all members
		//we do all actions before adding member to a list, this speeds things up a little
		PartySmallWindowAll window = new PartySmallWindowAll();
		window.setPartyList(_members);
		player.sendPacket(window);

		player.sendPacket(new SystemMessage(SystemMessage.YOU_HAVE_JOINED_S1S_PARTY).addString(_members.get(0).getName()));
		broadcastToPartyMembers(new SystemMessage(SystemMessage.S1_HAS_JOINED_THE_PARTY).addString(player.getName()));
		broadcastToPartyMembers(new PartySmallWindowAdd(player));

		synchronized (_members)
		{
			_members.add(player);
		}

		recalculatePartyData();

		player.updateEffectIcons();

		// Если партия уже в СС, то вновь прибывшем посылаем пакет открытия окна СС
		if(isInCommandChannel())
			player.sendPacket(new ExMPCCOpen());

		if(isInDimensionalRift())
			_dr.partyMemberInvited();
	}

	/**
	 * removes player from party
	 * @param player L2Player to remove
	 */
	public void removePartyMember(L2Player player)
	{
		if(player == null || !_members.contains(player))
			return;

		synchronized (_members)
		{
			_members.remove(player);
		}

		recalculatePartyData();

		if(player.isFestivalParticipant())
			SevenSignsFestival.getInstance().updateParticipants(player, this);

		// Отсылаемы вышедшему пакет закрытия СС
		if(isInCommandChannel())
			player.sendPacket(new ExMPCCClose());

		player.sendPacket(new SystemMessage(SystemMessage.YOU_HAVE_WITHDRAWN_FROM_THE_PARTY));
		player.sendPacket(new PartySmallWindowDeleteAll());
		player.setParty(null);

		broadcastToPartyMembers(new SystemMessage(SystemMessage.S1_HAS_LEFT_THE_PARTY).addString(player.getName()));
		broadcastToPartyMembers(new PartySmallWindowDelete(player));

		if(isInDimensionalRift())
			_dr.partyMemberExited(player);

		if(player.getDuel() != null)
			player.getDuel().onRemoveFromParty(player);

		if(_members.size() == 1)
		{
			L2Player lastMember = _members.get(0);

			if(lastMember.getDuel() != null)
				lastMember.getDuel().onRemoveFromParty(lastMember);

			// Если в партии остался 1 человек, то удаляем ее из СС
			if(isInCommandChannel())
				_commandChannel.removeParty(this);

			lastMember.setParty(null);
		}
	}

	/**
	 * Change party leader (used for string arguments)
	 * @param name имя нового лидера парти
	 */
	public void changePartyLeader(String name)
	{
		L2Player player = getPlayerByName(name);

		if(player == null)
			return;

		if(!_members.contains(player))
		{
			player.sendPacket(new SystemMessage(SystemMessage.YOU_CAN_TRANSFER_RIGHTS_ONLY_TO_ANOTHER_PARTY_MEMBER));
			return;
		}

		if(isLeader(player))
		{
			player.sendPacket(new SystemMessage(SystemMessage.YOU_CANNOT_TRANSFER_RIGHTS_TO_YOURSELF));
			return;
		}

		//Swap party members
		synchronized (_members)
		{
			L2Player temp;
			int p1 = _members.indexOf(player);
			temp = _members.get(0);
			_members.set(0, _members.get(p1));
			_members.set(p1, temp);
		}

		SystemMessage msg = new SystemMessage(SystemMessage.S1_HAS_BECOME_A_PARTY_LEADER);
		msg.addString(_members.get(0).getName());
		broadcastToPartyMembers(msg);
		broadcastToPartyMembers(new PartySmallWindowDeleteAll());
		broadcastToPartyMembers(new PartySmallWindowAll().setPartyList(_members));
		if(isInCommandChannel())
			_commandChannel.setChannelLeader(_members.get(0));
	}

	/**
	 * finds a player in the party by name
	 * @param name имя для поиска
	 * @return найденый L2Player или null если не найдено
	 */
	private L2Player getPlayerByName(String name)
	{
		for(L2Player pl : _members)
			if(pl.getName().equalsIgnoreCase(name))
				return pl;
		return null;
	}

	/**
	 * Oust player from party
	 * @param player L2Player которого выгоняют
	 */
	public void oustPartyMember(L2Player player)
	{
		if(!_members.contains(player))
			return;

		if(isLeader(player))
		{
			removePartyMember(player);
			if(_members.size() > 0)
			{
				SystemMessage msg = new SystemMessage(SystemMessage.S1_HAS_BECOME_A_PARTY_LEADER);
				msg.addString(_members.get(0).getName());
				broadcastToPartyMembers(msg);
				broadcastToPartyMembers(new PartySmallWindowUpdate(_members.get(0)));
			}
		}
		else
			removePartyMember(player);
	}

	/**
	 * Oust player from party Overloaded method that takes player's name as
	 * parameter
	 *
	 * @param name имя игрока для изгнания
	 */
	public void oustPartyMember(String name)
	{
		oustPartyMember(getPlayerByName(name));
	}

	/**
	 * distribute item(s) to party members
	 * @param player
	 * @param item
	 */
	public void distributeItem(L2Player player, L2ItemInstance item)
	{
		distributeItem(player, item, null);
	}

	public void distributeItem(L2Player player, L2ItemInstance item, L2NpcInstance fromNpc)
	{
		L2Player target = player;

		switch(_itemDistribution)
		{
			case ITEM_RANDOM:
			case ITEM_RANDOM_SPOIL:
				target = getRandomMemberInRange(player, item, 2700);
				break;
			case ITEM_ORDER:
			case ITEM_ORDER_SPOIL:
				target = getNextLooterInRange(player, item, 2700);
				break;
			case ITEM_LOOTER:
			default:
				target = player;
				break;
		}

		if(target == null)
		{
			item.dropToTheGround(player, fromNpc);
			return;
		}

		if(!target.getInventory().validateWeight(item))
		{
			target.sendActionFailed();
			target.sendPacket(new SystemMessage(SystemMessage.YOU_HAVE_EXCEEDED_THE_WEIGHT_LIMIT));
			item.dropToTheGround(target, fromNpc);
			return;
		}

		if(!target.getInventory().validateCapacity(item))
		{
			target.sendActionFailed();
			target.sendPacket(new SystemMessage(SystemMessage.YOUR_INVENTORY_IS_FULL));
			item.dropToTheGround(player, fromNpc);
			return;
		}

		if(item.getCount() == 1)
		{
			SystemMessage smsg;
			if(item.getEnchantLevel() > 0)
			{
				smsg = new SystemMessage(SystemMessage.YOU_HAVE_OBTAINED__S1S2);
				smsg.addNumber(item.getEnchantLevel());
				smsg.addItemName(item.getItemId());
			}
			else
			{
				smsg = new SystemMessage(SystemMessage.YOU_HAVE_OBTAINED_S1);
				smsg.addItemName(item.getItemId());
			}

			target.sendPacket(smsg);

			if(item.getEnchantLevel() > 0)
			{
				smsg = new SystemMessage(SystemMessage.S1_HAS_OBTAINED__S2S3);
				smsg.addString(target.getName());
				smsg.addNumber(item.getEnchantLevel());
				smsg.addItemName(item.getItemId());
			}
			else
			{
				smsg = new SystemMessage(SystemMessage.S1_HAS_OBTAINED_S2);
				smsg.addString(target.getName());
				smsg.addItemName(item.getItemId());
			}

			broadcastToPartyMembers(target, smsg);
		}
		else
		{
			SystemMessage smsg = new SystemMessage(SystemMessage.YOU_HAVE_OBTAINED_S2_S1);
			smsg.addItemName(item.getItemId());
			smsg.addNumber(item.getCount());
			target.sendPacket(smsg);

			smsg = new SystemMessage(SystemMessage.S1_HAS_OBTAINED_S3_S2);
			smsg.addString(target.getName());
			smsg.addItemName(item.getItemId());
			smsg.addNumber(item.getCount());
			broadcastToPartyMembers(target, smsg);
		}

		// Remove the L2ItemInstance from the world and send server->client GetItem packets
		item.pickupMe(target);

		L2ItemInstance item2 = target.getInventory().addItem(item);
		Log.LogItem(target, fromNpc, Log.GetItemInPaty, item2);

		InventoryUpdate iu = new InventoryUpdate();
		if(item2.getLastChange() == L2ItemInstance.ADDED)
			iu.addNewItem(item);
		else
			iu.addModifiedItem(item2);

		target.sendPacket(iu);
		target.sendChanges();
	}

	/**
	 * distribute adena to party members
	 * @param adena инстанс адены для распределения
	 */
	public void distributeAdena(L2ItemInstance adena, L2Player player)
	{
		distributeAdena(adena, null, player);
	}

	public void distributeAdena(L2ItemInstance adena, L2NpcInstance fromNpc, L2Player player)
	{
		SystemMessage smsg = new SystemMessage(SystemMessage.YOU_HAVE_OBTAINED_S1_ADENA);

		int totalAdena = adena.getCount();
		int _fromNpc;

		_fromNpc = fromNpc == null ? -1 : fromNpc.getObjectId();

		ArrayList<L2Player> _membersInRange = new ArrayList<L2Player>();

		if(adena.getCount() < _members.size())
			_membersInRange.add(player);
		else
			for(L2Player p : _members)
				if(p.equals(player) || player.isInRange(p, 2700) && !p.isDead())
					_membersInRange.add(p);

		int amount = totalAdena / _membersInRange.size();
		int ost = totalAdena % _membersInRange.size();

		for(L2Player member : _membersInRange)
		{
			smsg.addNumber(member.equals(player) ? amount + ost : amount);

			L2ItemInstance newAdena = ItemTable.getInstance().createItem(57, member.getObjectId(), _fromNpc, "L2Party.distributeAdena");
			newAdena.setCount(member.equals(player) ? amount + ost : amount);

			L2ItemInstance item2 = member.getInventory().addItem(newAdena);
			if(fromNpc == null)
				Log.LogItem(member, Log.GetItemInPaty, item2);
			else
				Log.LogItem(member, fromNpc, Log.GetItemInPaty, item2);

			InventoryUpdate iu = new InventoryUpdate();
			if(item2.getLastChange() == L2ItemInstance.ADDED)
				iu.addNewItem(adena);
			else
				iu.addModifiedItem(item2);
			member.sendPacket(smsg);
			member.sendPacket(iu);
		}
	}

	/**
	 * Distribute Experience and SP rewards to L2Player Party members in the known area of the last attacker.<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Get the L2Player owner of the L2SummonInstance (if necessary) </li>
	 * <li>Calculate the Experience and SP reward distribution rate </li>
	 * <li>Add Experience and SP to the L2Player </li><BR><BR>
	 *
	 * <FONT COLOR=#FF0000><B> <U>Caution</U> : This method DOESN'T GIVE rewards to L2PetInstance</B></FONT><BR><BR>
	 *
	 * @param xpReward The Experience reward to distribute
	 * @param spReward The SP reward to distribute
	 * @param rewardedMembers The list of L2Player to reward and LSummonInstance whose owner must be reward
	 * @param lastAttacker The L2Character that has killed the L2NpcInstance
	 *
	 */
	public void distributeXpAndSp(long xpReward, long spReward, ArrayList<L2Player> rewardedMembers, L2Character lastAttacker, byte targetLevel)
	{
		recalculatePartyData();

		if(Config.PARTY_EXP_CALCULATIONS == 2) //Experimental, must be tested
		{
			ArrayList<L2Player> mtr = new ArrayList<L2Player>();
			int minPartyLevel = -1;
			int maxPartyLevel = -1;
			int partyLvlSum = 0;

			for(L2Player member : rewardedMembers)
			{
				if(!lastAttacker.isInRange(member, 2700))
					continue;

				if(minPartyLevel == -1 || member.getLevel() < minPartyLevel)
					minPartyLevel = member.getLevel();

				if(maxPartyLevel == -1 || member.getLevel() > maxPartyLevel)
					maxPartyLevel = member.getLevel();

				partyLvlSum += member.getLevel();

				mtr.add(member);
			}

			double bonus = 1;
			if(maxPartyLevel - minPartyLevel < 20)
				bonus = getExpSpBonus(mtr.size());

			double XP = xpReward * bonus * Config.RATE_PARTY_XP;
			double SP = spReward * bonus * Config.RATE_PARTY_SP;

			for(L2Player member : mtr)
			{
				//double preCalculation = (mtr.size() * member.getLevel())/partyLvlSum;
				double preCalculation = mtr.size() * member.getLevel() * mtr.size() * member.getLevel() / partyLvlSum * partyLvlSum;
				double lvlPenalty = 1;

				if(member.getLevel() - 9 > targetLevel)
					lvlPenalty = Experience.penaltyModifier(member.getLevel() - targetLevel - Config.DIFF_LEVEL_PENALTY, 9);

				double memberXp = XP * preCalculation * lvlPenalty;
				double memberSp = SP * preCalculation * lvlPenalty;
				if(memberXp > xpReward)
					memberXp = xpReward;
				if(memberSp > spReward)
					memberSp = spReward;

				if(member.getPet() != null && member.getPet().isSummon)
				{
					float penalty = 1 - ((L2SummonInstance) member.getPet()).getExpPenalty();
					memberXp *= penalty;
					memberSp *= penalty;
				}

				member.addExpAndSp((int) memberXp, (int) memberSp);
			}
		}
		else if(Config.PARTY_EXP_CALCULATIONS == 1) //Old formula.
		{
			double bonus = getExpSpBonus(_members.size());
			double XP = xpReward * bonus * Config.RATE_PARTY_XP;
			double SP = spReward * bonus * Config.RATE_PARTY_SP;

			if(_partyDiffLevels > Config.PARTY_DIFF_LEVEL_PENALTY)
			{
				double mod = Experience.penaltyModifier(_partyDiffLevels - Config.PARTY_DIFF_LEVEL_PENALTY, 9);
				XP *= mod;
				SP *= mod;
			}

			for(L2Player member : rewardedMembers)
			{
				if(!lastAttacker.isInRange(member, 2700))
					continue;

				double sqLevel = member.getLevel() * member.getLevel();
				double preCalculation = sqLevel / _partySqLevelSum;

				double memberXp = XP * preCalculation;
				double memberSp = SP * preCalculation;
				if(memberXp > xpReward)
					memberXp = xpReward;
				if(memberSp > spReward)
					memberSp = spReward;

				if(member.getPet() != null && member.getPet().isSummon)
				{
					float penalty = 1 - ((L2SummonInstance) member.getPet()).getExpPenalty();
					memberXp *= penalty;
					memberSp *= penalty;
				}

				member.addExpAndSp((int) memberXp, (int) memberSp);
			}
		}

		recalculatePartyData();
	}

	/**
	 * refresh _partySqLevelSum
	 *
	 */
	public void recalculatePartyData()
	{
		_partySqLevelSum = 0;
		_partyLvl = 0;
		int min = 1000, max = 0;
		float rateExp = 0;
		float rateSp = 0;
		float rateDrop = 0;
		float rateAdena = 0;
		float rateSpoil = 0;
		byte count = 0;
		for(L2Player member : _members)
		{
			int level = member.getLevel();
			_partySqLevelSum += level * level;
			if(level > _partyLvl)
				_partyLvl = level;
			if(level < min)
				min = level;
			if(level > max)
				max = level;
			rateExp += member.getBonus().RATE_XP;
			rateSp += member.getBonus().RATE_SP;
			rateDrop += member.getBonus().RATE_DROP_ITEMS;
			rateAdena += member.getBonus().RATE_DROP_ADENA;
			rateSpoil += member.getBonus().RATE_DROP_SPOIL;
			count++;
		}
		_rateExp = rateExp / count;
		_rateSp = rateSp / count;
		_rateDrop = rateDrop / count;
		_rateAdena = rateAdena / count;
		_rateSpoil = rateSpoil / count;
		_partyDiffLevels = max - min;
	}

	private static final double[] bonuses = {
			1.00, 1.00, 1.30, 1.39, 1.50, 1.54, 1.58, 1.63, 1.67, 1.71 };

	private double getExpSpBonus(int size)
	{
		return bonuses[size];
	}

	public int getLevel()
	{
		return _partyLvl;
	}

	public int getLootDistribution()
	{
		return _itemDistribution;
	}

	public boolean isDistributeSpoilLoot()
	{
		boolean rv = false;

		if(_itemDistribution == ITEM_RANDOM_SPOIL || _itemDistribution == ITEM_ORDER_SPOIL)
			rv = true;

		return rv;
	}

	public boolean isInDimensionalRift()
	{
		return _dr != null;
	}

	public void setDimensionalRift(DimensionalRift dr)
	{
		_dr = dr;
	}

	public DimensionalRift getDimensionalRift()
	{
		return _dr;
	}

	public boolean isInCommandChannel()
	{
		return _commandChannel != null;
	}

	public L2CommandChannel getCommandChannel()
	{
		return _commandChannel;
	}

	public void setCommandChannel(L2CommandChannel channel)
	{
		_commandChannel = channel;
	}
}