package L2jFT.game.model;

import L2jFT.Config;
import L2jFT.game.ThreadPoolManager;
import L2jFT.game.ai.DefaultAI;
import L2jFT.game.model.instances.L2NpcInstance;
import L2jFT.game.tables.SpawnTable;

import java.util.Enumeration;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledFuture;
import java.util.logging.Logger;

/**
 * @author Diamond
 * @Date: 15/5/2007
 * @Time: 10:06:34
 */
public final class L2WorldRegion
{
	private static Logger _log = Logger.getLogger(L2WorldRegion.class.getName());

	ConcurrentHashMap<Integer, L2Object> _objects = null;
	ConcurrentHashMap<Integer, L2Character> _characters = null;
	ConcurrentHashMap<Integer, L2NpcInstance> _npcs = null;
	ConcurrentHashMap<Integer, L2Player> _players = null;
	ConcurrentHashMap<Integer, L2Playable> _playables = null;

	//Можно добавить...
	//FastList<L2MonsterInstance> _monsters = null;
	//FastList<L2NpcInstance> _folks = null;
	//FastList<L2Summon> _summons = null;

	private int _playersSize = 0;

	private int tileX, tileY, tileZ;
	private Boolean _active = false;
	private ScheduledFuture _neighborsTask = null;

	//private FastList<L2WorldRegion> _neighbors = new FastList<L2WorldRegion>();

	public L2WorldRegion(int pTileX, int pTileY, int pTileZ)
	{
		tileX = pTileX;
		tileY = pTileY;
		tileZ = pTileZ;
	}

	public class NeighborsTask implements Runnable
	{
		private boolean _isActivating;

		public NeighborsTask(boolean isActivating)
		{
			_isActivating = isActivating;
		}

		public void run()
		{
			if(_isActivating)
			{
				for(L2WorldRegion neighbor : getNeighbors())
					if(neighbor != null)
						neighbor.setActive(true);
			}
			else
				for(L2WorldRegion neighbor : getNeighbors())
					if(neighbor != null && neighbor.areNeighborsEmpty())
						neighbor.setActive(false);
		}
	}

	private synchronized void switchAI(Boolean isOn)
	{
		if(!isOn)
		{
			for(L2Character cha : getCharacters().values())
				if(cha.hasAI() && !cha.getAI().isGlobalAI())
				{
					cha.setTarget(null);
					cha.stopMove(null);
					cha.stopAllEffects();
					cha.clearAggroList();
					cha.getAI().stopAITask();
				}
		}
		else
			for(L2Character cha : getCharacters().values())
				if(cha.getAI() instanceof DefaultAI)
					cha.getAI().startAITask();
	}

	public Boolean isActive()
	{
		return _active;
	}

	public Boolean isEmpty()
	{
		return _playersSize == 0;
	}

	public Boolean areNeighborsEmpty()
	{
		if(!isEmpty())
			return false;
		for(L2WorldRegion neighbor : getNeighbors())
			if(neighbor != null && !neighbor.isEmpty())
				return false;
		return true;
	}

	public void setActive(Boolean value)
	{
		synchronized (_active)
		{
			if(_active == value)
				return;
			_active = value;
		}

		switchAI(value);

		if(value)
			L2World.addActiveRegion(this);
		else
			L2World.removeActiveRegion(this);

		if(Config.DEBUG)
			if(value)
				_log.warning("Starting Grid " + tileX + "," + tileY + "," + tileZ);
			else
				_log.warning("Stoping Grid " + tileX + "," + tileY + "," + tileZ);
	}

	private synchronized void startActivation()
	{
		if(_neighborsTask != null)
		{
			_neighborsTask.cancel(false);
			_neighborsTask = null;
		}
		_neighborsTask = ThreadPoolManager.getInstance().scheduleGeneral(new NeighborsTask(true), 5000);
	}

	private synchronized void startDeactivation()
	{
		if(_neighborsTask != null)
		{
			_neighborsTask.cancel(false);
			_neighborsTask = null;
		}
		_neighborsTask = ThreadPoolManager.getInstance().scheduleGeneral(new NeighborsTask(false), 5000);
	}

	public void addToPlayers(L2Object object, L2Character dropper)
	{
		L2Player player = null;
		if(object.isPlayer)
			player = (L2Player) object;

		// Если object - игрок, показать ему все видимые обьекты в регионе
		if(player != null)
			for(Enumeration<L2Object> i = getObjects().elements(); i.hasMoreElements();)
			{
				L2Object obj = i.nextElement();
				if(obj == null)
					continue;
				// Если это фэйк обсервера - не показывать.
				if(obj.inObserverMode())
				{
					L2WorldRegion[] neighbors = ((L2Player) obj).getObservNeighbors();
					if(neighbors == null)
						continue;
					boolean flag = false;
					for(L2WorldRegion neighbor : neighbors)
						if(neighbor != null && neighbor.equals(this))
						{
							flag = true;
							break;
						}
					if(flag)
						continue;
				}
				player.addVisibleObject(obj, dropper);
			}

		// Показать обьект всем игрокам в регионе
		for(Enumeration<L2Player> i = getPlayers().elements(); i.hasMoreElements();)
		{
			L2Player obj = i.nextElement();
			if(obj != null)
				obj.addVisibleObject(object, dropper);
		}
	}

	public void removeFromPlayers(L2Object object)
	{
		L2Player player = null;
		if(object.isPlayer)
			player = (L2Player) object;

		// Если object - игрок, убрать у него все видимые обьекты в регионе
		if(player != null)
			for(Enumeration<L2Object> i = getObjects().elements(); i.hasMoreElements();)
			{
				L2Object obj = i.nextElement();
				if(obj != null)
					player.removeVisibleObject(obj);
			}

		// Убрать обьект у всех игроков в регионе
		for(Enumeration<L2Player> i = getPlayers().elements(); i.hasMoreElements();)
		{
			L2Player obj = i.nextElement();
			if(obj != null)
				obj.removeVisibleObject(object);
		}
	}

	public void addVisibleObject(L2Object object)
	{
		if(object == null)
			return;
		Integer obj_id = object.getObjectId();
		getObjects().put(obj_id, object);
		if(object.isCharacter)
		{
			getCharacters().put(obj_id, (L2Character) object);
			if(object instanceof L2Playable)
			{
				getPlayables().put(obj_id, (L2Playable) object);
				if(object.isPlayer)
				{
					getPlayers().put(obj_id, (L2Player) object);
					_playersSize++;
					startActivation();
				}
			}
			else if(object instanceof L2NpcInstance)
			{
				getNpcs().put(obj_id, (L2NpcInstance) object);
				if(object.getAI() instanceof DefaultAI && !areNeighborsEmpty())
					object.getAI().startAITask();
			}
		}
	}

	public void removeVisibleObject(L2Object object, Boolean move)
	{
		if(object == null)
			return;
		Integer obj_id = object.getObjectId();
		getObjects().remove(obj_id, object);
		if(object.isCharacter)
		{
			getCharacters().remove(obj_id, object);
			if(object instanceof L2Playable)
			{
				getPlayables().remove(obj_id, object);
				if(object.isPlayer)
				{
					getPlayers().remove(obj_id, object);
					_playersSize--;
					startDeactivation();
				}
			}
			else if(object instanceof L2NpcInstance)
			{
				getNpcs().remove(obj_id, object);
				if(!move && object.getAI() instanceof DefaultAI && !object.getAI().isGlobalAI())
					object.getAI().stopAITask();
			}
		}
	}

	public L2WorldRegion[] getNeighbors()
	{
		return L2World.getNeighbors(tileX, tileY, tileZ);
	}

	public ConcurrentHashMap<Integer, L2Object> getObjects()
	{
		if(_objects == null)
			_objects = new ConcurrentHashMap<Integer, L2Object>(0);
		return _objects;
	}

	public ConcurrentHashMap<Integer, L2Character> getCharacters()
	{
		if(_characters == null)
			_characters = new ConcurrentHashMap<Integer, L2Character>(0);
		return _characters;
	}

	public ConcurrentHashMap<Integer, L2NpcInstance> getNpcs()
	{
		if(_npcs == null)
			_npcs = new ConcurrentHashMap<Integer, L2NpcInstance>(0);
		return _npcs;
	}

	public ConcurrentHashMap<Integer, L2Playable> getPlayables()
	{
		if(_playables == null)
			_playables = new ConcurrentHashMap<Integer, L2Playable>(0);
		return _playables;
	}

	public ConcurrentHashMap<Integer, L2Player> getPlayers()
	{
		if(_players == null)
		{
			_players = new ConcurrentHashMap<Integer, L2Player>(0);
			_playersSize = 0;
		}
		return _players;
	}

	public String getName()
	{
		return "(" + tileX + ", " + tileY + ", " + tileZ + ")";
	}

	public synchronized void deleteVisibleNpcSpawns()
	{
		if(_npcs != null)
			for(L2Object obj : _npcs.values())
				if(obj != null)
				{
					L2NpcInstance npc = (L2NpcInstance) obj;
					npc.deleteMe();
					L2Spawn spawn = npc.getSpawn();
					if(spawn != null)
					{
						spawn.stopRespawn();
						SpawnTable.getInstance().deleteSpawn(spawn, false);
					}
				}
	}

	/**
	 * Показывает игроку все видимые обьекты в регионе
	 */
	public void showObjectsToPlayer(L2Player player)
	{
		if(player != null)
			for(Enumeration<L2Object> i = getObjects().elements(); i.hasMoreElements();)
			{
				L2Object obj = i.nextElement();
				if(obj != null)
					player.addVisibleObject(obj, null);
			}
	}

	/**
	 * Убирает у игрока все видимые обьекты в регионе
	 */
	public void removeObjectsFromPlayer(L2Player player)
	{
		if(player != null)
			for(Enumeration<L2Object> i = getObjects().elements(); i.hasMoreElements();)
			{
				L2Object obj = i.nextElement();
				if(obj != null)
					player.removeVisibleObject(obj);
			}
	}

	/**
	 * Убирает обьект у всех игроков в регионе
	 */
	public void removePlayerFromOtherPlayers(L2Object object)
	{
		if(object != null)
			for(Enumeration<L2Player> i = getPlayers().elements(); i.hasMoreElements();)
			{
				L2Player obj = i.nextElement();
				if(obj != null)
					obj.removeVisibleObject(object);
			}
	}
}