package L2jFT.game.model;

import java.util.HashMap;
import java.util.Set;
import java.util.logging.Logger;

import javolution.util.FastList;
import L2jFT.extensions.listeners.MethodCollection;
import L2jFT.extensions.listeners.engine.DefaultListenerEngine;
import L2jFT.extensions.listeners.engine.ListenerEngine;
import L2jFT.extensions.listeners.events.L2Zone.L2ZoneEnterLeaveEvent;
import L2jFT.game.ThreadPoolManager;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.skills.Stats;
import L2jFT.game.skills.funcs.FuncAdd;
import L2jFT.game.skills.funcs.LambdaConst;
import L2jFT.game.tables.SkillTable;
import L2jFT.util.Location;
import L2jFT.util.Rnd;
import L2jFT.util.Util;

public class L2Zone
{
	private static Logger _log = Logger.getLogger(L2Zone.class.getName());

	public static enum ZoneType
	{
		Town,
		Castle,
		ClanHall,
		OlympiadStadia,
		MonsterDerbyTrack,
		CastleDefenderSpawn,
		Fishing,
		Underground,
		Siege,
		water,
		region,
		battle_zone,
		damage,
		instant_skill,
		mother_tree,
		peace_zone,
		poison,
		ssq_zone,
		swamp,
		no_escape,
		no_landing,
		no_restart,
		no_spawn
	}

	int _id;
	ZoneType _type;
	private String _name;
	private L2Territory _loc1;
	private L2Territory _loc2;
	private L2Territory _loc3;
	private L2Territory _restartPoints;
	private L2Territory _PKrestartPoints;
	private int _entering_message_no;
	private int _leaving_message_no;
	private boolean _active;
	L2Skill _skill;
	int _skillProb;
	long _unitTick;
	private long _initialDelay;
	private ZoneTarget _target;
	private int _index;
	private int _taxById;
	private long _restartTime;
	private String _blockedActions;
	public static final String BLOCKED_ACTION_PRIVATE_STORE = "private store";
	public static final String BLOCKED_ACTION_PRIVATE_WORKSHOP = "private workshop";
	private ListenerEngine<L2Zone> listenerEngine;
	private DamageTask damageTask;

	private final HashMap<L2Object, Integer> _objects = new HashMap<L2Object, Integer>();
	private final HashMap<L2Character, ZoneSkillTimer> _skillTimers = new HashMap<L2Character, ZoneSkillTimer>();

	public final static int location1 = 0x01;
	public final static int location2 = 0x02;
	public final static int location3 = 0x04;
	public final static int restratPoint = 0x08;
	public final static int restartPointPK = 0x10;

	/**
	 * Ордер в зонах, с ним мы и добавляем/убираем статы.
	 * TODO: сравнить ордер с оффом, пока от фонаря
	 */
	public final static int ZONE_STATS_ORDER = 0x40;

	/**
	 * Сообщение которое шлется при уроне от зоны (не скилла)
	 * К примеру на осадах. Пока это только 686 (You have received $s1 damage from the fire of magic.)
	 */
	private int messageNumber;

	/**
	 * Урон от зоны по хп
	 */
	private int damageOnHP;

	/**
	 * Урон от зоны по мп
	 */
	private int damageOnMP;

	/**
	 * Бонус\пеналька к скорости движения
	 */
	private int moveBonus;

	/**
	 * Бонус регенерации хп
	 */
	private int regenBonusHP;

	/**
	 * Бонус регенерации мп
	 */
	private int regenBonusMP;

	/**
	 * Раса на которую применим эффект
	 */
	private String affectRace;

	public L2Zone(int id)
	{
		_id = id;
	}

	public final int getId()
	{
		return _id;
	}

	@Override
	public final String toString()
	{
		return "zone '" + _id + "'";
	}

	public ZoneType getType()
	{
		return _type;
	}

	public void setType(ZoneType type)
	{
		_type = type;
	}

	public void setLoc1(L2Territory t)
	{
		_loc1 = t;
	}

	public void setLoc2(L2Territory t)
	{
		_loc2 = t;
	}

	public void setLoc3(L2Territory t)
	{
		_loc3 = t;
	}

	public void setRestartPoints(L2Territory t)
	{
		_restartPoints = t;
	}

	public void setPKRestartPoints(L2Territory t)
	{
		_PKrestartPoints = t;
	}

	public L2Territory getLoc1()
	{
		return _loc1;
	}

	public L2Territory getLoc2()
	{
		return _loc2;
	}

	public L2Territory getLoc3()
	{
		return _loc3;
	}

	public L2Territory getRestartPoints()
	{
		return _restartPoints;
	}

	public L2Territory getPKRestartPoints()
	{
		return _PKrestartPoints;
	}

	public Location getSpawn()
	{
		// При ошибке портанет во флоран
		if(_restartPoints == null)
		{
			_log.warning("L2Zone.getSpawn(): restartPoint not found for " + toString());
			return new Location(17817, 170079, -3530);
		}
		FastList<int[]> coords = _restartPoints.getCoords();
		int rnd = Rnd.get(coords.size());
		int[] coord = coords.get(rnd);
		return new Location(coord[0], coord[1], coord[2]);
	}

	public Location getPKSpawn()
	{
		if(_PKrestartPoints == null)
			return getSpawn();
		FastList<int[]> coords = _PKrestartPoints.getCoords();
		int rnd = Rnd.get(coords.size());
		int[] coord = coords.get(rnd);
		return new Location(coord[0], coord[1], coord[2]);
	}

	/**
	 * Проверяет находятся ли даные координаты в зоне.
	 * _loc1 - стандартная территория для зоны
	 * _loc2 - используется на осадах и в кланхоллах
	 * _loc3 - используется на осадах и в кланхоллах
	 * @param x координата
	 * @param y координата
	 * @return находятся ли координаты в любой из локаций
	 */
	public boolean checkIfInZone(int x, int y)
	{
		return _loc1 != null && _loc1.isInside(x, y) || _loc2 != null && _loc2.isInside(x, y) || _loc3 != null && _loc3.isInside(x, y);
	}

	public final double findDistanceToZone(L2Object obj, boolean includeZAxis)
	{
		return findDistanceToZone(obj.getX(), obj.getY(), obj.getZ(), includeZAxis);
	}

	public final double findDistanceToZone(int x, int y, int z, boolean includeZAxis)
	{
		double closestDistance = 99999999;
		if(_loc1 == null)
			return closestDistance;
		double distance;
		for(int[] coord : _loc1.getCoords())
		{
			distance = Util.calculateDistance(x, y, z, coord[0], coord[1], (coord[2] + coord[3]) / 2, includeZAxis);
			if(distance < closestDistance)
				closestDistance = distance;
		}
		return closestDistance;
	}

	public FastList<int[]> getCoords()
	{
		return _loc1.getCoords();
	}

	/**
	 * Обработка входа в территорию
	 * Ojbect всегда добавляется в список вне зависимости от активности территории.
	 * Если зона акивная, то обработается вход в зону
	 * @param object кто входит
	 * @param terr L2Territory которая вызвала вход
	 */
	public void doEnter(L2Object object, L2Territory terr)
	{
		Integer lvl = _objects.get(object);

		if(lvl == null) // Нас в зоне воообще нет
		{
			lvl = getTerritoryMask(terr);
			_objects.put(object, lvl);
		}
		else
		// Мы в зоне есть
		{
			int newVal = lvl | getTerritoryMask(terr);
			if(newVal == lvl)
			{
				_log.warning("Double territory enter"); // Если мы дважды зашли в одну территорию
				return;
			}
			_objects.put(object, newVal); // Добавилась новая территория, добавляем объект заново.
			return; // Но не даем дальше выполнятся т.к. мы уже в зоне
		}

		//System.out.println("Entering..." + this);

		if(!isActive())
			return;

		onZoneEnter(object);
	}

	/**
	 * Обработка выхода с зоны
	 * Object всегда убирается со списка вне зависимости от зоны
	 * Если зона активная, то обработается выход из зоны
	 * @param object кто выходит
	 * @param terr L2Territory которая вызвала выход
	 */
	public void doLeave(L2Object object, L2Territory terr)
	{
		Integer lvl = _objects.get(object);

		if(lvl != null) // Мы в зоне
		{
			int newVal = lvl & ~getTerritoryMask(terr);
			if(newVal == lvl)
			{
				_log.warning("double territory exit"); // Если мы дважды вышли с одной территории
				return;
			}
			if(newVal != 0) // Если остались другие территории, то мы просто удаляем ту с которой вышли
			{
				_objects.put(object, newVal);
				return; // Не продолжаем выполнение т.к. мы не вышли с зоны
			}
			_objects.remove(object); // Выход с зоны
		}
		else
		{
			_log.warning("Zone exit without zone enter???"); // Никогда не должно такого быть
			Thread.dumpStack();
		}

		//System.out.println("Exiting..." + this);

		if(!isActive())
			return;

		onZoneLeave(object);
	}

	/**
	 * Обработка входа в зону
	 * @param object кто входит
	 */
	private void onZoneEnter(L2Object object)
	{
		getListenerEngine().fireMethodInvoked(new L2ZoneEnterLeaveEvent(MethodCollection.L2ZoneObjectEnter, this, new L2Object[] { object }));

		checkEffects(object, true);

		addZoneStats((L2Character) object);

		if(object.isPlayer)
			((L2Player) object).doZoneCheck(_entering_message_no);
	}

	/**
	 * Обработка выхода из зоны
	 * @param object кто выходит
	 */
	private void onZoneLeave(L2Object object)
	{
		getListenerEngine().fireMethodInvoked(new L2ZoneEnterLeaveEvent(MethodCollection.L2ZoneObjectLeave, this, new L2Object[] { object }));

		checkEffects(object, false);

		removeZoneStats((L2Character) object);

		if(object.isPlayer)
			((L2Player) object).doZoneCheck(_leaving_message_no);
	}

	/**
	 * Добавляет статы зоне
	 * @param object обьект которому добавляется
	 */
	private void addZoneStats(L2Character object)
	{
		// Проверка цели
		if(_target != null)
			if(!checkTarget(object))
				return;

		// Скорость движения накладывается только на L2Playable
		// affectRace в базе не указан, если надо будет влияние, то поправим
		if(getMoveBonus() != 0)
			if(object instanceof L2Playable)
			{
				object.addStatFunc(new FuncAdd(Stats.RUN_SPEED, ZONE_STATS_ORDER, this, new LambdaConst(getMoveBonus())));
				object.sendChanges();
			}

		// Если раса указана, то это уже не NPC
		if(affectRace != null && !object.isPlayer)
			return;

		// Если у нас раса не "all"
		if(affectRace != null && !affectRace.equalsIgnoreCase("all"))
		{
			L2Player player = object.getPlayer();

			// если раса не подходит
			if(!player.getRace().toString().equalsIgnoreCase(affectRace))
				return;
		}

		// Если у нас есть что регенить
		if(getRegenBonusHP() != 0)
			object.addStatFunc(new FuncAdd(Stats.REGENERATE_HP_RATE, ZONE_STATS_ORDER, this, new LambdaConst(getRegenBonusHP())));

		// Если у нас есть что регенить
		if(getRegenBonusMP() != 0)
			object.addStatFunc(new FuncAdd(Stats.REGENERATE_MP_RATE, ZONE_STATS_ORDER, this, new LambdaConst(getRegenBonusMP())));
	}

	/**
	 * Убирает добавленые зоной статы
	 * @param object обьект у которого убирается
	 */
	private void removeZoneStats(L2Character object)
	{
		if(getRegenBonusHP() == 0 && getRegenBonusMP() == 0 && getMoveBonus() == 0)
			return;

		object.removeStatsOwner(this);

		if(getMoveBonus() > 0)
			object.sendChanges();
	}

	/**
	 * Применяет еффекты при входе\выходе из(в) зону
	 * @param object обьект
	 * @param enter вошел или вышел
	 */
	private void checkEffects(L2Object object, boolean enter)
	{
		if(_skill == null || _target == null)
			return;

		// Проверяет тип цели.
		// affectRace ну указан.
		if(!checkTarget(object))
			return;

		L2Character p = (L2Character) object;
		if(enter)
		{
			ZoneSkillTimer rem = _skillTimers.remove(p);
			if(rem != null)
				rem.setCanceled(true);

			ZoneSkillTimer t = new ZoneSkillTimer(p);
			_skillTimers.put(p, t);
			if(_initialDelay > 0)
				ThreadPoolManager.getInstance().scheduleGeneral(t, _initialDelay);
			else
				t.run();
		}
		else
		{
			ZoneSkillTimer t = _skillTimers.remove(p);
			if(t != null)
				t.setCanceled(true);
		}
	}

	/**
	 * Проверяет подходит ли обьект для вызвавшего действия
	 * @param object обьект
	 * @return подошел ли
	 */
	private boolean checkTarget(L2Object object)
	{
		switch(_target)
		{
			case pc:
				return object.isPlayer || object.isSummon || object.isPet;
			case only_pc:
				return object.isPlayer;
			case npc:
				return object.isNpc;
		}
		return false;
	}

	public Set<L2Object> getObjects()
	{
		return _objects.keySet();
	}

	/**
	 * Установка активности зоны
	 * @param value активна ли зона
	 */
	public void setActive(boolean value)
	{
		if(_active == value)
			return;

		synchronized (_objects)
		{
			_active = value;

			if(_loc1 != null)
				_loc1.setActive(value);
			if(_loc2 != null)
				_loc2.setActive(value);
			if(_loc3 != null)
				_loc3.setActive(value);

			for(L2Object obj : _objects.keySet())
				if(_active)
					onZoneEnter(obj);
				else
					onZoneLeave(obj);
		}

		setDamageTaskActive(value);
	}

	/**
	 * Запускает или останавливает таск урона в зоне.<br>
	 * <b>Таск <font color="red">не</font> запустится если урон по хп и мп равны 0</b>
	 * @param value запустить или остановить таск урона
	 */
	public void setDamageTaskActive(boolean value)
	{
		if(value && (getDamageOnHP() > 0 || getDamageOnMP() > 0))
			new DamageTask().start();
		else
			damageTask = null;
	}

	public boolean isActive()
	{
		return _active;
	}

	public final String getName()
	{
		return _name;
	}

	public final void setName(String name)
	{
		_name = name;
	}

	public final int getEnteringMessageId()
	{
		return _entering_message_no;
	}

	public final void setEnteringMessageId(int id)
	{
		_entering_message_no = id;
	}

	public final int getLeavingMessageId()
	{
		return _leaving_message_no;
	}

	public final void setLeavingMessageId(int id)
	{
		_leaving_message_no = id;
	}

	public final void setIndex(int index)
	{
		_index = index;
	}

	public final int getIndex()
	{
		return _index;
	}

	public final void setTaxById(int id)
	{
		_taxById = id;
	}

	public final Integer getTaxById()
	{
		return _taxById;
	}

	public void setSkill(String skill)
	{
		if(skill == null || skill.equalsIgnoreCase("null"))
			return;
		String[] sk = skill.split(";");
		setSkill(SkillTable.getInstance().getInfo(Short.parseShort(sk[0]), Short.parseShort(sk[1])));
	}

	public void setSkill(L2Skill skill)
	{
		_skill = skill;
	}

	public void setSkillProb(String chance)
	{
		if(chance == null)
		{
			_skillProb = -1;
			return;
		}
		_skillProb = Integer.parseInt(chance);
	}

	public void setUnitTick(String delay)
	{
		if(delay == null)
		{
			_unitTick = 1000;
			return;
		}

		_unitTick = Integer.parseInt(delay) * 1000;
	}

	public void setInitialDelay(String delay)
	{
		if(delay == null)
		{
			_initialDelay = 0;
			return;
		}

		_initialDelay = Integer.parseInt(delay) * 1000;
	}

	private class ZoneSkillTimer implements Runnable
	{
		private boolean canceled = false;
		private L2Character target;

		public ZoneSkillTimer(L2Character player)
		{
			target = player;
		}

		public void run()
		{
			if(canceled)
				return;

			if(Rnd.chance(_skillProb) && !target.isDead())
				_skill.getEffects(target, target, false);

			ThreadPoolManager.getInstance().scheduleGeneral(this, _unitTick);
		}

		public void setCanceled(boolean val)
		{
			canceled = val;
		}
	}

	public void setTarget(String target)
	{
		if(target == null)
			return;

		_target = ZoneTarget.valueOf(target);
	}

	public enum ZoneTarget
	{
		pc,
		npc,
		only_pc
	}

	public long getRestartTime()
	{
		return _restartTime;
	}

	public L2Skill getZoneSkill()
	{
		return _skill;
	}

	public ZoneTarget getZoneTarget()
	{
		return _target;
	}

	public void setRestartTime(long restartTime)
	{
		_restartTime = restartTime;
	}

	public void setBlockedActions(String blockedActions)
	{
		_blockedActions = blockedActions;
	}

	public boolean isActionBlocked(String action)
	{
		return _blockedActions != null && _blockedActions.contains(action);
	}

	public ListenerEngine<L2Zone> getListenerEngine()
	{
		if(listenerEngine == null)
			listenerEngine = new DefaultListenerEngine<L2Zone>(this);
		return listenerEngine;
	}

	/**
	 * Номер системного вообщения которое будет отослано игроку при нанесении урона зоной
	 * @return SystemMessage ID
	 */
	public int getMessageNumber()
	{
		return messageNumber;
	}

	/**
	 * Устанавлевает номер системного сообщение которое будет отослано игроку при уроне зоной
	 * @param messageNumber SystemMessage ID
	 */
	public void setMessageNumber(int messageNumber)
	{
		this.messageNumber = messageNumber;
	}

	/**
	 * Устанавлевает номер системного сообщение которое будет отослано игроку при уроне зоной
	 * @param number SystemMessage ID - Системное сообщение сторой
	 */
	public void setMessageNumber(String number)
	{
		setMessageNumber(number == null ? 0 : Integer.parseInt(number));
	}

	/**
	 * Сколько урона зона нанесет по хп
	 * @return количество урона
	 */
	public int getDamageOnHP()
	{
		return damageOnHP;
	}

	/**
	 * Установка урона по хп зоной
	 * @param damageOnHP сколько зона нанесет урона по хп
	 */
	public void setDamageOnHP(int damageOnHP)
	{
		this.damageOnHP = damageOnHP;
	}

	/**
	 * Установка урона по хп зоной. Парсит строку.
	 * @param damage урон в строке
	 */
	public void setDamageOnHP(String damage)
	{
		setDamageOnHP(damage == null ? 0 : Integer.parseInt(damage));
	}

	/**
	 * Сколько урона зона нанесет по мп
	 * @return количество урона
	 */
	public int getDamageOnMP()
	{
		return damageOnMP;
	}

	/**
	 * Установка урона по мп зоной
	 * @param damageOnMP сколько зона нанесет урона по мп
	 */
	public void setDamageOnMP(int damageOnMP)
	{
		this.damageOnMP = damageOnMP;
	}

	/**
	 * Установка урона по мп зоной. Парсит строку.
	 * @param damage урон в строке
	 */
	public void setDamageOnМP(String damage)
	{
		setDamageOnMP(damage == null ? 0 : Integer.parseInt(damage));
	}

	/**
	 * @return Бонус к скорости движения в зоне
	 */
	public int getMoveBonus()
	{
		return moveBonus;
	}

	/**
	 * Устанавливает бонус на движение в зоне
	 * @param moveBonus бонус
	 */
	public void setMoveBonus(int moveBonus)
	{
		this.moveBonus = moveBonus;
	}

	/**
	 * Устанавливает бонус на движение в зоне
	 * @param moveBonus бонус
	 */
	public void setMoveBonus(String moveBonus)
	{
		setMoveBonus(moveBonus == null ? 0 : Integer.parseInt(moveBonus));
	}

	/**
	 * Возвращает бонус регенерации хп в этой зоне
	 * @return Бонус регенарации хп в этой зоне
	 */
	public int getRegenBonusHP()
	{
		return regenBonusHP;
	}

	/**
	 * Устанавливает бонус регена хп в этой зоне. Парсит строку
	 * @param bonus бонус регена
	 */
	public void setRegenBonusHP(String bonus)
	{
		setRegenBonusHP(bonus == null ? 0 : Integer.parseInt(bonus));
	}

	/**
	 * Устанавливает бонус регенерации хп в этой зоне
	 * @param regenBonusHP бонус регена ХП
	 */
	public void setRegenBonusHP(int regenBonusHP)
	{
		this.regenBonusHP = regenBonusHP;
	}

	/**
	 * Возвращает бонус регенерации мп в этой зоне
	 * @return Бонус регенарации мп в этой зоне
	 */
	public int getRegenBonusMP()
	{
		return regenBonusMP;
	}

	/**
	 * Устанавливает бонус регенерации хп в этой зоне
	 * @param regenBonusMP бонус регена МП
	 */
	public void setRegenBonusMP(int regenBonusMP)
	{
		this.regenBonusMP = regenBonusMP;
	}

	/**
	 * Устанавливает бонус регена мп в этой зоне. Парсит строку
	 * @param bonus бонус регена
	 */
	public void setRegenBonusMP(String bonus)
	{
		setRegenBonusMP(bonus == null ? 0 : Integer.parseInt(bonus));
	}

	/**
	 * Возвращает расу на которую примени эффект. Раса может быть:
	 * <br><b>all</b> - любая раса
	 * <br><b>null</b> - проверка не применяется
	 * ну или любая из энумы Race
	 * @see L2jFT.game.model.base.Race
	 * @return раса
	 */
	public String getAffectRace()
	{
		return affectRace;
	}

	/**
	 * Устанавливает расу зоне на которую будут применятся эффекты. Раса может быть:
	 * <br><b>all</b> - любая раса
	 * <br><b>null</b> - проверка не применяется
	 * ну или любая из энумы Race
	 * @see L2jFT.game.model.base.Race
	 * @param affectRace строка с расой
	 */
	public void setAffectRace(String affectRace)
	{
		if(affectRace != null)
			affectRace = affectRace.toLowerCase().replace(" ", "");

		this.affectRace = affectRace;
	}

	/**
	 * Возвращает число для матрици "входимости" в даную территорию
	 * @param terr территория
	 * @return число
	 */
	public int getTerritoryMask(L2Territory terr)
	{
		if(terr == getLoc1())
			return location1;
		else if(terr == getLoc2())
			return location2;
		else if(terr == getLoc3())
			return location3;
		else if(terr == _restartPoints)
			return restratPoint;
		else if(terr == _PKrestartPoints)
			return restartPointPK;
		else
			throw new IllegalArgumentException("doEnter request from territory which current zone don't have");
	}

	/**
	 * Класс используется для нанесения урона всем кто в зоне.
	 */
	private class DamageTask implements Runnable
	{
		public void run()
		{
			if(damageTask != this || !isActive())
				return;

			Set<L2Object> objects = _objects.keySet();

			for(L2Object object : objects)
			{
				if(_target != null && !checkTarget(object)) // Если явно указана цель
					continue;

				if(_target == null && !(object instanceof L2Playable)) // Если цель не указана, то валим только игроков
					continue;

				L2Character target = (L2Character) object;

				int hp = getDamageOnHP();
				int mp = getDamageOnMP();
				int message = getMessageNumber();

				if(hp > 0)
				{
					target.reduceCurrentHp(hp, null);
					if(message > 0)
						target.sendPacket(new SystemMessage(message).addNumber(hp));
				}

				if(mp > 0)
				{
					target.reduceCurrentMp(mp);
					if(message > 0)
						target.sendPacket(new SystemMessage(message).addNumber(mp));
				}
			}

			ThreadPoolManager.getInstance().scheduleGeneral(this, 1000);
		}

		/**
		 * Запускает таймер нанесения урона зоной
		 */
		public void start()
		{
			damageTask = this;

			if(_initialDelay == 0)
			{
				run();
				return;
			}

			ThreadPoolManager.getInstance().scheduleGeneral(this, _initialDelay);
		}
	}
}
