package L2jFT.game.model;

import L2jFT.Config;
import L2jFT.database.DatabaseUtils;
import L2jFT.database.FiltredPreparedStatement;
import L2jFT.database.L2DatabaseFactory;
import L2jFT.database.ThreadConnection;
import L2jFT.game.idfactory.IdFactory;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.model.instances.L2ItemInstance.ItemClass;
import L2jFT.game.model.instances.L2ItemInstance.ItemLocation;
import L2jFT.game.tables.ItemTable;
import L2jFT.util.Log;

import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

public abstract class Warehouse
{
	public static enum WarehouseType
	{
		PRIVATE(1),
		CLAN(2),
		CASTLE(3),
		FREIGHT(4);

		private final int _type;

		private WarehouseType(final int type)
		{
			_type = type;
		}

		public int getPacketValue()
		{
			return _type;
		}
	}

	private static final Logger _log = Logger.getLogger(Warehouse.class.getName());

	public abstract int getOwnerId();

	public abstract ItemLocation getLocationType();

	public abstract WarehouseType getWarehouseType();

	public abstract String getLocationId();

	public abstract byte getLocationId(boolean addItem);

	public abstract void setLocationId(L2Player activeChar);

	public L2ItemInstance[] listItems(ItemClass clss)
	{
		final ArrayList<L2ItemInstance> items = new ArrayList<L2ItemInstance>();
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;

		try
		{
			String _select;
			switch(clss)
			{
				case EQUIPMENT:
				case CONSUMABLE:
				case MATHERIALS:
				case PIECES:
				case SPELLBOOKS:
				case RECIPIES:
				case MISC:
				case OTHER:
					_select = " AND class = '" + clss.name() + "'";
					break;
				case ALL:
				default:
					_select = "";
					break;
			}

			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT object_id FROM items WHERE owner_id=? AND loc=? " + _select + " ORDER BY name ASC LIMIT 200");
			statement.setInt(1, getOwnerId());
			statement.setString(2, getLocationType().name());
			rset = statement.executeQuery();

			L2ItemInstance item;
			while(rset.next())
			{
				final int objectId = rset.getInt(1);
				item = L2ItemInstance.restoreFromDb(objectId);
				if(item == null)
					continue;
				items.add(item);
			}
		}
		catch(final Exception e)
		{
			_log.log(Level.SEVERE, "could not restore warehouse:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
		return items.toArray(new L2ItemInstance[items.size()]);
	}

	public synchronized L2ItemInstance addItem(final L2ItemInstance newItem)
	{
		L2ItemInstance oldItem = null;
		// non-stackable items are simply added to DB
		if(!newItem.isStackable() || (oldItem = findItemId(newItem.getItemId())) == null)
		{
			newItem.setOwnerId(getOwnerId());
			newItem.setLocation(getLocationType(), getLocationId(true));
			newItem.setLastChange(L2ItemInstance.REMOVED);
			newItem.updateDatabase();
			return newItem;
		}
		Log.add("Warehouse|" + getOwnerId() + "|join item from-to|" + oldItem.getItemId() + "|" + newItem.getObjectId() + "|" + oldItem.getObjectId(), "items");
		oldItem.setCount(oldItem.getCount() + newItem.getCount());
		oldItem.setLastChange(L2ItemInstance.MODIFIED);
		newItem.setCount(0);
		newItem.setOwnerId(0);
		newItem.setLocation(ItemLocation.VOID, getLocationId(true));
		oldItem.updateDatabase(true);
		newItem.updateDatabase(true);
		L2World.removeObject(newItem);
		IdFactory.getInstance().releaseId(newItem.getObjectId());
		return oldItem;
	}

	public synchronized L2ItemInstance destroyItemByObj(final int objectId, final int count)
	{
		L2ItemInstance item = (L2ItemInstance) L2World.findObject(objectId);
		if(item == null)
			item = L2ItemInstance.restoreFromDb(objectId);

		if(item != null)
		{
			if(item.getCount() <= count)
			{
				item.setCount(0);
				item.setLocation(ItemLocation.VOID, getLocationId(true));
				item.setLastChange(L2ItemInstance.REMOVED);
				item.updateDatabase(true);
			}
			else
			{
				item.setCount(item.getCount() - count);
				item.setLastChange(L2ItemInstance.MODIFIED);
				item.updateDatabase(true);
			}
		}
		else if(Config.DEBUG)
			_log.fine("can't destroy " + count + " item(s) obj_id:" + objectId);
		return item;
	}

	public synchronized L2ItemInstance getItemByObj(final int objectId, final int count)
	{
		L2ItemInstance item = (L2ItemInstance) L2World.findObject(objectId);
		if(item == null)
			item = L2ItemInstance.restoreFromDb(objectId);
		if(item == null)
		{
			_log.fine("can't destroy " + count + " item(s) obj_id:" + objectId);
			return null;
		}
		if(item.getLocation() != ItemLocation.CLANWH && item.getLocation() != ItemLocation.WAREHOUSE && item.getLocation() != ItemLocation.FREIGHT)
		{
			_log.warning("WARNING get item not in WAREHOUSE via WAREHOUSE: item objid=" + item.getObjectId() + " ownerid=" + item.getOwnerId());
			return null;
		}
		if(item.getCount() <= count)
		{
			item.setLastChange(L2ItemInstance.MODIFIED);
			item.setLocation(ItemLocation.VOID, getLocationId(true));
			item.setWhFlag(true);
			item.updateDatabase(true);
			return item;
		}
		item.setCount(item.getCount() - count);
		item.setLastChange(L2ItemInstance.MODIFIED);
		item.updateDatabase(true);
		final L2ItemInstance Newitem = ItemTable.getInstance().createItem(item.getItem().getItemId(), getOwnerId(), objectId, "Warehouse.getItemByObj");
		Newitem.setCount(count);
		Log.add("Warehouse|" + getOwnerId() + "|Split item from-to|" + item.getItemId() + "|" + item.getObjectId() + "|" + Newitem.getObjectId(), "items");
		return Newitem;
	}

	public synchronized L2ItemInstance destroyItem(final int itemId, final int count)
	{
		final L2ItemInstance item = findItemId(itemId);
		if(item != null)
		{
			if(item.getCount() < count)
				return null;

			if(item.getCount() == count)
			{
				item.setCount(0);
				item.setLocation(ItemLocation.VOID, getLocationId(true));
				item.setLastChange(L2ItemInstance.REMOVED);
				item.updateDatabase(true);
			}
			else
			{
				item.setCount(item.getCount() - count);
				item.setLastChange(L2ItemInstance.MODIFIED);
				item.updateDatabase(true);
			}
		}
		else if(Config.DEBUG)
			_log.fine("can't destroy " + count + " item(s) " + itemId);
		return item;
	}

	public L2ItemInstance findItemId(final int itemId)
	{
		L2ItemInstance foundItem = null;
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT object_id FROM items " + "WHERE owner_id=? AND loc=? AND loc_data IN (" + getLocationId() + ") AND item_id=? ");
			statement.setInt(1, getOwnerId());
			statement.setString(2, getLocationType().name());
			statement.setInt(3, itemId);
			rset = statement.executeQuery();

			if(rset.next())
			{
				int objectId = rset.getInt(1);
				foundItem = L2ItemInstance.restoreFromDb(objectId);
			}
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not list warehouse:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
		return foundItem;
	}
}
