package L2jFT.game.model.entity;

import javolution.util.FastList;
import L2jFT.database.DatabaseUtils;
import L2jFT.database.FiltredPreparedStatement;
import L2jFT.database.L2DatabaseFactory;
import L2jFT.database.ThreadConnection;
import L2jFT.game.ThreadPoolManager;
import L2jFT.game.instancemanager.ZoneManager;
import L2jFT.game.model.L2Clan;
import L2jFT.game.model.L2Object;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2World;
import L2jFT.game.model.L2Zone;
import L2jFT.game.model.L2Zone.ZoneType;
import L2jFT.game.model.instances.L2ClanHallManagerInstance;
import L2jFT.game.model.instances.L2DoorInstance;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.network.serverpackets.PledgeShowInfoUpdate;
import L2jFT.game.tables.ClanTable;
import L2jFT.game.tables.DoorTable;
import L2jFT.game.tables.MapRegionTable;

import java.sql.ResultSet;
import java.util.Calendar;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public class ClanHall
{
	protected static Logger _log = Logger.getLogger(ClanHall.class.getName());

	int _ClanHallId = 0;
	private List<L2DoorInstance> _Doors;
	private List<String> _DoorDefault = new FastList<String>();
	private String _Name = "";
	private int _OwnerId = 0;
	private int _lease = 0;
	private String _desc = "";
	private String _location = "";
	private Calendar _paidUntil;
	private L2Zone _Zone;
	private int _grade;
	private int _price = 0;
	private List<ClanHallFunction> _functions = new FastList<ClanHallFunction>();

	// clan hall functions
	public static final int FUNC_TELEPORT = 1;
	public static final int FUNC_ITEM_CREATE = 2;
	public static final int FUNC_RESTORE_HP = 3;
	public static final int FUNC_RESTORE_MP = 4;
	public static final int FUNC_RESTORE_EXP = 5;
	public static final int FUNC_SUPPORT = 6;
	public static final int FUNC_CURTAIN = 7;
	public static final int FUNC_PLATFORM = 8;

	private class AutoTask implements Runnable
	{
		public void run()
		{
			if(getOwnerId() != 0)
				try
				{
					L2Clan clan = ClanTable.getInstance().getClan(getOwnerId());
					if(clan == null)
					{
						_log.warning("ClanHall[68]: clan == null");
						return;
					}

					L2ItemInstance adena = clan.getWarehouse().findItemId(57);

					if(getPaidUntil() - Calendar.getInstance().getTimeInMillis() > 0)
						ThreadPoolManager.getInstance().scheduleGeneral(new AutoTask(), getPaidUntil() - Calendar.getInstance().getTimeInMillis());
					else if(adena != null && adena.getCount() >= getLease())
					{
						clan.getWarehouse().destroyItem(57, getLease());
						updateRentTime();
						ThreadPoolManager.getInstance().scheduleGeneral(new AutoTask(), getPaidUntil() - Calendar.getInstance().getTimeInMillis());
						_log.fine("clan hall lease is gonna be deducted from " + getName() + " owner's cwh at " + _paidUntil.get(Calendar.DAY_OF_MONTH) + "/" + _paidUntil.get(Calendar.MONTH));
					}
					else
					{
						_log.fine("remove " + getName() + " clanhall, because there are not enough money");
						setOwner(null);
					}
				}
				catch(Exception e)
				{
					e.printStackTrace();
				}
		}
	}

	private void startAutoTask()
	{
		getPaidUntilCalendar().set(Calendar.MINUTE, 0);
		ThreadPoolManager.getInstance().scheduleGeneral(new AutoTask(), 1000);
	}

	public ClanHall(int clanHallId)
	{
		_ClanHallId = clanHallId;
		load();
	}

	/** Return true if object is inside the zone */
	public boolean checkIfInZone(L2Object obj)
	{
		return checkIfInZone(obj.getX(), obj.getY());
	}

	/** Return true if object is inside the zone */
	public boolean checkIfInZone(int x, int y)
	{
		return getZone().checkIfInZone(x, y);
	}

	public void openCloseDoors(boolean open)
	{
		for(L2DoorInstance door : getDoors())
			if(door != null)
				if(open)
					door.openMe();
				else
					door.closeMe();
	}

	public void setOwner(L2Clan clan)
	{
		// Remove old owner
		if(getOwnerId() > 0 && (clan == null || clan.getClanId() != getOwnerId()))
		{
			L2Clan oldOwner = ClanTable.getInstance().getClan(getOwnerId()); // Try to find clan instance 
			if(oldOwner != null)
				oldOwner.setHasHideout((byte) 0); // Unset has hideout flag for old owner
		}
		// Update in database
		updateOwnerInDB(clan);
	}

	/** Respawn all doors */
	public void spawnDoor()
	{
		spawnDoor(false);
	}

	/** Respawn all doors */
	public void spawnDoor(boolean isDoorWeak)
	{
		for(int i = 0; i < getDoors().size(); i++)
		{
			L2DoorInstance door = getDoors().get(i);
			if(door.getCurrentHp() <= 0)
			{
				door.decayMe(); // Kill current if not killed already
				door = DoorTable.parseList(_DoorDefault.get(i));
				if(isDoorWeak)
					door.setCurrentHp(door.getMaxHp() / 2);
				door.spawnMe(door.getX(), door.getY(), door.getZ());
				getDoors().set(i, door);
			}
			else if(door.isOpen())
				door.closeMe();
		}
	}

	private void load()
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rs = null;

		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();

			statement = con.prepareStatement("select * from clanhall where id = ?");
			statement.setInt(1, getId());
			rs = statement.executeQuery();

			while(rs.next())
			{
				_Name = rs.getString("name");
				_OwnerId = rs.getInt("ownerId");
				_price = rs.getInt("price");
				_lease = rs.getInt("lease");
				_desc = rs.getString("desc");
				_location = rs.getString("location");
				_paidUntil = Calendar.getInstance();
				_paidUntil.setTimeInMillis(rs.getLong("paidUntil"));
				_grade = rs.getInt("Grade");
			}
			DatabaseUtils.closeDatabaseSR(statement, rs);

			statement = con.prepareStatement("Select clan_id from clan_data where hasHideout = ?");
			statement.setInt(1, getId());
			rs = statement.executeQuery();

			while(rs.next())
				_OwnerId = rs.getInt("clan_id");
		}
		catch(Exception e)
		{
			System.out.println("Exception: ClanHall.load(): " + e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rs);
		}

		if(getOwnerId() == 0) // this should never happen, but one never knows ;)
			return;

		loadFunctions();
		startAutoTask();
	}

	private void loadFunctions()
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rs = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();

			statement = con.prepareStatement("Select * from clanhall_functions where hall_id = ?");
			statement.setInt(1, getId());
			rs = statement.executeQuery();

			while(rs.next())
				_functions.add(new ClanHallFunction(rs.getInt("type"), rs.getInt("lvl"), rs.getInt("lease"), rs.getLong("rate"), rs.getLong("endTime"), rs.getBoolean("inDebt")));

		}
		catch(Exception e)
		{
			_log.log(Level.SEVERE, "Exception: ClanHall.loadFunctions(): " + e.getMessage(), e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rs);
		}
	}

	private void updateOwnerInDB(L2Clan clan)
	{
		if(clan != null)
			_OwnerId = clan.getClanId(); // Update owner id property
		else
			_OwnerId = 0; // Remove owner

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("UPDATE clanhall SET ownerId=? WHERE id=?");
			statement.setInt(1, getOwnerId());
			statement.setInt(2, getId());
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("UPDATE clan_data SET hasHideout=0 WHERE hasHideout=?");
			statement.setInt(1, getId());
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("UPDATE clan_data SET hasHideout=? WHERE clan_id=?");
			statement.setInt(1, getId());
			statement.setInt(2, getOwnerId());
			statement.execute();

			// Announce to clan memebers
			if(clan != null)
			{
				clan.setHasHideout((byte) getId()); // Set has hideout flag for new owner
				clan.broadcastToOnlineMembers(new PledgeShowInfoUpdate(clan));
			}
			else
			{
				Calendar EndDate = Calendar.getInstance();
				EndDate.add(Calendar.DAY_OF_MONTH, 7); // Schedule to happen in 7 days
				statement = con.prepareStatement("REPLACE INTO auction (id, sellerId, sellerName, sellerClanName, itemType, itemId, itemObjectId, itemName, itemQuantity, startingBid, currentBid, endDate) VALUES (?,?,?,?,?,?,?,?,?,?,?,?)");
				statement.setInt(1, getId());
				statement.setInt(2, 0);
				statement.setString(3, "NPC");
				statement.setString(4, "NPC Clan");
				statement.setString(5, "ClanHall");
				statement.setInt(6, getId());
				statement.setInt(7, 0);
				statement.setString(8, getName());
				statement.setInt(9, 1);
				statement.setInt(10, getPrice());
				statement.setInt(11, 0);
				statement.setLong(12, EndDate.getTimeInMillis());
				statement.execute();
			}
		}
		catch(Exception e)
		{
			System.out.println("Exception: updateOwnerInDB(L2Clan clan): " + e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	public final int getId()
	{
		return _ClanHallId;
	}

	public final L2DoorInstance getDoor(int doorId)
	{
		if(doorId <= 0)
			return null;

		for(int i = 0; i < getDoors().size(); i++)
		{
			L2DoorInstance door = getDoors().get(i);
			if(door.getDoorId() == doorId)
				return door;
		}
		return null;
	}

	public final List<L2DoorInstance> getDoors()
	{
		if(_Doors == null)
			_Doors = new FastList<L2DoorInstance>();
		return _Doors;
	}

	public final String getName()
	{
		return _Name;
	}

	public final int getOwnerId()
	{
		return _OwnerId;
	}

	public final L2Clan getOwner()
	{
		return ClanTable.getInstance().getClan(_OwnerId);
	}

	public final L2Zone getZone()
	{
		if(_Zone == null)
			_Zone = ZoneManager.getInstance().getZoneByIndex(ZoneType.ClanHall, _ClanHallId, true);
		return _Zone;
	}

	public final int getPrice()
	{
		return _price;
	}

	public final int getLease()
	{
		return _lease;
	}

	public final String getDesc()
	{
		return _desc;
	}

	public final String getLocation()
	{
		return _location;
	}

	public final long getPaidUntil()
	{
		return _paidUntil.getTimeInMillis();
	}

	public final Calendar getPaidUntilCalendar()
	{
		return _paidUntil;
	}

	public final int getGrade()
	{
		return _grade;
	}

	public void banishForeigner(L2Player activeChar)
	{
		// Get players from this and nearest world regions
		for(L2Player player : L2World.getAroundPlayers(activeChar))
		{
			// Skip if player is in clan
			if(player.getClanId() == getOwnerId())
				continue;
			if(checkIfInZone(player))
				player.teleToLocation(MapRegionTable.getInstance().getTeleToLocation(player, MapRegionTable.TeleportWhereType.ClosestTown));
		}
	}

	public ClanHallFunction getFunction(int type)
	{
		for(int i = 0; i < _functions.size(); i++)
			if(_functions.get(i).getType() == type)
				return _functions.get(i);
		return null;
	}

	public void removeFunctions(int functionType)
	{
		for(int i = 0; i < _functions.size(); i++)
			if(_functions.get(i).getType() == functionType)
				_functions.remove(i);

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("DELETE FROM clanhall_functions WHERE hall_id=? AND type=?");
			statement.setInt(1, getId());
			statement.setInt(2, functionType);
			statement.execute();
		}
		catch(Exception e)
		{
			_log.log(Level.SEVERE, "Exception: ClanHall.removeFunctions(int functionType): " + e.getMessage(), e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	public boolean updateFunctions(int type, int lvl, int lease, long rate, long time, boolean addNew, L2ClanHallManagerInstance chManager)
	{
		L2Clan clan = ClanTable.getInstance().getClan(getOwnerId());
		if(clan == null)
			return false;

		L2ItemInstance adena = clan.getWarehouse().findItemId(57);
		if(adena == null)
			return false;

		int count = adena.getCount();

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();

			if(addNew)
			{
				if(count >= lease)
					clan.getWarehouse().destroyItem(57, lease);
				else
					return false;

				statement = con.prepareStatement("REPLACE INTO clanhall_functions (hall_id, type, lvl, lease, rate, endTime) VALUES (?,?,?,?,?,?)");
				statement.setInt(1, getId());
				statement.setInt(2, type);
				statement.setInt(3, lvl);
				statement.setInt(4, lease);
				statement.setLong(5, rate);
				statement.setLong(6, time);
				statement.execute();
				_functions.add(new ClanHallFunction(type, lvl, lease, rate, time, false));
			}
			else
			{
				if(count >= lease - getFunction(type).getLease())
				{
					if(lease - getFunction(type).getLease() > 0)
						clan.getWarehouse().destroyItem(57, lease - getFunction(type).getLease());
				}
				else
					return false;

				statement = con.prepareStatement("UPDATE clanhall_functions SET lvl=?, lease=? WHERE hall_id=? AND type=?");
				statement.setInt(1, lvl);
				statement.setInt(2, lease);
				statement.setInt(3, getId());
				statement.setInt(4, type);
				statement.execute();

				getFunction(type).setLvl(lvl);
				getFunction(type).setLease(lease);
			}
		}
		catch(Exception e)
		{
			_log.log(Level.SEVERE, "Exception: ClanHall.updateFunctions(int type, int lvl, int lease, long rate, long time, boolean addNew): " + e.getMessage(), e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}

		if(chManager != null)
			chManager.broadCastClanHallInfo();

		return true;
	}

	public void updateRentTime()
	{
		_paidUntil.setTimeInMillis(Calendar.getInstance().getTimeInMillis() + 604800000);
		_paidUntil.set(Calendar.MINUTE, 0);

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("UPDATE clanhall SET paidUntil=? WHERE id=?");
			statement.setLong(1, _paidUntil.getTimeInMillis());
			statement.setInt(2, getId());
			statement.execute();
		}
		catch(Exception e)
		{
			_log.log(Level.SEVERE, "Exception: ClanHall.updateRentTime(): " + e.getMessage(), e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	public class ClanHallFunction
	{
		private int _type;
		private int _lvl;
		protected int _fee;
		private long _rate;
		private Calendar _endDate;
		protected boolean _inDebt;

		public ClanHallFunction(int type, int lvl, int lease, long rate, long time, boolean inDebt)
		{
			_type = type;
			_lvl = lvl;
			_fee = lease;
			_rate = rate;
			_endDate = Calendar.getInstance();
			_endDate.setTimeInMillis(time);
			_inDebt = inDebt;
			StartAutoTaskForFunctions();
		}

		public int getType()
		{
			return _type;
		}

		public int getLvl()
		{
			return _lvl;
		}

		public int getLease()
		{
			return _fee;
		}

		public long getRate()
		{
			return _rate;
		}

		public void setLvl(int lvl)
		{
			_lvl = lvl;
		}

		public void setLease(int lease)
		{
			_fee = lease;
		}

		public long getEndTime()
		{
			return _endDate.getTimeInMillis();
		}

		public void updateRentTime(boolean inDebt)
		{
			_endDate.setTimeInMillis(Calendar.getInstance().getTimeInMillis() + getRate());

			ThreadConnection con = null;
			FiltredPreparedStatement statement = null;
			try
			{
				con = L2DatabaseFactory.getInstance().getConnection();

				statement = con.prepareStatement("UPDATE clanhall_functions SET endTime=?, inDebt=? WHERE type=? AND hall_id=?");
				statement.setLong(1, getEndTime() + getRate());
				statement.setInt(2, inDebt ? 1 : 0);
				statement.setInt(3, getType());
				statement.setInt(4, getId());
				statement.execute();
			}
			catch(Exception e)
			{
				_log.log(Level.SEVERE, "Exception: ClanHall.ClanHallFunction.updateRentTime(int functionType): " + e.getMessage(), e);
			}
			finally
			{
				DatabaseUtils.closeDatabaseCS(con, statement);
			}
		}

		private void StartAutoTaskForFunctions()
		{
			if(getOwnerId() != 0)
				try
				{
					L2Clan clan = ClanTable.getInstance().getClan(getOwnerId());
					if(clan == null)
					{
						_log.warning("ClanHall[690]: clan == null for hall " + _ClanHallId);
						return;
					}

					L2ItemInstance adena = clan.getWarehouse().findItemId(57);

					if(getEndTime() - Calendar.getInstance().getTimeInMillis() > 0)
						ThreadPoolManager.getInstance().scheduleGeneral(new AutoTaskForFunctions(), getEndTime() - Calendar.getInstance().getTimeInMillis());
					else if(adena != null && _inDebt && adena.getCount() >= _fee) // if player didn't pay before add extra fee
					{
						clan.getWarehouse().destroyItem(57, _fee);
						updateRentTime(false);
						ThreadPoolManager.getInstance().scheduleGeneral(new AutoTaskForFunctions(), getEndTime() - Calendar.getInstance().getTimeInMillis());
						_log.warning("deducted " + _fee + " adena from " + getName() + " owner's cwh for function type " + _type);
					}
					else if(!_inDebt)
					{
						_inDebt = true;
						updateRentTime(true);
						ThreadPoolManager.getInstance().scheduleGeneral(new AutoTaskForFunctions(), getEndTime() - Calendar.getInstance().getTimeInMillis());
					}
					else
					{
						removeFunctions(getType());
						_log.warning("remove function type " + _type + ", for " + getName() + " clanhall, because there are not enough money");
					}
				}
				catch(Exception e)
				{
					e.printStackTrace();
				}
		}

		private class AutoTaskForFunctions implements Runnable
		{
			public void run()
			{
				StartAutoTaskForFunctions();
			}
		}
	}
}
