package L2jFT.game.model.instances;

import javolution.util.FastList;
import L2jFT.game.ThreadPoolManager;
import L2jFT.util.Rnd;
import L2jFT.game.model.L2Character;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2Skill;
import L2jFT.game.network.serverpackets.MagicSkillUse;
import L2jFT.game.tables.SkillTable;

import java.util.ArrayList;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;

public class L2CubicInstance
{
	protected static final Logger _log = Logger.getLogger(L2CubicInstance.class.getName());

	public static final short STORM_CUBIC = 1;
	public static final short VAMPIRIC_CUBIC = 2;
	public static final short LIFE_CUBIC = 3;
	public static final short VIPER_CUBIC = 4;
	public static final short POLTERGEIST_CUBIC = 5;
	public static final short BINDING_CUBIC = 6;
	public static final short AQUA_CUBIC = 7;
	public static final short SPARK_CUBIC = 8;
	public static final short ATTRACTIVE_CUBIC = 9;

	protected final L2Player _owner;
	protected L2Character _target;

	protected int _id;
	protected int _level = 1;

	protected ArrayList<Integer> _skills = new ArrayList<Integer>();

	private Future _disappearTask;
	private Future _actionTask;

	protected Boolean ActionBlock = false;

	public L2CubicInstance(final L2Player owner, final int id, final int level, final int lifetime)
	{
		_owner = owner;
		_id = id;
		_level = level;

		switch(_id)
		{
			case STORM_CUBIC:
				_skills.add(4049);
				break;
			case VAMPIRIC_CUBIC:
				_skills.add(4050);
				break;
			case LIFE_CUBIC:
				_skills.add(4051);
				_actionTask = ThreadPoolManager.getInstance().scheduleGeneralAtFixedRate(new Heal(33), 0, 10000);
				break;
			case VIPER_CUBIC:
				_skills.add(4052);
				break;
			case POLTERGEIST_CUBIC:
				_skills.add(4053);
				_skills.add(4054);
				_skills.add(4055);
				break;
			case BINDING_CUBIC:
				_skills.add(4164);
				break;
			case AQUA_CUBIC:
				_skills.add(4165);
				break;
			case SPARK_CUBIC:
				_skills.add(4166);
				break;
			case ATTRACTIVE_CUBIC:
				_skills.add(5115);
				_skills.add(5116);
				break;
		}
		if(_disappearTask == null)
			_disappearTask = ThreadPoolManager.getInstance().scheduleGeneral(new Disappear(), lifetime); // disappear in 20 mins
	}

	public void doAction(L2Character target)
	{
		if(ActionBlock)
			return;
		if(_target == target || _owner == target)
			return;
		stopAction();
		_target = target;
		if(_id == ATTRACTIVE_CUBIC)
			_actionTask = ThreadPoolManager.getInstance().scheduleGeneralAtFixedRate(new Action(50), 0, 5000);
		else
			_actionTask = ThreadPoolManager.getInstance().scheduleGeneralAtFixedRate(new Action(33), 0, 8000);
	}

	public int getId()
	{
		return _id;
	}

	public void setLevel(final int level)
	{
		_level = level;
	}

	public void stopAction()
	{
		if(ActionBlock)
			return;
		_target = null;
		if(_actionTask != null)
		{
			_actionTask.cancel(true);
			_actionTask = null;
		}
	}

	public void cancelDisappear()
	{
		if(_disappearTask != null)
		{
			ActionBlock = false;
			_disappearTask.cancel(true);
			_disappearTask = null;
		}
	}

	private class Action implements Runnable
	{
		private int _chance;
		private Boolean First = true;

		Action(final int chance)
		{
			_chance = chance;
		}

		public void run()
		{
			if(First)
			{
				ActionBlock = true;
				First = false;
			}
			else
				ActionBlock = false;

			final L2Skill skill = SkillTable.getInstance().getInfo(_skills.get(Rnd.get(_skills.size())), _level);
			if(skill != null)
			{
				if(_owner.isDead() || _target == null || _target.isDead() || !_owner.isInRangeZ(_target, skill.getCastRange()))
				{
					stopAction();
					if(_owner.isDead())
					{
						_owner.delCubic(_id);
						_owner.broadcastUserInfo();
						cancelDisappear();
					}
					return;
				}
				try
				{
					if(Rnd.chance(_chance))
						if(skill.checkCondition(_owner, _target, false, false, true))
						{
							FastList<L2Character> targets = new FastList<L2Character>();
							targets.add(_target);
							skill.useSkill(_owner, targets);
							_owner.broadcastPacket(new MagicSkillUse(_owner, _target, skill.getId(), _level, 0, 0));
						}
				}
				catch(final Exception e)
				{
					_log.log(Level.SEVERE, "", e);
				}
			}
		}
	}

	private class Heal implements Runnable
	{
		private int _chance;

		Heal(final int chance)
		{
			_chance = chance;
		}

		public void run()
		{
			if(_owner.isDead())
			{
				stopAction();
				_owner.delCubic(_id);
				_owner.broadcastUserInfo();
				cancelDisappear();
				return;
			}
			try
			{
				if(Rnd.chance(_chance))
				{
					final L2Skill skill = SkillTable.getInstance().getInfo(_skills.get(Rnd.get(_skills.size())), _level);
					L2Character target = null;
					if(_owner.getCurrentHp() < _owner.getMaxHp())
						target = _owner;
					else if(_owner.getPet() != null && _owner.getPet().getCurrentHp() < _owner.getPet().getMaxHp())
						target = _owner.getPet();
					else if(_owner.getParty() != null)
						for(L2Player member : _owner.getParty().getPartyMembers())
							if(member != null && member.getCurrentHp() < member.getMaxHp() && (target == null || target.getMaxHp() - target.getCurrentHp() < member.getMaxHp() - member.getCurrentHp()) && _owner.isInRangeZ(member, skill.getCastRange()))
								target = member;
					if(target != null && skill != null)
					{
						FastList<L2Character> targets = new FastList<L2Character>();
						targets.add(target);
						skill.useSkill(_owner, targets);
						_owner.broadcastPacket(new MagicSkillUse(_owner, target, skill.getId(), _level, 0, 0));
					}
				}
			}
			catch(final Exception e)
			{}
		}
	}

	private class Disappear implements Runnable
	{
		Disappear()
		{}

		public void run()
		{
			stopAction();
			_owner.delCubic(_id);
			_owner.broadcastUserInfo();
		}
	}
}