package L2jFT.game.model.instances;

import java.sql.ResultSet;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;

import L2jFT.Config;
import L2jFT.database.DatabaseUtils;
import L2jFT.database.FiltredPreparedStatement;
import L2jFT.database.L2DatabaseFactory;
import L2jFT.database.ThreadConnection;
import L2jFT.extensions.scripts.Events;
import L2jFT.game.ItemsAutoDestroy;
import L2jFT.game.ThreadPoolManager;
import L2jFT.game.ai.CtrlIntention;
import L2jFT.game.instancemanager.CursedWeaponsManager;
import L2jFT.game.instancemanager.MercTicketManager;
import L2jFT.game.model.L2Character;
import L2jFT.game.model.L2Clan;
import L2jFT.game.model.L2Object;
import L2jFT.game.model.L2Playable;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2Skill;
import L2jFT.game.model.L2World;
import L2jFT.game.model.base.L2Augmentation;
import L2jFT.game.network.serverpackets.InventoryUpdate;
import L2jFT.game.network.serverpackets.ItemList;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.skills.funcs.Func;
import L2jFT.game.tables.ItemTable;
import L2jFT.game.templates.L2Armor;
import L2jFT.game.templates.L2EtcItem;
import L2jFT.game.templates.L2EtcItem.EtcItemType;
import L2jFT.game.templates.L2Item;
import L2jFT.game.templates.L2Weapon.WeaponType;
import L2jFT.util.CharPosition;
import L2jFT.util.Log;
import L2jFT.util.Rnd;

public final class L2ItemInstance extends L2Object
{
	private static final Logger _log = Logger.getLogger(L2ItemInstance.class.getName());

	/** Enumeration of locations for item */
	public static enum ItemLocation
	{
		VOID,
		INVENTORY,
		PAPERDOLL,
		WAREHOUSE,
		CLANWH,
		PET,
		PET_EQUIP,
		FREIGHT,
		MONSTER
	}

	/** Item types to select */
	public static enum ItemClass
	{
		/** List all deposited items */
		ALL,
		/** Weapons, Armor, Jevels, Arrows, Baits*/
		EQUIPMENT,
		/** Soul/Spiritshot, Potions, Scrolls */
		CONSUMABLE,
		/** Common craft matherials */
		MATHERIALS,
		/** Special (item specific) craft matherials */
		PIECES,
		/** Crafting recipies */
		RECIPIES,
		/** Skill learn books */
		SPELLBOOKS,
		/** Dyes, lifestones */
		MISC,
		/** All other */
		OTHER
	}

	/** ID of the owner */
	private int _owner_id;

	/** Время жизни шэдов вэпонов **/
	LifeTimeTask _shadowItemLifeTimeTask;
	private int _lifeTimeRemaining;
	private L2Player _shadowItemOwner;

	/** Quantity of the item */
	private long _count;

	/** ID of the item */
	private final int _itemId;

	/** Object L2Item associated to the item */
	private final L2Item _itemTemplate;

	/** Location of the item */
	private ItemLocation _loc;

	/** Slot where item is stored */
	private short _loc_data;

	/** Level of enchantment of the item */
	private int _enchantLevel;

	/** Price of the item for selling */
	private int _price_sell;

	/** Price of the item for buying */
	private int _price_buy;

	private short _count_sell;

	/** Wear Item */
	private boolean _wear;

	private L2Augmentation _augmentation = null;
	int _augmentationId;

	/** Custom item types (used loto, race tickets) */
	private int _type1;
	private int _type2;

	/** Item drop time for autodestroy task */
	private long _dropTime;

	/** Item drop time */
	private long _dropTimeOwner;

	/** owner of the dropped item */
	private L2Player _itemDropOwner;

	public static final byte CHARGED_NONE = 0;
	public static final byte CHARGED_SOULSHOT = 1;
	public static final byte CHARGED_SPIRITSHOT = 1;
	public static final byte CHARGED_BLESSED_SPIRITSHOT = 2;

	private byte _chargedSoulshot = CHARGED_NONE;
	private byte _chargedSpiritshot = CHARGED_NONE;

	private boolean _chargedFishtshot = false;

	public static final byte UNCHANGED = 0;
	public static final byte ADDED = 1;
	public static final byte REMOVED = 3;
	public static final byte MODIFIED = 2;
	private byte _lastChange = 2; //1 ??, 2 modified, 3 removed
	private boolean _existsInDb; // if a record exists in DB.
	private boolean _storedInDb; // if DB data is up-to-date.

	@SuppressWarnings("unchecked")
	private Future _lazyUpdateInDb;

	/** Task of delayed update item info in database */
	public class LazyUpdateInDb implements Runnable
	{
		LazyUpdateInDb()
		{}

		public void run()
		{
			updateInDb();
		}
	}

	private final int destination;
	private final int source;
	private final String create_type;
	private int create_time;

	private short _max_count_sell;

	private long _lastRechargeTime;

	private int _rechargeTime;

	private boolean _itemCreation;

	private int _bodypart;

	private boolean _whflag = false;

	/**
	 * Constructor of the L2ItemInstance from the objectId and the itemId.
	 * @param objectId : int designating the ID of the object in the world
	 * @param itemId : int designating the ID of the item
	 */
	public L2ItemInstance(int objectId, int itemId, int destination, int source, String create_type)
	{
		super(objectId);
		_itemId = itemId;
		_itemTemplate = ItemTable.getInstance().getTemplate(itemId);
		if(_itemId == 0 || _itemTemplate == null)
			throw new IllegalArgumentException();
		_count = 1;
		_loc = ItemLocation.VOID;
		_type1 = 0;
		_type2 = 0;
		_dropTime = 0;
		_dropTimeOwner = 0;
		_itemDropOwner = null;

		this.destination = destination;
		this.source = source;
		this.create_type = create_type;
		create_time = (int) (System.currentTimeMillis() / 1000L);
		_lifeTimeRemaining = isShadowItem() ? _itemTemplate.getDurability() : -1;

		_bodypart = _itemTemplate.getBodyPart();
	}

	/**
	 * Constructor of the L2ItemInstance from the objetId and the description of the item given by the L2Item.
	 * @param objectId : int designating the ID of the object in the world
	 * @param item : L2Item containing informations of the item
	 */
	public L2ItemInstance(int objectId, L2Item item, int destination, int source, String create_type)
	{
		super(objectId);
		_itemId = item.getItemId();
		_itemTemplate = item;
		if(_itemId == 0)
			throw new IllegalArgumentException();
		_count = 1;
		_loc = ItemLocation.VOID;

		_dropTime = 0;
		_dropTimeOwner = 0;
		_itemDropOwner = null;

		this.destination = destination;
		this.source = source;
		this.create_type = create_type;
		create_time = (int) (System.currentTimeMillis() / 1000L);
		_lifeTimeRemaining = isShadowItem() ? _itemTemplate.getDurability() : -1;

		_bodypart = _itemTemplate.getBodyPart();
	}

	public int getBodyPart()
	{
		return _bodypart;
	}

	public void setBodyPart(int bodypart)
	{
		_bodypart = bodypart;
	}

	/**
	 * Sets the ownerID of the item
	 * @param owner_id : int designating the ID of the owner
	 */
	public void setOwnerId(int owner_id)
	{
		if(owner_id == _owner_id)
			return;
		_owner_id = owner_id;
		_storedInDb = false;

		if(isShadowItem())
			for(L2Player p : L2World.getAllPlayers())
				if(p.getObjectId() == _owner_id)
					_shadowItemOwner = p;
	}

	public void setOwner(L2Character owner)
	{
		if(owner.isPlayer)
			_shadowItemOwner = (L2Player) owner;
	}

	/**
	 * Returns the ownerID of the item
	 * @return int : ownerID of the item
	 */
	public int getOwnerId()
	{
		return _owner_id;
	}

	/**
	 * Sets the location of the item
	 * @param loc : ItemLocation (enumeration)
	 */
	public void setLocation(ItemLocation loc)
	{
		setLocation(loc, (short) 0);
	}

	/**
	 * Sets the location of the item.<BR><BR>
	 * <U><I>Remark :</I></U> If loc and loc_data different from database, say datas not up-to-date
	 * @param loc : ItemLocation (enumeration)
	 * @param loc_data : int designating the slot where the item is stored or the village for freights
	 */
	public void setLocation(ItemLocation loc, short loc_data)
	{
		if(loc == _loc && loc_data == _loc_data)
			return;
		_loc = loc;
		_loc_data = loc_data;
		_storedInDb = false;
	}

	public ItemLocation getLocation()
	{
		return _loc;
	}

	/**
	 * Возвращает количество предметов в int, использовать только для отсылки клиенту
	 * В прочих случаях использовать getFullCount
	 * @return int
	 */
	public int getCount()
	{
		return (int) Math.min(_count, Integer.MAX_VALUE);
	}

	/**
	 * Возвращает количество предметов без приведения к int
	 * По возможности следует использовать именно его
	 * @return long
	 */
	public long getFullCount()
	{
		return _count;
	}

	/**
	 * Sets the quantity of the item.<BR><BR>
	 * <U><I>Remark :</I></U> If loc and loc_data different from database, say datas not up-to-date
	 * @param count : long
	 */
	public void setCount(long count)
	{
		if(!isStackable() && count > 1)
		{
			_count = 1;
			Log.IllegalPlayerAction(getPlayer(), "tried to stack unstackable item " + getItemId(), 0);
			return;
		}
		if(_count == count)
			return;
		//if(count > Integer.MAX_VALUE)
		//	count = Integer.MAX_VALUE;
		_count = count;
		_storedInDb = false;
	}

	/**
	 * Returns if item is equipable
	 * @return boolean
	 */
	public boolean isEquipable()
	{
		return _itemTemplate.getItemType() == EtcItemType.BAIT || !(getBodyPart() == 0 || _itemTemplate instanceof L2EtcItem);
	}

	/**
	 * Returns if item is equipped
	 * @return boolean
	 */
	public boolean isEquipped()
	{
		return _loc == ItemLocation.PAPERDOLL || _loc == ItemLocation.PET_EQUIP;
	}

	/**
	 * Returns the slot where the item is stored
	 * @return int
	 */
	public short getEquipSlot()
	{
		return _loc_data;
	}

	/**
	 * Returns the characteristics of the item
	 * @return L2Item
	 */
	public L2Item getItem()
	{
		return _itemTemplate;
	}

	public int getCustomType1()
	{
		return _type1;
	}

	public int getCustomType2()
	{
		return _type2;
	}

	public void setCustomType1(int newtype)
	{
		_type1 = newtype;
	}

	public void setCustomType2(int newtype)
	{
		_type2 = newtype;
	}

	public void setDropTime(long time)
	{
		_dropTime = time;
	}

	public long getDropTime()
	{
		return _dropTime;
	}

	public void setDropTimeOwner(long time)
	{
		_dropTimeOwner = time;
	}

	public long getDropTimeOwner()
	{
		return _dropTimeOwner;
	}

	public void setItemDropOwner(L2Player owner)
	{
		_itemDropOwner = owner;
	}

	public L2Player getItemDropOwner()
	{
		return _itemDropOwner;
	}

	public boolean isWear()
	{
		return _wear;
	}

	public void setWear(boolean newwear)
	{
		_wear = newwear;
	}

	/**
	 * Returns the type of item
	 * @return Enum
	 */
	public Enum getItemType()
	{
		return _itemTemplate.getItemType();
	}

	/**
	 * Returns the ID of the item
	 * @return int
	 */
	public int getItemId()
	{
		return _itemId;
	}

	/**
	 * Returns the reference price of the item
	 * @return int
	 */
	public int getReferencePrice()
	{
		return _itemTemplate.getReferencePrice();
	}

	/**
	 * Returns the price of the item for selling
	 * @return int
	 */
	public int getPriceToSell()
	{
		return _price_sell;
	}

	/**
	 * Sets the price of the item for selling
	 * <U><I>Remark :</I></U> If loc and loc_data different from database, say datas not up-to-date
	 * @param price : int designating the price
	 */
	public void setPriceToSell(int price)
	{
		_price_sell = price;
		_storedInDb = false;
	}

	/**
	 * Returns the price of the item for buying
	 * @return int
	 */
	public int getPriceToBuy()
	{
		return _price_buy;
	}

	/**
	 * Sets the price of the item for buying
	 * <U><I>Remark :</I></U> If loc and loc_data different from database, say datas not up-to-date
	 * @param price : int
	 */
	public void setPriceToBuy(int price)
	{
		_price_buy = price;
		_storedInDb = false;
	}

	public void setCountToSell(short count)
	{
		_count_sell = count;
	}

	public short getCountToSell()
	{
		return _count_sell;
	}

	public void setMaxCountToSell(short count)
	{
		_max_count_sell = count;
	}

	public short getMaxCountToSell()
	{
		return _max_count_sell;
	}

	public void setLastRechargeTime(long lastRechargeTime)
	{
		_lastRechargeTime = lastRechargeTime;
	}

	public long getLastRechargeTime()
	{
		return _lastRechargeTime;
	}

	public void setRechargeTime(long rechargeTime)
	{
		_lastRechargeTime = rechargeTime;
	}

	public long getRechargeTime()
	{
		return _rechargeTime;
	}

	public void setItemCreation(boolean itemCreation)
	{
		_itemCreation = itemCreation;
	}

	public boolean getItemCreation()
	{
		return _itemCreation;
	}

	/**
	 * Returns the last change of the item
	 * @return int
	 */
	public int getLastChange()
	{
		return _lastChange;
	}

	/**
	 * Sets the last change of the item
	 * @param lastChange : int
	 */
	public void setLastChange(byte lastChange)
	{
		_lastChange = lastChange;
	}

	/**
	 * Returns if item is stackable
	 * @return boolean
	 */
	public boolean isStackable()
	{
		return _itemTemplate.isStackable();
	}

	/* (non-Javadoc)
	 * @see L2jFT.game.model.L2Object#onAction(L2jFT.game.model.L2Player)
	 */
	@Override
	public void onAction(L2Player player)
	{
		if(Events.onAction(player, this))
			return;

		if(player.isCursedWeaponEquipped() && CursedWeaponsManager.getInstance().isCursed(_itemId))
			return;

		// this causes the validate position handler to do the pickup if the location is reached.
		// mercenary tickets can only be picked up by the castle owner.	

		int _castleId = MercTicketManager.getInstance().getTicketCastleId(_itemId);

		if(_castleId > 0)
		{
			if((player.getClanPrivileges() & L2Clan.CP_CS_MERCENARIES) == L2Clan.CP_CS_MERCENARIES)
			{
				if(player.isInParty())
					player.sendMessage("You cannot pickup mercenaries while in a party.", "Вы не можете подбирать наемников, находясь в группе");
				else
					player.getAI().setIntention(CtrlIntention.AI_INTENTION_PICK_UP, this);
			}
			else
				player.sendMessage("You don't have rights to do this.");

			player.setTarget(this);
			player.getAI().setIntention(CtrlIntention.AI_INTENTION_IDLE, this, null);
			// Send a Server->Client ActionFailed to the L2Player in order to avoid that the client wait another packet	
			player.sendActionFailed();
		}
		else
			player.getAI().setIntention(CtrlIntention.AI_INTENTION_PICK_UP, this, null);
	}

	/**
	 * Returns the level of enchantment of the item
	 * @return int
	 */
	public int getEnchantLevel()
	{
		return _enchantLevel;
	}

	/**
	 * Sets the level of enchantment of the item
	 * @param enchantLevel level of enchant
	 */
	public void setEnchantLevel(int enchantLevel)
	{
		if(_enchantLevel == enchantLevel)
			return;
		_enchantLevel = enchantLevel;
		_storedInDb = false;
	}

	/**
	 * Returns the physical defense of the item
	 * @return int
	 */
	public int getPDef()
	{
		if(_itemTemplate instanceof L2Armor)
			return ((L2Armor) _itemTemplate).getPDef();
		return 0;
	}

	/**
	 * Returns false cause item can't be attacked
	 * @return boolean false
	 */
	@Override
	public boolean isAutoAttackable(@SuppressWarnings("unused")
	L2Character attacker)
	{
		return false;
	}

	/**
	 * Returns whether this item is augmented or not
	 * @return true if augmented
	 */
	public boolean isAugmented()
	{
		return _augmentation != null;
	}

	/**
	 * Returns the augmentation object for this item
	 * @return augmentation
	 */
	public L2Augmentation getAugmentation()
	{
		return _augmentation;
	}

	public int getAugmentationId()
	{
		return _augmentationId;
	}

	/**
	 * Sets a new augmentation
	 * @param augmentation
	 * @return return true if successfully
	 */
	public boolean setAugmentation(L2Augmentation augmentation)
	{
		// there shall be no previous augmentation..
		if(_augmentation != null)
			return false;
		_augmentation = augmentation;
		_augmentationId = augmentation.getAugmentationId();
		setLastChange(MODIFIED);
		return true;
	}

	/**
	 * Remove the augmentation
	 *
	 */
	public void removeAugmentation()
	{
		if(_augmentation == null)
			return;
		_augmentation.deleteAugmentationData();
		_augmentationId = 0;
		_augmentation = null;
		setLastChange(MODIFIED);
	}

	/**
	 * Returns the type of charge with SoulShot of the item.
	 * @return int (CHARGED_NONE, CHARGED_SOULSHOT)
	 */
	public byte getChargedSoulshot()
	{
		return _chargedSoulshot;
	}

	/**
	 * Returns the type of charge with SpiritShot of the item
	 * @return int (CHARGED_NONE, CHARGED_SPIRITSHOT, CHARGED_BLESSED_SPIRITSHOT)
	 */
	public byte getChargedSpiritshot()
	{
		return _chargedSpiritshot;
	}

	public boolean getChargedFishshot()
	{
		return _chargedFishtshot;
	}

	/**
	 * Sets the type of charge with SoulShot of the item
	 * @param type : int (CHARGED_NONE, CHARGED_SOULSHOT)
	 */
	public void setChargedSoulshot(byte type)
	{
		_chargedSoulshot = type;
	}

	/**
	 * Sets the type of charge with SpiritShot of the item
	 * @param type : int (CHARGED_NONE, CHARGED_SPIRITSHOT, CHARGED_BLESSED_SPIRITSHOT)
	 */
	public void setChargedSpiritshot(byte type)
	{
		_chargedSpiritshot = type;
	}

	public void setChargedFishshot(boolean type)
	{
		_chargedFishtshot = type;
	}

	/**
	 * This function basically returns a set of functions from
	 * L2Item/L2Armor/L2Weapon, but may add additional
	 * functions, if this particular item instance is enhanched
	 * for a particular player.
	 * @param player : L2Character designating the player
	 * @return Func[]
	 */
	public Func[] getStatFuncs(L2Character player)
	{
		return _itemTemplate.getStatFuncs(this, player);
	}

	/**
	 * Updates database.<BR><BR>
	 * <U><I>Concept : </I></U><BR>
	 *
	 * <B>IF</B> the item exists in database :
	 * <UL>
	 *		<LI><B>IF</B> the item has no owner, or has no location, or has a null quantity : remove item from database</LI>
	 *		<LI><B>ELSE</B> : update item in database</LI>
	 * </UL>
	 *
	 * <B> Otherwise</B> :
	 * <UL>
	 *		<LI><B>IF</B> the item hasn't a null quantity, and has a correct location, and has a correct owner : insert item in database</LI>
	 * </UL>
	 */
	public void updateDatabase()
	{
		updateDatabase(false);
	}

	public synchronized void updateDatabase(boolean commit)
	{
		if(_existsInDb)
		{
			if(_owner_id == 0 || _loc == ItemLocation.VOID || _count == 0)
				removeFromDb();
			else if(Config.LAZY_ITEM_UPDATE && isStackable())
			{
				if(commit)
				{
					// cancel lazy update task if need
					if(_lazyUpdateInDb != null)
					{
						_lazyUpdateInDb.cancel(true);
						_lazyUpdateInDb = null;
					}
					updateInDb();
					L2World.increaseUpdateItemCount();
				}
				else if(_lazyUpdateInDb == null)
				{
					_lazyUpdateInDb = ThreadPoolManager.getInstance().scheduleGeneral(new LazyUpdateInDb(), Config.LAZY_ITEM_UPDATE_TIME);
					L2World.increaseLazyUpdateItem();
				}
			}
			else
			{
				updateInDb();
				L2World.increaseUpdateItemCount();
			}
		}
		else
		{
			if(_count == 0)
				return;
			if(_loc == ItemLocation.VOID || _owner_id == 0)
				return;
			insertIntoDb();
		}
	}

	/**
	 * Returns a L2ItemInstance stored in database from its objectID
	 * @param objectId : int designating the objectID of the item
	 * @return L2ItemInstance
	 */
	public static L2ItemInstance restoreFromDb(int objectId)
	{
		L2ItemInstance inst = null;
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		FiltredPreparedStatement statement2 = null;
		ResultSet rset2 = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT owner_id, object_id, item_id, count, enchant_level, loc, loc_data, price_sell, price_buy, custom_type1, custom_type2, source, destination, create_type, create_time, shadow_life_time FROM items WHERE object_id = ?");
			statement.setInt(1, objectId);
			rset = statement.executeQuery();

			if(rset.next())
			{
				int owner_id = rset.getInt("owner_id");
				int item_id = rset.getInt("item_id");
				long count = rset.getLong("count");
				ItemLocation loc = ItemLocation.valueOf(rset.getString("loc"));
				byte loc_data = rset.getByte("loc_data");
				int enchant_level = rset.getInt("enchant_level");
				int custom_type1 = rset.getInt("custom_type1");
				int custom_type2 = rset.getInt("custom_type2");
				int price_sell = rset.getInt("price_sell");
				int price_buy = rset.getInt("price_buy");
				L2Item item = ItemTable.getInstance().getTemplate(item_id);
				if(item == null)
				{
					_log.severe("Item item_id=" + item_id + " not known, object_id=" + objectId);
					return null;
				}
				inst = new L2ItemInstance(objectId, item, rset.getInt("destination"), rset.getInt("source"), rset.getString("create_type"));
				inst._existsInDb = true;
				inst._storedInDb = true;
				inst._owner_id = owner_id;
				inst._count = count;
				inst._enchantLevel = enchant_level;
				inst._type1 = custom_type1;
				inst._type2 = custom_type2;
				inst._loc = loc;
				inst._loc_data = loc_data;
				inst._price_sell = price_sell;
				inst._price_buy = price_buy;
				inst._lifeTimeRemaining = rset.getInt("shadow_life_time");

				inst.create_time = rset.getInt("create_time");
			}
			else
			{
				_log.severe("Item object_id=" + objectId + " not found");
				return null;
			}

			//load augmentation
			statement2 = con.prepareStatement("SELECT attributes,skill,level FROM augmentations WHERE item_id=?");
			statement2.setInt(1, objectId);
			rset2 = statement2.executeQuery();
			if(rset2.next())
				inst.setAugmentation(new L2Augmentation(inst, rset2.getInt("attributes"), rset2.getInt("skill"), rset2.getInt("level"), false));

		}
		catch(Exception e)
		{
			_log.log(Level.SEVERE, "Could not restore item " + objectId + " from DB: " + e.getMessage());
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
			DatabaseUtils.closeResultSet(rset2);
			DatabaseUtils.closeStatement(statement2);
		}
		return inst;
	}

	/**
	 * Update the database with values of the item
	 */
	void updateInDb()
	{
		if(_wear)
			return;
		if(_storedInDb)
			return;

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("UPDATE items SET owner_id=?,count=?,loc=?,loc_data=?,enchant_level=?,price_sell=?,price_buy=?,custom_type1=?,custom_type2=?,destination=?,source=?,create_type=?,create_time=?,shadow_life_time=? WHERE object_id = ?");
			statement.setInt(1, _owner_id);
			statement.setLong(2, _count);
			statement.setString(3, _loc.name());
			statement.setInt(4, _loc_data);
			statement.setInt(5, getEnchantLevel());
			statement.setInt(6, _price_sell);
			statement.setInt(7, _price_buy);
			statement.setInt(8, getCustomType1());
			statement.setInt(9, getCustomType2());
			statement.setInt(10, destination);
			statement.setInt(11, source);
			statement.setString(12, create_type);
			statement.setInt(13, create_time);
			statement.setInt(14, getLifeTimeRemaining());
			statement.setInt(15, getObjectId());
			statement.executeUpdate();

			_existsInDb = true;
			_storedInDb = true;
		}
		catch(Exception e)
		{
			_log.log(Level.SEVERE, "Could not update item " + getObjectId() + " itemID " + _itemId + " count " + getCount() + " owner " + _owner_id + " in DB:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	/**
	 * Insert the item in database
	 */
	private void insertIntoDb()
	{
		if(_wear)
			return;
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("INSERT DELAYED INTO items (owner_id,item_id,count,loc,loc_data,enchant_level,price_sell,price_buy,object_id,custom_type1,custom_type2,destination,source,create_type,create_time,shadow_life_time,name,class) VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");
			statement.setInt(1, _owner_id);
			statement.setInt(2, _itemId);
			statement.setLong(3, _count);
			statement.setString(4, _loc.name());
			statement.setInt(5, _loc_data);
			statement.setInt(6, getEnchantLevel());
			statement.setInt(7, _price_sell);
			statement.setInt(8, _price_buy);
			statement.setInt(9, getObjectId());
			statement.setInt(10, _type1);
			statement.setInt(11, _type2);
			statement.setInt(12, destination);
			statement.setInt(13, source);
			statement.setString(14, create_type);
			statement.setInt(15, create_time);
			statement.setInt(16, getLifeTimeRemaining());
			statement.setString(17, getItem().getName());
			statement.setString(18, getItemClass().name());
			statement.executeUpdate();

			_existsInDb = true;
			_storedInDb = true;

			L2World.increaseInsertItemCount();
		}
		catch(Exception e)
		{
			_log.warning("Could not insert item " + getObjectId() + "; itemID=" + _itemId + "; count=" + getCount() + "; owner=" + _owner_id + "; source=" + source + "; destination=" + destination + "; create_type=" + create_type + "; exception: " + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	/**
	 * Delete item from database
	 */
	private void removeFromDb()
	{
		if(_wear)
			return;

		// cancel lazy update task if need
		if(Config.LAZY_ITEM_UPDATE && _lazyUpdateInDb != null)
		{
			_lazyUpdateInDb.cancel(true);
			_lazyUpdateInDb = null;
		}

		if(!_whflag)
			removeAugmentation();

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("DELETE FROM items WHERE object_id=" + _objectId + " limit 1");
			statement.executeUpdate();

			_existsInDb = false;
			_storedInDb = false;

			L2World.increaseDeleteItemCount();
		}
		catch(Exception e)
		{
			_log.log(Level.SEVERE, "Could not delete item " + getObjectId() + " in DB:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	public boolean isBlessedEnchantScroll()
	{
		switch(_itemId)
		{
			case 6569:
			case 6570:
			case 6571:
			case 6572:
			case 6573:
			case 6574:
			case 6575:
			case 6576:
			case 6577:
			case 6578:
				return true;
		}
		return false;
	}

	public boolean isCrystallEnchantScroll()
	{
		switch(_itemId)
		{
			case 731:
			case 732:
			case 949:
			case 950:
			case 953:
			case 954:
			case 957:
			case 958:
			case 961:
			case 962:
				return true;
		}
		return false;
	}

	public short getEnchantCrystalId(L2ItemInstance scroll)
	{
		int scrollItemId = scroll.getItemId();
		short crystalId = 0;
		switch(_itemTemplate.getCrystalType())
		{
			case L2Item.CRYSTAL_A:
				crystalId = 1461;
				break;
			case L2Item.CRYSTAL_B:
				crystalId = 1460;
				break;
			case L2Item.CRYSTAL_C:
				crystalId = 1459;
				break;
			case L2Item.CRYSTAL_D:
				crystalId = 1458;
				break;
			case L2Item.CRYSTAL_S:
				crystalId = 1462;
		}

		for(short scrollId : getEnchantScrollId())
			if(scrollItemId == scrollId)
				return crystalId;

		return 0;
	}

	public short[] getEnchantScrollId()
	{
		if(_itemTemplate.getType2() == L2Item.TYPE2_WEAPON)
			switch(_itemTemplate.getCrystalType())
			{
				case L2Item.CRYSTAL_A:
					return new short[] { 729, 6569, 731 };
				case L2Item.CRYSTAL_B:
					return new short[] { 947, 6571, 949 };
				case L2Item.CRYSTAL_C:
					return new short[] { 951, 6573, 953 };
				case L2Item.CRYSTAL_D:
					return new short[] { 955, 6575, 957 };
				case L2Item.CRYSTAL_S:
					return new short[] { 959, 6577, 961 };
			}
		else if(_itemTemplate.getType2() == L2Item.TYPE2_SHIELD_ARMOR || _itemTemplate.getType2() == L2Item.TYPE2_ACCESSORY)
			switch(_itemTemplate.getCrystalType())
			{
				case L2Item.CRYSTAL_A:
					return new short[] { 730, 6570, 732 };
				case L2Item.CRYSTAL_B:
					return new short[] { 948, 6572, 950 };
				case L2Item.CRYSTAL_C:
					return new short[] { 952, 6574, 954 };
				case L2Item.CRYSTAL_D:
					return new short[] { 956, 6576, 958 };
				case L2Item.CRYSTAL_S:
					return new short[] { 960, 6578, 962 };
			}
		return new short[] { 0, 0, 0 };
	}

	/**
	 * Returns if item is available for manipulation
	 * @return boolean
	 */
	public boolean isAvailable(L2Player player, boolean allowAdena)
	{
		return !isEquipped() && _itemTemplate.getType2() != 3 && (_itemTemplate.getType2() != 4 || _itemTemplate.getType1() != 1)
		// TODO: what does this mean?
		&& (player.getPet() == null || getObjectId() != player.getPet().getControlItemId())
		// Not Control item of currently summoned pet
		&& player.getEnchantScroll() != this && (allowAdena || getItemId() != 57);
	}

	/**
	 * Return true if item is hero-item
	 * @return boolean
	 */
	public boolean isHeroItem()
	{
		int myid = _itemId;
		return myid >= 6611 && myid <= 6621 || myid == 6842;
	}

	/**
	 * Return true if item is strider-item (bugle)
	 * @return boolean
	 */
	public boolean isStriderItem()
	{
		int myid = _itemId;
		return myid >= 4422 && myid <= 4424;
	}

	/**
	 * Return true if item can be destroyed
	 * @return boolean
	 */
	public boolean canBeDestroyed(L2Player player)
	{
		//is hero item?
		if(isHeroItem())
			return false;

		if(isStriderItem() && player.isMounted())
			return false;

		if(CursedWeaponsManager.getInstance().isCursed(_itemId))
			return false;

		/** TODO: fill conditions ( by Styx to Styx ;) ) */

		return _itemTemplate.isDestroyable();
	}

	/**
	 * Return true if item can be dropped
	 * @return boolean
	 */
	public boolean canBeDropped(L2Player player)
	{
		//is hero item?
		if(isHeroItem())
			return false;

		if(isAugmented())
			return false;

		//is quest item?
		if(_itemTemplate.getType2() == L2Item.TYPE2_QUEST)
			return false;

		if(isStriderItem() && player.isMounted())
			return false;

		if(CursedWeaponsManager.getInstance().isCursed(_itemId))
			return false;

		/** TODO: fill conditions ( by Styx to Styx ;) ) */

		return _itemTemplate.isDropable();
	}

	/**
	 * Return true if item can be trade or sold
	 * @return boolean
	 */
	public boolean canBeTraded(L2Player player)
	{
		//is hero item?
		if(isHeroItem())
			return false;

		if(isStriderItem() && player.isMounted())
			return false;

		if(isAugmented())
			return false;

		if(CursedWeaponsManager.getInstance().isCursed(_itemId))
			return false;

		/** TODO: fill conditions ( by Styx to Styx ;) ) */

		return _itemTemplate.isTradeable();
	}

	/**
	 * Можно ли положить на клановый склад или передать фрейтом
	 * @param player
	 * @return
	 */
	public boolean canBeStored(L2Player player)
	{
		if(isStriderItem() && player.isMounted())
			return false;

		if(CursedWeaponsManager.getInstance().isCursed(_itemId))
			return false;

		return _itemTemplate.isTradeable();
	}

	/**
	 * Return true if item can be crystallized
	 * @return boolean
	 */
	public boolean canBeCrystallized(L2Player player)
	{
		//is hero item?
		if(isHeroItem())
			return false;

		//crystallizable?
		if(!_itemTemplate.isCrystallizable())
			return false;

		//can player crystallize?
		int level = player.getSkillLevel(L2Skill.SKILL_CRYSTALLIZE);
		if(level < 1 || _itemTemplate.getCrystalType() > level)
		{
			player.sendPacket(new SystemMessage(SystemMessage.CANNOT_CRYSTALLIZE_CRYSTALLIZATION_SKILL_LEVEL_TOO_LOW));
			player.sendActionFailed();
			return false;
		}

		if(isStriderItem() && player.isMounted())
			return false;

		if(CursedWeaponsManager.getInstance().isCursed(_itemId))
			return false;

		/** TODO: fill conditions ( by Styx to Styx ;) ) */

		return true;
	}

	/**
	 * Return true if item can be enchanted
	 * @return boolean
	 */
	public boolean canBeEnchanted(@SuppressWarnings("unused")
	L2Player player)
	{
		//is hero item?
		if(isHeroItem())
			return false;

		if(getItemType() == WeaponType.ROD)
			return false;

		if(CursedWeaponsManager.getInstance().isCursed(_itemId))
			return false;

		if(isShadowItem())
			return false;

		/** TODO: fill conditions */
		return true;
	}

	/**
	 * Returns the item in String format
	 * @return String
	 */
	@Override
	public String toString()
	{
		return _itemTemplate.toString();
	}

	public boolean isNightLure()
	{
		return _itemId >= 8505 && _itemId <= 8513 || _itemId == 8485;
	}

	public void notifyEquipped(boolean equipped)
	{
		if(!equipped)
		{
			_shadowItemLifeTimeTask = null;
			return;
		}

		if(_shadowItemLifeTimeTask == null) // Если вещь уже надета метод может вызываться из refreshListeners
			setLifeTimeRemaining(getLifeTimeRemaining() - 1);

		if(!needsDestruction())
		{
			_shadowItemLifeTimeTask = new LifeTimeTask();
			ThreadPoolManager.getInstance().scheduleGeneral(_shadowItemLifeTimeTask, 60000);
		}
	}

	public boolean isShadowItem()
	{
		return _itemTemplate.isShadowItem();
	}

	public boolean isAltSeed()
	{
		return _itemTemplate.isAltSeed();
	}

	/**
	 * true означает завершить таск, false продолжить
	 */
	public boolean needsDestruction()
	{
		if(!isShadowItem())
			return true;

		// Если владельца нет в игре, вернуть true, что завершит таск
		if(_shadowItemOwner == null)
			return true;

		if(getLifeTimeRemaining() <= 0)
		{
			_shadowItemOwner.getInventory().unEquipItem(this);
			_shadowItemOwner.getInventory().destroyItem(this, getCount(), true, true);
			_shadowItemOwner.sendPacket(new SystemMessage(SystemMessage.S1_HAS_DISAPPEARED).addItemName(_itemTemplate.getItemId()));
			_shadowItemOwner.sendChanges();
			_shadowItemOwner.sendPacket(new ItemList(_shadowItemOwner, false)); // перестраховка
			_shadowItemOwner.broadcastUserInfo();
			return true;
		}
		return false;
	}

	public int getLifeTimeRemaining()
	{
		return _lifeTimeRemaining;
	}

	public void setLifeTimeRemaining(int lt)
	{
		//_log.info("Shadow Weapon: Old lifetime value: "+_lifeTimeRemaining+", new: "+lt);
		_lifeTimeRemaining = lt;

		if(_shadowItemOwner == null)
			return;

		if(_lifeTimeRemaining == 10 || _lifeTimeRemaining == 5 || _lifeTimeRemaining == 1 || _lifeTimeRemaining == 0)
		{
			SystemMessage sm;
			if(_lifeTimeRemaining == 10)
				sm = new SystemMessage(SystemMessage.S1S_REMAINING_MANA_IS_NOW_10);
			else if(_lifeTimeRemaining == 5)
				sm = new SystemMessage(SystemMessage.S1S_REMAINING_MANA_IS_NOW_5);
			else if(_lifeTimeRemaining == 1)
				sm = new SystemMessage(SystemMessage.S1S_REMAINING_MANA_IS_NOW_1);
			else
				sm = new SystemMessage(SystemMessage.S1S_REMAINING_MANA_IS_NOW_0);
			sm.addItemName(getItemId());
			_shadowItemOwner.sendPacket(sm);
		}

		_storedInDb = false;

		InventoryUpdate iu = new InventoryUpdate();
		iu.addModifiedItem(this);
		_shadowItemOwner.sendPacket(iu);
	}

	public class LifeTimeTask implements Runnable
	{
		public void run()
		{
			if(_shadowItemLifeTimeTask != this)
				return;

			if(!isEquipped())
				return;

			if(!_shadowItemOwner.isOnline())
			{
				_shadowItemOwner = null;
				return;
			}

			setLifeTimeRemaining(getLifeTimeRemaining() - 1);

			if(needsDestruction())
				return;

			ThreadPoolManager.getInstance().scheduleGeneral(this, 60000); //У шэдовов 1 цикл = 60 сек.
		}
	}

	public void dropToTheGround(L2Player lastAttacker, L2NpcInstance dropper)
	{
		if(dropper == null)
		{
			CharPosition dropPos = Rnd.coordsRandomize(lastAttacker, 70);
			dropMe(lastAttacker, dropPos.x, dropPos.y, dropPos.z);
			ItemsAutoDestroy.getInstance().addItem(this);
			return;
		}

		CharPosition dropPos = Rnd.coordsRandomize(dropper, 70);

		// Init the dropped L2ItemInstance and add it in the world as a visible object at the position where mob was last
		dropMe(dropper, dropPos.x, dropPos.y, dropPos.z);

		// Add drop to auto destroy item task
		if(getItemId() >= 8600 && getItemId() <= 8614)
			ItemsAutoDestroy.getInstance().addHerb(this);
		else if(Config.AUTODESTROY_ITEM_AFTER > 0)
			ItemsAutoDestroy.getInstance().addItem(this);

		// activate nonowner penalty
		setItemDropOwner(lastAttacker);
		L2Character MostHated = dropper.getMostHated();
		if(MostHated instanceof L2Playable)
			setItemDropOwner(MostHated.getPlayer());
		if(dropper instanceof L2RaidBossInstance)
			setDropTimeOwner(System.currentTimeMillis() + 285000);
		else
			setDropTimeOwner(System.currentTimeMillis());
	}

	public boolean isDestroyable()
	{
		return true;
	}

	public void setWhFlag(boolean whflag)
	{
		_whflag = whflag;
	}

	public void deleteMe()
	{
		removeFromDb();
		decayMe();
		L2World.removeObject(this);
	}

	public ItemClass getItemClass()
	{
		return _itemTemplate.getItemClass();
	}
}