package L2jFT.game.model.instances;

import javolution.util.FastList;
import javolution.util.FastMap;
import L2jFT.Config;
import L2jFT.game.ThreadPoolManager;
import L2jFT.game.instancemanager.CursedWeaponsManager;
import L2jFT.game.model.L2Character;
import L2jFT.game.model.L2Drop;
import L2jFT.game.model.L2DropData;
import L2jFT.game.model.L2Manor;
import L2jFT.game.model.L2Party;
import L2jFT.game.model.L2Playable;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2Summon;
import L2jFT.game.model.base.Experience;
import L2jFT.game.model.base.ItemToDrop;
import L2jFT.game.model.instances.L2ItemInstance.ItemLocation;
import L2jFT.game.model.quest.QuestState;
import L2jFT.game.network.serverpackets.InventoryUpdate;
import L2jFT.game.network.serverpackets.SocialAction;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.tables.ItemTable;
import L2jFT.game.templates.L2Item;
import L2jFT.game.templates.L2NpcTemplate;
import L2jFT.util.CharPosition;
import L2jFT.util.MinionList;
import L2jFT.util.Rnd;
import L2jFT.util.Util;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledFuture;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This class manages all Monsters.
 *
 * L2MonsterInstance :<BR><BR>
 * <li>L2MinionInstance</li>
 * <li>L2RaidBossInstance </li>
 *
 * @version $Revision: 1.20.4.6 $ $Date: 2005/04/06 16:13:39 $
 */
public class L2MonsterInstance extends L2NpcInstance
{
	final static Logger _log = Logger.getLogger(L2MonsterInstance.class.getName());

	protected final class RewardInfo
	{
		protected L2Character _attacker;
		protected int _dmg = 0;

		public RewardInfo(final L2Character attacker, final int dmg)
		{
			_attacker = attacker;
			_dmg = dmg;
		}

		public void addDamage(int dmg)
		{
			if(dmg < 0)
				dmg = 0;

			_dmg += dmg;
		}

		@Override
		public int hashCode()
		{
			return _attacker.getObjectId();
		}
	}

	public final class AbsorberInfo
	{
		/** The attacker L2Character concerned by this AbsorberInfo of this L2NpcInstance */
		L2Player _absorber;
		int _crystalId;
		double _absorbedHP;

		/**
		 * Constructor of AbsorberInfo.<BR><BR>
		 */
		AbsorberInfo(final L2Player attacker, final int crystalId, final double absorbedHP)
		{
			_absorber = attacker;
			_crystalId = crystalId;
			_absorbedHP = absorbedHP;
		}

		/**
		 * Return the Identifier of the absorber L2Character.<BR><BR>
		 */
		@Override
		public int hashCode()
		{
			return _absorber.getObjectId();
		}
	}

	private Boolean _dead = false;

	/** True if an over-hit enabled skill has successfully landed on the L2NpcInstance */
	private boolean _overhit;

	/** Stores the extra (over-hit) damage done to the L2NpcInstance when the attacker uses an over-hit enabled skill */
	private double _overhitDamage;

	/** Stores the attacker who used the over-hit enabled skill on the L2NpcInstance */
	private L2Character _overhitAttacker;

	protected final MinionList minionList;

	@SuppressWarnings("unchecked")
	private ScheduledFuture minionMaintainTask;

	private static final int MONSTER_MAINTENANCE_INTERVAL = 1000;

	/** Максимальный уровень мобов */
	private static final int MONSTER_MAX_LEVEL = 100;

	private HashSet<L2ItemInstance> _inventory = new HashSet<L2ItemInstance>();

	/** crops */
	private L2ItemInstance[] _harvestItems;
	private L2Item _seeded;

	/** True if a Soul Crystal was successfully used on the L2NpcInstance */
	private boolean _absorbed;

	/** The table containing all L2Player that successfully absorbed the soul of this L2NpcInstance */
	private ConcurrentHashMap<L2Player, AbsorberInfo> _absorbersList;

	/** Table containing all Items that a Dwarf can Sweep on this L2NpcInstance */
	private L2ItemInstance[] _sweepItems;

	/** Static tables containing all mobIDs that can have their souls absorbed, divided by max soul crystal level */
	private static final int[] _absorbingMOBS_level4 = {
			20583, 20584, 20585, 20586, 20587, 20588, 20636, 20637, 20638, 20639,
			20640, 20641, 20642, 20643, 20644, 20645 };
	private static final int[] _absorbingMOBS_level8 = {
			20646, 20647, 20648, 20649, 20650, 21006, 21007, 21008 };
	private static final int[] _absorbingMOBS_level10 = {
			20627, 20628, 20629, 20674, 20761, 20762, 20821, 20823, 20826, 20827,
			20828, 20829, 20830, 20831, 20858, 20859, 20860, 21009, 21010, 21062,
			21063, 21068, 21070 };
	private static final int[] _absorbingMOBS_level12 = {
			29022, 25163, 25269, 25453, 25328, 25109, 29020, 25283, 25286, 22215,
			22216, 22217, 29065 };
	private static final int[] _absorbingMOBS_level13 = {
			25338, 29019, 25319, 29028, 29046, 29047 };

	/** Soul Crystal Basic Informations */
	// First ID of each soul crystal
	private static final short _REDCRYSTAL_OFFSET = 4629;
	private static final short _REDCRYSTAL_LVL11 = 5577;
	private static final short _REDCRYSTAL_LVL12 = 5580;
	private static final short _REDCRYSTAL_LVL13 = 5908;
	private static final short _GREENCRYSTAL_OFFSET = 4640;
	private static final short _GREENCRYSTAL_LVL11 = 5578;
	private static final short _GREENCRYSTAL_LVL12 = 5581;
	private static final short _GREENCRYSTAL_LVL13 = 5911;
	private static final short _BLUECRYSTAL_OFFSET = 4651;
	private static final short _BLUECRYSTAL_LVL11 = 5579;
	private static final short _BLUECRYSTAL_LVL12 = 5582;
	private static final short _BLUECRYSTAL_LVL13 = 5914;

	// Max number of levels a soul crystal may reach
	private static final short _MAX_CRYSTALS_LEVEL = 13;
	// Chances for each event of a soul crystal
	private static final float _LEVEL_CHANCE = 0.325f;
	/** End of Soul Crystal Basic Informations */

	// For ALT_GAME_MATHERIALSDROP
	protected static final L2DropData[] _matdrop = new L2DropData[] {
	//                                           Item              Price Chance
			new L2DropData(1864, 1, 1, 50000, 1), // Stem              100   5%
			new L2DropData(1865, 1, 1, 25000, 1), // Varnish           200   2.5%
			new L2DropData(1866, 1, 1, 16666, 1), // Suede             300   1.6666%
			new L2DropData(1867, 1, 1, 33333, 1), // Animal Skin       150   3.3333%
			new L2DropData(1868, 1, 1, 50000, 1), // Thread            100   5%
			new L2DropData(1869, 1, 1, 25000, 1), // Iron Ore          200   2.5%
			new L2DropData(1870, 1, 1, 25000, 1), // Coal              200   2.5%
			new L2DropData(1871, 1, 1, 25000, 1), // Charcoal          200   2.5%
			new L2DropData(1872, 1, 1, 50000, 1), // Animal Bone       150   5%
			new L2DropData(1873, 1, 1, 10000, 1), // Silver Nugget     500   1%
			new L2DropData(1874, 1, 1, 1666, 20), // Oriharukon Ore    3000  0.1666%
			new L2DropData(1875, 1, 1, 1666, 20), // Stone of Purity   3000  0.1666%
			new L2DropData(1876, 1, 1, 5000, 20), // Mithril Ore       1000  0.5%
			new L2DropData(1877, 1, 1, 1000, 20), // Adamantite Nugget 5000  0.1%
			new L2DropData(4039, 1, 1, 833, 40), //  Mold Glue         6000  0.0833%
			new L2DropData(4040, 1, 1, 500, 40), //  Mold Lubricant    10000 0.05%
			new L2DropData(4041, 1, 1, 217, 40), //  Mold Hardener     23000 0.0217%
			new L2DropData(4042, 1, 1, 417, 40), //  Enria             12000 0.0417%
			new L2DropData(4043, 1, 1, 833, 40), //  Asofe             6000  0.0833%
			new L2DropData(4044, 1, 1, 833, 40) //   Thons             6000  0.0833%
	};

	protected static final L2DropData[] _herbs = new L2DropData[] {
	//                                            Name      Chance
			// HP/MP herbs
			new L2DropData(8600, 1, 1, 150000, 1), // of Life            10%
			new L2DropData(8601, 1, 1, 60000, 1), //  Greater of Life    4%
			new L2DropData(8602, 1, 1, 16000, 1), //  Superior of Life   0.8%
			new L2DropData(8603, 1, 1, 150000, 1), // of Mana            10%
			new L2DropData(8604, 1, 1, 60000, 1), //  Greater of Mana    4%
			new L2DropData(8605, 1, 1, 16000, 1), //  Superior of Mana   0.8%
			//new L2DropData(8614, 1, 1, 2000, 1), // f Recovery         0.2%

			// Speed herb
			new L2DropData(8611, 1, 1, 150000, 1), // of Speed           15%

			// Warrior herbs
			new L2DropData(8606, 1, 1, 150000, 1), // of Power           15%
			new L2DropData(8608, 1, 1, 150000, 1), // of Atk. Spd.       15%
			new L2DropData(8610, 1, 1, 150000, 1), // of Critical Attack 15%
			//new L2DropData(8612, 1, 1, 2000, 1), //   of Warrior       0.2%

			// Mystic herbs
			new L2DropData(8607, 1, 1, 150000, 1), // of Magic           15%
			new L2DropData(8609, 1, 1, 150000, 1), // of Casting Speed   15%
	//new L2DropData(8613, 1, 1, 2000, 1), //     of Mystic          0.2%
	};

	protected static final L2DropData[] _lifestones = new L2DropData[] {
			new L2DropData(8723, 1, 1, 100, 44, 46), // Life Stone: level 46
			new L2DropData(8724, 1, 1, 100, 47, 49), // Life Stone: level 49
			new L2DropData(8725, 1, 1, 100, 50, 52), // Life Stone: level 52
			new L2DropData(8726, 1, 1, 100, 53, 55), // Life Stone: level 55
			new L2DropData(8727, 1, 1, 100, 56, 58), // Life Stone: level 58
			new L2DropData(8728, 1, 1, 100, 59, 61), // Life Stone: level 61
			new L2DropData(8729, 1, 1, 100, 62, 66), // Life Stone: level 64
			new L2DropData(8730, 1, 1, 100, 67, 72), // Life Stone: level 67
			new L2DropData(8731, 1, 1, 100, 73, 75), // Life Stone: level 70
			new L2DropData(8732, 1, 1, 100, 76, MONSTER_MAX_LEVEL), // Life Stone: level 76
			new L2DropData(8733, 1, 1, 100, 44, 46), // Mid-Grade Life Stone: level 46
			new L2DropData(8734, 1, 1, 100, 47, 49), // Mid-Grade Life Stone: level 49
			new L2DropData(8735, 1, 1, 100, 50, 52), // Mid-Grade Life Stone: level 52
			new L2DropData(8736, 1, 1, 100, 53, 55), // Mid-Grade Life Stone: level 55
			new L2DropData(8737, 1, 1, 100, 56, 58), // Mid-Grade Life Stone: level 58
			new L2DropData(8738, 1, 1, 100, 59, 61), // Mid-Grade Life Stone: level 61
			new L2DropData(8739, 1, 1, 100, 62, 66), // Mid-Grade Life Stone: level 64
			new L2DropData(8740, 1, 1, 100, 67, 72), // Mid-Grade Life Stone: level 67
			new L2DropData(8741, 1, 1, 100, 73, 75), // Mid-Grade Life Stone: level 70
			new L2DropData(8742, 1, 1, 100, 76, MONSTER_MAX_LEVEL), // Mid-Grade Life Stone: level 76
			new L2DropData(8743, 1, 1, 100, 44, 46), // High-Grade Life Stone: level 46
			new L2DropData(8744, 1, 1, 100, 47, 49), // High-Grade Life Stone: level 49
			new L2DropData(8745, 1, 1, 100, 50, 52), // High-Grade Life Stone: level 52
			new L2DropData(8746, 1, 1, 100, 53, 55), // High-Grade Life Stone: level 55
			new L2DropData(8747, 1, 1, 100, 56, 58), // High-Grade Life Stone: level 58
			new L2DropData(8748, 1, 1, 100, 59, 61), // High-Grade Life Stone: level 61
			new L2DropData(8749, 1, 1, 100, 62, 66), // High-Grade Life Stone: level 64
			new L2DropData(8750, 1, 1, 100, 67, 72), // High-Grade Life Stone: level 67
			new L2DropData(8751, 1, 1, 100, 73, 75), // High-Grade Life Stone: level 70
			new L2DropData(8752, 1, 1, 100, 76, MONSTER_MAX_LEVEL) // High-Grade Life Stone: level 76
	};

	protected static final L2DropData[] _toplifestones = new L2DropData[] {
			new L2DropData(8753, 1, 1, 1000000, 44, 46), // Top-Grade Life Stone: level 46
			new L2DropData(8754, 1, 1, 1000000, 47, 49), // Top-Grade Life Stone: level 49
			new L2DropData(8755, 1, 1, 1000000, 50, 52), // Top-Grade Life Stone: level 52
			new L2DropData(8756, 1, 1, 1000000, 53, 55), // Top-Grade Life Stone: level 55
			new L2DropData(8757, 1, 1, 1000000, 56, 58), // Top-Grade Life Stone: level 58
			new L2DropData(8758, 1, 1, 1000000, 59, 61), // Top-Grade Life Stone: level 61
			new L2DropData(8759, 1, 1, 1000000, 62, 66), // Top-Grade Life Stone: level 64
			new L2DropData(8760, 1, 1, 1000000, 67, 72), // Top-Grade Life Stone: level 67
			new L2DropData(8761, 1, 1, 1000000, 73, 75), // Top-Grade Life Stone: level 70
			new L2DropData(8762, 1, 1, 1000000, 76, MONSTER_MAX_LEVEL) // Top-Grade Life Stone: level 76
	};

	protected static final L2DropData[] _raiditems = new L2DropData[] {
			new L2DropData(8158, 1, 3, 300000, 40, 49), // Skull of the Dead: C-Grade
			new L2DropData(8159, 1, 3, 300000, 50, 60), // Skull of the Dead: B-Grade
			new L2DropData(8160, 1, 3, 300000, 61, 76), // Skull of the Dead: A-Grade
			new L2DropData(8161, 1, 3, 300000, 40, 49), // Dragon's Heart: C-Grade
			new L2DropData(8162, 1, 3, 300000, 50, 60), // Dragon's Heart: B-Grade
			new L2DropData(8163, 1, 3, 300000, 61, 76), // Dragon's Heart: A-Grade
			new L2DropData(8164, 1, 3, 300000, 40, 49), // Egg of Earth: C-Grade
			new L2DropData(8165, 1, 3, 300000, 50, 60), // Egg of Earth: B-Grade
			new L2DropData(8166, 1, 3, 300000, 61, 76), // Egg of Earth: A-Grade
			new L2DropData(8167, 1, 3, 300000, 40, 49), // False Nucleus of Life: C-Grade
			new L2DropData(8168, 1, 3, 300000, 50, 60), // False Nucleus of Life: B-Grade
			new L2DropData(8169, 1, 3, 300000, 61, 76), // False Nucleus of Life: A-Grade
			new L2DropData(8170, 1, 3, 300000, 40, 49), // Angel's Soul: C-Grade
			new L2DropData(8171, 1, 3, 300000, 50, 60), // Angel's Soul: B-Grade
			new L2DropData(8172, 1, 3, 300000, 61, 76), // Angel's Soul: A-Grade
			new L2DropData(8176, 1, 3, 300000, 40, 76)  // Destruction Tombstone
	};

	/**
	 * load event drop from table
	 * structure:
	 * item
	 * quantity
	 * chance
	 * minlevel
	 * TODO: Надо будет сделать нормальный интерфейс для динамического изменения списка
	 */
	protected static ArrayList<L2DropData> _eventdrop = new ArrayList<L2DropData>();

	static
	{
		/*
		 _eventdrop.add(new L2DropData(5556, 1, 2, 40000, 10));
		 _eventdrop.add(new L2DropData(5557, 1, 1, 40000, 10));
		 _eventdrop.add(new L2DropData(5558, 1, 1, 80000, 10));
		 _eventdrop.add(new L2DropData(5559, 1, 1, 12500, 10));
		 */
	}

	/**
	 * Constructor of L2MonsterInstance (use L2Character and L2NpcInstance constructor).<BR><BR>
	 *
	 * <B><U> Actions</U> :</B><BR><BR>
	 * <li>Call the L2Character constructor to set the _template of the L2MonsterInstance (copy skills from template to object and link _calculators to NPC_STD_CALCULATOR) </li>
	 * <li>Set the name of the L2MonsterInstance</li>
	 * <li>Create a RandomAnimation Task that will be launched after the calculated delay if the server allow it </li><BR><BR>
	 *
	 * @param objectId Identifier of the object to initialized
	 * @param template to apply to the NPC
	 */
	public L2MonsterInstance(int objectId, L2NpcTemplate template)
	{
		super(objectId, template);
		minionList = new MinionList(this);

		isMonster = true;
	}

	@Override
	public boolean isMovementDisabled()
	{
		// Невозможность ходить для этих мобов
		return getNpcId() == 18344 || getNpcId() == 18345 || super.isMovementDisabled();
	}

	@Override
	public boolean isLethalImmune()
	{
		return getNpcId() == 22215 || getNpcId() == 22216 || getNpcId() == 22217 || super.isLethalImmune();
	}

	/**
	 * Return True if the attacker is not another L2MonsterInstance.<BR><BR>
	 */
	@Override
	public boolean isAutoAttackable(L2Character attacker)
	{
		return !attacker.isMonster;
	}

	/**
	 * Return True if the L2MonsterInstance is Agressive (aggroRange > 0).<BR><BR>
	 */
	@Override
	public boolean isAggressive()
	{
		return !_unAggred && getTemplate().aggroRange > 0;
	}

	@Override
	public void onSpawn()
	{
		_dead = false;
		super.onSpawn();
		spawnMinions();

		// Clear mob spoil, absorbs, seed
		setSpoiled(false);
		_sweepItems = null;
		_absorbed = false;
		_absorbersList = null;
		_seeded = null;
	}

	protected int getMaintenanceInterval()
	{
		return MONSTER_MAINTENANCE_INTERVAL;
	}

	public void spawnMinions()
	{
		if(getTemplate().getMinionData() != null)
			try
			{
				minionMaintainTask = ThreadPoolManager.getInstance().scheduleGeneral(new Runnable(){
					public void run()
					{
						try
						{
							minionList.maintainMinions();
						}
						catch(Throwable e)
						{
							_log.log(Level.SEVERE, "", e);
						}
					}
				}, getMaintenanceInterval());
			}
			catch(NullPointerException e)
			{}
	}

	@Override
	public void callMinions()
	{
		if(minionList.hasMinions())
			for(L2MinionInstance minion : minionList.getSpawnedMinions())
			{
				double distance = getDistance(minion.getX(), minion.getY());
				if(distance > 200)
					// Move the minion to the new coords
					if(!minion.isInCombat() && !minion.isDead())
					{
						// Calculate a new random coord for the minion based on the master's coord
						CharPosition pos = minionList.getMinionPosition();

						if(distance > 1000)
							minion.teleToLocation(pos.x, pos.y, pos.z);
						else
							minion.moveToLocation(pos.x, pos.y, pos.z, 0, true);
					}
			}
	}

	public synchronized void callMinionsToAssist(L2Character attacker)
	{
		if(minionList.hasMinions())
		{
			FastList<L2MinionInstance> minions = minionList.getSpawnedMinions();
			if(minions != null)
				for(FastList.Node<L2MinionInstance> n = minions.head(), end = minions.tail(); (n = n.getNext()) != end;)
				{
					if(n == null)
						break;

					L2MinionInstance minion = n.getValue();

					// Trigger the aggro condition of the minion
					if(minion != null && !minion.isInCombat() && !minion.isDead())
						minion.addDamage(attacker, 1);
				}
		}
	}

	public void setDead(boolean dead)
	{
		_dead = dead;
	}

	public void removeMinions()
	{
		if(minionMaintainTask != null)
			minionMaintainTask.cancel(true);

		minionList.maintainLonelyMinions();
	}

	@Override
	public FastList<L2MinionInstance> getSpawnedMinions()
	{
		return minionList.getSpawnedMinions();
	}

	public int getTotalSpawnedMinionsInstances()
	{
		return minionList.countSpawnedMinions();
	}

	public int getTotalSpawnedMinionsGroups()
	{
		return minionList.lazyCountSpawnedMinionsGroups();
	}

	public void notifyMinionDied(L2MinionInstance minion)
	{
		minionList.removeSpawnedMinion(minion);
	}

	public void notifyMinionSpawned(L2MinionInstance minion)
	{
		minionList.addSpawnedMinion(minion);
	}

	@Override
	public boolean hasMinions()
	{
		return minionList.hasMinions();
	}

	@Override
	public void doDie(L2Character killer)
	{
		if(minionMaintainTask != null)
			minionMaintainTask.cancel(true);

		synchronized (_dead)
		{
			if(_dead)
				return;
			_dead = true;
		}

		if(this instanceof L2ChestInstance && !((L2ChestInstance) this).isFake())
		{
			super.doDie(killer);
			return;
		}

		// Notify the Quest Engine of the L2NpcInstance death if necessary
		try
		{
			if(killer instanceof L2Playable)
			{
				// Get the L2Player that killed the L2NpcInstance
				final L2Player player = killer.getPlayer();
				if(player == null) // маловероятно но бывает
					return;
				List<QuestState> quests = new FastList<QuestState>();

				if(Config.KILL_COUNTER)
				{
					player.incrementKillsCounter(getNpcId());
					getTemplate().killscount++;
				}

				if(player.getParty() != null)
				{
					Map<String, List<QuestState>> tempMap = new FastMap<String, List<QuestState>>();

					for(L2Player pl : player.getParty().getPartyMembers())
					{
						QuestState[] questsForKills = pl.getQuestsForKills(this);

						if(questsForKills == null)
							continue;

						for(QuestState qs : questsForKills)
							if(qs.getQuest().isParty())
							{
								if(!qs.isCompleted() && !pl.isDead() && Util.checkIfInRange(1150, this, pl, true))
									if(tempMap.get(qs.getQuest().getName()) != null)
										tempMap.get(qs.getQuest().getName()).add(qs);
									else
									{
										List<QuestState> tempList = new FastList<QuestState>();
										tempList.add(qs);
										tempMap.put(qs.getQuest().getName(), tempList);
									}
							}
							else if(pl == player)
								quests.add(qs);
					}

					for(List<QuestState> list : tempMap.values())
						quests.add((QuestState) list.toArray()[Rnd.nextInt(list.size())]);
				}
				else if(player.getQuestsForKills(this) != null)
					for(QuestState qs : player.getQuestsForKills(this))
						quests.add(qs);

				// Notify the Quest Engine of the L2NpcInstance death
				for(QuestState qs : quests)
					qs.getQuest().notifyKill(this, qs);
				if(getFactionId().equals("varka_silenos_clan"))
					player.setVarka(0);
				if(getFactionId().equals("ketra_orc_clan"))
					player.setKetra(0);
			}
		}
		catch(final Exception e)
		{
			_log.log(Level.SEVERE, "", e);
		}

		// Distribute Exp and SP rewards to L2Player (including Summon owner) that hit the L2NpcInstance and to their Party members
		try
		{
			calculateRewards(killer);
		}
		catch(final Exception e)
		{
			_log.log(Level.SEVERE, "", e);
		}

		// Enhance soul crystals of the attacker if this L2NpcInstance had its soul absorbed
		try
		{
			if(killer instanceof L2Playable)
				levelSoulCrystals(killer.getPlayer());
		}
		catch(final Exception e)
		{
			_log.log(Level.SEVERE, "", e);
		}

		super.doDie(killer);
	}

	private void calculateRewards(final L2Character lastAttacker)
	{
		final L2Player player = lastAttacker.getPlayer();
		if(player == null || _aggroList.size() == 0)
			return;

		final int npcID = getTemplate().npcId;

		FastMap<L2Character, RewardInfo> rewards = new FastMap<L2Character, RewardInfo>().setShared(true);

		for(final AggroInfo info : _aggroList.values())
		{
			L2Character attacker = info.attacker;
			L2Character owner = null;

			if(attacker.isSummon)
				owner = attacker.getOwner();

			int damage = info.damage;

			if(npcID <= 0 || damage <= 1)
				continue;

			synchronized (this)
			{
				RewardInfo reward = rewards.get(attacker);
				if(attacker.isPlayer || attacker.isPet)
				{
					if(reward == null)
						rewards.put(attacker, new RewardInfo(attacker, damage));
					else
						reward.addDamage(damage);
				}
				else if(owner != null)
				{
					//TODO Проверить нужно ли нам учитывать разницу в уровне суммона и хозяина.
					int levelsDiff = owner.getLevel() - attacker.getLevel();
					if(levelsDiff < 0)
						levelsDiff = 0;
					damage = damage - levelsDiff * 50;
					if(damage < 0)
						damage = 0;
					reward = rewards.get(owner);
					if(reward == null)
						rewards.put(owner, new RewardInfo(owner, damage));
					else
						reward.addDamage(damage);
				}
			}
		}

		for(FastMap.Entry<L2Character, RewardInfo> e = rewards.head(), end = rewards.tail(); e != null && (e = e.getNext()) != end && e != null;)
		{
			final RewardInfo reward = e.getValue();
			if(reward == null)
				continue;

			final L2Character attacker = reward._attacker;
			L2Character master = attacker;

			if(master == null)
				continue;

			if(master.isDead())
				continue;

			L2Party attackerParty = null;
			if(master.isPlayer)
				attackerParty = ((L2Player) master).getParty();

			final int maxHp = getMaxHp();

			if(attackerParty == null)
			{
				int damage = reward._dmg;
				float penalty = 0;
				if(master.isPlayer)
				{
					final L2Summon summon = master.getPet();
					if(summon != null && summon.isSummon)
						penalty = ((L2SummonInstance) summon).getExpPenalty();
				}

				if(damage > maxHp)
					damage = maxHp;

				if(damage > 0)
				{
					final int diff = master.getLevel() - getLevel();

					long xp = 0;
					long sp = 0;

					if(master.knowsObject(this))
					{
						final long[] tmp = calculateExpAndSp(diff, damage);
						xp = tmp[0];
						xp *= 1 - penalty;
						sp = tmp[1];
					}

					if(isOverhit() && player == getOverhitAttacker())
					{
						int overHitExp = calculateOverhitExp(xp);
						player.sendPacket(new SystemMessage(SystemMessage.OVER_HIT));
						player.sendPacket(new SystemMessage(SystemMessage.ACQUIRED_S1_BONUS_EXPERIENCE_THROUGH_OVER_HIT).addNumber(overHitExp));
						xp += overHitExp;
					}

					master.addExpAndSp(xp, sp);
				}
				rewards.remove(master);
			}
			else
			{
				int partyDmg = 0;
				float partyMul = 1.f;

				final ArrayList<L2Player> rewardedMembers = new ArrayList<L2Player>();

				for(final L2Player partyMember : attackerParty.getPartyMembers())
				{
					if(partyMember == null || partyMember.isDead())
						continue;

					final RewardInfo ai = rewards.get(partyMember);
					if(ai != null)
						partyDmg += ai._dmg;

					final L2Summon pet = partyMember.getPet();
					if(pet != null)
						//final RewardInfo aiPet = rewards.get(pet);
						//if(aiPet != null)
						//	partyDmg += aiPet.dmg;
						rewards.remove(pet);

					if(!partyMember.isDead())
						rewardedMembers.add(partyMember);

					rewards.remove(partyMember);
				}

				if(partyDmg < maxHp)
					partyMul = (float) partyDmg / maxHp;
				else
					partyDmg = maxHp;

				if(partyDmg > 0)
				{
					long xp = 0;
					long sp = 0;

					if(master.knowsObject(this))
					{
						final int diff = attackerParty.getLevel() - getLevel();
						final long[] tmp = calculateExpAndSp(diff, partyDmg);
						xp = tmp[0];
						sp = tmp[1];
					}

					xp *= partyMul;
					sp *= partyMul;

					// Check for an over-hit enabled strike
					// (When in party, the over-hit exp bonus is given to the whole party and splitted proportionally through the party members)
					if(isOverhit() && player == getOverhitAttacker())
					{
						int overHitExp = calculateOverhitExp(xp);
						attackerParty.broadcastToPartyMembersInRange(player, new SystemMessage(SystemMessage.ACQUIRED_S1_BONUS_EXPERIENCE_THROUGH_OVER_HIT).addNumber(overHitExp), 2700);
						xp += overHitExp;
					}

					attackerParty.distributeXpAndSp(xp, sp, rewardedMembers, lastAttacker, getLevel());
				}
			}
		}

		rewards = null;

		// Check the drop of a cursed weapon
		CursedWeaponsManager.getInstance().dropAttackable(this, player);

		// Manage Base, Quests and Special Events drops of the L2NpcInstance
		doItemDrop(lastAttacker);

		// Manage Sweep drops of the L2NpcInstance
		if(isSpoiled())
			doSweepDrop(lastAttacker);

		if(!isRaid && !(this instanceof L2MinionInstance)) // С рейдов падают только топовые лайфстоны, с миньонов вообще ничего дополнительно не падает
		{
			final int leveldiff = calculateLevelDiffForDrop();
			final double chancemod = leveldiff > 1 ? Experience.penaltyModifier(leveldiff - 1, 9) : 1.;
			final double RATE_HP = ((L2NpcTemplate) _template).rateHp;

			// Дополнительный дроп материалов
			if(Config.ALT_GAME_MATHERIALSDROP && chancemod > 0 && (!isSeeded() || _seeded.isAltSeed()))
				for(L2DropData d : _matdrop)
					if(getLevel() >= d.getMinLevel() && Rnd.get(1, L2Drop.MAX_CHANCE) <= d.getChance() * Config.RATE_DROP_ITEMS * chancemod * (RATE_HP < 1 ? .5 : 1.))
						dropItem(lastAttacker.getPlayer(), d.getItemId(), RATE_HP < 1 ? 1 : (int) RATE_HP);

			// Эвентовый дроп, TODO: сделать интерфейс динамического изменения
			if(chancemod > 0)
				for(L2DropData evd : _eventdrop)
					if(getLevel() >= evd.getMinLevel() && Rnd.get(1, L2Drop.MAX_CHANCE) <= evd.getChance() * Config.RATE_DROP_ITEMS * chancemod * (RATE_HP < 1 ? .5 : 1.))
						dropItem(lastAttacker.getPlayer(), evd.getItemId(), (int) RATE_HP);

			// Хербы
			if(((L2NpcTemplate) _template).isDropHerbs)
				for(L2DropData h : _herbs)
					if(Rnd.get(1, L2Drop.MAX_CHANCE) <= h.getChance() * chancemod)
						dropItem(lastAttacker.getPlayer(), h.getItemId(), 1);

			// Лайфстоуны
			if(chancemod > 0)
				for(L2DropData l : _lifestones)
					if(getLevel() >= l.getMinLevel() && getLevel() <= l.getMaxLevel() && Rnd.get(1, L2Drop.MAX_CHANCE) <= l.getChance() * Config.RATE_DROP_ITEMS * chancemod * RATE_HP)
						dropItem(lastAttacker.getPlayer(), l.getItemId(), 1);
		}
		else if(isRaid)
		{
			// Лайфстоуны с рейдов
			for(L2DropData l : _toplifestones)
				if(getLevel() >= l.getMinLevel() && getLevel() <= l.getMaxLevel() && Rnd.get(1, L2Drop.MAX_CHANCE) <= l.getChance() * Config.RATE_DROP_ITEMS)
					dropItem(lastAttacker.getPlayer(), l.getItemId(), 1);
			// предметы для изучения клановых скилов (не падают с эпиков)
			if(!(this instanceof L2BossInstance))
				for(L2DropData l : _raiditems)
					if(getLevel() >= l.getMinLevel() && getLevel() <= l.getMaxLevel() && Rnd.get(1, L2Drop.MAX_CHANCE) <= l.getChance() * Config.RATE_DROP_ITEMS)
						dropItem(lastAttacker.getPlayer(), l.getItemId(), Rnd.get(l.getMinDrop(), l.getMinDrop()));
		}

		_aggroList.clear();
	}

	private int calculateLevelDiffForDrop()
	{
		if(Config.DEEPBLUE_DROP_RULES)
		{
			int maxLevel = 0, minLevel = 1000;
			final int mobLevel = getLevel();
			// Check to prevent very high level player to nearly kill mob and let low level player do the last hit.
			if(getAggroListSize() > 0)
				for(final L2Character atkChar : getAggroList().keySet())
					if(atkChar != null)
					{
						if(atkChar.getLevel() > maxLevel)
							maxLevel = atkChar.getLevel();
						if(atkChar.getLevel() < minLevel)
							minLevel = atkChar.getLevel();
					}

			// According to official data (Prima), deep blue mobs are 9 or more levels below players
			if(maxLevel - mobLevel >= 9)
				return maxLevel - mobLevel - 8;
			else if(minLevel <= mobLevel - 9 && Config.OVER_LEVEL_PENALTY)
				return mobLevel - minLevel - 8;
		}

		return 0;
	}

	public void giveItem(L2ItemInstance item, boolean store)
	{
		if(_inventory == null)
			_inventory = new HashSet<L2ItemInstance>();

		_inventory.add(item);

		if(store)
		{
			item.setOwnerId(getNpcId());
			item.setLocation(ItemLocation.MONSTER);
			item.updateDatabase();
		}
	}

	public HashSet<L2ItemInstance> getInventory()
	{
		if(_inventory == null)
			_inventory = new HashSet<L2ItemInstance>();

		return _inventory;
	}

	public void truncateInventory()
	{
		_inventory.clear();
	}

	@Override
	public void onRandomAnimation()
	{
		// Action id для живности 1-3
		broadcastPacket(new SocialAction(getObjectId(), Rnd.get(1, 3)));
	}

	@Override
	public int getKarma()
	{
		return 0;
	}

	/**
	 * Activate the absorbed soul condition on the L2NpcInstance.<BR><BR>
	 */
	public void absorbSoul()
	{
		_absorbed = true;
	}

	/**
	 * Return True if the L2NpcInstance had his soul absorbed.<BR><BR>
	 */
	public boolean isAbsorbed()
	{
		return _absorbed;
	}

	/**
	 * Adds an attacker that successfully absorbed the soul of this L2NpcInstance into the _absorbersList.<BR><BR>
	 *
	 * params:  attacker  - a valid L2Player
	 *      condition - an integer indicating the event when mob dies. This should be:
	 *              = 0   - "the crystal scatters";
	 *              = 1   - "the crystal failed to absorb. nothing happens";
	 *              = 2   - "the crystal resonates because you got more than 1 crystal on you";
	 *              = 3   - "the crystal cannot absorb the soul because the mob level is too low";
	 *              = 4   - "the crystal successfuly absorbed the soul";
	 */
	public void addAbsorber(final L2Player attacker, final int crystalId)
	{
		// The attacker must not be null
		if(attacker == null)
			return;

		// This L2NpcInstance must be of one type in the _absorbingMOBS_levelXX tables.
		// OBS: This is done so to avoid triggering the absorbed conditions for mobs that can't be absorbed.
		boolean validTarget = false;
		for(final int mobId : _absorbingMOBS_level4)
			if(getNpcId() == mobId)
			{
				validTarget = true;
				break;
			}
		if(!validTarget)
			for(final int mobId : _absorbingMOBS_level8)
				if(getNpcId() == mobId)
				{
					validTarget = true;
					break;
				}
		if(!validTarget)
			for(final int mobId : _absorbingMOBS_level10)
				if(getNpcId() == mobId)
				{
					validTarget = true;
					break;
				}
		if(!validTarget)
			for(final int mobId : _absorbingMOBS_level12)
				if(getNpcId() == mobId)
				{
					validTarget = true;
					break;
				}
		if(!validTarget)
			for(final int mobId : _absorbingMOBS_level13)
				if(getNpcId() == mobId)
				{
					validTarget = true;
					break;
				}
		if(!validTarget)
			return;

		// If we have no _absorbersList initiated, do it
		AbsorberInfo ai = null;
		if(_absorbersList == null)
			_absorbersList = new ConcurrentHashMap<L2Player, AbsorberInfo>();
		else
			ai = _absorbersList.get(attacker);

		// If the L2Character attacker isn't already in the _absorbersList of this L2NpcInstance, add it
		if(ai == null)
		{
			ai = new AbsorberInfo(attacker, crystalId, getCurrentHp());
			_absorbersList.put(attacker, ai);
		}
		else
		{
			ai._absorber = attacker;
			ai._crystalId = crystalId;
			ai._absorbedHP = getCurrentHp();
		}

		absorbSoul();
	}

	/**
	 * Calculate the leveling chance of Soul Crystals based on the attacker that killed this L2NpcInstance
	 *
	 * @param attacker The player that last hitted (killed) this L2NpcInstance
	 *
	 */
	private void levelSoulCrystals(final L2Character attacker)
	{
		// Only L2Player can absorb a soul
		if(attacker == null || !attacker.isPlayer)
		{
			_absorbed = false;
			_absorbersList = null;
			return;
		}

		// Init some useful vars
		boolean success = false;
		boolean levelPartyCrystals = false;
		final int thisMobId = getNpcId();
		final L2Player killer = (L2Player) attacker;

		// Check if this L2NpcInstance isn't within any of the groups of mobs that can be absorbed
		int minCrystalLevel = 0;
		int maxCrystalLevel = 0;
		for(final int mobId : _absorbingMOBS_level4)
			if(thisMobId == mobId)
			{
				maxCrystalLevel = 4;
				success = true;
				break;
			}
		if(!success)
			for(final int mobId : _absorbingMOBS_level8)
				if(thisMobId == mobId)
				{
					maxCrystalLevel = 8;
					success = true;
					break;
				}
		if(!success)
			for(final int mobId : _absorbingMOBS_level10)
				if(thisMobId == mobId)
				{
					maxCrystalLevel = 10;
					success = true;
					break;
				}
		if(!success)
			for(final int mobId : _absorbingMOBS_level12)
				if(thisMobId == mobId)
				{
					minCrystalLevel = 10;
					maxCrystalLevel = 12;
					success = true;
					levelPartyCrystals = true;
					break;
				}
		if(!success)
			for(final int mobId : _absorbingMOBS_level13)
				if(thisMobId == mobId)
				{
					minCrystalLevel = 12;
					maxCrystalLevel = 13;
					success = true;
					levelPartyCrystals = true;
					break;
				}

		// If this is not a valid L2NpcInstance, clears the _absorbersList and just return
		if(!success)
		{
			_absorbed = false;
			_absorbersList = null;
			return;
		}

		// If this mob is a boss, then skip some checkings
		success = true;
		if(!levelPartyCrystals)
		{
			// Fail if this L2NpcInstance isn't absorbed or there's no one in its _absorbersList
			if(!_absorbed || _absorbersList == null)
			{
				_absorbed = false;
				_absorbersList = null;
				return;
			}

			// Fail if the killer isn't in the _absorbersList of this L2NpcInstance and mob is not boss
			final AbsorberInfo ai = _absorbersList.get(killer);
			if(ai == null || ai._absorber.getObjectId() != killer.getObjectId())
				success = false;

			// Check if the soul crystal was used when HP of this L2NpcInstance wasn't higher than half of it
			if(ai != null && ai._absorbedHP > getMaxHp() / 2)
				success = false;

			if(!success)
			{
				_absorbed = false;
				_absorbersList = null;
				return;
			}
		}

		_absorbed = false;
		_absorbersList = null;

		// Init some more info now
		final float dice = Rnd.get();

		// Now we got four choices:
		// 1- The Monster level is too low for the crystal. Nothing happens.
		// 2- Everything is correct, but it failed. Nothing happens. (57.5%)
		// 3- Everything is correct, but it failed. The crystal scatters. A sound event is played. (10%)
		// 4- Everything is correct, the crystal level up. A sound event is played. (32.5%)

		ArrayList<L2Player> players;
		if(levelPartyCrystals && killer.isInParty())
			players = killer.getParty().getPartyMembers();
		else
			(players = new ArrayList<L2Player>()).add(killer);

		for(final L2Player player : players)
		{
			int oldCrystalId = 0;
			int newCrystalId = 0;
			short crystalsCount = 0;
			byte crystalLevel = 0;
			boolean canIncreaseCrystal = false;
			boolean resonated = false;

			// Check how many soul crystals the player has in his inventory and which soul crystal he has
			for(final L2ItemInstance item : player.getInventory().getItems())
			{
				final int itemId = item.getItemId();
				if(!(itemId == _REDCRYSTAL_LVL11 || itemId == _BLUECRYSTAL_LVL11 || itemId == _GREENCRYSTAL_LVL11 || itemId == _REDCRYSTAL_LVL12 || itemId == _BLUECRYSTAL_LVL12 || itemId == _GREENCRYSTAL_LVL12 || itemId == _REDCRYSTAL_LVL13 || itemId == _BLUECRYSTAL_LVL13 || itemId == _GREENCRYSTAL_LVL13))
					if(!(itemId >= _REDCRYSTAL_OFFSET || itemId >= _GREENCRYSTAL_OFFSET || itemId >= _BLUECRYSTAL_OFFSET) || !(itemId <= _REDCRYSTAL_OFFSET + 10 || itemId <= _GREENCRYSTAL_OFFSET + 10 || itemId <= _BLUECRYSTAL_OFFSET + 10))
						continue;

				if(++crystalsCount > 1)
				{
					resonated = true;
					break;
				}

				int crystalOffset = 0;
				for(crystalLevel = 0; crystalLevel <= _MAX_CRYSTALS_LEVEL; crystalLevel++)
				{
					// Get a crystal offset to calculate the new crystal ID
					// If the crystal level is way too high for this mob, say that we can't increase it
					// Define crystals' IDs based on the crystalOffset found

					crystalOffset = getNextLevelCrystalId(itemId, crystalLevel);
					if(crystalOffset == 0)
						continue;

					canIncreaseCrystal = crystalLevel >= minCrystalLevel && crystalLevel < maxCrystalLevel;
					oldCrystalId = itemId;
					newCrystalId = crystalOffset;
					break;
				}
			}

			if(crystalsCount < 1)
				continue;

			// The player has more than one soul crystal with him, the crystal resonates and we skip this player
			if(resonated)
			{
				// player.sendPacket(new PlaySound("ItemSound2.broken_key"));
				player.sendPacket(new SystemMessage(SystemMessage.THE_SOUL_CRYSTALS_CAUSED_RESONATION_AND_FAILED_AT_ABSORBING_A_SOUL));
				continue;
			}

			// The soul crystal stage of the player is way too high, refuse to increase it
			if(!canIncreaseCrystal)
			{
				player.sendPacket(new SystemMessage(SystemMessage.THE_SOUL_CRYSTAL_IS_REFUSING_TO_ABSORB_A_SOUL));
				continue;
			}

			// If the killer succeeds or it is a boss mob, level up the crystal
			if(levelPartyCrystals || dice < _LEVEL_CHANCE)
			{
				final L2ItemInstance oldCrystal = player.getInventory().getItemByItemId(oldCrystalId);
				if(oldCrystal == null)
					continue;

				final L2ItemInstance newCrystal = ItemTable.getInstance().createItem(newCrystalId, player.getObjectId(), getNpcId(), "L2NpcInstance.levelSoulCrystals");
				InventoryUpdate iu = new InventoryUpdate();
				iu.addItem(player.getInventory().addItem(newCrystal));
				iu.addItem(player.getInventory().destroyItem(oldCrystal, 1, true, true));
				// player.sendPacket(new PlaySound("ItemSound3.sys_enchant_success"));
				player.sendPacket(new SystemMessage(SystemMessage.EARNED_S1).addItemName(newCrystalId));
				player.sendPacket(new SystemMessage(SystemMessage.THE_SOUL_CRYSTAL_SUCCEEDED_IN_ABSORBING_A_SOUL));
				player.sendPacket(iu);
				continue;
			}

			player.sendPacket(new SystemMessage(SystemMessage.THE_SOUL_CRYSTAL_IS_REFUSING_TO_ABSORB_A_SOUL));
		}
	}

	private int getNextLevelCrystalId(int crystalId, final int crystalLevel)
	{
		if(crystalLevel > 9)
		{
			if(crystalId == _REDCRYSTAL_OFFSET + 10)
				return _REDCRYSTAL_LVL11;
			if(crystalId == _GREENCRYSTAL_OFFSET + 10)
				return _GREENCRYSTAL_LVL11;
			if(crystalId == _BLUECRYSTAL_OFFSET + 10)
				return _BLUECRYSTAL_LVL11;
			if(crystalId == _REDCRYSTAL_LVL11)
				return _REDCRYSTAL_LVL12;
			if(crystalId == _GREENCRYSTAL_LVL11)
				return _GREENCRYSTAL_LVL12;
			if(crystalId == _BLUECRYSTAL_LVL11)
				return _BLUECRYSTAL_LVL12;
			if(crystalId == _REDCRYSTAL_LVL12)
				return _REDCRYSTAL_LVL13;
			if(crystalId == _GREENCRYSTAL_LVL12)
				return _GREENCRYSTAL_LVL13;
			if(crystalId == _BLUECRYSTAL_LVL12)
				return _BLUECRYSTAL_LVL13;
		}
		if(crystalId == _REDCRYSTAL_OFFSET + crystalLevel || crystalId == _GREENCRYSTAL_OFFSET + crystalLevel || crystalId == _BLUECRYSTAL_OFFSET + crystalLevel)
			return ++crystalId;

		return 0;
	}

	public L2ItemInstance[] takeHarvest()
	{
		synchronized (_harvestItems)
		{
			final L2ItemInstance[] harvest = _harvestItems;
			_harvestItems = null;
			_seeded = null;
			return harvest;
		}
	}

	public void setSeeded(final short id)
	{
		_seeded = ItemTable.getInstance().getTemplate(id);
		// Количество всходов зависит от множителя HP конкретного моба 
		int count = (int) getTemplate().rateHp;
		// Для мобов с множителем HP < 1 число всходов = 1
		if(count == 0)
			count = 1;
		final List<L2ItemInstance> harvested = new FastList<L2ItemInstance>();
		final L2ItemInstance crops = ItemTable.getInstance().createItem(L2Manor.getInstance().getCropType(id), 0, getNpcId(), "L2NpcInstance.setSeeded");
		// Количество всходов от xHP до (xHP + xHP/2)  
		crops.setCount(Rnd.get(count * Config.RATE_MANOR, (int) (1.5 * count * Config.RATE_MANOR)));
		harvested.add(crops);

		if(harvested.size() > 0)
			_harvestItems = harvested.toArray(new L2ItemInstance[harvested.size()]);
	}

	public boolean isSeeded()
	{
		return _seeded != null;
	}

	/** True if a Dwarf has used Spoil on this L2NpcInstance */
	private boolean _isSpoiled;

	/**
	 * Return True if this L2NpcInstance has drops that can be sweeped.<BR><BR>
	 */
	public boolean isSpoiled()
	{
		return _isSpoiled;
	}

	/**
	 * Set the spoil state of this L2NpcInstance.<BR><BR>
	 */
	public void setSpoiled(boolean isSpoiled)
	{
		_isSpoiled = isSpoiled;
	}

	public void doItemDrop(L2Character lastAttacker)
	{
		L2Character mostHated = getMostHated();
		if(mostHated != null)
			lastAttacker = mostHated;

		final L2Player player = lastAttacker.getPlayer();

		if(player == null)
			return;

		final int levelDiff = calculateLevelDiffForDrop();

		if(getTemplate().getDropData() != null)
		{
			final FastList<ItemToDrop> drops = getTemplate().getDropData().rollDrop(levelDiff, this instanceof L2RaidBossInstance, player);
			for(final ItemToDrop drop : drops)
			{
				if(_seeded != null && !_seeded.isAltSeed() && drop.itemId != 57 && drop.itemId != 6360 && drop.itemId != 6361 && drop.itemId != 6362)
					continue;
				dropItem(player, drop.itemId, drop.count);
			}
		}

		final FastList<L2DropData> qdrops = new FastList<L2DropData>();

		player.fillQuestDrops(this, qdrops);

		for(final L2DropData drop : qdrops)
		{
			if(drop.isSweep())
				continue;
			dropItem(player, drop.getItemId(), Rnd.get(drop.getMinDrop(), drop.getMaxDrop()));
		}

		for(final L2ItemInstance drop : getInventory())
			if(drop != null)
			{
				player.sendMessage("This item belonged to other player: " + drop.getItem().getName(), "Эта вещь принадлежала другому игроку: " + drop.getItem().getName());
				dropItem(player, drop);
			}
		truncateInventory();
	}

	private void doSweepDrop(final L2Character lastAttacker)
	{
		final L2Player player = lastAttacker.getPlayer();

		if(player == null)
			return;

		final int levelDiff = calculateLevelDiffForDrop();

		final ArrayList<L2ItemInstance> spoiled = new ArrayList<L2ItemInstance>();

		if(getTemplate().getDropData() != null)
		{
			final FastList<ItemToDrop> spoils = getTemplate().getDropData().rollSpoil(levelDiff, this instanceof L2RaidBossInstance, player);
			for(final ItemToDrop spoil : spoils)
			{
				final L2ItemInstance dropit = ItemTable.getInstance().createItem(spoil.itemId, lastAttacker.getObjectId(), getNpcId(), "L2NpcInstance.doSweepDrop");
				dropit.setCount(spoil.count);
				spoiled.add(dropit);
			}
		}

		final FastList<L2DropData> qspoils = new FastList<L2DropData>();
		player.fillQuestDrops(this, qspoils);
		for(final L2DropData spoil : qspoils)
		{
			if(!spoil.isSweep())
				continue;
			final L2ItemInstance dropit = ItemTable.getInstance().createItem(spoil.getItemId(), lastAttacker.getObjectId(), getNpcId(), "L2NpcInstance.doSweepDrop");
			dropit.setCount(Rnd.get(spoil.getMinDrop(), spoil.getMaxDrop()));
			spoiled.add(dropit);
		}

		if(spoiled.size() > 0)
			_sweepItems = spoiled.toArray(new L2ItemInstance[spoiled.size()]);
	}

	private long[] calculateExpAndSp(long diff, final long damage)
	{
		long xp = (long) (getExpReward() / getTemplate().baseHpMax * damage);
		long sp = (long) (getSpReward() / getTemplate().baseHpMax * damage);

		if(Config.OVER_LEVEL_PENALTY)
			diff = Math.abs(diff);

		if(diff > 5)
		{
			final double mod = Math.pow(.83, diff - 5);
			xp *= mod;
			sp *= mod;
		}

		if(xp < 0)
			xp = 0;
		if(sp < 0)
			sp = 0;

		return new long[] { xp, sp };
	}

	@Override
	public void setOverhitEnabled(final boolean status)
	{
		_overhit = status;
	}

	@Override
	public void setOverhitValues(final L2Character attacker, final double damage)
	{
		// Calculate the over-hit damage
		// Ex: mob had 10 HP left, over-hit skill did 50 damage total, over-hit damage is 40
		final double overhitDmg = (getCurrentHp() - damage) * -1;
		if(overhitDmg < 0)
		{
			// we didn't killed the mob with the over-hit strike. (it wasn't really an over-hit strike)
			// let's just clear all the over-hit related values
			setOverhitEnabled(false);
			_overhitDamage = 0;
			_overhitAttacker = null;
			return;
		}
		setOverhitEnabled(true);
		_overhitDamage = overhitDmg;
		_overhitAttacker = attacker;
	}

	public L2Character getOverhitAttacker()
	{
		return _overhitAttacker;
	}

	public double getOverhitDamage()
	{
		return _overhitDamage;
	}

	@Override
	public boolean isOverhit()
	{
		return _overhit;
	}

	public int calculateOverhitExp(final long normalExp)
	{
		double overhitPercentage = getOverhitDamage() * 100 / getMaxHp();
		if(overhitPercentage > 25)
			overhitPercentage = 25;
		final double overhitExp = overhitPercentage / 100 * normalExp;
		final int bonusOverhit = (int) Math.round(overhitExp);
		return bonusOverhit;
	}

	/**
	 * Return True if a Dwarf use Sweep on the L2NpcInstance and if item can be spoiled.<BR><BR>
	 */
	public boolean isSweepActive()
	{
		return _sweepItems != null && _sweepItems.length > 0;
	}

	/**
	 * Return table containing all L2ItemInstance that can be spoiled.<BR><BR>
	 */
	public L2ItemInstance[] takeSweep()
	{
		synchronized (_sweepItems)
		{
			final L2ItemInstance[] sweep = _sweepItems;
			_sweepItems = null;
			return sweep;
		}
	}

	@Override
	public boolean isInvul()
	{
		return false;
	}
}
