package L2jFT.game.model.quest;

import javolution.util.FastList;
import javolution.util.FastMap;
import L2jFT.Config;
import L2jFT.database.DatabaseUtils;
import L2jFT.database.FiltredPreparedStatement;
import L2jFT.database.L2DatabaseFactory;
import L2jFT.database.ThreadConnection;
import L2jFT.game.instancemanager.QuestManager;
import L2jFT.game.model.L2Character;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2Skill;
import L2jFT.game.model.instances.L2NpcInstance;
import L2jFT.game.network.serverpackets.NpcHtmlMessage;
import L2jFT.game.network.serverpackets.QuestList;
import L2jFT.game.tables.NpcTable;
import L2jFT.game.templates.L2NpcTemplate;
import L2jFT.util.Files;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.sql.ResultSet;
import java.util.Collection;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

public abstract class Quest
{
	public static class Drop
	{
		public Integer chance;
		public Integer condition;
		public FastList<Short> itemList = new FastList<Short>();
		public Integer maxcount;

		public Drop(Integer _condition, Integer _maxcount, Integer _chance)
		{
			condition = _condition;
			maxcount = _maxcount;
			chance = _chance;
		}

		public void addItem(Short item)
		{
			itemList.add(item);
		}
	}

	public static enum QuestEventType
	{
		MOB_TARGETED_BY_SKILL(true), // control the first dialog shown by NPCs when
		MOBGOTATTACKED(true), // onTalk action from start npcs
		// attacked by someone
		MOBKILLED(true), // onTalk action from npcs participating in a quest
		NPC_FIRST_TALK(false), // onAttack action triggered when a mob gets
		// they are clicked (some quests must override
		// the default npc action)
		QUEST_START(true), // onKill action triggered when a mob gets killed.
		QUEST_TALK(true); // onSkillUse action triggered when a
		// character uses a skill on a mob

		// control whether this event type is allowed for the same npc template
		// in multiple quests
		// or if the npc must be registered in at most one quest for the
		// specified event
		private boolean _allowMultipleRegistration;

		QuestEventType(boolean allowMultipleRegistration)
		{
			_allowMultipleRegistration = allowMultipleRegistration;
		}

		public boolean isMultipleRegistrationAllowed()
		{
			return _allowMultipleRegistration;
		}
	}

	protected static Logger _log = Logger.getLogger(Quest.class.getName());

	/** HashMap containing events from String value of the event */
	private static Map<String, Quest> _allEventsS = new FastMap<String, Quest>();

	/** HashMap containing lists of timers from the name of the timer */
	private static Map<String, FastList<QuestTimer>> _allEventTimers = new FastMap<String, FastList<QuestTimer>>();

	/**
	 * Create a record in database for quest.<BR><BR>
	 * <U><I>Actions :</I></U><BR>
	 * Use fucntion createQuestVarInDb() with following parameters :<BR>
	 * <LI>QuestState : parameter sq that puts in fields of database :
	 * 	 <UL type="square">
	 *		 <LI>char_id : ID of the player</LI>
	 *		 <LI>name : name of the quest</LI>
	 *	 </UL>
	 * </LI>
	 * <LI>var : string "&lt;state&gt;" as the name of the variable for the quest</LI>
	 * <LI>val : string corresponding at the ID of the state (in fact, initial state)</LI>
	 * @param qs : QuestState
	 */
	public static void createQuestInDb(QuestState qs)
	{
		createQuestVarInDb(qs, "<state>", qs.getStateId());
	}

	/**
	 * Insert in the database the quest for the player.
	 * @param qs : QuestState pointing out the state of the quest
	 * @param var : String designating the name of the variable for the quest
	 * @param value : String designating the value of the variable for the quest
	 */
	public static void createQuestVarInDb(QuestState qs, String var, String value)
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("REPLACE INTO character_quests (char_id,name,var,value) VALUES (?,?,?,?)");
			statement.setInt(1, qs.getPlayer().getObjectId());
			statement.setString(2, qs.getQuest().getName());
			statement.setString(3, var);
			statement.setString(4, value);
			statement.executeUpdate();
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not insert char quest:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	/**
	 * Delete the player's quest from database.
	 * @param qs : QuestState pointing out the player's quest
	 */
	public static void deleteQuestInDb(QuestState qs)
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("DELETE FROM character_quests WHERE char_id=? AND name=?");
			statement.setInt(1, qs.getPlayer().getObjectId());
			statement.setString(2, qs.getQuest().getName());
			statement.executeUpdate();
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not delete char quest:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	/**
	 * Delete a variable of player's quest from the database.
	 * @param qs : object QuestState pointing out the player's quest
	 * @param var : String designating the variable characterizing the quest
	 */
	public static void deleteQuestVarInDb(QuestState qs, String var)
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("DELETE FROM character_quests WHERE char_id=? AND name=? AND var=?");
			statement.setInt(1, qs.getPlayer().getObjectId());
			statement.setString(2, qs.getQuest().getName());
			statement.setString(3, var);
			statement.executeUpdate();
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not delete char quest:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	/**
	 * Return collection view of the values contains in the allEventS
	 * @return Collection<Quest>
	 */
	public static Collection<Quest> findAllEvents()
	{
		return _allEventsS.values();
	}

	/**
	 * Add quests to the L2PCInstance of the player.<BR><BR>
	 * <U><I>Action : </U></I><BR>
	 * Add state of quests, drops and variables for quests in the HashMap _quest of L2Player
	 * @param player : Player who is entering the world
	 */
	public static void playerEnter(L2Player player)
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		FiltredPreparedStatement invalidQuestData = null;
		FiltredPreparedStatement invalidQuestDataVar = null;
		ResultSet rset = null;
		try
		{
			// Get list of quests owned by the player from database
			con = L2DatabaseFactory.getInstance().getConnection();

			invalidQuestData = con.prepareStatement("DELETE FROM character_quests WHERE char_id=? and name=?");
			invalidQuestDataVar = con.prepareStatement("delete FROM character_quests WHERE char_id=? and name=? and var=?");
			statement = con.prepareStatement("SELECT name,value FROM character_quests WHERE char_id=? AND var=?");
			statement.setInt(1, player.getObjectId());
			statement.setString(2, "<state>");
			rset = statement.executeQuery();
			while(rset.next())
			{

				// Get ID of the quest and ID of its state
				String questId = rset.getString("name");
				String stateId = rset.getString("value");

				// Search quest associated with the ID
				Quest q = QuestManager.getInstance().getQuest(questId);
				if(q == null)
				{
					_log.finer("Unknown quest " + questId + " for player " + player.getName());
					if(Config.AUTODELETE_INVALID_QUEST_DATA)
					{
						invalidQuestData.setInt(1, player.getObjectId());
						invalidQuestData.setString(2, questId);
						invalidQuestData.executeUpdate();
					}
					continue;
				}

				// Create an object State containing the state of the quest
				State state = q._states.get(stateId);
				if(state == null)
				{
					_log.finer("Unknown state " + state + " in quest " + questId + " for player " + player.getName());
					if(Config.AUTODELETE_INVALID_QUEST_DATA)
					{
						invalidQuestData.setInt(1, player.getObjectId());
						invalidQuestData.setString(2, questId);
						invalidQuestData.executeUpdate();
					}
					continue;
				}
				// Create a new QuestState for the player that will be added to the player's list of quests
				new QuestState(q, player, state);
			}
			invalidQuestData.close();
			DatabaseUtils.closeDatabaseSR(statement, rset);

			// Get list of quests owned by the player from the DB in order to add variables used in the quest.
			statement = con.prepareStatement("SELECT name,var,value FROM character_quests WHERE char_id=?");
			statement.setInt(1, player.getObjectId());
			rset = statement.executeQuery();
			while(rset.next())
			{
				String questId = rset.getString("name");
				String var = rset.getString("var");
				String value = rset.getString("value");
				// Get the QuestState saved in the loop before
				QuestState qs = player.getQuestState(questId);
				if(qs == null)
				{
					_log.finer("Lost variable " + var + " in quest " + questId + " for player " + player.getName());
					if(Config.AUTODELETE_INVALID_QUEST_DATA)
					{
						invalidQuestDataVar.setInt(1, player.getObjectId());
						invalidQuestDataVar.setString(2, questId);
						invalidQuestDataVar.setString(3, var);
						invalidQuestDataVar.executeUpdate();
					}
					continue;
				}
				// Add parameter to the quest
				qs.setInternal(var, value);
			}
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not insert char quest:", e);
		}
		finally
		{
			DatabaseUtils.closeStatement(invalidQuestData);
			DatabaseUtils.closeStatement(invalidQuestDataVar);
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}

		// events
		for(String name : _allEventsS.keySet())
			player.processQuestEvent(name, "enter");
	}

	/**
	 * Update informations regarding quest in database.<BR>
	 * <U><I>Actions :</I></U><BR>
	 * <LI>Get ID state of the quest recorded in object qs</LI>
	 * <LI>Save in database the ID state (with or without the star) for the variable called "&lt;state&gt;" of the quest</LI>
	 * @param qs : QuestState
	 */
	public static void updateQuestInDb(QuestState qs)
	{
		String val = qs.getStateId();
		updateQuestVarInDb(qs, "<state>", val);
	}

	/**
	 * Update the value of the variable "var" for the quest.<BR><BR>
	 * <U><I>Actions :</I></U><BR>
	 * The selection of the right record is made with :
	 * <LI>char_id = qs.getPlayer().getObjectID()</LI>
	 * <LI>name = qs.getQuest().getName()</LI>
	 * <LI>var = var</LI>
	 * <BR><BR>
	 * The modification made is :
	 * <LI>value = parameter value</LI>
	 * @param qs : Quest State
	 * @param var : String designating the name of the variable for quest
	 * @param value : String designating the value of the variable for quest
	 */
	public static void updateQuestVarInDb(QuestState qs, String var, String value)
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("UPDATE character_quests SET value=? WHERE char_id=? AND name=? AND var = ?");
			statement.setString(1, value);
			statement.setInt(2, qs.getPlayer().getObjectId());
			statement.setString(3, qs.getQuest().getName());
			statement.setString(4, var);
			statement.executeUpdate();
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not update char quest:", e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
		if(var.equals("cond"))
			qs.getPlayer().sendPacket(new QuestList());
	}

	protected final String _descr;

	protected State _initialState;

	protected final String _name;

	protected final boolean _party;

	protected final int _questId;

	protected Map<String, State> _states;

	/**
	 * (Constructor)Add values to class variables and put the quest in HashMaps. 
	 * @param questId : int pointing out the ID of the quest
	 * @param name : String corresponding to the name of the quest
	 * @param descr : String for the description of the quest
	 * @param party : boolean for the party quest drop
	 */
	public Quest(int questId, String name, String descr, boolean party)
	{
		_questId = questId;
		_name = name;
		_descr = descr;
		_party = party;
		_states = new FastMap<String, State>();
		if(questId != 0)
			QuestManager.getInstance().addQuest(this);
		else
			_allEventsS.put(name, this);
	}

	/**
	 * Add this quest to the list of quests that the passed mob will respond to
	 * for Attack Events.<BR>
	 * <BR>
	 * 
	 * @param attackId
	 * @return int : attackId
	 */
	public L2NpcTemplate addAttackId(int attackId)
	{
		return addEventId(attackId, Quest.QuestEventType.MOBGOTATTACKED);
	}

	/**
	 * Add this quest to the list of quests that the passed mob will respond to
	 * for the specified Event type.<BR>
	 * <BR>
	 * 
	 * @param npcId : id of the NPC to register
	 * @param eventType : type of event being registered
	 * @return int : npcId
	 */
	public L2NpcTemplate addEventId(int npcId, QuestEventType eventType)
	{
		try
		{
			L2NpcTemplate t = NpcTable.getTemplate(npcId);
			if(t != null)
				t.addQuestEvent(eventType, this);
			return t;
		}
		catch(Exception e)
		{
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Add the quest to the NPC's first-talk (default action dialog)
	 * 
	 * @param npcId
	 * @return L2NpcTemplate : Start NPC
	 */
	public L2NpcTemplate addFirstTalkId(int npcId)
	{
		return addEventId(npcId, Quest.QuestEventType.NPC_FIRST_TALK);
	}

	/**
	 * Add this quest to the list of quests that the passed mob will respond to
	 * for Kill Events.<BR>
	 * <BR>
	 * 
	 * @param killId
	 * @return int : killId
	 */
	public L2NpcTemplate addKillId(int killId)
	{
		return addEventId(killId, Quest.QuestEventType.MOBKILLED);
	}

	/**
	 * Add this quest to the list of quests that the passed npc will respond to
	 * for Skill-Use Events.<BR>
	 * <BR>
	 * 
	 * @param npcId : ID of the NPC
	 * @return int : ID of the NPC
	 */
	public L2NpcTemplate addSkillUseId(int npcId)
	{
		return addEventId(npcId, Quest.QuestEventType.MOB_TARGETED_BY_SKILL);
	}

	/**
	 * Add the quest to the NPC's startQuest
	 * 
	 * @param npcId
	 * @return L2NpcTemplate : Start NPC
	 */
	public L2NpcTemplate addStartNpc(int npcId)
	{
		return addEventId(npcId, Quest.QuestEventType.QUEST_START);
	}

	/**
	 * Add a state to the quest
	 * @param state
	 * @return state added
	 */
	public State addState(State state)
	{
		_states.put(state.getName(), state);
		return state;
	}

	/**
	 * Add this quest to the list of quests that the passed npc will respond to
	 * for Talk Events.<BR>
	 * <BR>
	 * 
	 * @param talkId : ID of the NPC
	 * @return int : ID of the NPC
	 */
	public L2NpcTemplate addTalkId(int talkId)
	{
		return addEventId(talkId, Quest.QuestEventType.QUEST_TALK);
	}

	public void cancelQuestTimer(String name, L2NpcInstance npc, L2Player player)
	{
		QuestTimer timer = getQuestTimer(name, npc, player);
		if(timer != null)
			timer.cancel();
	}

	/**
	 * Return description of the quest
	 * @return String
	 */
	public String getDescr()
	{
		return _descr;
	}

	/**
	 * Return initial state of the quest
	 * @return State
	 */
	public State getInitialState()
	{
		return _initialState;
	}

	/**
	 * Return name of the quest
	 * @return String
	 */
	public String getName()
	{
		return _name;
	}

	/**
	 * Return a QuestPcSpawn for the given player instance
	 */
	public final QuestPcSpawn getPcSpawn(L2Player player)
	{
		return QuestPcSpawnManager.getInstance().getPcSpawn(player);
	}

	/**
	 * Return ID of the quest
	 * @return int
	 */
	public int getQuestIntId()
	{
		return _questId;
	}

	public QuestTimer getQuestTimer(String name, L2NpcInstance npc, L2Player player)
	{
		if(_allEventTimers.get(name) == null)
			return null;
		for(QuestTimer timer : _allEventTimers.get(name))
			if(timer.isMatch(this, name, npc, player))
				return timer;
		return null;
	}

	public FastList<QuestTimer> getQuestTimers(String name)
	{
		return _allEventTimers.get(name);
	}

	/**
	 * Return party state of quest
	 * @return String
	 */
	public boolean isParty()
	{
		return _party;
	}

	/**
	 * Add a new QuestState to the database and return it.
	 * @param player
	 * @return QuestState : QuestState created
	 */
	public QuestState newQuestState(L2Player player)
	{
		QuestState qs = new QuestState(this, player, getInitialState());
		Quest.createQuestInDb(qs);
		return qs;
	}

	public final boolean notifyAttack(L2NpcInstance npc, QuestState qs)
	{
		String res = null;
		try
		{
			res = onAttack(npc, qs);
		}
		catch(Exception e)
		{
			return showError(qs.getPlayer(), e);
		}
		return showResult(qs.getPlayer(), res);
	}

	public final boolean notifyDeath(L2NpcInstance killer, L2Character victim, QuestState qs)
	{
		String res = null;
		try
		{
			res = onDeath(killer, victim, qs);
		}
		catch(Exception e)
		{
			return showError(qs.getPlayer(), e);
		}
		return showResult(qs.getPlayer(), res);
	}

	public final boolean notifyEvent(String event, QuestState qs)
	{
		String res;
		try
		{
			res = onEvent(event, qs);
		}
		catch(Exception e)
		{
			return showError(qs.getPlayer(), e);
		}
		return showResult(qs.getPlayer(), res);
	}

	// override the default NPC dialogs when a quest defines this for the given NPC
	public final boolean notifyFirstTalk(L2NpcInstance npc, L2Player player)
	{
		String res = null;
		try
		{
			res = onFirstTalk(npc, player);
		}
		catch(Exception e)
		{
			return showError(player, e);
		}
		player.setLastQuestNpcObject(npc.getObjectId());
		// if the quest returns text to display, display it. Otherwise, use the
		// default npc text.
		if(res != null && res.length() > 0)
			return showResult(player, res);
		npc.showChatWindow(player, 0);
		return true;
	}

	public final boolean notifyKill(L2NpcInstance npc, QuestState qs)
	{
		String res = null;
		try
		{
			res = onKill(npc, qs);
		}
		catch(Exception e)
		{
			return showError(qs.getPlayer(), e);
		}
		return showResult(qs.getPlayer(), res);
	}

	public final boolean notifySkillUse(L2NpcInstance npc, L2Player caster, L2Skill skill)
	{
		String res = null;
		try
		{
			res = onSkillUse(npc, caster, skill);
		}
		catch(Exception e)
		{
			return showError(caster, e);
		}
		return showResult(caster, res);
	}

	public final boolean notifyTalk(L2NpcInstance npc, QuestState qs)
	{
		String res = null;
		try
		{
			res = onTalk(npc, qs);
		}
		catch(Exception e)
		{
			return showError(qs.getPlayer(), e);
		}
		qs.getPlayer().setLastQuestNpcObject(npc.getObjectId());
		return showResult(qs.getPlayer(), res);
	}

	@SuppressWarnings("unused")
	public String onAttack(L2NpcInstance npc, QuestState qs)
	{
		return null;
	}

	@SuppressWarnings("unused")
	public String onDeath(L2NpcInstance killer, L2Character victim, QuestState qs)
	{
		return null;
	}

	@SuppressWarnings("unused")
	public String onEvent(String event, QuestState qs)
	{
		return null;
	}

	@SuppressWarnings("unused")
	public String onFirstTalk(L2NpcInstance npc, L2Player player)
	{
		return null;
	}

	@SuppressWarnings("unused")
	public String onKill(L2NpcInstance npc, QuestState qs)
	{
		return null;
	}

	@SuppressWarnings("unused")
	public String onSkillUse(L2NpcInstance npc, L2Player caster, L2Skill skill)
	{
		return null;
	}

	@SuppressWarnings("unused")
	public String onTalk(L2NpcInstance npc, QuestState qs)
	{
		return null;
	}

	public void removeQuestTimer(QuestTimer timer)
	{
		if(timer == null)
			return;
		FastList<QuestTimer> timers = getQuestTimers(timer.getName());
		if(timers == null)
			return;
		timers.remove(timer);
	}

	/**
	 * Set the initial state of the quest with parameter "state"
	 * @param state
	 */
	public void setInitialState(State state)
	{
		_initialState = state;
	}

	/**
	 * Show message error to player who has an access level greater than 0
	 * 
	 * @param player : L2PcInstance
	 * @param t : Throwable
	 * @return boolean
	 */
	private boolean showError(L2Player player, Throwable t)
	{
		_log.log(Level.WARNING, "", t);
		if(player.getPlayerAccess().IsGM)
		{
			StringWriter sw = new StringWriter();
			PrintWriter pw = new PrintWriter(sw);
			t.printStackTrace(pw);
			pw.close();
			String res = "<html><body><title>Script error</title>" + sw.toString() + "</body></html>";
			return showResult(player, res);
		}
		return false;
	}

	/**
	 * Show HTML file to client
	 * @param fileName
	 * @return String : message sent to client
	 */
	public String showHtmlFile(L2Player player, String fileName)
	{
		return showHtmlFile(player, fileName, null, (String[]) null); //(String[]) - затык для вызова корректного метода.
	}

	public String showHtmlFile(L2Player player, String fileName, String toReplace, String replaceWith)
	{
		return showHtmlFile(player, fileName, new String[] { toReplace }, new String[] { replaceWith });
	}

	public String showHtmlFile(L2Player player, String fileName, String toReplace[], String replaceWith[])
	{
		String questId = getName();
		String content = Files.read("data/jscript/" + getDescr().toLowerCase() + "/" + questId + "/" + fileName, player);

		// for new directory structure aka _number_questname
		if(content == null)
			content = Files.read("data/jscript/" + getDescr().toLowerCase() + "/_" + questId + "/" + fileName, player);

		if(content == null)
			content = Files.read("data/jscript/quests/" + questId + "/" + fileName, player);

		// for new directory structure
		if(content == null)
			content = Files.read("data/jscript/quests/_" + questId + "/" + fileName, player);

		// for scripts
		if(content == null)
			content = Files.read("data/scripts/" + getClass().getPackage().toString().replace(".", "/").substring(8) + "/" + fileName, player);

		if(content == null)
			content = "Can't find file '" + fileName + "'";

		if(player != null && player.getTarget() != null)
			content = content.replaceAll("%objectId%", String.valueOf(player.getTarget().getObjectId()));

		// Make a replacement inside before sending html to client
		if(toReplace != null && replaceWith != null && toReplace.length == replaceWith.length)
			for(int i = 0; i < toReplace.length; i++)
				content = content.replaceAll(toReplace[i], replaceWith[i]);

		// Send message to client if message not empty
		if(content != null && player != null)
		{
			NpcHtmlMessage npcReply = new NpcHtmlMessage(5);
			npcReply.setHtml(content);
			player.sendPacket(npcReply);
		}

		return content;
	}

	/**
	 * Show a message to player.<BR><BR>
	 * <U><I>Concept : </I></U><BR>
	 * 3 cases are managed according to the value of the parameter "res" :<BR>
	 * <LI><U>"res" ends with string ".html" :</U> an HTML is opened in order to be shown in a dialog box</LI>
	 * <LI><U>"res" starts with "<html>" :</U> the message hold in "res" is shown in a dialog box</LI>
	 * <LI><U>otherwise :</U> the message hold in "res" is shown in chat box</LI>
	 * @param qs : QuestState 
	 * @param res : String pointing out the message to show at the player
	 * @return boolean
	 */
	private boolean showResult(L2Player player, String res)
	{
		if(res == null)
			return true;
		if(res.endsWith(".htm"))
			showHtmlFile(player, res);
		else if(res.startsWith("<html>"))
		{
			NpcHtmlMessage npcReply = new NpcHtmlMessage(5);
			npcReply.setHtml(res);
			player.sendPacket(npcReply);
		}
		else
			//FIXME Случайно не это выводит чарам дебаг при прохождении квестов?
			player.sendMessage(res);

		return false;
	}

	/**
	 * Add a timer to the quest, if it doesn't exist already
	 * 
	 * @param name: name of the timer (also passed back as "event" in
	 *            onAdvEvent)
	 * @param time: time in ms for when to fire the timer
	 * @param npc: npc associated with this timer (can be null)
	 * @param player: player associated with this timer (can be null)
	 */
	public void startQuestTimer(String name, long time, L2NpcInstance npc, L2Player player)
	{
		// Add quest timer if timer doesn't already exist
		FastList<QuestTimer> timers = getQuestTimers(name);
		// no timer exists with the same name, at all
		if(timers == null)
		{
			timers = new FastList<QuestTimer>();
			timers.add(new QuestTimer(this, name, time, npc, player));
			_allEventTimers.put(name, timers);
		}
		// a timer with this name exists, but may not be for the same set of npc
		// and player
		else // if there exists a timer with this name, allow the timer only if
		// the [npc, player] set is unique
		// nulls act as wildcards
		if(getQuestTimer(name, npc, player) == null)
			timers.add(new QuestTimer(this, name, time, npc, player));
	}
}
