package L2jFT.game.model.quest;

import javolution.util.FastList;
import javolution.util.FastMap;
import L2jFT.Config;
import L2jFT.game.GameTimeController;
import L2jFT.util.Files;
import L2jFT.util.Log;
import L2jFT.util.Rnd;
import L2jFT.game.instancemanager.QuestManager;
import L2jFT.game.model.L2Character;
import L2jFT.game.model.L2DropData;
import L2jFT.game.model.L2Object;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2World;
import L2jFT.game.model.base.Experience;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.model.instances.L2MonsterInstance;
import L2jFT.game.model.instances.L2NpcInstance;
import L2jFT.game.network.serverpackets.InventoryUpdate;
import L2jFT.game.network.serverpackets.NpcSay;
import L2jFT.game.network.serverpackets.PlaySound;
import L2jFT.game.network.serverpackets.QuestList;
import L2jFT.game.network.serverpackets.StatusUpdate;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.network.serverpackets.TutorialCloseHtml;
import L2jFT.game.network.serverpackets.TutorialEnableClientEvent;
import L2jFT.game.network.serverpackets.TutorialShowHtml;
import L2jFT.game.network.serverpackets.TutorialShowQuestionMark;
import L2jFT.game.network.serverpackets.ExShowQuestMark;
import L2jFT.game.tables.ItemTable;
import L2jFT.game.templates.L2Item;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

public final class QuestState
{
	private static HashMap<String, String> _globalVars = new HashMap<String, String>();

	protected static Logger _log = Logger.getLogger(Quest.class.getName());

	public static String getGlobal(String name)
	{
		return _globalVars.get(name);
	}

	public static void setGlobal(String name, String value)
	{
		_globalVars.put(name, value);
	}

	/** List of drops needed for quest according to the mob */
	private Map<Integer, FastList<L2DropData>> _drops;

	/** Boolean flag letting QuestStateManager know to exit quest when cleaning up */
	private boolean _IsExitQuestOnCleanUp = false;

	/** Player who engaged the quest */
	private L2Player _player;

	/** Quest associated to the QuestState */
	private Quest _quest;

	/** State of the quest */
	private State _state;

	/** List of couples (variable for quest,value of the variable for quest) */
	private Map<String, String> _vars;

	/**
	 * Constructor of the QuestState : save the quest in the list of quests of the player.<BR/><BR/>
	 *
	 * <U><I>Actions :</U></I><BR/>
	 * <LI>Save informations in the object QuestState created (Quest, Player, Completion, State)</LI>
	 * <LI>Add the QuestState in the player's list of quests by using setQuestState()</LI>
	 * <LI>Add drops gotten by the quest</LI>
	 * <BR/>
	 * @param quest : quest associated with the QuestState
	 * @param player : L2Player pointing out the player
	 * @param state : state of the quest
	 */
	QuestState(Quest quest, L2Player player, State state)
	{
		_quest = quest;
		_player = player;

		// Save the state of the quest for the player in the player's list of quest onwed
		_player.setQuestState(this);

		// set the state of the quest
		_state = state;

		// add drops from state of the quest
		if(state != null)
		{
			Map<Integer, FastList<L2DropData>> new_drops = state.getDrops();
			if(new_drops != null)
			{
				_drops = new FastMap<Integer, FastList<L2DropData>>();
				_drops.putAll(new_drops);
			}
		}
	}

	/**
	 * Add XP and SP as quest reward
	 * @param exp
	 * @param sp
	 */
	public void addExpAndSp(int exp, int sp)
	{
		_player.addExpAndSp((int) (exp * Config.RATE_QUESTS_REWARD * _player.getBonus().RATE_QUESTS_REWARD), (int) (sp * Config.RATE_QUESTS_REWARD * _player.getBonus().RATE_QUESTS_REWARD));
	}

	/**
	 * Add player to get notification of characters death
	 * @param character : L2Character of the character to get notification of death
	 */
	public void addNotifyOfDeath(L2Character character)
	{
		if(character == null)
			return;
		character.addNotifyQuestOfDeath(this);
	}

	/**
	 * Add drop to the list of drops for the quest
	 * @param npcId : int pointing out the NPC given the drop
	 * @param itemId : int pointing out the ID of the item dropped
	 * @param chance : int pointing out the chance to get the drop
	 */
	public void addQuestDrop(int npcId, int itemId, int chance)
	{
		if(_drops == null)
			_drops = new FastMap<Integer, FastList<L2DropData>>();
		L2DropData d = new L2DropData();
		d.setItemId((short) itemId);
		d.setChance(chance);
		d.setQuestID(_quest.getName());
		d.addStates(new String[] { _state.getName() });
		FastList<L2DropData> lst = _drops.get(npcId);
		if(lst != null)
			lst.add(d);
		else
		{
			lst = new FastList<L2DropData>();
			lst.add(d);
			_drops.put(npcId, lst);
		}
	}

	public void addRadar(int x, int y, int z)
	{
		_player.radar.addMarker(x, y, z);
	}

	/**
	 * Clear all drops from the class variable "drops"
	 */
	public void clearQuestDrops()
	{
		_drops = null;
	}

	public void clearRadar()
	{
		_player.radar.removeAllMarkers();
	}

	public void closeTutorialHtml()
	{
		_player.sendPacket(new TutorialCloseHtml());
		if(Config.DEBUG)
			_log.fine("close Tutorial Html.");
	}

	public void dropItem(L2MonsterInstance npc, L2Player player, int itemId, int count)
	{
		npc.dropItem(player, itemId, count);
	}

	/**
	 * Destroy element used by quest when quest is exited
	 * @param repeatable
	 * @return QuestState
	 */
	public QuestState exitCurrentQuest(boolean repeatable)
	{
		// Clean drops
		if(_drops != null)
		{
			InventoryUpdate iu = new InventoryUpdate();
			// Go through values of class variable "drops" pointing out mobs that drop for quest
			for(Iterator<FastList<L2DropData>> i = _drops.values().iterator(); i.hasNext();)
			{
				List<L2DropData> lst = i.next();
				for(L2DropData d : lst)
				{
					int itemId = d.getItemId();
					// Get [item from] / [presence of the item in] the inventory of the player
					L2ItemInstance item = _player.getInventory().getItemByItemId(itemId);
					if(item == null || itemId == 57)
						continue;
					int count = item.getCount();
					// If player has the item in inventory, destroy it (if not gold)
					iu.addItem(_player.getInventory().destroyItemByItemId(itemId, count, false, true));
					_player.getWarehouse().destroyItem(itemId, count);
				}
			}
			// Send new inventory list to player
			_player.sendPacket(iu);
			_drops = null;
		}

		// If quest is repeatable, delete quest from list of quest of the player and from database (quest CAN be created again => repeatable)
		if(repeatable)
		{
			_player.delQuestState(_quest.getName());
			Quest.deleteQuestInDb(this);
			_vars = null;
		}
		else
		{ // Otherwise, delete variables for quest and update database (quest CANNOT be created again => not repeatable)
			if(_vars != null)
				for(String var : _vars.keySet().toArray(new String[_vars.size()]))
					unset(var);
			setState(new State("Completed", _quest));
			Quest.updateQuestInDb(this);
		}
		_player.sendPacket(new QuestList());
		return this;
	}

	public void exitQuest(int id)
	{
		QuestState qs = _player.getQuestState(QuestManager.getInstance().getQuest(id).getName());
		if(qs != null)
			qs.exitCurrentQuest(true);
	}

	/**
	 * Add quest drops to the parameter "drops"
	 * @param npc : L2NpcInstance killed
	 * @param drops : List of drops of the L2NpcInstance
	 */
	public void fillQuestDrops(L2NpcInstance npc, FastList<L2DropData> drops)
	{
		if(_drops == null)
			return;
		// Get drops of the NPC recorded in class variable "drops"
		FastList<L2DropData> lst = _drops.get(npc.getTemplate().npcId);
		// If drops of the NPC in class variable "drops" exist, add them to parameter "drops"
		if(lst != null)
			drops.addAll(lst);
		// Get drops for all NPC recorded in class variable "drops"
		lst = _drops.get(0); // all mobs
		// If drops for all NPC in class variable "drops" exist, add them to parameter "drops"
		if(lst != null)
			drops.addAll(lst);
	}

	/**
	 * Return the value of the variable of quest represented by "var"
	 * @param var : name of the variable of quest
	 * @return Object
	 */
	public Object get(String var)
	{
		if(_vars == null)
			return null;
		return _vars.get(var);
	}

	/**
	 * Return list of drops needed for the quest in concordance with mobs
	 * @return FastMap
	 */
	public Map<Integer, FastList<L2DropData>> getDrops()
	{
		return _drops;
	}

	/**
	 * Return the level of enchantment on the weapon of the player(Done specifically for weapon SA's)
	 * @param itemId : ID of the item to check enchantment
	 * @return int
	 */
	public int getEnchantLevel(int itemId)
	{
		L2ItemInstance enchanteditem = _player.getInventory().getItemByItemId(itemId);
		if(enchanteditem == null)
			return 0;
		return enchanteditem.getEnchantLevel();
	}

	/**
	 * Return the number of ticks from the GameTimeController
	 * @return int
	 */
	public int getGameTicks()
	{
		return GameTimeController.getGameTicks();
	}

	/**
	 * Return the value of the variable of quest represented by "var"
	 * @param var : String designating the variable for the quest
	 * @return int
	 */
	public int getInt(String var)
	{
		int varint = 0;
		try
		{
			varint = Integer.parseInt(_vars.get(var));
		}
		catch(Exception e)
		{
			_log.finer(getPlayer().getName() + ": variable " + var + " isn't an integer: " + varint + e);
		}
		return varint;
	}

	/**
	 * Return item number which is equipped in selected slot
	 * @return int
	 */
	public int getItemEquipped(int loc)
	{
		return getPlayer().getInventory().getPaperdollItemId(loc);
	}

	/**
	 * NOTE: This is to be deprecated; replaced by
	 * Quest.getPcSpawn(L2PcInstance) For now, I shall leave it as is Return a
	 * QuestPcSpawn for current player instance
	 */
	//TODO снести
	public final QuestPcSpawn getPcSpawn()
	{
		return QuestPcSpawnManager.getInstance().getPcSpawn(getPlayer());
	}

	/**
	 * Return the L2Player
	 * @return L2Player
	 */
	public L2Player getPlayer()
	{
		return _player;
	}

	/**
	 * Return the quest
	 * @return Quest
	 */
	public Quest getQuest()
	{
		return _quest;
	}

	/**
	 * Return the quantity of one sort of item hold by the player
	 * @param itemId : ID of the item wanted to be count
	 * @return int
	 */
	public int getQuestItemsCount(int itemId)
	{
		int count = 0;
		for(L2ItemInstance item : _player.getInventory().getItems())
			if(item != null && item.getItemId() == itemId)
				count += item.getCount();
		return count;
	}

	/**
	 * Return the QuestTimer object with the specified name
	 * @return QuestTimer<BR> Return null if name does not exist
	 */
	public final QuestTimer getQuestTimer(String name)
	{
		return getQuest().getQuestTimer(name, null, getPlayer());

	}

	/**
	 * Возвращает случайное значение<BR>
	 * Будьте внимательны, этот метод возвращает число от 0 до max.<br>
	 * Где 0 включительно, а max <b>исключительно</b>.<br>
	 * Этот метод устаревший, используйте Rnd.chance() или Rnd.get()
	 * @param max : max value for randomisation
	 * @return int
	 */
	@Deprecated
	public int getRandom(int max)
	{
		return Rnd.get(max);
	}

	/**
	 * Return the state of the quest
	 * @return State
	 */
	public State getState()
	{
		return _state;
	}

	/**
	 * Return ID of the state of the quest
	 * @return String
	 */
	public String getStateId()
	{
		return _state.getName();
	}

	/**
	 * Give item/reward to the player
	 * @param itemId
	 * @param count
	 */
	public void giveItems(int itemId, int count)
	{
		giveItems(itemId, count, 0);
	}

	public void giveItems(int itemId, int count, int enchantlevel)
	{
		if(count <= 0)
			count = 1;

		// If item for reward is gold (ID=57), modify count with rate for quest reward
		if(itemId == 57)
			count = (int) (count * Config.RATE_QUESTS_REWARD * _player.getBonus().RATE_QUESTS_REWARD);

		// Get template of item
		L2Item template = ItemTable.getInstance().getTemplate(itemId);
		if(template == null)
			return;

		InventoryUpdate iu = new InventoryUpdate();
		if(template.isStackable())
		{
			L2ItemInstance item = ItemTable.getInstance().createItem(itemId, _player.getObjectId(), 0, "QuestState.giveItems");

			// Set quantity of item
			item.setCount(count);

			// Add items to player's inventory
			iu.addItem(_player.getInventory().addItem(item));

			if(enchantlevel > 0)
				item.setEnchantLevel(enchantlevel);

			Log.LogItem(_player, Log.GetQuestItem, item);
		}
		else
			for(int i = 0; i < count; i++)
			{
				L2ItemInstance item = ItemTable.getInstance().createItem(itemId, _player.getObjectId(), 0, "QuestState.giveItems");

				// Set quantity of item
				item.setCount(1);

				// Add items to player's inventory
				iu.addItem(_player.getInventory().addItem(item));

				if(enchantlevel > 0)
					item.setEnchantLevel(enchantlevel);

				Log.LogItem(_player, Log.GetQuestItem, item);
			}

		// If item for reward is gold, send message of gold reward to client
		if(template.getItemId() == 57)
			_player.sendPacket(new SystemMessage(SystemMessage.EARNED_S1_ADENA).addNumber(count));
		else if(count > 1)
		{
			SystemMessage smsg = new SystemMessage(SystemMessage.EARNED_S2_S1_s);
			smsg.addItemName(template.getItemId());
			smsg.addNumber(count);
			_player.sendPacket(smsg);
		}
		else
		{
			SystemMessage smsg = new SystemMessage(SystemMessage.EARNED_S1);
			smsg.addItemName(template.getItemId());
			_player.sendPacket(smsg);
		}

		_player.sendPacket(iu);
		StatusUpdate su = new StatusUpdate(_player.getObjectId());
		su.addAttribute(StatusUpdate.CUR_LOAD, _player.getCurrentLoad());
		_player.sendPacket(su);
	}

	/**
	 * Return true if quest completed, false otherwise
	 * @return boolean
	 */
	public boolean isCompleted()
	{
		return getStateId().equals("Completed");
	}

	/**
	 * Return true if quest is to exited on clean up by QuestStateManager
	 * @return boolean
	 */
	public final boolean isExitQuestOnCleanUp()
	{
		return _IsExitQuestOnCleanUp;
	}

	/**
	 * Return true if quest started, false otherwise
	 * @return boolean
	 */
	public boolean isStarted()
	{
		if(getStateId().equals("Start") || getStateId().equals("Completed"))
			return false;
		return true;
	}

	public void killNpcByObjectId(int _objId)
	{
		L2Object obj = L2World.findObject(_objId);
		if(obj instanceof L2NpcInstance)
			((L2NpcInstance) obj).doDie(null);
		else
			_log.warning("Attemp to kill object that is not npc in quest " + getQuest().getQuestIntId());
	}

	public void monsterSay(L2NpcInstance npc, int chatType, String text)
	{
		NpcSay cs = null;
		if(npc == null)
			return;
		cs = new NpcSay(npc, chatType, text);
		List<L2Player> nearbyPlayers = new FastList<L2Player>();
		switch(chatType)
		{
			case 0:
				for(L2Player player : npc.getAroundPlayers(1500))
					nearbyPlayers.add(player);
				for(L2Player nearbyPlayer : nearbyPlayers)
				{
					if(nearbyPlayer == null)
						continue;
					nearbyPlayer.sendPacket(cs);
				}
				break;
			case 1:
				if(Config.GLOBAL_CHAT < Experience.MAX_LEVEL + 1)
					for(L2Player player : L2World.getAllPlayers())
						nearbyPlayers.add(player);
				else
					for(L2Player player : L2World.getAroundPlayers(npc))
						nearbyPlayers.add(player);
				for(L2Player nearbyPlayer : nearbyPlayers)
				{
					if(nearbyPlayer == null)
						continue;
					nearbyPlayer.sendPacket(cs);
				}
				break;
			case 2:
				_player.sendPacket(cs);
				break;
		}
	}

	public void onTutorialClientEvent(int number)
	{
		if(Config.TUTORIAL_ON_OFF)
			_player.sendPacket(new TutorialEnableClientEvent(number));
		else if(Config.DEBUG)
			_log.fine("Enable Tutorial Client Event: " + number + ".(Tutorial off)");
		if(Config.DEBUG)
			_log.fine("Enable Tutorial Client Event: " + number + ".");
	}

	/**
	 * Send a packet in order to play sound at client terminal
	 * @param sound
	 */
	public void playSound(String sound)
	{
		_player.sendPacket(new PlaySound(sound));
	}

	public void playTutorialVoice(String voice)
	{
		if(Config.TUTORIAL_ON_OFF)
			_player.sendPacket(new PlaySound(2, voice, 0, 0, _player.getX(), _player.getY(), _player.getZ()));
		else if(Config.DEBUG)
			_log.fine("play Tutorial Voice: " + voice + ".(Tutorial off)");
		if(Config.DEBUG)
			_log.fine("play Tutorial Voice: " + voice + ".");
	}

	public void removeRadar(int x, int y, int z)
	{
		_player.radar.removeMarker(x, y, z);
	}

	/**
	 * Return value of parameter "val" after adding the couple (var,val) in class variable "vars".<BR><BR>
	 * <U><I>Actions :</I></U><BR>
	 * <LI>Initialize class variable "vars" if is null</LI>
	 * <LI>Initialize parameter "val" if is null</LI>
	 * <LI>Add/Update couple (var,val) in class variable FastMap "vars"</LI>
	 * <LI>If the key represented by "var" exists in FastMap "vars", the couple (var,val) is updated in the database. The key is known as
	 * existing if the preceding value of the key (given as result of function put()) is not null.<BR>
	 * If the key doesn't exist, the couple is added/created in the database</LI>
	 * @param var : String indicating the name of the variable for quest
	 * @param val : String indicating the value of the variable for quest
	 * @return String (equal to parameter "val")
	 */
	public String set(String var, String val)
	{
		if(_vars == null)
			_vars = new FastMap<String, String>();
		if(val == null)
			val = "";
		// FastMap.put() returns previous value associated with specified key, or null if there was no mapping for key.
		String old = _vars.put(var, val);
		if(old != null)
			Quest.updateQuestVarInDb(this, var, val);
		else
			Quest.createQuestVarInDb(this, var, val);
		if(var.equals("cond"))
			_player.sendPacket(new QuestList());
		return val;
	}

	/**
	 * Add parameter used in quests.
	 * @param var : String pointing out the name of the variable for quest
	 * @param val : String pointing out the value of the variable for quest
	 * @return String (equal to parameter "val")
	 */
	String setInternal(String var, String val)
	{
		if(_vars == null)
			_vars = new FastMap<String, String>();
		if(val == null)
			val = "";
		_vars.put(var, val);
		return val;
	}

	/**
	 * Return the QuestTimer object with the specified name
	 * @return QuestTimer<BR> Return null if name does not exist
	 */
	public void setIsExitQuestOnCleanUp(boolean isExitQuestOnCleanUp)
	{
		_IsExitQuestOnCleanUp = isExitQuestOnCleanUp;
	}

	/**
	 * Return state of the quest after its initialization.<BR><BR>
	 * <U><I>Actions :</I></U>
	 * <LI>Remove drops from previous state</LI>
	 * <LI>Set new state of the quest</LI>
	 * <LI>Add drop for new state</LI>
	 * <LI>Update information in database</LI>
	 * <LI>Send packet QuestList to client</LI>
	 * @param state
	 * @return object
	 */
	public Object setState(State state)
	{
		if(state == null)
			return null;

		// remove drops from previous state
		_drops = new FastMap<Integer, FastList<L2DropData>>();

		// set new state
		_state = state;

		// add drops from new state
		if(!isCompleted())
		{
			Map<Integer, FastList<L2DropData>> new_drops = state.getDrops();
			if(new_drops != null)
			{
				if(_drops == null)
					_drops = new FastMap<Integer, FastList<L2DropData>>();
				_drops.putAll(new_drops);
			}
		}

		if(getQuest().getQuestIntId() < 999 && isStarted())
			getPlayer().sendPacket(new ExShowQuestMark(getQuest().getQuestIntId()));

		Quest.updateQuestInDb(this);
		QuestList ql = new QuestList();
		_player.sendPacket(ql);
		return state;
	}

	public void showQuestionMark(int number)
	{
		if(Config.TUTORIAL_ON_OFF)
			_player.sendPacket(new TutorialShowQuestionMark(number));
		else if(Config.DEBUG)
			_log.fine("show Question Mark: " + number + ".(Tutorial off)");
		if(Config.DEBUG)
			_log.fine("show Question Mark: " + number + ".");
	}

	public void showTutorialHTML(String html)
	{
		if(!Config.TUTORIAL_ON_OFF)
		{
			if(Config.DEBUG)
				_log.fine("show Tutorial HTML: " + html + ".(Tutorial off)");
			return;
		}

		String _rutext = Files.read("data/html-ru/tutorial/guide/" + html, _player);
		if(_rutext == null)
		{
			String _text = Files.read("data/html/tutorial/guide/" + html, _player);
			if(_text != null)
				_player.sendPacket(new TutorialShowHtml(_text));
			else
				_text = "<html><head><body><center><font color=\"LEVEL\">Error:<br> Html " + html + " not found.</font></center><br><br><a action=\"bypass -h TE00\">Close Window</a></body></html>";
			_player.sendPacket(new TutorialShowHtml(_text));
		}
		else
			_player.sendPacket(new TutorialShowHtml(_rutext));
		if(Config.DEBUG)
			_log.fine("show Tutorial HTML: " + html + ".");
	}

	/**
	 * Start a timer for quest.<BR><BR>
	 * @param name<BR> The name of the timer. Will also be the value for event of onEvent
	 * @param time<BR> The milisecond value the timer will elapse
	 */
	public void startQuestTimer(String name, long time)
	{
		getQuest().startQuestTimer(name, time, null, getPlayer());
	}

	/**
	 * Удаляет указанные предметы из инвентаря игрока, и обновляет инвентарь
	 * @param itemId : id удаляемого предмета
	 * @param count : число удаляемых предметов<br>
	 * Если count передать -1, то будут удалены все указанные предметы.
	 * @return Количество удаленных предметов
	 */
	public int takeItems(int itemId, int count)
	{
		// Get object item from player's inventory list
		L2ItemInstance item = _player.getInventory().getItemByItemId(itemId);
		if(item == null)
			return 0;
		// Tests on count value in order not to have negative value
		if(count < 0 || count > item.getCount())
			count = item.getCount();

		InventoryUpdate iu = new InventoryUpdate();
		// Destroy the quantity of items wanted
		iu.addItem(_player.getInventory().destroyItemByItemId(itemId, count, false, true));
		// Send message of destruction to client
		if(itemId == 57)
			_player.sendPacket(new SystemMessage(SystemMessage.S1_ADENA_DISAPPEARED).addNumber(count));
		else if(count == 1)
			_player.sendPacket(new SystemMessage(SystemMessage.S1_HAS_DISAPPEARED).addItemName(itemId));
		else
			_player.sendPacket(new SystemMessage(SystemMessage.S2_S1_HAS_DISAPPEARED).addItemName(itemId).addNumber(count));

		// Create and send new inventory list to client
		_player.sendPacket(iu);
		return count;
	}

	/**
	 * Remove the variable of quest from the list of variables for the quest.<BR><BR>
	 * <U><I>Concept : </I></U>
	 * Remove the variable of quest represented by "var" from the class variable FastMap "vars" and from the database.
	 * @param var : String designating the variable for the quest to be deleted
	 * @return String pointing out the previous value associated with the variable "var"
	 */
	public String unset(String var)
	{
		if(_vars == null)
			return null;
		String old = _vars.remove(var);
		if(old != null)
			Quest.deleteQuestVarInDb(this, var);
		return old;
	}
}
