package L2jFT.game.network;

import java.nio.ByteBuffer;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.concurrent.ScheduledFuture;
import java.util.logging.Level;
import java.util.logging.Logger;

import L2jFT.Config;
import L2jFT.database.DatabaseUtils;
import L2jFT.database.FiltredPreparedStatement;
import L2jFT.database.L2DatabaseFactory;
import L2jFT.database.ThreadConnection;
import L2jFT.extensions.scripts.Scripts;
import L2jFT.game.LoginServerThread;
import L2jFT.game.RecipeController;
import L2jFT.game.ThreadPoolManager;
import L2jFT.game.LoginServerThread.SessionKey;
import L2jFT.game.instancemanager.CursedWeaponsManager;
import L2jFT.game.model.CharSelectInfoPackage;
import L2jFT.game.model.L2Clan;
import L2jFT.game.model.L2Object;
import L2jFT.game.model.L2Party;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2RecipeList;
import L2jFT.game.model.L2World;
import L2jFT.game.model.entity.Olympiad;
import L2jFT.game.network.serverpackets.L2GameServerPacket;
import L2jFT.game.tables.ClanTable;

import com.l2jserver.mmocore.network.MMOClient;
import com.l2jserver.mmocore.network.MMOConnection;

/**
 * Represents a client connected on Game Server
 * @author KenM
 */
public final class L2GameClient extends MMOClient<MMOConnection<L2GameClient>>
{
	protected static Logger _log = Logger.getLogger(L2GameClient.class.getName());

	public static int devCharId;
	public GameCrypt _crypt;
	private double _bonus;
	public GameClientState _state;
	private int _upTryes = 0;
	private long _upLastConnection = 0;
	private boolean _checking = false;
	private boolean _passed = false;

	public static enum GameClientState
	{
		CONNECTED,
		AUTHED,
		IN_GAME
	}

	class AutoSaveTask implements Runnable
	{
		public void run()
		{
			try
			{
				L2Player player = getActiveChar();
				if(player != null)
					saveCharToDisk(player);
				else if(getConnection() == null || getConnection().isClosed())
					_autoSaveInDB.cancel(false);
			}
			catch(Throwable e)
			{
				_log.severe(e.toString());
			}
		}
	}

	private String _loginName;
	private L2Player _activeChar;
	private SessionKey _sessionId;
	private final MMOConnection<L2GameClient> _connection;
	final ScheduledFuture _autoSaveInDB;

	//private byte[] _filter;

	private int revision = 0;
	private boolean _gameGuardOk = false;

	private ArrayList<Integer> _charSlotMapping = new ArrayList<Integer>();

	public L2GameClient(MMOConnection<L2GameClient> con)
	{
		super(con);
		_state = GameClientState.CONNECTED;
		_sessionId = new SessionKey(-1, -1, -1, -1);
		_connection = con;
		_crypt = new GameCrypt();
		_autoSaveInDB = ThreadPoolManager.getInstance().scheduleGeneralAtFixedRate(new AutoSaveTask(), 5 * 60 * 1000L, 15 * 60 * 1000L);
	}

	@Override
	protected void onDisconection()
	{
		if(getLoginName() == null || getLoginName().equals("") || _state != GameClientState.IN_GAME && _state != GameClientState.AUTHED)
		{
			_autoSaveInDB.cancel(true);
			//closeNow(true);
			return;
		}
		try
		{
			LoginServerThread.getInstance().removeAccount(getLoginName());
			_autoSaveInDB.cancel(false);
			if(_activeChar != null) // this should only happen on connection loss
			{
				L2Player player = _activeChar;
				_activeChar = null;

				if(player.isCastingNow())
					player.abortCast();

				if(player.isFestivalParticipant())
				{
					L2Party playerParty = player.getParty();
					if(playerParty != null)
						playerParty.broadcastMessageToPartyMembers(player.getName() + " has been removed from the upcoming festival.");
				}

				CursedWeaponsManager.getInstance().doLogout(player);

				if(player.isInOlympiadMode())
					Olympiad.getInstance().unRegisterNoble(player); //TODO так работать не будет

				// Вызов всех хэндлеров, определенных в скриптах
				for(String handler : Scripts.onPlayerExit)
				{
					String[] list = handler.split(":");
					String _class = list[0];
					String _method = list[1];
					player.callScripts(_class, _method, new Object[] { player });
				}

				if(player.getPet() != null)
					player.getPet().unSummon();

				saveCharToDisk(player);
				player.deleteMe(false);
				if(player.getNetConnection() != null)
				{
					player.getNetConnection().closeNow(false);
					player.setNetConnection(null);
				}
				_activeChar = null;
			}
		}
		catch(Exception e1)
		{
			_log.log(Level.WARNING, "error while disconnecting client", e1);
		}
		finally
		{
			LoginServerThread.getInstance().sendLogout(getLoginName());
		}
	}

	/**
	 *
	 */
	public static void saveCharToDisk(L2Player cha)
	{
		try
		{
			cha.getInventory().updateDatabase(true);
			cha.store();
			storeRecipeBook(cha);
		}
		catch(Exception e)
		{
			_log.warning("Error saving player character: " + e);
			e.printStackTrace();
		}
	}

	public void deleteFromClan(L2Player cha)
	{
		L2Clan clan = cha.getClan();
		if(clan != null)
			clan.removeClanMember(cha.getObjectId());
	}

	public static void deleteFromClan(int charId, int clanId)
	{
		if(clanId == 0)
			return;
		L2Clan clan = ClanTable.getInstance().getClan(clanId);
		if(clan != null)
			clan.removeClanMember(charId);
	}

	public void markRestoredChar(int charslot) throws Exception
	{
		//have to make sure active character must be nulled
		if(getActiveChar() != null)
		{
			saveCharToDisk(getActiveChar());
			if(Config.DEBUG)
				_log.fine("active Char saved");
			_activeChar = null;
		}

		int objid = getObjectIdForSlot(charslot);
		if(objid < 0)
			return;
		
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("UPDATE characters SET deletetime=0 WHERE obj_id=?");
			statement.setInt(1, objid);
			statement.execute();
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "data error on restore char:", e);
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	public void markToDeleteChar(int charslot) throws Exception
	{
		//have to make sure active character must be nulled
		if(getActiveChar() != null)
		{
			saveCharToDisk(getActiveChar());
			if(Config.DEBUG)
				_log.fine("active Char saved");
			_activeChar = null;
		}

		int objid = getObjectIdForSlot(charslot);
		if(objid < 0)
			return;
		
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("UPDATE characters SET deletetime=? WHERE obj_id=?");
			statement.setLong(1, (int) (System.currentTimeMillis() / 1000));
			statement.setInt(2, objid);
			statement.execute();
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "data error on update deletime char:", e);
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	public void deleteChar(int charslot) throws Exception
	{
		//have to make sure active character must be nulled
		if(getActiveChar() != null)
		{
			saveCharToDisk(getActiveChar());
			if(Config.DEBUG)
				_log.fine("active Char saved");
			_activeChar = null;
		}

		int objid = getObjectIdForSlot(charslot);
		if(objid == -1)
			return;

		deleteCharByObjId(objid);
	}

	public static void deleteCharByObjId(int objid)
	{
		if(objid < 0)
			return;
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("DELETE FROM characters WHERE obj_Id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM character_shortcuts WHERE char_obj_id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM character_macroses WHERE char_obj_id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE pets FROM pets, items WHERE pets.item_obj_id=items.object_id AND items.owner_id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM items WHERE owner_id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM character_skills WHERE char_obj_Id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM character_effects_save WHERE char_obj_Id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM character_skills_save WHERE char_obj_Id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM character_quests WHERE char_id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM character_recipebook WHERE char_id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM character_friends WHERE char_id=? or friend_id = ?");
			statement.setInt(1, objid);
			statement.setInt(2, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM seven_signs WHERE char_obj_id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM character_subclasses WHERE char_obj_id=?");
			statement.setInt(1, objid);
			statement.execute();
			DatabaseUtils.closeStatement(statement);
			statement = null;
		}
		catch(Exception e)
		{
			_log.warning("data error on delete char:" + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	public L2Player loadCharFromDisk(int charslot)
	{
		Integer objectId = getObjectIdForSlot(charslot);
		if(objectId == -1)
			return null;

		L2Object object = L2World.findObject(objectId);
		if(object != null)
			if(object.isPlayer)
			{
				L2Player player = (L2Player) object;
				//_log.warning(player.toFullString() + " tried to make a clone.");
				player.logout(false, false);
				//player.deleteMe(false);
				LoginServerThread.getInstance().sendLogout(getLoginName());
			}

		L2Player character = L2Player.load(objectId);
		if(character != null)
		{
			restoreRecipeBook(character);

			// preinit some values for each login
			character.setRunning(); // running is default
			character.standUp(); // standing is default

			character.updateStats();
			character.setOnlineStatus(true);
			setActiveChar(character);
			character.restoreBonus();
		}
		else
			_log.warning("could not restore in slot:" + charslot);

		return character;
	}

	/**
	 * @param cha
	 * @param saveFile
	 */
	private static void storeRecipeBook(L2Player cha)
	{
		int recipesCount = cha.recipesCount();
		if(recipesCount == 0)
			return;

		int obj_id = cha.getObjectId();
		L2RecipeList[] commonRecipeBook = cha.getCommonRecipeBook();
		L2RecipeList[] dwarvenRecipeBook = cha.getDwarvenRecipeBook();

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("DELETE FROM character_recipebook WHERE char_id=?");
			statement.setInt(1, obj_id);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("INSERT INTO character_recipebook (char_id, id) values(?,?)");
			for(L2RecipeList recipe : commonRecipeBook)
			{
				statement.setInt(1, obj_id);
				statement.setInt(2, recipe.getId());
				statement.execute();
			}
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("REPLACE INTO character_recipebook (char_id, id) values(?,?)");
			for(L2RecipeList recipe : dwarvenRecipeBook)
			{
				statement.setInt(1, obj_id);
				statement.setInt(2, recipe.getId());
				statement.execute();
			}
		}
		catch(Exception e)
		{
			_log.log(Level.WARNING, "could not store recipes:", e);
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	/**
	 * @param cha
	 * @param saveFile
	 */
	public static void restoreRecipeBook(L2Player cha)
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT id FROM character_recipebook WHERE char_id=?");
			statement.setInt(1, cha.getObjectId());
			rset = statement.executeQuery();

			while(rset.next())
			{
				int id = rset.getInt("id");
				L2RecipeList recipe = RecipeController.getInstance().getRecipeList(id - 1);
				cha.registerRecipeList(recipe);
			}
		}
		catch(Exception e)
		{
			_log.warning("count not recipe skills:" + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
	}

	/**
	 * @param charslot
	 * @return
	 */
	public int getObjectIdForSlot(int charslot)
	{
		if(charslot < 0 || charslot >= _charSlotMapping.size())
		{
			_log.warning(getLoginName() + " tried to modify Character in slot " + charslot + " but no characters exits at that slot.");
			return -1;
		}
		return _charSlotMapping.get(charslot);
	}

	@Override
	public MMOConnection<L2GameClient> getConnection()
	{
		return _connection;
	}

	public L2Player getActiveChar()
	{
		return _activeChar;
	}

	/**
	 * @return Returns the sessionId.
	 */
	public SessionKey getSessionId()
	{
		return _sessionId;
	}

	public String getLoginName()
	{
		return _loginName;
	}

	public void setLoginName(String loginName)
	{
		_loginName = loginName;
	}

	/**
	 * @param cha
	 */
	public void setActiveChar(L2Player cha)
	{
		_activeChar = cha;
		if(cha != null)
		{
			// we store the connection in the player object so that external
			// events can directly send events to the players client
			// might be changed later to use a central event management and distribution system
			_activeChar.setNetConnection(this);

			// update world data
			L2World.addObject(_activeChar);
		}
	}

	/**
	 * @param sessionKey
	 */
	public void setSessionId(SessionKey sessionKey)
	{
		_sessionId = sessionKey;
	}

	/**
	 * @param chars
	 */
	public void setCharSelection(CharSelectInfoPackage[] chars)
	{
		_charSlotMapping.clear();

		for(CharSelectInfoPackage element : chars)
		{
			int objectId = element.getObjectId();
			_charSlotMapping.add(objectId);
		}
	}

	/**
	 * @return Returns the revision.
	 */
	public int getRevision()
	{
		return revision;
	}

	/**
	 * @param revision The revision to set.
	 */
	public void setRevision(int revision)
	{
		this.revision = revision;
	}

	public void setGameGuardOk(boolean gameGuardOk)
	{
		_gameGuardOk = gameGuardOk;
	}

	public boolean isGameGuardOk()
	{
		return _gameGuardOk;
	}

	@Override
	public boolean encrypt(final ByteBuffer buf, final int size)
	{
		_crypt.encrypt(buf.array(), buf.position(), size);
		buf.position(buf.position() + size);
		return true;
	}

	@Override
	public boolean decrypt(ByteBuffer buf, int size)
	{
		_crypt.decrypt(buf.array(), buf.position(), size);
		return true;
	}

	public void sendPacket(L2GameServerPacket gsp)
	{
		if(getConnection() == null)
			return;
		getConnection().sendPacket(gsp);
		gsp.runImpl();
	}

	public String getIpAddr()
	{
		try
		{
			return _connection.getSocketChannel().socket().getInetAddress().getHostAddress();
		}
		catch(NullPointerException e)
		{
			return "Disconnected";
		}
	}

	public byte[] enableCrypt()
	{
		byte[] key = BlowFishKeygen.getRandomKey();
		_crypt.setKey(key);
		return key;
	}

	public double getBonus()
	{
		return _bonus;
	}

	public void setBonus(double bonus)
	{
		_bonus = bonus;
	}

	public GameClientState getState()
	{
		return _state;
	}

	public void setState(GameClientState state)
	{
		_state = state;
	}

	public void addUPTryes()
	{
		if(_upLastConnection != 0 && System.currentTimeMillis() - _upLastConnection > 1000)
			_upTryes = 0;
		_upTryes++;
		_upLastConnection = System.currentTimeMillis();
	}

	public int getUPTryes()
	{
		if(_upLastConnection != 0 && System.currentTimeMillis() - _upLastConnection > 1000)
		{
			_upTryes = 0;
			_upLastConnection = System.currentTimeMillis();
		}
		return _upTryes;
	}

	public void setChecking(boolean val)
	{
		_checking = val;
	}

	public boolean isChecking()
	{
		return _checking;
	}

	public void setPassed(boolean val)
	{
		_passed = val;
	}

	public boolean isPassed()
	{
		return _passed;
	}
}
