package L2jFT.game.network.clientpackets;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.logging.Logger;

import L2jFT.Config;
import L2jFT.game.ItemsAutoDestroy;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.network.serverpackets.InventoryUpdate;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.tables.GmListTable;
import L2jFT.game.templates.L2EtcItem.EtcItemType;
import L2jFT.util.Log;
import L2jFT.util.Util;

public class RequestDropItem extends L2GameClientPacket
{
	private static String _C__12_REQUESTDROPITEM = "[C] 12 RequestDropItem";
	private static Logger _log = Logger.getLogger(RequestDropItem.class.getName());

	private int _objectId;
	private int _count;
	private int _x;
	private int _y;
	private int _z;

	/**
	 * packet type id 0x12
	 *
	 * sample
	 *
	 * 12
	 * 09 00 00 40 		// object id
	 * 01 00 00 00 		// count ??
	 * fd e7 fe ff 		// x
	 * e5 eb 03 00 		// y
	 * bb f3 ff ff 		// z
	 *
	 * format:		cdd ddd
	 */
	@Override
	public void readImpl()
	{
		_objectId = readD();
		_count = readD();
		_x = readD();
		_y = readD();
		_z = readD();
	}

	@Override
	public void runImpl()
	{
		L2Player activeChar = getClient().getActiveChar();
		if(activeChar == null)
			return;

		if(_count == 0 || _x == 0 || _y == 0 || _z == 0)
		{
			activeChar.sendActionFailed();
			return;
		}

		if(!Config.ALLOW_DISCARDITEM)
		{
			activeChar.sendMessage("Unable to discard item, use 'destroy item' instead", "Выбрасывать предметы запрещено, удаляйте ненужные вещи");
			return;
		}

		if(_count < 0)
		{
			Util.handleIllegalPlayerAction(activeChar, "RequestDropItem[72]", "Negative count of items to drop", 2);
			return;
		}

		if(activeChar.isDead())
		{
			activeChar.sendActionFailed();
			return;
		}

		if(activeChar.getPrivateStoreType() != L2Player.STORE_PRIVATE_NONE)
		{
			activeChar.sendPacket(new SystemMessage(SystemMessage.WHILE_OPERATING_A_PRIVATE_STORE_OR_WORKSHOP_YOU_CANNOT_DISCARD_DESTROY_OR_TRADE_AN_ITEM));
			return;
		}

		if(activeChar.isTransactionInProgress())
		{
			sendPacket(new SystemMessage(SystemMessage.NOTHING_HAPPENED));
			return;
		}

		L2ItemInstance oldItem = activeChar.getInventory().getItemByObjectId(_objectId);
		if(oldItem == null)
		{
			_log.warning(activeChar.getName() + ":tried to drop an item that is not in the inventory ?!?:" + _objectId);
			return;
		}

		if(activeChar.isFishing())
		{
			activeChar.sendPacket(new SystemMessage(SystemMessage.YOU_CANNOT_DO_THAT_WHILE_FISHING));
			return;
		}

		if(activeChar.isActionsDisabled() || activeChar.isSitting() || activeChar.isDropDisabled())
		{
			activeChar.sendActionFailed();
			return;
		}

		if(oldItem.getItemType() == EtcItemType.QUEST)
		{
			activeChar.sendPacket(new SystemMessage(SystemMessage.THAT_ITEM_CANNOT_BE_DISCARDED));
			return;
		}

		if(!oldItem.canBeDropped(activeChar))
		{
			activeChar.sendPacket(new SystemMessage(SystemMessage.THIS_ITEM_CANNOT_BE_DISCARDED));
			return;
		}

		if(oldItem.isWear())
		{
			activeChar.sendActionFailed();
			_log.warning(activeChar.getName() + ":tried to drop an item that is wear:" + _objectId);
			return;
		}

		int oldCount = oldItem.getCount();
		if(Config.DEBUG)
			_log.fine("requested drop item " + _objectId + "(" + oldCount + ") at " + _x + "/" + _y + "/" + _z);

		if(oldCount < _count)
		{
			activeChar.sendActionFailed();
			if(Config.DEBUG)
				_log.finest(activeChar.getObjectId() + ":player tried to drop more items than he has");
			return;
		}

		if(!activeChar.isInRangeSq(_x, _y, 22500) || Math.abs(_z - activeChar.getZ()) > 50)
		{
			if(Config.DEBUG)
				_log.finest(activeChar.getObjectId() + ": trying to drop too far away");
			activeChar.sendPacket(new SystemMessage(SystemMessage.TOO_FAR_TO_DISCARD));
			return;
		}

		if(activeChar.getEnchantScroll() != null && oldItem.getObjectId() == activeChar.getEnchantScroll().getObjectId())
		{
			activeChar.sendActionFailed();
			if(Config.DEBUG)
				_log.finest(activeChar.getObjectId() + ":player tried to drop an enchant scroll he was using");
			return;
		}

		if(oldItem.isEquipped())
		{
			L2ItemInstance weapon = activeChar.getActiveWeaponInstance();
			L2ItemInstance[] unequipped = activeChar.getInventory().unEquipItemInBodySlotAndRecord(oldItem.getItem().getBodyPart(), oldItem);

			for(L2ItemInstance uneq : unequipped)
			{
				if(uneq == null)
					continue;

				SystemMessage sm;

				if(oldItem.getEnchantLevel() > 0)
				{
					sm = new SystemMessage(SystemMessage.EQUIPMENT_OF__S1_S2_HAS_BEEN_REMOVED);
					sm.addNumber(oldItem.getEnchantLevel());
					sm.addItemName(oldItem.getItemId());
				}
				else
					sm = new SystemMessage(SystemMessage.S1__HAS_BEEN_DISARMED).addItemName(oldItem.getItemId());

				activeChar.sendPacket(sm);

				if(weapon != null && uneq == weapon)
				{
					uneq.setChargedSpiritshot(L2ItemInstance.CHARGED_NONE);
					uneq.setChargedSoulshot(L2ItemInstance.CHARGED_NONE);
				}
			}

			activeChar.refreshExpertisePenalty();
			activeChar.broadcastUserInfo();
			activeChar.sendPacket(new InventoryUpdate(new ArrayList<L2ItemInstance>(Arrays.asList(unequipped))));
		}

		L2ItemInstance dropedItem = activeChar.getInventory().dropItem(_objectId, _count);

		if(dropedItem == null)
		{
			activeChar.sendActionFailed();
			return;
		}

		InventoryUpdate iu = new InventoryUpdate();
		iu.addItem(oldItem);
		activeChar.sendPacket(iu);

		if(Config.DEBUG)
			_log.fine("dropping " + _objectId + " item(" + _count + ") at: " + _x + " " + _y + " " + _z);

		dropedItem.dropMe(activeChar, _x, _y, _z);
		ItemsAutoDestroy.getInstance().addItem(dropedItem);

		activeChar.disableDrop(1000);

		Log.LogItem(activeChar, Log.Drop, dropedItem);

		if(dropedItem.getItemId() == 57 && dropedItem.getCount() >= 1000000)
		{
			String msg = "Character (" + activeChar.getName() + ") has dropped (" + dropedItem.getCount() + ")adena at (" + _x + "," + _y + "," + _z + ")";
			_log.warning(msg);
			GmListTable.broadcastMessageToGMs(msg);
		}

		activeChar.updateStats();
	}

	@Override
	public String getType()
	{
		return _C__12_REQUESTDROPITEM;
	}
}