package L2jFT.game.network.serverpackets;

import L2jFT.Config;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.entity.SevenSigns;
import L2jFT.game.model.entity.SevenSignsFestival;
import L2jFT.game.templates.StatsSet;

import java.util.logging.Logger;

/**
 * Seven Signs Record Update
 *
 * packet type id 0xf5
 * format:
 *
 * c cc	(Page Num = 1 -> 4, period)
 *
 * 1: [ddd cc dd ddd c ddd c]
 * 2: [hc [cd (dc (S))]
 * 3: [ccc (cccc)]
 * 4: [(cchh)]
 *
 * @author Tempy
 *
 */
public class SSQStatus extends L2GameServerPacket
{
	private static Logger _log = Logger.getLogger(SSQStatus.class.getName());

	private static final String _S__F5_RecordUpdate = "[S] F5 SSQStatus";
	private L2Player _player;
	private int _page;

	public SSQStatus(L2Player player, int recordPage)
	{
		_player = player;
		_page = recordPage;
	}

	@Override
	final public void runImpl()
	{
	// no long-running tasks
	}

	@Override
	protected final void writeImpl()
	{
		writeC(0xF5);

		writeC(_page);
		writeC(SevenSigns.getInstance().getCurrentPeriod()); // current period?

		switch(_page)
		{
			case 1:
				// [ddd cc dd ddd c ddd c]	
				writeD(SevenSigns.getInstance().getCurrentCycle());

				int currentPeriod = SevenSigns.getInstance().getCurrentPeriod();

				switch(currentPeriod)
				{
					case SevenSigns.PERIOD_COMP_RECRUITING:
						writeD(1183);
						break;
					case SevenSigns.PERIOD_COMPETITION:
						writeD(1176);
						break;
					case SevenSigns.PERIOD_COMP_RESULTS:
						writeD(1184);
						break;
					case SevenSigns.PERIOD_SEAL_VALIDATION:
						writeD(1177);
						break;
				}

				switch(currentPeriod)
				{
					case SevenSigns.PERIOD_COMP_RECRUITING:
					case SevenSigns.PERIOD_COMP_RESULTS:
						writeD(1287);
						break;
					case SevenSigns.PERIOD_COMPETITION:
					case SevenSigns.PERIOD_SEAL_VALIDATION:
						writeD(1286);
						break;
				}

				writeC(SevenSigns.getInstance().getPlayerCabal(_player));
				writeC(SevenSigns.getInstance().getPlayerSeal(_player));

				writeD(SevenSigns.getInstance().getPlayerStoneContrib(_player)); // Seal Stones Turned-In
				writeD(SevenSigns.getInstance().getPlayerAdenaCollect(_player)); // Ancient Adena to Collect

				long dawnStoneScore = SevenSigns.getInstance().getCurrentStoneScore(SevenSigns.CABAL_DAWN);
				long dawnFestivalScore = SevenSigns.getInstance().getCurrentFestivalScore(SevenSigns.CABAL_DAWN);
				long dawnTotalScore = dawnStoneScore + dawnFestivalScore;

				long duskStoneScore = SevenSigns.getInstance().getCurrentStoneScore(SevenSigns.CABAL_DUSK);
				long duskFestivalScore = SevenSigns.getInstance().getCurrentFestivalScore(SevenSigns.CABAL_DUSK);
				long duskTotalScore = duskStoneScore + duskFestivalScore;

				long totalStoneScore = duskStoneScore + dawnStoneScore;
				totalStoneScore = totalStoneScore == 0 ? 1 : totalStoneScore; // Prevents divide by zero errors when competition begins.

				/*
				 * Scoring seems to be proportionate to a set base value, so base this on 
				 * the maximum obtainable score from festivals, which is 500.
				 */
				long duskStoneScoreProp = duskStoneScore * 500 / totalStoneScore;
				long dawnStoneScoreProp = dawnStoneScore * 500 / totalStoneScore;

				long totalOverallScore = duskTotalScore + dawnTotalScore;
				totalOverallScore = totalOverallScore == 0 ? 1 : totalOverallScore; // Prevents divide by zero errors when competition begins.

				long dawnPercent = dawnTotalScore * 100 / totalOverallScore;
				long duskPercent = duskTotalScore * 100 / totalOverallScore;

				if(Config.DEBUG)
				{
					_log.info("Dusk Stone Score: " + duskStoneScore + " - Dawn Stone Score: " + dawnStoneScore);
					_log.info("Dusk Festival Score: " + duskFestivalScore + " - Dawn Festival Score: " + dawnFestivalScore);
					_log.info("Dusk Score: " + duskTotalScore + " - Dawn Score: " + dawnTotalScore);
					_log.info("Overall Score: " + totalOverallScore);
					_log.info("");
					_log.info("Dusk Prop: " + duskStoneScore / totalStoneScore * 500 + " - Dawn Prop: " + dawnStoneScore / totalStoneScore * 500);
					_log.info("Dusk %: " + duskPercent + " - Dawn %: " + dawnPercent);
				}

				/* DUSK */
				writeD((int) duskStoneScoreProp); // Seal Stone Score
				writeD((int) duskFestivalScore); // Festival Score
				writeD((int) duskStoneScoreProp + (int) duskFestivalScore); // Total Score

				writeC((int) duskPercent); // Dusk %

				/* DAWN */
				writeD((int) dawnStoneScoreProp); // Seal Stone Score
				writeD((int) dawnFestivalScore); // Festival Score
				writeD((int) dawnStoneScoreProp + (int) dawnFestivalScore); // Total Score

				writeC((int) dawnPercent); // Dawn %
				break;
			case 2:
				// c cc hc [cd (dc (S))]
				if(SevenSigns.getInstance().isSealValidationPeriod())
					writeH(0);
				else
					writeH(1);

				writeC(5); // Total number of festivals

				for(int i = 0; i < 5; i++)
				{
					writeC(i + 1); // Current client-side festival ID
					writeD(SevenSignsFestival.FESTIVAL_LEVEL_SCORES[i]);

					int duskScore = SevenSignsFestival.getInstance().getHighestScore(SevenSigns.CABAL_DUSK, i);
					int dawnScore = SevenSignsFestival.getInstance().getHighestScore(SevenSigns.CABAL_DAWN, i);

					// Dusk Score \\
					writeD(duskScore);

					StatsSet highScoreData = SevenSignsFestival.getInstance().getHighestScoreData(SevenSigns.CABAL_DUSK, i);
					String[] partyMembers = highScoreData.getString("members").split(",");

					if(partyMembers != null)
					{
						writeC(partyMembers.length);

						for(String partyMember : partyMembers)
							writeS(partyMember);
					}
					else
						writeC(0);

					// Dawn Score \\
					writeD(dawnScore);

					highScoreData = SevenSignsFestival.getInstance().getHighestScoreData(SevenSigns.CABAL_DAWN, i);
					partyMembers = highScoreData.getString("members").split(",");

					if(partyMembers != null)
					{
						writeC(partyMembers.length);

						for(String partyMember : partyMembers)
							writeS(partyMember);
					}
					else
						writeC(0);
				}
				break;
			case 3:
				// c cc [ccc (cccc)]
				writeC(10); // Minimum limit for winning cabal to retain their seal
				writeC(35); // Minimum limit for winning cabal to claim a seal
				writeC(3); // Total number of seals

				for(int i = 1; i < 4; i++)
				{
					int dawnProportion = SevenSigns.getInstance().getSealProportion(i, SevenSigns.CABAL_DAWN);
					int duskProportion = SevenSigns.getInstance().getSealProportion(i, SevenSigns.CABAL_DUSK);
					int totalProportion = dawnProportion + duskProportion;

					totalProportion = totalProportion < 1 ? 1 : totalProportion; // Prevents divide by zero errors.

					if(Config.DEBUG)
						_log.info(SevenSigns.getSealName(i, true) + " = Dawn Prop: " + dawnProportion + "(" + dawnProportion / totalProportion * 100 + "%)" + ", Dusk Prop: " + duskProportion + "(" + duskProportion / totalProportion * 100 + "%)");

					writeC(i);
					writeC(SevenSigns.getInstance().getSealOwner(i));
					writeC((duskProportion * 100 / totalProportion));
					writeC((dawnProportion * 100 / totalProportion));
				}
				break;
			case 4:
				// c cc [(cchh)]

				writeC(SevenSigns.getInstance().getCabalHighestScore()); // Overall predicted winner
				writeC(3); // Total number of seals

				for(int i = 1; i < 4; i++)
				{
					int dawnProportion = SevenSigns.getInstance().getSealProportion(i, SevenSigns.CABAL_DAWN);
					int duskProportion = SevenSigns.getInstance().getSealProportion(i, SevenSigns.CABAL_DUSK);
					int totalProportion = dawnProportion + duskProportion;

					int sealOwner = SevenSigns.getInstance().getSealOwner(i);

					writeC(i);
					writeC(sealOwner);

					/*
					 * 1289 = 10% or more voted for owned seal
					 * 1290 = 35% or more voted for non-owned seal
					 * 1291 = 10% or less voted for owned seal
					 * 1292 = 35% or less voted for non-owned seal
					 */
					if(sealOwner != SevenSigns.CABAL_NULL)
					{
						if(totalProportion >= 10)
							writeH(1289);
						else
							writeH(1291);
					}
					else if(totalProportion >= 35)
						writeH(1290);
					else
						writeH(1292);

					// Shows a short description of the seal status when not Seal Validation.
					if(SevenSigns.getInstance().isSealValidationPeriod())
						writeH(1);
					else
						writeH(0);
				}

				break;
		}
	}

	/* (non-Javadoc)
	 * @see L2jFT.game.network.serverpackets.L2GameServerPacket#getType()
	 */
	@Override
	public String getType()
	{
		return _S__F5_RecordUpdate;
	}
}
