package L2jFT.game.tables;

import javolution.util.FastMap;
import javolution.util.FastMap.Entry;
import L2jFT.Config;
import L2jFT.database.DatabaseUtils;
import L2jFT.database.FiltredPreparedStatement;
import L2jFT.database.L2DatabaseFactory;
import L2jFT.database.ThreadConnection;
import L2jFT.game.idfactory.IdFactory;
import L2jFT.game.instancemanager.SiegeManager;
import L2jFT.game.model.L2Alliance;
import L2jFT.game.model.L2Clan;
import L2jFT.game.model.L2ClanMember;
import L2jFT.game.model.L2Player;
import L2jFT.game.network.serverpackets.PledgeShowMemberListDeleteAll;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.util.Util;

import java.sql.ResultSet;
import java.util.logging.Logger;

public class ClanTable
{
	private static final Logger _log = Logger.getLogger(ClanTable.class.getName());

	private static ClanTable _instance;

	private final FastMap<Integer, L2Clan> _clans = new FastMap<Integer, L2Clan>();
	private final FastMap<Integer, L2Alliance> _alliances = new FastMap<Integer, L2Alliance>();

	public static ClanTable getInstance()
	{
		if(_instance == null)
			new ClanTable();
		return _instance;
	}

	public L2Clan[] getClans()
	{
		return _clans.values().toArray(new L2Clan[_clans.size()]);
	}

	public L2Alliance[] getAlliances()
	{
		return _alliances.values().toArray(new L2Alliance[_alliances.size()]);
	}

	private ClanTable()
	{
		_instance = this;
		restorewars();
	}

	public L2Clan getClan(int clanId)
	{
		if(clanId <= 0)
			return null;
		synchronized (_clans)
		{
			L2Clan clan = _clans.get(new Integer(clanId));
			if(clan == null)
				clan = new L2Clan(clanId);
			if(clan.getMembersCount() > 0 && clan.getLeader() != null)
				_clans.put(clan.getClanId(), clan);
			else
				clan = null;
			return clan;
		}
	}

	public L2Alliance getAlliance(int allyId)
	{
		if(allyId <= 0)
			return null;
		L2Alliance ally = _alliances.get(new Integer(allyId));
		if(ally == null)
		{
			ally = new L2Alliance(allyId);
			if(ally.getMembersCount() > 0 && ally.getLeader() != null)
				_alliances.put(ally.getAllyId(), ally);
			else
			{
				ally = null;
				_log.warning("Got request for non-existent ally " + allyId);
				Thread.dumpStack();
			}
		}
		return ally;
	}

	public L2Clan getClanByName(String clanName)
	{
		if(!Util.isMatchingRegexp(clanName, Config.CLAN_NAME_TEMPLATE))
			return null;

		for(Entry<Integer, L2Clan> n = _clans.head(), end = _clans.tail(); (n = n.getNext()) != end;)
			if(n.getValue().getName().equalsIgnoreCase(clanName))
				return n.getValue();

		L2Clan clan = null;
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet result = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT clan_id FROM clan_data WHERE clan_name = ?");
			statement.setString(1, clanName);
			result = statement.executeQuery();
			if(result.next())
			{
				int clanId = result.getInt("clan_id");
				clan = getClan(clanId);
			}
		}
		catch(Exception e)
		{
			_log.warning("data error on clan item '" + clanName + "': " + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, result);
		}
		return clan;
	}

	public L2Alliance getAllyByName(String allyName)
	{
		if(!Util.isMatchingRegexp(allyName, Config.ALLY_NAME_TEMPLATE))
			return null;

		for(Entry<Integer, L2Alliance> n = _alliances.head(), end = _alliances.tail(); (n = n.getNext()) != end;)
			if(n.getValue().getAllyName().equalsIgnoreCase(allyName))
				return n.getValue();

		L2Alliance ally = null;
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet result = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT ally_id FROM ally_data WHERE ally_name = ?");
			statement.setString(1, allyName);
			result = statement.executeQuery();
			if(result.next())
			{
				int allyId = result.getInt("ally_id");
				ally = getAlliance(allyId);
			}
		}
		catch(Exception e)
		{
			_log.warning("data error on ally item '" + allyName + "': " + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, result);
		}
		return ally;
	}

	public L2Clan createClan(L2Player player, String clanName)
	{
		L2Clan clan = null;

		if(getClanByName(clanName) == null)
		{
			L2ClanMember leader = new L2ClanMember(player);
			leader.setPlayerInstance(player);
			clan = new L2Clan(IdFactory.getInstance().getNextId(), clanName, leader);
			clan.store();
			player.setClan(clan);
			player.setPowerGrade(6);
			_clans.put(clan.getClanId(), clan);
		}

		return clan;
	}

	public void dissolveClan(L2Player player)
	{
		L2Clan clan = player.getClan();
		long curtime = System.currentTimeMillis();
		SiegeManager.getInstance().removeSiegeSkills(player);
		SystemMessage sm = new SystemMessage(SystemMessage.YOU_HAVE_RECENTLY_BEEN_DISMISSED_FROM_A_CLAN_YOU_ARE_NOT_ALLOWED_TO_JOIN_ANOTHER_CLAN_FOR_24_HOURS);
		for(L2Player clanMember : clan.getOnlineMembers(null))
		{
			clanMember.setClan(null);
			clanMember.setTitle(null);
			clanMember.sendPacket(new PledgeShowMemberListDeleteAll());
			clanMember.broadcastUserInfo();
			clanMember.sendPacket(sm);
			clanMember.setLeaveClanTime(curtime);
		}
		clan.flush();
		deleteClanFromDb(clan.getClanId());
		_clans.remove(clan.getClanId());
		player.sendPacket(new SystemMessage(SystemMessage.CLAN_HAS_DISPERSED));
		player.setDeleteClanTime(curtime);
	}

	public void deleteClanFromDb(int clanId)
	{
		long curtime = System.currentTimeMillis();
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("UPDATE characters SET clanid=0,title='',pledge_type=0,pledge_rank=0,lvl_joined_academy=0,apprentice=0,leaveclan=? WHERE clanid=?");
			statement.setLong(1, curtime / 1000);
			statement.setInt(2, clanId);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM clan_data WHERE clan_id=?");
			statement.setInt(1, clanId);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM clan_subpledges WHERE clan_id=?");
			statement.setInt(1, clanId);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM clan_privs WHERE clan_id=?");
			statement.setInt(1, clanId);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM clan_skills WHERE clan_id=?");
			statement.setInt(1, clanId);
			statement.execute();

		}
		catch(Exception e)
		{
			_log.warning("could not dissolve clan:" + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	public L2Alliance createAlliance(L2Player player, String allyName)
	{
		L2Alliance alliance = null;

		if(getAllyByName(allyName) == null)
		{
			L2Clan leader = player.getClan();
			alliance = new L2Alliance(IdFactory.getInstance().getNextId(), allyName, leader);
			alliance.store();
			_alliances.put(alliance.getAllyId(), alliance);

			player.getClan().setAllyId(alliance.getAllyId());
			for(L2Player temp : player.getClan().getOnlineMembers(null))
				temp.broadcastUserInfo();
			if(Config.DEBUG)
				_log.fine("New ally created: " + alliance.getAllyId() + " " + alliance.getAllyName());

		}

		return alliance;
	}

	public void dissolveAlly(L2Player player)
	{
		int allyId = player.getAllyId();
		SystemMessage sm = new SystemMessage(SystemMessage.YOU_HAVE_WITHDRAWN_FROM_THE_ALLIANCE);
		for(L2Clan member : player.getAlliance().getMembers())
		{
			member.setAllyId(0);
			member.broadcastClanStatus();
			member.broadcastToOnlineMembers(sm);
			member.setLeavedAlly();
		}
		deleteAllyFromDb(allyId);
		_alliances.remove(allyId);
		player.sendPacket(new SystemMessage(SystemMessage.THE_ALLIANCE_HAS_BEEN_DISSOLVED));
		player.getClan().setDissolvedAlly();
	}

	public void deleteAllyFromDb(int allyId)
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("UPDATE clan_data SET ally_id=0 WHERE ally_id=?");
			statement.setInt(1, allyId);
			statement.execute();
			DatabaseUtils.closeStatement(statement);

			statement = con.prepareStatement("DELETE FROM ally_data WHERE ally_id=?");
			statement.setInt(1, allyId);
			statement.execute();
		}
		catch(Exception e)
		{
			_log.warning("could not dissolve clan:" + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
	}

	public void startClanWar(L2Clan clan1, L2Clan clan2)
	{
		// clan1 is declaring war against clan2
		clan1.setEnemyClan(clan2);
		clan2.setAttackerClan(clan1);
		clan1.broadcastClanStatus();
		clan2.broadcastClanStatus();
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();

			statement = con.prepareStatement("REPLACE INTO clan_wars (clan1, clan2) VALUES(?,?)");
			statement.setInt(1, clan1.getClanId());
			statement.setInt(2, clan2.getClanId());
			statement.execute();
		}
		catch(Exception e)
		{
			_log.warning("could not store clan war data:" + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
		clan1.broadcastToOnlineMembers(new SystemMessage(SystemMessage.CLAN_WAR_HAS_BEEN_DECLARED_AGAINST_S1_CLAN_IF_YOU_ARE_KILLED_DURING_THE_CLAN_WAR_BY_MEMBERS_OF_THE_OPPOSING_CLAN_THE_EXPERIENCE_PENALTY_WILL_BE_REDUCED_TO_1_4_OF_NORMAL).addString(clan2.getName()));
		clan2.broadcastToOnlineMembers(new SystemMessage(SystemMessage.S1_CLAN_HAS_DECLARED_CLAN_WAR).addString(clan1.getName()));
	}

	public void stopClanWar(L2Clan clan1, L2Clan clan2)
	{
		// clan1 is ceases war against clan2
		clan1.deleteEnemyClan(clan2);
		clan2.deleteAttackerClan(clan1);

		clan1.broadcastClanStatus();
		clan2.broadcastClanStatus();
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("DELETE FROM clan_wars WHERE clan1=? AND clan2=?");
			statement.setInt(1, clan1.getClanId());
			statement.setInt(2, clan2.getClanId());
			statement.execute();
		}
		catch(Exception e)
		{
			_log.warning("could not delete war data:" + e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseCS(con, statement);
		}
		clan1.broadcastToOnlineMembers(new SystemMessage(SystemMessage.THE_WAR_AGAINST_S1_CLAN_HAS_BEEN_STOPPED).addString(clan2.getName()));
		clan2.broadcastToOnlineMembers(new SystemMessage(SystemMessage.S1_CLAN_HAS_STOPPED_THE_WAR).addString(clan1.getName()));
	}

	private void restorewars()
	{
		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		ResultSet rset = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();

			statement = con.prepareStatement("SELECT clan1, clan2 FROM clan_wars");
			rset = statement.executeQuery();
			L2Clan clan1;
			L2Clan clan2;
			while(rset.next())
			{
		        clan1 = getClan(rset.getInt("clan1"));
				clan2 = getClan(rset.getInt("clan2"));
				if(clan1 != null && clan2 != null)
				{
					clan1.setEnemyClan(clan2);
					clan2.setAttackerClan(clan1);
				}
			}
		}
		catch(Exception e)
		{
			_log.warning("could not restore clan wars data:");
			e.printStackTrace();
		}
		finally
		{
			DatabaseUtils.closeDatabaseCSR(con, statement, rset);
		}
	}
}