package L2jFT.game.tables;

import javolution.util.FastMap;
import L2jFT.database.DatabaseUtils;
import L2jFT.database.FiltredPreparedStatement;
import L2jFT.database.FiltredStatement;
import L2jFT.database.L2DatabaseFactory;
import L2jFT.database.ThreadConnection;
import L2jFT.game.model.L2Clan;
import L2jFT.game.model.L2Player;
import L2jFT.game.model.L2Skill;
import L2jFT.game.model.L2SkillLearn;
import L2jFT.game.model.base.ClassId;
import L2jFT.game.model.base.L2EnchantSkillLearn;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;

@SuppressWarnings( { "nls", "unqualified-field-access", "boxing" })
public class SkillTreeTable
{
	private static final Logger _log = Logger.getLogger(SkillTreeTable.class.getName());

	private static final SkillTreeTable _instance = new SkillTreeTable();

	private static FastMap<ClassId, ArrayList<L2SkillLearn>> _skillTrees;
	private static ArrayList<FastMap<Integer, FastMap<Integer, L2SkillLearn>>> _skillCostTable;
	private static FastMap<Integer, ArrayList<L2EnchantSkillLearn>> _enchant;
	private static ArrayList<L2SkillLearn> _fishingSkills;
	private static ArrayList<L2SkillLearn> _clanSkills;

	public static SkillTreeTable getInstance()
	{
		return _instance;
	}

	/**
	 * Return the minimum level needed to have this Expertise.<BR><BR>
	 *
	 * @param grade The grade level searched
	 *
	 */
	public static short getExpertiseLevel(int grade)
	{
		if(grade <= 0)
			return 0;

		for(L2SkillLearn sl : SkillTreeTable._skillTrees.get(ClassId.paladin))
			if(sl.id == 239 && sl.skillLevel == grade)
				return sl.minLevel;

		throw new Error("Expertise not found for grade " + grade);
	}

	public static int getMinSkillLevel(int skillID, ClassId classID, int skillLVL)
	{
		if(skillID > 0 && skillLVL > 0)
			for(L2SkillLearn sl : SkillTreeTable._skillTrees.get(classID))
				if(sl.skillLevel == skillLVL && sl.id == skillID)
					return sl.minLevel;

		return 0;
	}

	private SkillTreeTable()
	{
		_skillTrees = new FastMap<ClassId, ArrayList<L2SkillLearn>>();
		_fishingSkills = new ArrayList<L2SkillLearn>();
		_clanSkills = new ArrayList<L2SkillLearn>();
		int classId = 0;
		int count = 0;

		ThreadConnection con = null;
		FiltredPreparedStatement statement = null;
		FiltredPreparedStatement statement2 = null;
		ResultSet classlist = null, skilltree = null;
		try
		{
			con = L2DatabaseFactory.getInstance().getConnection();
			statement = con.prepareStatement("SELECT * FROM class_list ORDER BY id");
			statement2 = con.prepareStatement("SELECT class_id, skill_id, level, name, sp, min_level, rep FROM skill_trees where class_id=? AND class_id >= 0 ORDER BY skill_id, level");
			classlist = statement.executeQuery();
			while(classlist.next())
			{
				ArrayList<L2SkillLearn> list = new ArrayList<L2SkillLearn>();
				int parentClassId = classlist.getInt("parent_id");
				classId = classlist.getInt("id");

				statement2.setInt(1, classId);
				skilltree = statement2.executeQuery();
				if(parentClassId > -1)
				{
					ArrayList<L2SkillLearn> parentList = _skillTrees.get(ClassId.values()[parentClassId]);
					list.addAll(parentList);
				}

				addSkills(con, skilltree, list);

				_skillTrees.put(ClassId.values()[classId], list);
				count += list.size();
				_log.fine("SkillTreeTable: skill tree for class " + classId + " has " + list.size() + " skills");
			}
			DatabaseUtils.closeDatabaseSR(statement, classlist);
			statement = null;
			classlist = null;
			DatabaseUtils.closeDatabaseSR(statement2, skilltree);
			loadFishingSkills(con);
			loadClanSkills(con);
			loadEnchants(con);
		}
		catch(Exception e)
		{
			_log.log(Level.SEVERE, "error while creating skill tree for classId " + classId, e);
		}
		finally
		{
			DatabaseUtils.closeDatabaseSR(statement, classlist);
			DatabaseUtils.closeDatabaseSR(statement2, skilltree);
			DatabaseUtils.closeConnection(con);
		}

		loadSkillCostTable();

		_log.config("SkillTreeTable: Loaded " + count + " skills.");
		_log.config("SkillTreeTable: Loaded " + _fishingSkills.size() + " fishing skills.");
		_log.config("SkillTreeTable: Loaded " + _clanSkills.size() + " clan skills.");
		_log.config("SkillTreeTable: Loaded " + _enchant.size() + " enchanted skills.");
	}

	private void loadFishingSkills(ThreadConnection con) throws SQLException
	{
		FiltredPreparedStatement statement = null;
		ResultSet skilltree = null;
		try
		{
			statement = con.prepareStatement("SELECT class_id, skill_id, level, name, sp, min_level, rep FROM skill_trees WHERE class_id=-1 ORDER BY skill_id, level");
			skilltree = statement.executeQuery();
			addSkills(con, skilltree, _fishingSkills);
		}
		finally
		{
			DatabaseUtils.closeDatabaseSR(statement, skilltree);
		}
	}

	private void loadClanSkills(ThreadConnection con) throws SQLException
	{
		FiltredPreparedStatement statement = null;
		ResultSet skilltree = null;
		try
		{
			statement = con.prepareStatement("SELECT class_id, skill_id, level, name, sp, min_level, rep FROM skill_trees WHERE class_id=-2 ORDER BY skill_id, level");
			skilltree = statement.executeQuery();
			addSkills(con, skilltree, _clanSkills);
		}
		finally
		{
			DatabaseUtils.closeDatabaseSR(statement, skilltree);
		}
	}

	private void addSkills(ThreadConnection con, ResultSet skilltree, ArrayList<L2SkillLearn> dest) throws SQLException
	{
		while(skilltree.next())
		{
			short id = skilltree.getShort("skill_id");
			byte lvl = skilltree.getByte("level");
			String name = skilltree.getString("name");
			byte minLvl = skilltree.getByte("min_level");
			int cost;
			short itemId = 0;
			int itemCount = 0;
			if(skilltree.getInt("class_id") == -2)
				cost = skilltree.getInt("rep");
			else
				cost = skilltree.getInt("sp");
			FiltredPreparedStatement statement2 = con.prepareStatement("SELECT item_id, item_count FROM skill_spellbooks WHERE skill_id=? AND level=?");
			statement2.setInt(1, id);
			statement2.setInt(2, lvl);
			ResultSet itemIdCount = statement2.executeQuery();
			if(itemIdCount.next())
			{
				itemId = itemIdCount.getShort("item_id");
				itemCount = itemIdCount.getInt("item_count");
			}
			statement2.close();
			L2SkillLearn skl = new L2SkillLearn(id, lvl, minLvl, name, cost, itemId, itemCount, skilltree.getInt("class_id") == -1, skilltree.getInt("class_id") == -2);
			dest.add(skl);
		}
	}

	@SuppressWarnings("unchecked")
	private void loadSkillCostTable()
	{
		_skillCostTable = new ArrayList<FastMap<Integer, FastMap<Integer, L2SkillLearn>>>(ClassId.values().length + 1);
		for(ClassId cid : ClassId.values())
			_skillCostTable.add(cid.getId(), new FastMap<Integer, FastMap<Integer, L2SkillLearn>>());

		for(ClassId classId : _skillTrees.keySet())
		{
			FastMap<Integer, FastMap<Integer, L2SkillLearn>> skt = _skillCostTable.get(classId.getId());
			while(classId != null)
			{
				ArrayList<L2SkillLearn> lst = _skillTrees.get(classId);
				for(L2SkillLearn skl : lst)
				{
					FastMap<Integer, L2SkillLearn> skillmap = skt.get((int) skl.getId());
					if(skillmap == null)
					{
						skillmap = new FastMap<Integer, L2SkillLearn>();
						skt.put((int) skl.getId(), skillmap);
					}
					skillmap.put((int) skl.getLevel(), skl);
				}
				classId = classId.getParent();
			}
		}
	}

	private void loadEnchants(ThreadConnection con) throws SQLException
	{
		_enchant = new FastMap<Integer, ArrayList<L2EnchantSkillLearn>>();
		FiltredStatement statement = null;
		ResultSet skilltree = null;
		try
		{
			statement = con.createStatement();
			skilltree = statement.executeQuery("SELECT skill_id, level, name, base_lvl, enchant_type, sp, min_skill_lvl FROM enchant_skill_trees ORDER BY skill_id, level");
			int cnt = 0;
			int unimplemented = 0;
			while(skilltree.next())
			{
				int id = skilltree.getInt("skill_id");
				short level = skilltree.getShort("level");
				String name = skilltree.getString("name");
				int base = skilltree.getInt("base_lvl");
				String type = skilltree.getString("enchant_type");
				int sp = skilltree.getInt("sp");
				int min = skilltree.getInt("min_skill_lvl");
				L2EnchantSkillLearn e = new L2EnchantSkillLearn(id, level, name, type, min, base, sp);

				ArrayList<L2EnchantSkillLearn> t = _enchant.get(id);
				if(t == null)
					t = new ArrayList<L2EnchantSkillLearn>();
				t.add(e);
				_enchant.put(id, t);

				try
				{
					SkillTable.getInstance().getInfo(id, base + level - 100).setDisplayLevel(level);
				}
				catch(NullPointerException x)
				{
					unimplemented++;
				}

				cnt++;
			}
			_log.config("SkillTreeTable: Loaded " + cnt + " skills enchants.");
			_log.config("SkillTreeTable: Found " + unimplemented + " unimplemented skills enchants.");
		}
		finally
		{
			DatabaseUtils.closeDatabaseSR(statement, skilltree);
		}
	}

	public void reloadEnchant()
	{
		for(ArrayList<L2EnchantSkillLearn> t : _enchant.values())
			if(t != null)
				for(L2EnchantSkillLearn e : t)
					try
					{
						SkillTable.getInstance().getInfo(e.getId(), e.getBaseLevel() + e.getLevel() - 100).setDisplayLevel((short) e.getLevel());
					}
					catch(NullPointerException x)
					{}
	}

	public L2SkillLearn[] getAvailableSkills(L2Player cha, ClassId classId)
	{
		ArrayList<L2SkillLearn> result = new ArrayList<L2SkillLearn>();
		ArrayList<L2SkillLearn> skills = _skillTrees.get(classId);
		if(skills == null)
		{
			// the skilltree for this class is undefined, so we give an empty list
			_log.warning("Skilltree for class " + classId + " is not defined !");
			return new L2SkillLearn[0];
		}

		Collection<L2Skill> oldSkills = cha.getAllSkills();
		for(L2SkillLearn temp : skills)
			if(temp.minLevel <= cha.getLevel())
			{
				boolean knownSkill = false;
				for(L2Skill s : oldSkills)
				{
					if(knownSkill)
						break;
					if(s.getId() == temp.id)
					{
						knownSkill = true;
						if(s.getLevel() == temp.skillLevel - 1)
							// this is the next level of a skill that we know
							result.add(temp);
					}
				}
				if(!knownSkill && temp.skillLevel == 1)
					// this is a new skill
					result.add(temp);
			}
		return result.toArray(new L2SkillLearn[result.size()]);
	}

	public L2SkillLearn[] getAvailableClanSkills(L2Clan clan)
	{
		ArrayList<L2SkillLearn> result = new ArrayList<L2SkillLearn>();
		ArrayList<L2SkillLearn> skills = _clanSkills;

		if(skills == null)
			return new L2SkillLearn[0];

		L2Skill[] oldSkills = clan.getAllSkills();

		for(L2SkillLearn temp : skills)
			if(temp.minLevel <= clan.getLevel())
			{
				boolean knownSkill = false;

				for(int j = 0; j < oldSkills.length && !knownSkill; j++)
					if(oldSkills[j].getId() == temp.id)
					{
						knownSkill = true;

						if(oldSkills[j].getLevel() == temp.skillLevel - 1)
							// this is the next level of a skill that we know
							result.add(temp);
					}

				if(!knownSkill && temp.skillLevel == 1)
					// this is a new skill
					result.add(temp);
			}

		return result.toArray(new L2SkillLearn[result.size()]);
	}

	public ArrayList<L2EnchantSkillLearn> getAvailableEnchants(L2Player cha)
	{
		ArrayList<L2EnchantSkillLearn> result = new ArrayList<L2EnchantSkillLearn>();
		Collection<L2Skill> skills = cha.getAllSkills();
		if(skills.size() == 0)
			return result;

		for(L2Skill s : skills)
		{
			ArrayList<L2EnchantSkillLearn> enchants = _enchant.get(s.getId().intValue());
			if(enchants != null)
				for(L2EnchantSkillLearn e : enchants)
					if(s.getLevel() == e.getMinSkillLevel() || s.getLevel() == e.getMinSkillLevel() + e.getBaseLevel() - 100)
						result.add(e);
			//System.out.println(e.getName() + " : " + e.get_type());
		}

		return result;
	}

	public static L2EnchantSkillLearn getSkillEnchant(Short skillid, short level)
	{
		ArrayList<L2EnchantSkillLearn> enchants = _enchant.get(skillid.intValue());
		if(enchants == null)
			return null;
		for(L2EnchantSkillLearn e : enchants)
			if(e.getLevel() == level)
				return e;
		return null;
	}

	public static L2SkillLearn getSkillLearn(short skillid, short level, ClassId classid)
	{
		if(_fishingSkills != null)
			for(L2SkillLearn tmp : _fishingSkills)
				if(tmp.id == skillid && tmp.skillLevel == level)
					return tmp;

		if(_clanSkills != null)
			for(L2SkillLearn tmp : _clanSkills)
				if(tmp.id == skillid && tmp.skillLevel == level)
					return tmp;

		//		for(ArrayList<L2SkillLearn> tmp1 : _skillTrees.get(classid))
		for(L2SkillLearn tmp : _skillTrees.get(classid))
			if(tmp.id == skillid && tmp.skillLevel == level)
				return tmp;

		return null;
	}

	public L2SkillLearn[] getAvailableFishingSkills(L2Player cha)
	{
		ArrayList<L2SkillLearn> result = new ArrayList<L2SkillLearn>();
		if(_fishingSkills == null)
		{
			_log.warning("Fishing skills not defined!");
			return new L2SkillLearn[0];
		}

		Collection<L2Skill> oldSkills = cha.getAllSkills();

		for(L2SkillLearn temp : _fishingSkills)
			if(temp.minLevel <= cha.getLevel())
			{
				boolean knownSkill = false;
				for(L2Skill s : oldSkills)
				{
					if(knownSkill)
						break;
					if(s.getId() == temp.id)
					{
						knownSkill = true;
						if(s.getLevel() == temp.skillLevel - 1)
							result.add(temp);
					}
				}

				if(!knownSkill && temp.skillLevel == 1)
					result.add(temp);
			}
		return result.toArray(new L2SkillLearn[result.size()]);
	}

	public byte getMinLevelForNewSkill(L2Player cha, ClassId classId)
	{
		byte minlevel = 0;
		//ArrayList<L2SkillLearn> result = new ArrayList<L2SkillLearn>();
		ArrayList<L2SkillLearn> skills = _skillTrees.get(classId);
		if(skills == null)
		{
			// the skilltree for this class is undefined, so we give an empty list
			_log.warning("Skilltree for class " + classId + " is not defined !");
			return minlevel;
		}

		//L2Skill[] oldSkills = cha.getAllSkills();

		for(L2SkillLearn temp : skills)
			if(temp.minLevel > cha.getLevel())
				if(minlevel == 0 || temp.minLevel < minlevel)
					minlevel = temp.minLevel;
		return minlevel;
	}

	public int getSkillCost(L2Player player, L2Skill skill)
	{
		// TODO убрать
		switch(skill.getId())
		{
			case (short) 1312:
			case (short) 1313:
			case (short) 1314:
			case (short) 1315:
			case (short) 1368:
			case (short) 1369:
			case (short) 1370:
			case (short) 1371:
			case (short) 1372:
				return 0;
		}

		FastMap<Integer, FastMap<Integer, L2SkillLearn>> skt = _skillCostTable.get(player.getActiveClass());
		if(skt == null)
			return Integer.MAX_VALUE;
		FastMap<Integer, L2SkillLearn> skillmap = skt.get((int) skill.getId());
		if(skillmap == null)
			return Integer.MAX_VALUE;
		L2SkillLearn skl = skillmap.get(1 + Math.max(player.getSkillLevel(skill.getId()), 0));
		if(skl == null)
			return Integer.MAX_VALUE;
		return skl.getSpCost();
	}

	public int getSkillRepCost(L2Clan clan, L2Skill skill)
	{
		int min = 100000000;
		int lvl = clan.getLeader().getPlayer().getSkillLevel(skill.getId());

		if(lvl > 0)
			lvl += 1;
		else
			lvl = 1;
		if(_clanSkills != null)
			for(L2SkillLearn tmp : _clanSkills)
			{
				if(tmp.id != skill.getId())
					continue;
				if(tmp.skillLevel != lvl)
					continue;
				if(tmp.minLevel > clan.getLevel())
					continue;
				min = Math.min(min, Math.round(tmp._repCost));
			}
		return min;
	}

	/**
	 * Возвращает true если скилл может быть изучен данным классом
	 * @param player
	 * @param skill_id
	 * @param skill_level
	 * @return true/false
	 */
	public boolean isSkillPossible(L2Player player, short skillid, short level)
	{
		for(L2SkillLearn tmp : _clanSkills)
			if(tmp.id == skillid && tmp.skillLevel <= level)
				return true;

		ArrayList<L2SkillLearn> skills = _skillTrees.get(ClassId.values()[player.getActiveClass()]);
		for(L2SkillLearn skilllearn : skills)
			if(skilllearn.id == skillid && skilllearn.skillLevel <= level)
				return true;

		return false;
	}
}