package L2jFT.game.templates;

import L2jFT.util.Rnd;
import L2jFT.game.model.L2Character;
import L2jFT.game.model.L2Skill;
import L2jFT.game.model.instances.L2ItemInstance;
import L2jFT.game.network.serverpackets.SystemMessage;
import L2jFT.game.skills.Env;
import L2jFT.game.skills.funcs.Func;
import L2jFT.game.skills.funcs.FuncTemplate;
import L2jFT.game.tables.SkillTable;

import java.util.ArrayList;

public final class L2Weapon extends L2Item
{
	private final int _soulShotCount;
	private final int _spiritShotCount;
	private final int _pDam;
	private final int _rndDam;
	private final int _critical;
	private final double _hitModifier;
	private final int _avoidModifier;
	private final int _shieldDef;
	private final double _shieldDefRate;
	private final int _atkSpeed;
	private final int _atkReuse;
	private final int _mpConsume;
	private final int _mDam;
	private L2Skill _itemSkill = null; // for passive skill 
	private L2Skill _enchant4Skill = null; // skill that activates when item is enchanted +4 (for duals) 
	private final boolean _sa;

	public enum WeaponType
	{
		NONE(1, "Shield"),
		SWORD(2, "Sword"),
		BLUNT(3, "Blunt"),
		DAGGER(4, "Dagger"),
		BOW(5, "Bow"),
		POLE(6, "Pole"),
		ETC(7, "Etc"),
		FIST(8, "Fist"),
		DUAL(9, "Dual Sword"),
		DUALFIST(10, "Dual Fist"),
		BIGSWORD(11, "Big Sword"), // Two Handed Swords
		PET(12, "Pet"),
		ROD(13, "Rod"),
		BIGBLUNT(14, "Big Blunt");

		private final int _id;
		private final String _name;

		private WeaponType(int id, String name)
		{
			_id = id;
			_name = name;
		}

		public int mask()
		{
			return 1 << _id;
		}

		@Override
		public String toString()
		{
			return _name;
		}
	}

	/**
	 * Constructor for Weapon.<BR><BR>
	 * <U><I>Variables filled :</I></U><BR>
	 * <LI>_soulShotCount & _spiritShotCount</LI>
	 * <LI>_pDam & _mDam & _rndDam</LI>
	 * <LI>_critical</LI>
	 * <LI>_hitModifier</LI>
	 * <LI>_avoidModifier</LI>
	 * <LI>_shieldDes & _shieldDefRate</LI>
	 * <LI>_atkSpeed & _AtkReuse</LI>
	 * <LI>_mpConsume</LI>
	 * @param type : L2ArmorType designating the type of armor
	 * @param set : StatsSet designating the set of couples (key,value) caracterizing the armor
	 * @see L2Item constructor
	 */
	public L2Weapon(WeaponType type, StatsSet set)
	{
		super(type, set);
		_soulShotCount = set.getInteger("soulshots");
		_spiritShotCount = set.getInteger("spiritshots");
		_pDam = set.getInteger("p_dam");
		_rndDam = set.getInteger("rnd_dam");
		_critical = set.getInteger("critical");
		_hitModifier = set.getDouble("hit_modify");
		_avoidModifier = set.getInteger("avoid_modify");
		_shieldDef = set.getInteger("shield_def");
		_shieldDefRate = set.getDouble("shield_def_rate");
		_atkSpeed = set.getInteger("atk_speed");
		_atkReuse = set.getInteger("atk_reuse", type == WeaponType.BOW ? 1500 : 0);
		_mpConsume = set.getInteger("mp_consume");
		_mDam = set.getInteger("m_dam");
		_sa = _icon.endsWith("1_0");

		int sId = set.getInteger("item_skill_id");
		int sLv = set.getInteger("item_skill_lvl");
		if(sId > 0 && sLv > 0)
			_itemSkill = SkillTable.getInstance().getInfo(sId, sLv);

		sId = set.getInteger("enchant4_skill_id");
		sLv = set.getInteger("enchant4_skill_lvl");
		if(sId > 0 && sLv > 0)
			_enchant4Skill = SkillTable.getInstance().getInfo(sId, sLv);
	}

	/**
	 * Returns the type of Weapon
	 * @return L2WeaponType
	 */
	@Override
	public WeaponType getItemType()
	{
		return (WeaponType) super.type;
	}

	/**
	 * Returns the ID of the Etc item after applying the mask.
	 * @return int : ID of the Weapon
	 */
	@Override
	public int getItemMask()
	{
		return getItemType().mask();
	}

	/**
	 * Returns the quantity of SoulShot used.
	 * @return int
	 */
	public int getSoulShotCount()
	{
		return _soulShotCount;
	}

	/**
	 * Returns the quatity of SpiritShot used.
	 * @return int
	 */
	public int getSpiritShotCount()
	{
		return _spiritShotCount;
	}

	/**
	 * Returns the physical damage.
	 * @return int
	 */
	public int getPDamage()
	{
		return _pDam;
	}

	/**
	 * Returns the random damage inflicted by the weapon
	 * @return int
	 */
	public int getRandomDamage()
	{
		return _rndDam;
	}

	/**
	 * Returns the attack speed of the weapon
	 * @return int
	 */
	public int getAttackSpeed()
	{
		return _atkSpeed;
	}

	/**
	 * Return the Attack Reuse Delay of the L2Weapon.<BR><BR>
	 * @return int
	 */
	public int getAttackReuseDelay()
	{
		return _atkReuse;
	}

	/**
	 * Returns the avoid modifier of the weapon
	 * @return int
	 */
	public int getAvoidModifier()
	{
		return _avoidModifier;
	}

	/**
	 * Returns the rate of critical hit
	 * @return int
	 */
	public int getCritical()
	{
		return _critical;
	}

	/**
	 * Returns the hit modifier of the weapon
	 * @return double
	 */
	public double getHitModifier()
	{
		return _hitModifier;
	}

	/**
	 * Returns the magical damage inflicted by the weapon
	 * @return int
	 */
	public int getMDamage()
	{
		return _mDam;
	}

	/**
	 * Returns the MP consumption with the weapon
	 * @return int
	 */
	public int getMpConsume()
	{
		return _mpConsume;
	}

	/**
	 * Returns the shield defense of the weapon
	 * @return int
	 */
	public int getShieldDef()
	{
		return _shieldDef;
	}

	/**
	 * Returns the rate of shield defense of the weapon
	 * @return double
	 */
	public double getShieldDefRate()
	{
		return _shieldDefRate;
	}

	/**
	 * Returns array of Func objects containing the list of functions used by the weapon 
	 * @param instance : L2ItemInstance pointing out the weapon
	 * @param player : L2Character pointing out the player
	 * @return Func[] : array of functions
	 */
	@Override
	public Func[] getStatFuncs(L2ItemInstance instance, L2Character player)
	{
		ArrayList<Func> funcs = new ArrayList<Func>();
		if(_funcTemplates != null)
			for(FuncTemplate t : _funcTemplates)
			{
				Env env = new Env();
				env.character = player;
				env.item = instance;
				Func f = t.getFunc(env, instance);
				if(f != null)
					funcs.add(f);
			}
		return funcs.toArray(new Func[funcs.size()]);
	}

	public void getEffect(boolean critNotCast, L2Character effector, L2Character effected, boolean offensive)
	{
		if(effector == null || effected == null)
			return;

		if(_skill == null)
			return;

		if(_skillOnCritNotCast != critNotCast || _offensive != offensive)
			return;

		if(Rnd.chance(_chance))
			return;

		_skill.getEffects(effector, effected, true);

		effector.sendPacket(new SystemMessage(SystemMessage.S1_HAS_SUCCEEDED).addString("Your weapon's special ability"));
	}

	/**
	 * @return passive skill linked to that weapon 
	 */
	@Override
	public L2Skill getSkill()
	{
		return _itemSkill;
	}

	/**
	 *
	 * @return skill that player get when has equipped weapon +4  or more  (for duals SA) 
	 */
	public L2Skill getEnchant4Skill()
	{
		return _enchant4Skill;
	}

	public boolean isSa()
	{
		return _sa;
	}
}